@@ -2,26 +2,27 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	Feature maintenance by  Matti Aarnio <mea@nic.funet.fi> 1991-2000
+ *	Feature maintenance by  Matti Aarnio <mea@nic.funet.fi> 1991-2003
  *
  */
 
 /* Sendmail -- a sendmail compatible interface to ZMailer */
 
-#include "hostenv.h"
+#include "mailer.h"
+
 #include <stdio.h>
 #include <sysexits.h>
 #include <ctype.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <pwd.h>
 #include "mail.h"
 #include "zmsignal.h"
 #ifdef HAVE_FCNTL_H
 # include <fcntl.h>
 #endif
 #include <sys/file.h>
+#include <errno.h>
 #include "zsyslog.h"
 
 #include "zmalloc.h"
@@ -144,7 +145,7 @@ main(argc, argv)
 	const char *configfile, *mailpriority, *pooption;
 	char * path;
 	FILE	*vfp;
-	struct passwd *pwd;
+	struct Zpasswd *pwd;
 	char	*LC_ctype;
 	char	buf[8192];
 	char   *buf2 = NULL;
@@ -165,6 +166,9 @@ main(argc, argv)
 		progname = argv[0];
 	umask(022);
 
+
+	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
+
 	mailpriority = getenv("MAILPRIORITY");
 
 	/* Pick up the sender's idea about C-chartype.. */
@@ -420,10 +424,10 @@ otherprog:
 		const char *av[30+1];
 		path = emalloc((unsigned)(strlen(mailbin)+1+strlen(SMTPSERVER)+1));
 		sprintf(path, "%s/%s", mailbin, SMTPSERVER);
-		av[0] = "smtp-in";
+		av[0] = "sendmail-smtp-in";
 		av[1] = "-i";
 		n = 2;
-		cp = getzenv("SMTPOPTIONS"); /* Normal smtp-server options */
+		cp = (char*) getzenv("SMTPOPTIONS"); /* Normal smtp-server options */
 		/* pass on suidness */
 		s = strtok(cp, " \t\"'");
 		do {
@@ -521,7 +525,7 @@ otherprog:
 		  else
 #endif /* HAVE_FCHOWN */
 		    if (from == NULL) {
-		      if (uid != 0 && (pwd = getpwuid(uid)) != NULL)
+		      if (uid != 0 && (pwd = zgetpwuid(uid)) != NULL)
 			from = pwd->pw_name;
 		      else if ((cp = getlogin()) != NULL)
 			from = cp;
@@ -552,7 +556,7 @@ otherprog:
 #else
 		  mktemp(verbfile);
 		  if (*verbfile == '\0' ||
-		      (vfd = open(verbfile, O_CREAT|O_RDWR, 0600)) < 0)
+		      (vfd = open(verbfile, O_CREAT|O_EXCL|O_RDWR, 0600)) < 0)
 #endif
 		    {
 		      fprintf(stderr,
@@ -604,7 +608,7 @@ otherprog:
 		fprintf(mfp,"env-end\n");
 
 		if (fflush(mfp) == EOF
-/* #ifdef NFSFSYNC */  /* This is propably ALWAYS a good idea.. */
+/* #ifdef NFSFSYNC */  /* This is probably ALWAYS a good idea.. */
 		    || fsync(fileno(mfp)) < 0
 /* #endif */	/* NFSFSYNC */
 		    || ferror(mfp)) {
@@ -724,20 +728,25 @@ otherprog:
 		  exit(EX_UNAVAILABLE);
 		}
 		if (vfd >= 0) {
+		  long filepos = 0;
 		  int sleeplimit = 260000; /* 260 000 seconds is circa 3d */
 
 		  if (fork() > 0) exit(EX_OK); /* Let the child to do
 						  the trace printout */
 		  if (vfd != 0) close(0);
 		  if (vfd != 1) close(1);
-		  vfp = fdopen(vfd, "r");
+		  vfp = fopen(verbfile, "r");
+		  if (vfp) fseek(vfp, filepos, SEEK_SET);
 		  while (vfp != NULL) {
 		    while (fgets(buf, sizeof buf, vfp) != NULL) {
 		      fprintf(stderr, "%s", buf);
 		      if (STREQN(buf, "scheduler done", 14))
 			sleeplimit = 20; /* 20 seconds to death */
 		    }
-		    clearerr(vfp);
+		    filepos = ftell(vfp);
+		    fclose(vfp);
+		    vfp = fopen(verbfile, "r");
+		    if (vfp) fseek(vfp, filepos, SEEK_SET);
 		    sleep(1);
 		    --sleeplimit;
 		    if (sleeplimit < 0)
@@ -760,11 +769,12 @@ FILE *
 deadletter(uid)
 	int uid;
 {
-	struct passwd *pwd;
+	struct Zpasswd *pwd;
 	FILE *fp;
 	char	buf[8192];
 
-	if ((pwd = getpwuid(uid)) == NULL) {
+	pwd = zgetpwuid(uid);
+	if (pwd == NULL) {
 		fprintf(stderr, "%s: can't save mail!\n", zmailer);
 		exit(EX_NOUSER);
 		/* NOTREACHED */
