@@ -35,7 +35,7 @@ extern char *    strnsave __((const char *s, const size_t n));
 
 /* cfgets.c */
 #ifdef _SFIO_H
-extern int cfgets __((char *, int, Sfio_t *));
+extern int csfgets __((char *, int, Sfio_t *));
 #endif
 
 /* cleanenv.c */
@@ -61,8 +61,8 @@ extern char *dottedquad __((struct in_addr *inp));
 extern int eopen __((const char *s, int f, int m));
 extern int eread __((int fd, char *buf, int len));
 extern int epipe __((int fdarr[2]));
-extern univptr_t emalloc  __((size_t len));
-extern univptr_t erealloc __((univptr_t buf, size_t len));
+extern void *emalloc  __((size_t len));
+extern void *erealloc __((void *buf, size_t len));
 #ifdef S_IFMT
 extern int efstat __((int fd, struct stat *stbuf));
 extern int estat  __((const char *path, struct stat *stbuf));
@@ -110,12 +110,10 @@ extern unsigned long parse_interval __((char *str, char **retp));
 /* prversion.c */
 extern void prversion __((const char *prgname));
 
-#ifdef USE_ZGETPWNAM
 /* pwdgrp.c */
-extern struct passwd	*zgetpwnam __((const char *name));
-extern struct passwd	*zgetpwuid __((const char *uid));
-extern struct group	*zgetgrnam __((const char *name));
-#endif
+extern struct Zpasswd	*zgetpwnam __((const char *name));
+extern struct Zpasswd	*zgetpwuid __((const uid_t uid));
+extern struct Zgroup	*zgetgrnam __((const char *name));
 
 /* ranny.c */
 extern u_int ranny __((u_int m));
@@ -135,20 +133,34 @@ extern token822 * scan822utext __((const char **cpp, size_t n,
 
 /* selfaddrs.c */
 
+#ifndef __Usockaddr__
+typedef union {
+    struct sockaddr_in v4;
+#ifdef INET6
+    struct sockaddr_in6 v6;
+#endif
+} Usockaddr;
+#define __Usockaddr__
+#endif
+
+
 struct sockaddr; /* a "forward" declaration */
 struct addrinfo; /* a "forward" declaration */
 
 extern void stashmyaddresses  __((const char *host));
 #ifdef SOCK_STREAM
-extern int  loadifaddresses   __((struct sockaddr ***));
-extern int  matchmyaddress    __((struct sockaddr *));
+extern int  loadifaddresses   __((Usockaddr ***));
+extern int  matchmyaddress    __((Usockaddr *));
 #endif
 #ifdef EAI_AGAIN   /* have 'struct addrinfo' */
 extern int  matchmyaddresses  __((struct addrinfo *));
 #endif
 
 /* zgetifaddress.c */
-extern int zgetifaddress __((int af, const char *ifname, struct sockaddr *));
+extern int zgetifaddress __((int af, const char *ifname, Usockaddr *));
+
+/* zgetbindaddr.c */
+extern int zgetbindaddr __((char *ifname, const int af, Usockaddr *));
 
 /* splay.c */
 /* .... much ... in  "splay.h" */
@@ -201,3 +213,19 @@ extern const char *Version;
 extern const char *VersionNumb;
 extern const char *CC_user;
 extern const char *CC_pwd;
+
+/* zshmmibattach.c */
+extern int  Z_SHM_MIB_Attach      __((int rw));
+extern int  Z_SHM_MIB_is_attached __((void)); /* True if we do have the segment */
+extern void Z_SHM_MIB_Detach      __((void)); /* automatic atexit() handling */
+
+extern struct MIB_MtaEntry *MIBMtaEntry; /* public MIB block pointer, either
+					    private data before attach call,
+					    or possibly shared data after the
+					    call... */
+extern long Z_SHM_FileSysFreeSpace __((void)); /* in KILOBYTES */
+
+/* fdstatfs.c */
+extern long used_fd_statfs __((int fd));
+extern long free_fd_statfs __((int fd));
+
