@@ -87,9 +87,12 @@ struct policy_user {
 #define P_A_TrustWhosOn		21
 #define P_A_RcptDnsRBL		22
 #define P_A_TestRcptDnsRBL	23
+#define P_A_WarnRcptDnsRBL	24
+#define P_A_Filtering           25
+#define P_A_MaxSameIpSource	26
 
 #define P_A_FirstAttr	        2
-#define P_A_LastAttr	        23
+#define P_A_LastAttr	        26
 /* Note: Attribute codes outside range 1..31 cause problems at policystate
          processing!  If you ever need modify these, fix the  policytest.c,
 	 and  policytest.h: struct policystate { char values[]; } array,
@@ -138,7 +141,10 @@ static char *_KA[] = {
 	"trustrecipients",
 	"trust-whoson",
 	"rcpt-dns-rbl",
-	"test-rcpt-dns-rbl"
+	"test-rcpt-dns-rbl",
+	"warn-rcpt-dns-rbl",
+	"filtering",
+	"maxsameipsource"
 };
 #define KA(x) ((((x)>0)&&((x)<=P_A_LastAttr))?_KA[(x) & 0xFF]:"??")
 
