@@ -33,6 +33,8 @@ struct spblk {
 
 struct sptree {
 	struct spblk	*root;		/* root node */
+	struct sptree	*symbols;	/* If this db needs symbol support,
+					   here is another sptree for those */
 	int		eltscnt;	/* How many elements in this tree */
 	int		lookups;	/* number of splookup()s */
 	int		lkpcmps;	/* number of lookup comparisons */
@@ -56,6 +58,18 @@ extern void sp_delete __((struct spblk *n, struct sptree *q)); /* delete node
 extern void        sp_null __((struct sptree *));
 extern const char *sp_stats __((struct sptree *q));/* return tree statistics */
 extern spkey_t     symbol __((const void *s));	/* build this into a symbol */
+extern spkey_t     symbol_lookup __((const void *s));
+extern spkey_t     symbol_db        __((const void *, struct sptree *));
+extern spkey_t     symbol_lookup_db __((const void *, struct sptree *));
+extern spkey_t     symbol_db_mem    __((const void *, int, struct sptree *));
+extern spkey_t symbol_lookup_db_mem __((const void *, int, struct sptree *));
+extern void	   symbol_free_db __((const void *, struct sptree *));
+extern void	   symbol_null_db __((struct sptree *));
+
+extern struct spblk *lookup_incoresp __((const char *, struct sptree *));
+extern int      add_incoresp __((const char *, const char *, struct sptree *));
+extern int     addd_incoresp __((const char *, const void *, struct sptree *));
+
 extern const char *pname  __((spkey_t id));
 #ifdef MALLOC_TRACE
 extern int  icpname    __((struct spblk *spl));
