@@ -3,7 +3,7 @@
  *	This will be free software, but only when it is finished.
  */
 /*
- *	A plenty of changes, copyright Matti Aarnio 1990-1999
+ *	A plenty of changes, copyright Matti Aarnio 1990-2003
  */
 
 #ifndef _Z_TA_H_
@@ -20,13 +20,16 @@
 #include <sfio.h>
 
 
+#define EX_DEFERALL 100 /* Outside <sysexits.h> codes */
+
+
 struct taddress {
 	struct taddress	*link;		/* next sender / sender for this rcpt */
 	const char	*channel;
 	const char	*host;
 	const char	*user;
 	const char	*misc;		/* expected to be uid privilege */
-	const char	**routermxes;	/* [mea] hostpointers from router */
+	struct taddress *ta_next;	/* global freeup chain */
 };
 
 #define _DSN_NOTIFY_SUCCESS	0x001
@@ -54,12 +57,14 @@ typedef enum {
 } CONVERTMODE;
 
 struct rcpt {
-	struct rcpt	*next;
+	struct rcpt	*next;		/* next rcpt in this group */
+	struct rcpt	*rp_next;	/* ctlclose() link-chain */
 	struct taddress	*addr;		/* addr.link is the sender address */
 	const char	*orcpt;		/*  DSN  ORCPT=  string */
 	const char	*inrcpt;	/* "DSN" INRCPT= string */
 	const char	*infrom;	/* "DSN" INFROM= string */
 	const char	*notify;	/*  DSN  NOTIFY= flags  */
+	const char	*ezmlm;		/* "DSN" EZMLM=  string */
 	int		notifyflgs;
 	time_t		deliverby;	/*  RFC 2852  DELIVERBY */
 	int		deliverbyflgs;
@@ -74,6 +79,7 @@ struct rcpt {
 					   Thus enabling rewrite of the
 					   header on the transport. */
 	char		***newmsgheadercvt; /* the rewrite results */
+	char		*top_received;
 	int		id;		/* the index of this address */
 	int		lockoffset;	/* the index of the address lock */
 	int		headeroffset;
@@ -97,6 +103,7 @@ struct ctldesc {
 	const char	*verbose;	/* file for verbose logging */
 	const char	*envid;		/* DSN ENVID data */
 	const char	*dsnretmode;	/* DSN RET=-mode */
+	const char	*taspoolid;
 	time_t		msgmtime;	/* Message file arrival time */
 	long		msgbodyoffset;	/* offset of message body in msgfile */
 	long		msgsizeestimate; /* Estimate of the msg size */
@@ -109,17 +116,18 @@ struct ctldesc {
 	long		contentsize;	/* message file size */
 	long		*offset;	/* array of indices into contents */
 	struct taddress	*senders;	/* list of sender addresses */
+	struct taddress *ta_chain;	/* freeup chain for these */
 	struct rcpt	*recipients;	/* list of selected recipients */
+	struct rcpt	*rp_chain;	/* freeup chain for these */
 	int		rcpnts_total;	/* number of recipients, total */
 	int		rcpnts_remaining;/* .. how many yet to deliver */
 	int		rcpnts_failed  ;/* .. how many failed ones */
 	long		format;		/* _CF_FORMAT_xxx */
 	char		***msgheaders;	/* pointer to all msg headers */
 	char		***msgheaderscvt; /* converted headers */
-#ifdef	HAVE_MMAP
 	const char	*let_buffer;	/* MMAP()ed memory area containing */
 	const char	*let_end;	/* the mail -- and its end..	   */
-#endif
+	int		let_buffer_size; /* != 0 If it is 'malloc()ed! */
 };
 
 
@@ -150,9 +158,10 @@ struct mimestate {
 
 
 /* ctlopen.c: */
+extern int	       ta_use_mmap;
 extern void            ctlfree __((struct ctldesc *dp, void *anyp));
 extern void           *ctlrealloc __((struct ctldesc *dp, void *anyp, size_t size));
-extern struct ctldesc *ctlopen __((const char *file, const char *channel, const char *host, int *exitflag, int (*selectaddr)(const char *, const char *, void *), void *saparam, int (*matchrouter)(const char *, struct taddress *, void *), void *mrparam));
+extern struct ctldesc *ctlopen __((const char *file, const char *channel, const char *host, int *exitflag, int (*selectaddr)(const char *, const char *, void *), void *saparam));
 extern void            ctlclose __((struct ctldesc *dp));
 extern int	       ctlsticky __((const char *spec_host, const char *addr_host, void *cbparam));
 
@@ -165,10 +174,11 @@ extern void            notary_setwtt __(( const char *host ));
 extern void            notary_setwttip __(( const char *ip ));
 extern void            notary_settaid __(( const char *name, int ));
 extern void            notary_setcvtmode __(( CONVERTMODE ));
+extern void	       notaryflush __(( void ));
 #if defined(HAVE_STDARG_H)
-extern void	       diagnostic __((struct rcpt *rp, int rc, int timeout, const char *fmt, ... ));
+extern void	       diagnostic __((FILE *verboselog, struct rcpt *rp, int rc, int timeout, const char *fmt, ... ));
 #else
-extern void	       diagnostic __((/* struct rcpt *, int, int, char *,... */));
+extern void	       diagnostic __((/* FILE *verboselog, struct rcpt *, int, int, char *,... */));
 #endif
 
 
@@ -194,7 +204,9 @@ extern int append_header __((struct rcpt *rp, const char *fmt, ...));
 extern int append_header __(());
 #endif
 extern struct cte_data *parse_content_encoding __((const char *cte_line));
+extern void             free_content_encoding  __((struct cte_data *cte));
 extern struct ct_data  *parse_content_type __((const char *ct_line));
+extern void		free_content_type  __((struct ct_data *ct));
 extern void output_content_type __((struct rcpt *rp, struct ct_data *ct, char **oldpos));
 extern int check_conv_prohibit __((struct rcpt *rp));
 extern int cte_check __((struct rcpt *rp));
@@ -202,6 +214,7 @@ extern char **has_header __((struct rcpt *rp, const char *keystr));
 extern void delete_header __((struct rcpt *rp, char **hdrp));
 extern int  downgrade_charset __((struct rcpt *rp, FILE *verboselog));
 extern int  downgrade_headers __((struct rcpt *rp, CONVERTMODE downgrade, FILE *verboselog));
+extern int  header_received_for_clause __((struct rcpt *rp, int rcptcnt, FILE *verboselog));
 extern int qp_to_8bit __((struct rcpt *rp));
 
 /* mime2headers.c */
@@ -237,9 +250,4 @@ extern void tasyslog __((struct rcpt *rp, int xdelay, const char *wtthost, const
 
 extern int getmyuucename __((char *, int));
 
-/* nonblocking.c */
-extern int  fd_nonblockingmode __((int fd));
-extern int  fd_blockingmode __((int fd));
-extern void fd_restoremode __((int fd, int mode));
-
 #endif
