@@ -6,17 +6,19 @@
 #include <stdio.h>
 #include "hostenv.h"
 #include "mailer.h"
+#include "libz.h"
 
 /*
  * Abnormal death: blabber about why we are croaking, then die.
  */
 
+extern char *progname;
+
 void
 die(status, message)
 	int status;
-	char *message;
+	const char *message;
 {
-	extern char *progname;
 #ifdef	MALLOC_TRACE
 	int i;
 
