@@ -91,7 +91,7 @@ int
 zgetifaddress(af, ifname, sap)
      int af;
      const char *ifname;
-     struct sockaddr * sap;
+     Usockaddr * sap;
 {
 	int i;
 
@@ -100,6 +100,8 @@ zgetifaddress(af, ifname, sap)
 	{
 	  struct ifaddrs *ifar = NULL, *ifa;
 
+	  if (sap == NULL) return -1; /* Bad param! */
+
 	  i = getifaddrs( &ifar );
 	  if (i < 0) {
 	    free(sap);
@@ -123,24 +125,18 @@ zgetifaddress(af, ifname, sap)
 		continue; /* Nope.. */
 
 	      if (sa->sa_family == AF_INET) {
-		struct sockaddr_in *si4 = (void*)sap;
-
-		if (si4 == NULL) break; /* Bad param! */
 
 		/* pick the whole sockaddr package! */
-		memcpy(si4, sa, sizeof(*si4));
+		memcpy(sap, sa, sizeof(struct sockaddr_in));
 		i = 0; /* Found! */
 		break;
 	      }
 
 #if defined(AF_INET6) && defined(INET6)
 	      if (sa->sa_family == AF_INET6) {
-		struct sockaddr_in6 *si6 = (void*)sap;
-
-		if (si6 == NULL) break;
 
 		/* pick the whole sockaddr package! */
-		memcpy(si6, sa, sizeof(*si6));
+		memcpy(sap, sa, sizeof(struct sockaddr_in6));
 		i = 0;
 		break;
 	      }
