@@ -23,8 +23,8 @@
 #endif
 #include "libc.h"
 
-int
-getmyhostname(namebuf, len)
+static int
+_getmyhostname(namebuf, len)
 	char *namebuf;
 	int len;
 {
@@ -62,7 +62,7 @@ getmyhostname(namebuf, len)
 	    break; /* We try it again if we fail here with EAI_AGAIN */
 	}
 	if (ai != NULL) {
-	  if (ai->ai_canonname != NULL)
+	  if (ai->ai_canonname)
 	    strncpy(namebuf, ai->ai_canonname, len);
 	  namebuf[len-1] = 0;
 	  freeaddrinfo(ai);
@@ -131,3 +131,25 @@ getmyhostname(namebuf, len)
 	return 0;
 #endif
 }
+
+
+int
+getmyhostname(namebuf, len)
+	char *namebuf; 
+	int len;
+{
+	int rc;
+	static char *savedname=NULL;
+
+	if (savedname) {
+		strncpy(namebuf, savedname, len);
+		namebuf[len - 1] = 0;
+		return 0;
+	}
+	rc = _getmyhostname(namebuf, len);
+	if (rc == 0) {
+		savedname=(char*)malloc(strlen(namebuf)+1);
+		strcpy(savedname, namebuf);
+	}
+	return rc;
+}
