@@ -2,7 +2,10 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	SFIO version by Matti Aarnio, copyright 1999-2000
+ *	Some modifications  by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ *
+ *	SFIO version by Matti Aarnio, copyright 1999-2003
  */
 
 /*LINTLIBRARY*/
@@ -23,8 +26,12 @@
 #endif
 #include <sys/file.h>
 #include <sys/socket.h>
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 
 #include "mail.h"
+#include "zmsignal.h"
 
 #include "listutils.h"
 #include "libc.h"
@@ -95,7 +102,8 @@ eqrename(from, to)
 
 #else	/* !HAVE_RENAME */
 	
-	if ((unlink(to) < 0 && errno != ENOENT) || (link(from, to) < 0)) {
+	if ((unlink(to) < 0 && errno != ENOENT) ||
+	    (link(from, to) < 0)) {
 	  return -1;
 	}
 
@@ -119,11 +127,12 @@ eqrename(from, to)
 extern int mail_priority;
 
 /*
- * Makes a temporary file under the postoffice, based on a file name template.
- * The last '%' character of the file name passed will be overwritten with
- * different suffix characters until the open() succeeds or we have exhausted
- * the search space.  Note: a single process cannot hold more than number-of-
- * suffix-characters message files at once.
+ * Makes a temporary file under the postoffice, based on a file name
+ * template.  The last '%' character of the file name passed will be
+ * overwritten with different suffix characters until the open()
+ * succeeds or we have exhausted the search space.  Note: a single
+ * process cannot hold more than number-of-suffix-characters message
+ * files at once.
  */
 
 Sfio_t *
@@ -135,8 +144,10 @@ _sfmail_fopen(filenamep)
 	Sfio_t *fp;
 	int fd, eno;
 
-	if (postoffice == NULL && (postoffice = getzenv("POSTOFFICE")) == NULL)
+	if (postoffice == NULL &&
+	    (postoffice = getzenv("POSTOFFICE")) == NULL)
 	  postoffice = POSTOFFICE;
+
 	path = mail_alloc(strlen(postoffice)+strlen(*filenamep)+2);
 	sprintf(path, "%s/%s", postoffice, *filenamep);
 	for (cp = *filenamep; *cp != '\0' && *cp != '%'; ++cp)
@@ -150,7 +161,7 @@ _sfmail_fopen(filenamep)
 	eno = 0;
 	for (suffix = SUFFIXCHARS; *suffix != 0; ++suffix) {
 		if (cp == NULL)
-			sleep(2);	/* hope something happens meanwhile */
+			sleep(2); /* hope something happens meanwhile */
 		else if (*suffix != ' ') {
 			*cp = *suffix;
 			strcpy(cp+1, post);
@@ -178,9 +189,9 @@ _sfmail_fopen(filenamep)
 
 /*
  * Link from-file to a file given by the to-file template.
- * The last '%' character of the to-file name passed will be overwritten with
- * different suffix characters until the link() succeeds or we have exhausted
- * the search space.
+ * The last '%' character of the to-file name passed will be overwritten
+ * with different suffix characters until the link() succeeds or we have
+ * exhausted the search space.
  */
 
 int
@@ -192,8 +203,10 @@ sfmail_link(from, tonamep)
 	const char *suffix, *post;
 	int eno;
 
-	if (postoffice == NULL && (postoffice = getzenv("POSTOFFICE")) == NULL)
+	if (postoffice == NULL &&
+	    (postoffice = getzenv("POSTOFFICE")) == NULL)
 		postoffice = POSTOFFICE;
+
 	path = mail_alloc(strlen(postoffice)+strlen(*tonamep)+2);
 	sprintf(path, "%s/%s", postoffice, *tonamep);
 	for (cp = *tonamep; *cp != '\0' && *cp != '%'; ++cp)
@@ -225,8 +238,8 @@ sfmail_link(from, tonamep)
 }
 
 /*
- * Open a message file of the specified type and initialize generic envelope
- * information (i.e. the file position on return may not be 0).
+ * Open a message file of the specified type and initialize generic
+ * envelope information (i.e. the file position on return may not be 0).
  */
 
 Sfio_t *
@@ -248,30 +261,32 @@ sfmail_open(type)
 	cp = (host == NULL) ? "I" : host ;
 	scratch = mail_alloc(strlen(PUBLICDIR)+strlen(cp)+3+1+10);
 
-	sprintf(scratch, "%s/%7s:%d%%", PUBLICDIR, cp, (int)getpid());
+	sprintf(scratch, "%s/%.19s:%d%%", PUBLICDIR, cp, (int)getpid());
 
 	fp = _sfmail_fopen(&scratch);
 	if (fp == NULL) {
-		eno = errno;
-		fprintf(stderr, "sfmail_fopen(\"%s\", \"w+\"): errno %d\n",
-			scratch, errno);
-		mail_free(scratch);
-		errno = eno;
-		return NULL;
+	  eno = errno;
+	  fprintf(stderr, "sfmail_fopen(\"%s\", \"w+\"): errno %d\n",
+		  scratch, errno);
+	  mail_free(scratch);
+	  errno = eno;
+	  return NULL;
 	}
 
-	/* Determine a unique id associated with the file (inode number) */
+	/* Determine a unique id associated with the file
+	   (inode number) */
 
 	fn = sffileno(fp);
 	if (fstat(fn, &stbuf) < 0) {
-		eno = errno;
-		fprintf(stderr, "fstat(\"%s\"): errno %d\n", scratch, errno);
-		mail_free(scratch);
-		errno = eno;
-		return NULL;
+	  eno = errno;
+	  fprintf(stderr, "fstat(\"%s\"): errno %d\n", scratch, errno);
+	  mail_free(scratch);
+	  errno = eno;
+	  return NULL;
 	}
 
-	/* Rename the scratch file to the message file name based on the id */
+	/* Rename the scratch file to the message file name
+	   based on the id */
 
 	if (type == NULL)
 		type = MSG_RFC822;
@@ -281,12 +296,14 @@ sfmail_open(type)
 #endif
 
 	/* Extend when need! */
-
+	
 	if (fn >= mail_nfiles) {
 	  int nfile = fn+1;
 	  if (mail_file == NULL) {
-	    mail_file = (char**)mail_alloc((u_int)(sizeof(char*) * nfile));
-	    mail_type = (char**)mail_alloc((u_int)(sizeof(char*) * nfile));
+	    mail_file = (char**)mail_alloc((u_int)(sizeof(char*) *
+						   nfile));
+	    mail_type = (char**)mail_alloc((u_int)(sizeof(char*) *
+						   nfile));
 	  } else {
 	    mail_file = (char**)mail_realloc((char*)mail_file,
 					     (sizeof(char*) * nfile));
@@ -303,19 +320,21 @@ sfmail_open(type)
 	mail_file[fn] = scratch;
 	mail_type[fn] = strdup(type);
 
-	/* Grab preferences from the environment to initialize the envelope */
+	/* Grab preferences from the environment to initialize
+	   the envelope */
 
 #ifndef	notype
 	if (type != NULL && *type != '\0')
-		sfprintf(fp, "type %s\n", type);
+	  sfprintf(fp, "type %s\n", type);
 #endif
 	cp = getenv("FULLNAME");
 	if (cp != NULL)
-		sfprintf(fp, "fullname %s\n",
-			fullname(cp, namebuf, sizeof namebuf, (char *)NULL));
+	  sfprintf(fp, "fullname %s\n",
+		   fullname(cp, namebuf, sizeof namebuf, (char *)NULL));
 	cp = getenv("PRETTYLOGIN");
 	if (cp != NULL)
-		sfprintf(fp, "loginname %s\n", cp);
+	  sfprintf(fp, "loginname %s\n", cp);
+
 	/*
 	 * If the postoffice lives elsewhere, put our hostname
 	 * in the Received-from header, to aid in message tracing.
@@ -389,43 +408,36 @@ sfmail_abort(fp)
 }
 
 /*
- * Close the message file on the indicated stream and submit it to the mailer.
+ * Close the message file on the indicated stream and submit it to
+ * the mailer.
  */
 
 int sfmail_close(fp)
 	Sfio_t *fp;
 {
-	int ino;
-	time_t mtime;
-
-	return _sfmail_close_(fp, &ino, &mtime);
+	return _sfmail_close_(fp, NULL, NULL);
 }
 
 
 static int routersubdirhash = -1;
 
-int
-_sfmail_close_(fp,inop, mtimep)
+static int
+_sfmail_close__(fp,inop, mtimep, async)
 	Sfio_t *fp;
-	int *inop;
+	long *inop;
 	time_t *mtimep;
+	int async;
 {
 	char *message, *nmessage, *type, *ftype;
 	const char *routerdir;
+	const char *inputdirs;
 	char *s = NULL;
 	struct stat stb;
+	int disable_routerdirhash = 0;
 	int fn;
 	long ino;
 	char subdirhash[6];
 
-	if (routersubdirhash < 0) {
-	  const char *ss = getzenv("ROUTERDIRHASH");
-	  if (ss && *ss == '1')
-	    routersubdirhash = 1;
-	  else
-	    routersubdirhash = 0;
-	}
-
 	if (postoffice == NULL) {
 		fprintf(stderr, "mail_close: called out of order!\n");
 		errno = EINVAL;
@@ -456,11 +468,6 @@ _sfmail_close_(fp,inop, mtimep)
 	}
 	ino = stb.st_ino;
 
-	if (routersubdirhash > 0) {
-	  sprintf(subdirhash, "%c/", (int)('A' + (ino % 26)));
-	} else
-	  *subdirhash = 0;
-
 	/*
 	 * *** NFS users beware ***
 	 * the fsync() between fflush() and fclose() may be mandatory
@@ -469,20 +476,24 @@ _sfmail_close_(fp,inop, mtimep)
 	 */
 
 
-	if (sfsync(fp) != 0) {
+	while (sfsync(fp) != 0) {
+	  if (errno == EINTR || errno == EAGAIN)
+	    continue;
 	  mail_free(message);
 	  if (ftype) mail_free(ftype);
 	  errno = EIO;
 	  return -1;
 	}
 #ifdef HAVE_FSYNC
-	while (fsync(fn) < 0) {
-	  if (errno == EINTR || errno == EAGAIN)
-	    continue;
-	  if (ftype) mail_free(ftype);
-	  mail_free(message);
-	  errno = EIO;
-	  return -1;
+	if (!async) {
+	  while (fsync(fn) < 0) {
+	    if (errno == EINTR || errno == EAGAIN)
+	      continue;
+	    if (ftype) mail_free(ftype);
+	    mail_free(message);
+	    errno = EIO;
+	    return -1;
+	  }
 	}
 #endif
 	if (sfclose(fp) == EOF) {
@@ -492,24 +503,76 @@ _sfmail_close_(fp,inop, mtimep)
 	  return -1;
 	}
 
-	routerdir = ROUTERDIR;
+	inputdirs = getzenv("INPUTDIRS");
+
+	routerdir = NULL;
 	nmessage  = NULL;
 	s         = NULL;
-	if (mail_priority) {
+
+	if (inputdirs) {
+	  int i = mail_priority;
+	  const char *rd = inputdirs;
+	  const char *ord = NULL;
+#ifdef HAVE_ALLOCA
+	  nmessage = alloca(strlen(postoffice)+strlen(inputdirs)+3+
+			    9+4+strlen(type));
+#else
+	  nmessage = mail_realloc(nmessage,
+				  strlen(postoffice)+strlen(inputdirs)+3+
+				  9+4+strlen(type));
+#endif
+	  /* There are some defined!   A ":" separated list of strings */
+
+	  /* mail_priority == 1 pics first, 2 pics second, ..
+	     if segments run out, last one is kept at  rd     */
+
+	  while (i-- && (s = strchr(rd,':'))) {
+	    *s = 0;
+	    sprintf(nmessage, "%s/%s", postoffice, rd);
+	    *s = ':';
+	    if ((stat(nmessage,&stb) < 0) || !S_ISDIR(stb.st_mode)) {
+	      rd = s+1;
+	      continue;	/* Not ok -- not a dir, for example */
+	    }
+	    ord = rd;
+	    rd = s+1;
+	  }
+	  
+	  /* Here we are when there is only one entry in the inputdirs:*/
+	  if (s == NULL && i > 0 && *rd != 0) {
+	    if (s) *s = 0;
+	    sprintf(nmessage, "%s/%s", postoffice, rd);
+	    if (s) *s = ':';
+	    /* Is it a valid directory ? */
+	    if ((stat(nmessage,&stb) == 0) && S_ISDIR(stb.st_mode))
+	      ord = rd; /* IT IS ! */
+	  }
+	  routerdir = ord;
+	  if (ord) disable_routerdirhash = 1;
+	}
+
+
+	if (!routerdir && !mail_priority)
+	  routerdir = ROUTERDIR;
+
+	if (mail_priority && !routerdir) {
 	  /* We are asked to place the mail somewhere else */
 	  const char *routerdirs = getzenv("ROUTERDIRS");
+	  routerdir = ROUTERDIR;
 	  if (routerdirs) {
 	    int i = mail_priority;
 	    const char *rd = routerdirs;
 	    const char *ord = routerdir;
 #ifdef HAVE_ALLOCA
-	    nmessage = alloca(strlen(postoffice)+strlen(routerdirs)+3+
-			      9+4+strlen(type));
+	    nmessage = alloca(strlen(postoffice)+strlen(routerdirs)+
+			      3+9+4+strlen(type));
 #else
-	    nmessage = mail_alloc(strlen(postoffice)+strlen(routerdirs)+3+
-				  9+4+strlen(type));
+	    nmessage = mail_realloc(nmessage,
+				    strlen(postoffice)+strlen(routerdirs)+
+				    3+9+4+strlen(type));
 #endif
-	    /* There are some defined!   A ":" separated list of strings */
+	    /* There are some defined!
+	       A ":" separated list of strings */
 
 	    /* mail_priority == 1 pics first, 2 pics second, ..
 	       if segments run out, last one is kept at  rd     */
@@ -526,7 +589,8 @@ _sfmail_close_(fp,inop, mtimep)
 	      rd = s+1;
 	    }
 
-	    /* Here we are when there is only one entry in the routerdirs: */
+	    /* Here we are when there is only one entry in
+	       the routerdirs: */
 	    if (s == NULL && i > 0 && *rd != 0) {
 	      if (s) *s = 0;
 	      sprintf(nmessage, "%s/%s", postoffice, rd);
@@ -538,14 +602,34 @@ _sfmail_close_(fp,inop, mtimep)
 	    routerdir = ord;
 	  }
 	}
+
+	if (routersubdirhash < 0) {
+	  const char *ss;
+	  if (disable_routerdirhash)
+	    ss = getzenv("INPUTDIRHASH");
+	  else
+	    ss = getzenv("ROUTERDIRHASH");
+
+	  if (ss && *ss == '1')
+	    routersubdirhash = 1;
+	  else
+	    routersubdirhash = 0;
+	}
+
+	if (routersubdirhash > 0) {
+	  sprintf(subdirhash, "%c/", (int)('A' + (ino % 26)));
+	} else
+	  *subdirhash = 0;
+
 	/* Assert postoffice != NULL */
 	if (nmessage == NULL) {
 #ifdef HAVE_ALLOCA
 	  nmessage = alloca(strlen(postoffice)+strlen(routerdir)+
 			    9+4+2+1+strlen(type));
 #else
-	  nmessage = mail_alloc(strlen(postoffice)+strlen(routerdir)+
-				9+4+2+1+strlen(type));
+	  nmessage = mail_realloc(nmessage,
+				  strlen(postoffice)+strlen(routerdir)+
+				  9+4+2+1+strlen(type));
 #endif
 	  sprintf(nmessage, "%s/%s/%s%ld%s", postoffice, routerdir,
 		  subdirhash, ino ,type);
@@ -572,6 +656,67 @@ _sfmail_close_(fp,inop, mtimep)
 
 	stat(nmessage, &stb);
 
+#ifdef AF_UNIX
+	do {
+
+	  const char *routernotify;
+	  char buf[1000];
+	  int notifysocket;
+	  struct sockaddr_un sad;
+#ifndef MSG_NOSIGNAL
+	  RETSIGTYPE (*oldsig)__((int));
+#endif
+
+	  if (disable_routerdirhash)
+	    routernotify = getzenv("INPUTNOTIFY");
+	  else
+	    routernotify = getzenv("ROUTERNOTIFY");
+
+	  if (!routernotify) break;
+
+	  memset(&sad, 0, sizeof(sad));
+	  sad.sun_family = AF_UNIX;
+	  strncpy(sad.sun_path, routernotify, sizeof(sad.sun_path));
+	  sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
+
+	  notifysocket = socket(PF_UNIX, SOCK_DGRAM, 0);
+	  if (notifysocket < 0) {
+	    perror("notifysocket: socket(PF_UNIX)");
+	    break;
+	  }
+
+	  fcntl(notifysocket, F_SETFL, O_NONBLOCK);
+
+	  s = strchr(routerdir,':');
+	  if (s) *s = 0;
+
+	  sprintf(buf, "NEW %s %s%ld%s",
+		  routerdir, subdirhash, ino, type);
+
+	  if (s) *s = ':';
+
+#ifndef MSG_NOSIGNAL
+	  SIGNAL_HANDLESAVE(SIGPIPE, SIG_IGN, oldsig);
+#endif
+
+	  sendto(notifysocket, buf, strlen(buf),
+#ifdef MSG_NOSIGNAL
+		 MSG_NOSIGNAL|
+#endif
+#ifdef MSG_DONTWAIT
+		 MSG_DONTWAIT
+#endif
+		 ,
+		 (struct sockaddr *)&sad, sizeof(sad));
+
+	  close(notifysocket);
+#ifndef MSG_NOSIGNAL
+	  SIGNAL_HANDLE(SIGPIPE, oldsig);
+#endif
+
+	} while (0);
+#endif /* AF_UNIX */
+
 #ifndef HAVE_ALLOCA
 	mail_free(nmessage);
 #endif
@@ -579,23 +724,49 @@ _sfmail_close_(fp,inop, mtimep)
 	if (ftype) mail_free(ftype);
 
 	if (inop != NULL)
-	  *inop   = (int)    stb.st_ino;
+	  *inop   = (long) stb.st_ino;
 	if (mtimep != NULL)
 	  *mtimep = (time_t) stb.st_mtime;
 
+
 	return 0;
 }
 
+
+
+int
+_sfmail_close_(fp,inop, mtimep)
+	Sfio_t *fp;
+	long *inop;
+	time_t *mtimep;
+{
+	return _sfmail_close__(fp, inop, mtimep, 0);
+}
+
+ int
+_sfmail_close_async(fp,inop, mtimep, async)
+	Sfio_t *fp;
+	long *inop;
+	time_t *mtimep;
+	int async;
+{
+	return _sfmail_close__(fp, inop, mtimep, async);
+}
+
+
+
+
 /*
  * Close the message file on the indicated stream, and submit
  * it to alternate directory. (For smtpserver->scheduler messages,
  * for example.)
  */
 
-int
-sfmail_close_alternate(fp,where,suffix)
+static int
+sfmail_close_alternate_(fp,where,suffix,async)
 	Sfio_t *fp;
 	const char *where, *suffix;
+	int async;
 {
 	char *message, *nmessage, *msgbase;
 	char *type, *ftype;
@@ -603,18 +774,20 @@ sfmail_close_alternate(fp,where,suffix)
 	int fn;
 
 	if (postoffice == NULL) {
-		fprintf(stderr, "sfmail_close_alternate: called out of order!\n");
-		errno = EINVAL;
-		return -1;
+	  fprintf(stderr,
+		  "sfmail_close_alternate: called out of order!\n");
+	  errno = EINVAL;
+	  return -1;
 	}
 	if (fp == NULL) {
-		errno = EBADF;
-		return -1;
+	  errno = EBADF;
+	  return -1;
 	}
 	fn = sffileno(fp);
 	fstat(fn, &stbuf);
 	if (fn >= mail_nfiles)
-		abort(); /* Usage error -- no such fileno in our use! */
+	  abort(); /* Usage error -- no such fileno in our use! */
+
 	message = mail_file[fn];
 	if (message == NULL) {
 		errno = ENOENT;
@@ -635,20 +808,24 @@ sfmail_close_alternate(fp,where,suffix)
 	 */
 
 
-	if (sfsync(fp) != 0) {
+	while (sfsync(fp) != 0) {
+	  if (errno == EINTR || errno == EAGAIN)
+	    continue;
 	  mail_free(message);
 	  if (ftype) mail_free(ftype);
 	  errno = EIO;
 	  return -1;
 	}
 #ifdef HAVE_FSYNC
-	while (fsync(fn) < 0) {
-	  if (errno == EINTR || errno == EAGAIN)
-	    continue;
-	  if (ftype) mail_free(ftype);
-	  mail_free(message);
-	  errno = EIO;
-	  return -1;
+	if (!async) {
+	  while (fsync(fn) < 0) {
+	    if (errno == EINTR || errno == EAGAIN)
+	      continue;
+	    if (ftype) mail_free(ftype);
+	    mail_free(message);
+	    errno = EIO;
+	    return -1;
+	  }
 	}
 #endif
 	if (sfclose(fp) == EOF) {
@@ -665,10 +842,11 @@ sfmail_close_alternate(fp,where,suffix)
 	else
 		++msgbase;
 
-	nmessage  = NULL;
 	/* Assert postoffice != NULL */
+
 	nmessage = mail_alloc(strlen(postoffice)+1+strlen(where)+1+
 			      20+strlen(suffix)+1+strlen(type));
+
 	sprintf(nmessage, "%s/%s/%ld%s%s",
 		postoffice, where, (long)stbuf.st_ino, suffix, type);
 
@@ -688,3 +866,22 @@ sfmail_close_alternate(fp,where,suffix)
 	if (ftype) mail_free(ftype);
 	return 0;
 }
+
+int
+sfmail_close_alternate(fp,where,suffix)
+	Sfio_t *fp;
+	const char *where, *suffix;
+{
+	return sfmail_close_alternate_(fp, where, suffix, 0);
+}
+
+
+int
+sfmail_close_alternate_async(fp,where,suffix,async)
+	Sfio_t *fp;
+	const char *where, *suffix;
+	int async;
+{
+	return sfmail_close_alternate_(fp, where, suffix, async);
+}
+
