@@ -1,5 +1,5 @@
 /*  Author: Mark Moraes <moraes@csri.toronto.edu> */
-/* $Id: malloc.h,v 1.16 1993/05/23 03:38:27 moraes Exp $ */
+/* $Id: malloc.h,v 1.2 1999/12/29 19:37:35 mea Exp $ */
 #ifndef __XMALLOC_H__
 #define __XMALLOC_H__
 
@@ -39,7 +39,7 @@ extern univptr_t __memalign __proto((size_t, size_t, const char *, int));
 extern univptr_t __emalloc __proto((size_t, const char *, int));
 extern univptr_t __ecalloc __proto((size_t, size_t, const char *, int));
 extern univptr_t __erealloc __proto((univptr_t, size_t, const char *, int));
-extern char *__strdup __proto((const char *, const char *, int));
+extern char *___strdup __proto((const char *, const char *, int));
 extern char *__strsave __proto((const char *, const char *, int));
 extern void __free __proto((univptr_t, const char *, int));
 extern void __cfree __proto((univptr_t, const char *, int));
@@ -52,7 +52,7 @@ extern void __cfree __proto((univptr_t, const char *, int));
 #define emalloc(x)		__emalloc((x), __FILE__, __LINE__)
 #define ecalloc(x, n)		__ecalloc((x), (n), __FILE__, __LINE__)
 #define erealloc(p, x)		__erealloc((p), (x), __FILE__, __LINE__)
-#define strdup(p)		__strdup((p), __FILE__, __LINE__)
+#define strdup(p)		___strdup((p), __FILE__, __LINE__)
 #define strsave(p)		__strsave((p), __FILE__, __LINE__)
 /* cfree and free are identical */
 #define cfree(p)		__free((p), __FILE__, __LINE__)
