@@ -1,6 +1,7 @@
 /*
  *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Some functions Copyright 1991-2003 Matti Aarnio.
  */
 
 #include "hostenv.h"
@@ -14,6 +15,7 @@
 #include "interpret.h"
 #include "io.h"		/* redefines stdio routines */
 #include "shconfig.h"
+#include "sysexits.h"
 
 #include "libz.h"
 #include "libc.h"
@@ -45,6 +47,7 @@ extern int wait();
 #define	HZ	60
 #endif	/* HZ */
 
+#include <errno.h>
 extern int errno;
 
 #define CSARGS2 __((conscell *, conscell *))
@@ -59,6 +62,7 @@ static conscell *sh_length	CSARGS2;
 static conscell *sh_last	CSARGS2;
 static conscell *sh_lappend     CSARGS2;
 static conscell *sh_lreplace    CSARGS2;
+static conscell *sh_ifssplit    CSARGS2;
 extern conscell *sh_glob	CSARGS2; /* expand.c */
 
 #define CSARGV2 __((int argc, const char *argv[]))
@@ -95,6 +99,7 @@ struct shCmd builtins[] = {
 {	"elements",	NULL,	sh_elements,	NULL,	SH_ARGV		},
 {	"get",		NULL,	sh_get,		NULL,	SH_ARGV		},
 {	"length",	NULL,	sh_length,	NULL,	SH_ARGV		},
+{	"ifssplit",	NULL,	sh_ifssplit,	NULL,	SH_ARGV		},
 {	"[",		sh_test,	NULL,	NULL,	0		},
 {	"test",		sh_test,	NULL,	NULL,	0		},
 {	"echo",		sh_echo,	NULL,	NULL,	0		},
@@ -252,7 +257,7 @@ sh_lappend(avl, il)
 	stickymem = omem;
 	return NULL; /* Be quiet, don't force the caller
 			to store the result into heap just
-		        for latter discard... */
+		        for later discard... */
 }
 
 /*
@@ -370,7 +375,7 @@ sh_lreplace(avl, il)
 	if (d) {
 	  cdr(data) = cdr(d);
 	  cdr(d) = NULL;	/* Disconnect and discard old data */
-	  /* s_free_tree(d); -- GC cleans it out latter .. */
+	  /* s_free_tree(d); -- GC cleans it out later .. */
 	}
 
 	UNGCPRO4;
@@ -378,7 +383,7 @@ sh_lreplace(avl, il)
 	stickymem = omem;
 	return NULL; /* Be quiet, don't force the caller
 			to store the result into heap just
-		        for latter discard... */
+		        for later discard... */
 }
 
 
@@ -451,6 +456,7 @@ sh_get(avl, il)
 	return d;
 }
 
+/* This is NUMBER OF ELEMENTS in a chain! */
 static conscell *
 sh_length(avl, il)
 	conscell *avl, *il;
@@ -458,7 +464,8 @@ sh_length(avl, il)
 	char buf[10];
 	int len = 0;
 
-	if ((il = cdar(avl)) && LIST(il)) {
+	il = cdar(avl);
+	if (il && LIST(il)) {
 		for (il = car(il); il != NULL; il = cdr(il))
 			++len;
 	}
@@ -469,38 +476,70 @@ sh_length(avl, il)
 	return avl;
 }
 
+/* This splits incoming string at WHITESPACEs to a list of strings */
+
+static conscell *
+sh_ifssplit(avl, il)
+	conscell *avl, *il;
+{
+	int len, l;
+	conscell *tmp, *reply, *t;
+	const char *p, *p0;
+	GCVARS1;
+
+	if (ifs == NULL)
+		ifs_flush();
+
+	il = cdar(avl);
+	/* 'il' is supposedly a single string */
+
+	if (!il || !STRING(il)) return il;
+	p0  = il->cstring;
+	len = il->slen;
+	if (!p0) return il;
+
+
+	reply = t = NULL;
+	GCPRO1(reply);
+
+	for (;len > 0; ++p0, --len) {
+
+	  if (WHITESPACE(*p0)) continue;
+	  p = p0;
+	  while (len > 0 && !WHITESPACE(*p)) ++p, --len;
+
+	  l = p - p0; /* Will always be at least 1 ! */
+
+	  tmp = newstring(dupnstr(p0, l), l);
+	  tmp->flags |= QUOTEDSTRING|ELEMENT;
+
+	  p0 = p;
+
+	  if (!t)
+	    reply = t = tmp;
+	  else {
+	    cdr(t) = tmp;
+	    t = tmp;
+	  }
+
+	}
+
+	UNGCPRO1;
+	return reply;
+}
+
 /* returns -- return ALWAYS a string [mea] */
 conscell *
 sh_returns(avl, il, statusp)
 	conscell *avl, *il;
 	int *statusp;
 {
-	int n;
-	char *cp;
-
 	if ((il = cdar(avl)) && LIST(il) && cdr(il) == NULL)
 		return il;
 	else if (il == NULL)
 		return NULL;
-	else if (STRING(il)) {
-		for (cp = il->string; *cp != 0; ++cp) {
-		  int c = (*cp) & 0xFF;
-		  if (!isascii(c) || !isdigit(c))
-		    break;
-		}
-		if (*cp)
-			return il;
-		n = atoi(il->string);
-		if (n < 0) {
-			fprintf(stderr, "%s: %s: %d\n", car(avl)->string,
-					NEGATIVE_VALUE, n);
-			n = 1;
-		}
-		*statusp = n;
-		return NULL;
-	}
-	/* NOTREACHED */
-	return NULL;
+	else
+		return il;
 }
 
 conscell *
@@ -508,27 +547,32 @@ sh_return(avl, il, statusp)
 	conscell *avl, *il;
 	int *statusp;
 {
-	int n;
-	char *cp;
+	int n, l;
+	const char *cp;
 
 	if ((il = cdar(avl)) && LIST(il) && cdr(il) == NULL)
 		return il;
 	else if (il == NULL)
 		return NULL;
 	else if (STRING(il)) {
-		for (cp = il->string; *cp != 0; ++cp) {
+		cp = il->cstring;
+		l  = il->slen;
+		n  = 0;
+		for (;l > 0; ++cp, --l) {
 		  int c = (*cp) & 0xFF;
 		  if (!isascii(c) || !isdigit(c))
 		    break;
+		  n = n * 10 + c - '0';
 		}
-		if (*cp != 0)
+		if (l > 0) /* all not consumed! */
 			return il;
-		n = atoi(il->string);
+#if 0 /* can't happen... */
 		if (n < 0) {
 			fprintf(stderr, "%s: %s: %d\n", car(avl)->string,
 					NEGATIVE_VALUE, n);
 			n = 1;
 		}
+#endif
 		*statusp = n;
 		return NULL;
 	}
@@ -609,7 +653,7 @@ sh_cd(argc, argv)
 		dir = argv[1];
 	} else {
 		fprintf(stderr, USAGE_CD, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	if (!(dir[0] == '/' ||
 	      (dir[0] == '.' &&
@@ -632,7 +676,7 @@ sh_cd(argc, argv)
 		return 0;
 
 	fprintf(stderr, "%s: %s: %s\n", argv[0], dir, strerror(errno));
-	return 1;
+	return 2;
 }
 
 static int
@@ -655,103 +699,139 @@ sh_hash(argc, argv)
 	return 0;
 }
 
+
+/*
+
+The "man sh" tells us:
+
+     read name ...
+          One line is read from the standard input and, using the
+          internal  field separator, IFS (normally space or tab),
+          to delimit word boundaries, the first word is  assigned
+          to  the first name, the second word to the second name,
+          and so forth, with leftover words assigned to the  last
+          name.   Lines can be continued using \newline.  Charac-
+          ters other than newline can be quoted by preceding them
+          with a backslash.  These backslashes are removed before
+          words are assigned to names, and no  interpretation  is
+          done  on the character that follows the backslash.  The
+          return code is 0, unless an EOF is encountered.
+
+We add a note:
+
+	  We ignore backslash quoting completely, no line continuations!
+
+	  If there are fewer input "words" than parameter names,
+	  rest of the names will be assigned empty strings.
+
+This is tested with ZMSH script:
+
+  oldIFS=$IFS;IFS=":";while read user pass rest;do echo "u: $user";echo "p: $pass";echo "r: $rest";done < /etc/passwd;IFS=$oldIFS
+
+which shows:
+
+  u: root
+  p: x
+  r: 0:0:root:/root:/bin/bash
+...etc...
+
+*/
+
+/* Store a char into buffer, allocate it, if necessary. */
+
+static void sh_read_cput __((int, int *, char **, int *));
+static void sh_read_cput(c, ip, bufp, sizep)
+	int c, *ip, *sizep;
+	char **bufp;
+{
+	char *buf   = *bufp;
+	int bufsize = *sizep;
+	int i       = *ip;
+
+	if (i >= bufsize || !buf) {
+	  /* Need to grow ... */
+	  bufsize <<= 1;
+	  buf = erealloc(buf, bufsize+3);
+	}
+
+	buf[i]   = c;
+
+	buf[++i] = 0; /* Keep the string terminated.. */
+
+	*ip = i;
+	*sizep = bufsize;
+	*bufp  = buf;
+}
+
 static int
 sh_read(argc, argv)
 	int argc;
 	const char *argv[];
 {
-	static char *buf = NULL;
-	static u_int bufsize;
-	char *cp, *value = NULL, *bp;
-	int flag, offset, eoinp = 0;
+	static char *buf     = NULL;
+	static u_int bufsize = (BUFSIZ / 2) - 12;
+
+	int eoinp, i, flag;
 
 	if (argc == 1) {
-		fprintf(stderr, USAGE_READ, argv[0]);
-		return 1;
+	    fprintf(stderr, USAGE_READ, argv[0]);
+	    return EX_USAGE;
 	}
 	if (ifs == NULL)
-		ifs_flush();
+	    ifs_flush();
+
 	--argc, ++argv;
-	if (buf == NULL) {
-		bufsize = BUFSIZ - 24;
-		buf = emalloc(bufsize);
+
+	eoinp = 0;
+
+	for ( ; argc > 0; ) {
+
+	    int c;
+
+	    i = 0; /* Collect one variable full at the time into
+		      the buffer.  Reuse the buffer for next variable
+		      content. */
+
+	    flag = 0; /* 0 == SKIP IFS;  0 !=  ARGC rules */
+
+	    for (;;) {
+		c = fgetc(stdin);
+		if (c == EOF) break;
+		++eoinp;
+		if (c == '\r' || c == 0) continue; /* ignore those */
+		if (c == '\n') break;
+		if (!flag) {
+		    /* Skipping leading IFSs */
+		    if (WHITESPACE(c)) continue;
+		    /* Non IFS character! */
+		    flag = 1;
+		}
+		if (argc > 1) {
+		    /* Scan over all non-IFS chars - stop at IFS chars */
+		    if (WHITESPACE(c)) break;
+		} else {
+		    /* Scan until end of line (or EOF), as
+		       the last variable collects the tail... */
+		    ;
+		}
+		/* Store this character into the buffer */
+		sh_read_cput(c, &i, &buf, &bufsize);
+	    }
+
+	    /* Store collected value */
+	    if (argc > 0 && buf && i > 0) {
+		--argc;
+		v_set(*argv++, buf);
+	    }
+
+	    if ((c == EOF && i == 0) || c == '\n') break; /* Urgh.. EOF */
+
 	}
-	flag = 0;
-	bp = NULL;
-	buf[0] = '\0';
-	while (argc > 0) {
-		if (fgets(buf, bufsize, stdin) == NULL) {
-		  eoinp = 1;
-		  break;
-		}
-		for (cp = buf; argc > 0 && *cp != '\0'; ) {
-			while (*cp != '\0' && WHITESPACE((unsigned)*cp))
-				++cp;
-			if (*cp == '\0') {
-				if (cp > buf && *(cp-1) != '\n' &&
-				    (offset = cp - buf)		&&
-				    (bufsize = 2*bufsize)	&&
-				    (buf = erealloc(buf, bufsize)) &&
-				    fgets(cp, bufsize/2, stdin) != NULL) {
-					cp = buf + offset;
-					continue;
-				} else
-					break;
-			}
-			if (!flag) {
-				bp = cp;
-				value = cp;
-			}
-			if (argc == 1)
-				flag = 1;
-			while (*cp == '\0' || !WHITESPACE((unsigned)*cp)) {
-				if (*cp == '\0') {
-					if (cp > buf && *(cp-1) != '\n' &&
-					    (offset = cp - buf)		&&
-					    (bufsize = 2*bufsize)	&&
-					    (buf = erealloc(buf, bufsize)) &&
-					    fgets(cp, bufsize/2, stdin)!=NULL) {
-						cp = buf + offset;
-						continue;
-					} else
-						break;
-				}
-				if (*cp == '\\' && *(cp+1) != '\0') { /* bug */
-					if (*++cp == '\n') {
-						*cp++ = '\0'; /* defeat above */
-						continue;
-					}
-				}
-				*bp++ = *cp++;
-			}
-			if (argc > 1) {
-				--argc;
-				if (*cp == '\0') {
-					*bp = '\0';	/* bp might == cp */
-					v_set(*argv++, value);
-					bp = value;
-					break;
-				} else {
-					*bp = '\0';
-					v_set(*argv++, value);
-					++cp;
-				}
-			} else if (*cp++ != '\0')
-				*bp++ = ' ';
-		}
-		if (bp > value) {
-			if (*(bp-1) == ' ')
-				*--bp = '\0';
-			else
-				*bp = '\0';
-			v_set(*argv++, value), --argc;
-		}
-	}
-	/* if (buf[0] == '\0') return 1; */
-	if (eoinp) return 1;
-eoinput:
+
+	/* Left-over variables to be zeroed: */
 	while (argc-- > 0) v_set(*argv++, "");
-	return 0;
+
+	return (0 == eoinp);
 }
 
 int
@@ -767,9 +847,9 @@ sh_include(argc, argv)
 	char *dir, *buf, *path = NULL;
 	const char *rpath = NULL;
 
-	if (argc < 2 || argv[1][0] == '\0') {
+	if (argc != 2 || argv[1][0] == '\0') {
 		fprintf(stderr, USAGE_INCLUDE, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	fd = -1;
 	if (strchr(argv[1], '/') == NULL) {
@@ -810,7 +890,7 @@ sh_include(argc, argv)
 	if (status != (int) stbuf.st_size) {
 		perror("read");
 		close(fd);
-		return 1;
+		return 2;
 	}
 	close(fd);
 	buf[(u_int)stbuf.st_size] = '\0';
@@ -927,9 +1007,9 @@ sh_getopts(argc, argv)
 
 	if (argc < 3) {
 		fprintf(stderr, USAGE_GETOPTS, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
-	opterr = 0;
+	zopterr = 0;
 	--argc, ++argv;
 	optstring = argv[0];
 	--argc, ++argv;
@@ -946,14 +1026,14 @@ sh_getopts(argc, argv)
 			}
 		*av = NULL;
 	}
-	c = getopt(argc, (char*const*)argv,
-		   optstring);	/* this MUST be our getopt() */
-	sprintf(buf, "%d", optind);
+	c = zgetopt(argc, (char*const*)argv,
+		    optstring);	/* this MUST be our getopt() */
+	sprintf(buf, "%d", zoptind);
 	v_set(OPTIND, buf);
 	if (c == EOF)
 		return 1;
-	if (optarg != NULL)	/* our getopt() makes this reliable */
-		v_set(OPTARG, optarg);
+	if (zoptarg != NULL)	/* our zgetopt() makes this reliable */
+		v_set(OPTARG, zoptarg);
 	sprintf(buf, "%c", c);
 	v_set(name, buf);
 	return 0;
@@ -969,7 +1049,7 @@ sh_shift(argc, argv)
 
 	if (argc > 2 || (argc == 2 && !isdigit(argv[1][0]))) {
 		fprintf(stderr, USAGE_SHIFT, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	if (argc == 2) {
 		if ((n = atoi(argv[1])) < 1)
@@ -992,7 +1072,7 @@ sh_umask(argc, argv)
 
 	if (argc > 2 || (argc == 2 && !isdigit(argv[1][0]))) {
 		fprintf(stderr, USAGE_UMASK, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	if (argc == 1) {
 		mask = umask(077);
@@ -1124,7 +1204,7 @@ sh_unset(argc, argv)
 
 	if (argc == 1) {
 		fprintf(stderr, USAGE_UNSET, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	av0 = argv[0];
 	--argc, ++argv;
@@ -1170,7 +1250,7 @@ sh_wait(argc, argv)
 
 	if (argc > 2 || (argc == 2 && !isdigit(argv[1][0]))) {
 		fprintf(stderr, USAGE_WAIT, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	while ((pid = wait(&status)) > 0) {
 		if (argc == 2 && pid == atoi(argv[1]))
@@ -1198,7 +1278,7 @@ sh_times(argc, argv)
 
 	if (argc > 1) {
 		fprintf(stderr, USAGE_TIMES, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	if (times(&foo) < 0)
 		return 1;
@@ -1256,7 +1336,7 @@ sh_sleep(argc, argv)
 {
 	if (argc != 2 || atoi(argv[1]) <= 0) {
 		fprintf(stderr, USAGE_SLEEP, argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	sleep(atoi(argv[1]));
 	return 0;
