@@ -51,8 +51,11 @@ int std_feof(fp) FILE *fp; { return feof(fp); }
 
 long std_ftell(fp) FILE *fp; { return ftell(fp); }
 
+#if 0
 char *std_gets(s) char *s; { return gets(s); }
+#endif
 
+int std_fgetc(fp) FILE *fp; { return fgetc(fp); }
 char *std_fgets(s, n, fp) char *s; u_int n; FILE *fp; { return fgets(s, n, fp); }
 
 int std_puts(s) const char *s; { return puts(s); }
@@ -77,6 +80,7 @@ int std_fwrite(b,s,n,fp) const char *b; u_int s, n; FILE *fp;{ return fwrite(b,s
 
 /* functionally equivalent (to stdio) routines to deal with string buffers */
 
+#if 0
 char *
 siogets(s)
 	register char *s;
@@ -94,6 +98,22 @@ siogets(s)
 	*s = '\0';
 	return (s == os) ? NULL : os;
 }
+#endif
+
+int
+siofgetc(fp)
+	FILE *fp;
+{
+	register struct siobuf *siop = siofds[FILENO(fp)];
+
+	/* assert siop != NULL */
+	if (siop->sb_ptr + siop->sb_cnt < siop->sb_base + siop->sb_bufsiz) {
+	  register int ch = (int)*siop->sb_ptr++;
+	  if (ch == '\0')  return EOF;
+	  return ch;
+	}
+	return EOF;
+}
 
 char *
 siofgets(s, n, fp)
@@ -106,8 +126,10 @@ siofgets(s, n, fp)
 	char *os = s;
 
 	/* assert siop != NULL */
-	while (--n > 0 && siop->sb_ptr + siop->sb_cnt < siop->sb_base + siop->sb_bufsiz) {
-	  if ((ch = (int)*siop->sb_ptr++) == '\0')
+	while (--n > 0 &&
+	       siop->sb_ptr + siop->sb_cnt < siop->sb_base + siop->sb_bufsiz) {
+	  ch = (int)*siop->sb_ptr++;
+	  if (ch == '\0')
 	    break;
 	  *s++ = ch;
 	  if (ch == '\n')
@@ -371,7 +393,7 @@ register struct siobuf *siop;
 #endif
 #define	ARG() \
 	_ulong = flags&LONGINT ? va_arg(argp, long) : \
-	    flags&SHORTINT ? va_arg(argp, short) : va_arg(argp, int);
+	    flags&SHORTINT ? (short)va_arg(argp, int) : va_arg(argp, int);
 
 #define	todigit(c)	((c) - '0')
 #define	tochar(n)	((n) + '0')
@@ -387,6 +409,8 @@ register struct siobuf *siop;
 #define	ZEROPAD		0x20		/* zero (as opposed to blank) pad */
 #define	HEXPREFIX	0x40		/* add 0x or 0X prefix */
 
+#ifndef HAVE_MEMCHR
+
 static const char *mymemchr __((const char *s, int, int));
 static const char *
 mymemchr(s, c, n)
@@ -400,6 +424,12 @@ mymemchr(s, c, n)
 	return NULL;
 }
 
+#else
+
+#define mymemchr memchr
+
+#endif
+
 
 STATIC int
 vsiodoprnt(siop, fmt0, argp)
