@@ -1,7 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Some functions Copyright 1991-2000 Matti Aarnio.
+ *	Some functions Copyright 1991-2002 Matti Aarnio.
  */
 
 /*
@@ -21,8 +21,14 @@
 #include <pwd.h>			/* for run_homedir() */
 #include <grp.h>			/* for run_grpmems() */
 #include <errno.h>
+#ifdef HAVE_SYS_UN_H
+#include <sys/socket.h>
+#include <sys/un.h>
+#endif
 
+#include "zsyslog.h"
 #include "shmmib.h"
+#include "sysexits.h"
 
 extern struct MIB_MtaEntry *MIBMtaEntry;
 
@@ -41,6 +47,18 @@ extern struct MIB_MtaEntry *MIBMtaEntry;
 # endif /* HAVE_NDIR_H */
 #endif /* HAVE_DIRENT_H */
 
+#ifdef HAVE_DIRFD
+#   define Z_DIRFD(dir) dirfd(dir)
+#else
+# ifdef USE_DIRFD_DD_FD
+#   define Z_DIRFD(dir) ((dir).dd_fd)
+# else
+#  ifdef USE_DIRFD_D_FD
+#   define Z_DIRFD(dir) ((dir).dd_fd)
+#  endif
+# endif
+#endif
+
 
 #ifdef HAVE_SYS_RESOURCE_H
 #ifdef linux
@@ -49,28 +67,8 @@ extern struct MIB_MtaEntry *MIBMtaEntry;
 #include <sys/resource.h>
 #endif
 
-#ifdef  HAVE_WAITPID
-# include <sys/wait.h>
-#else
-# ifdef HAVE_WAIT3
-#  include <sys/wait.h> /* Has BSD wait3() */
-# else
-#  ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
-#   include <sys/wait.h>
-#  else /* Not POSIX.1 compatible, lets fake it.. */
-extern int wait();
-#  endif
-# endif
-#endif
-
-#ifndef WEXITSTATUS
-# define WEXITSTATUS(s) (((s) >> 8) & 0377)
-#endif
-#ifndef WSIGNALSTATUS
-# define WSIGNALSTATUS(s) ((s) & 0177)
-#endif
-
 #include "zmsignal.h"
+
 #include "zsyslog.h"
 #include "mail.h"
 #include "interpret.h"
@@ -80,15 +78,28 @@ extern int wait();
 
 #include "prototypes.h"
 
+#ifdef	HAVE_LOCKF
+#ifdef	F_OK
+#undef	F_OK
+#endif	/* F_OK */
+#ifdef	X_OK
+#undef	X_OK
+#endif	/* X_OK */
+#ifdef	W_OK
+#undef	W_OK
+#endif	/* W_OK */
+#ifdef	R_OK
+#undef	R_OK
+#endif	/* R_OK */
+#endif	/* HAVE_LOCKF */
+
+
 #ifndef	_IOFBF
 #define	_IOFBF	0
 #endif	/* !_IOFBF */
 
 
 extern const char *traps[];
-extern int nobody;
-extern struct group  *getgrnam __((const char *));
-extern struct passwd *getpwnam __((const char *));
 extern time_t time __((time_t *));
 extern int routerdirloops;
 
@@ -168,7 +179,7 @@ struct router_child {
   int   readsize, readout;
 
   int   statloc;
-#if defined(HAVE_WAIT3) && defined(HAVE_SYS_RESOURCE_H)
+#if defined(HAVE_SYS_RESOURCE_H)
   struct rusage r;
 #endif
 
@@ -181,11 +192,6 @@ struct router_child routerchilds[MAXROUTERCHILDS];
 extern int nrouters;
 extern const char *logfn;
 
-/* ../transports/libta/nonblocking.c */
-extern int  fd_nonblockingmode __((int fd));
-extern int  fd_blockingmode __((int fd));
-extern void fd_restoremode __((int fd, int mode));
-
 /* ../scheduler/pipes.c */
 extern int  pipes_create         __((int *tochild, int *fromchild));
 extern void pipes_close_parent   __((int *tochild, int *fromchild));
@@ -200,60 +206,130 @@ extern int  resources_query_pipesize __((int fildes));
 
 static void child_server __((int tofd, int frmfd));
 static int  rd_doit __((const char *filename, const char *dirs));
-static int  parent_reader __((void));
+static int  parent_reader __((int waittime));
+static void notify_reader __((int sock));
+
+
+static int notifysocket = -1;
+static time_t notifysocket_reinit = 1;
+
+
+static void notifysock_init()
+{
+	while (notifysocket < 0) {
+
+	  struct sockaddr_un sad;
+	  int on = 1;
+	  char *notifysock;
+	  int oldumask;
+
+	  notifysock = (char *)getzenv("ROUTERNOTIFY");
+	  if (!notifysock) {
+	    notifysocket_reinit = 0;
+	    return;
+	  }
+
+#ifdef  AF_UNIX
+
+	  memset(&sad, 0, sizeof(sad));
+	  sad.sun_family = AF_UNIX;
+	  strncpy(sad.sun_path, notifysock, sizeof(sad.sun_path));
+	  sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
+
+	  notifysocket = socket(PF_UNIX, SOCK_DGRAM, 0);
+	  if (notifysocket < 0) {
+	    perror("notifysocket: socket(PF_UNIX)");
+	    break;
+	  }
+
+	  setsockopt(notifysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+	  /* In case that one already exists.. */
+	  unlink(sad.sun_path);
+
+	  oldumask = umask(0555);
+
+	  if (bind(notifysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
+	    perror("bind:UNIX notify socket");
+	    umask(oldumask);
+	    close(notifysocket);
+	    notifysocket = -1;
+	    break;
+	  }
+	  umask(oldumask);
+
+	  fd_nonblockingmode(notifysocket);
+
+#if defined(F_SETFD)
+	  fcntl(notifysocket, F_SETFD, 1); /* close-on-exec */
+#endif
+#endif /* AF_UNIX */
+	  break;
+	}
+
+	if (notifysocket < 0) {
+	  time(&notifysocket_reinit);
+	  notifysocket_reinit += 10;
+	} else
+	  notifysocket_reinit = 0;
+}
 
 
 static int  start_child   __((int idx));
 static int  start_child (i)
      const int i;
 {
-  int pid;
-  int tofd[2], frmfd[2];
+	int pid;
+	int tofd[2], frmfd[2];
 
-  if (pipes_create(tofd, frmfd) < 0)
-    return -1; /* D'uh :-( */
+	if (pipes_create(tofd, frmfd) < 0)
+	  return -1; /* D'uh :-( */
 
-  pid = fork();
-  if (pid == 0) { /* child */
+	pid = fork();
+	if (pid == 0) { /* child */
 
-    int idx;
+	  int idx;
 
-    pipes_to_child_fds(tofd,frmfd);
-    for (idx = resources_query_nofiles(); idx >= 3; --idx)
-	close(idx);
+	  pipes_to_child_fds(tofd,frmfd);
+	  for (idx = resources_query_nofiles(); idx >= 3; --idx)
+	    close(idx);
 
 #if 0
-    resources_maximize_nofiles();
+	  resources_maximize_nofiles();
 #endif
 
-    zcloselog();
-    /* Each (sub-)process does openlog() all by themselves */
-    zopenlog("router", LOG_PID, LOG_MAIL);
+	  zcloselog();
+	  /* Each (sub-)process does openlog() all by themselves */
+	  zopenlog("router", LOG_PID, LOG_MAIL);
 
-    child_server(0, 1);
+	  child_server(0, 1);
 
-    exit(1);
+	  exit(0);
 
-  } else if (pid < 0) { /* fork failed - yell and forget it! */
-    close(tofd[0]);  close(tofd[1]);
-    close(frmfd[0]); close(frmfd[1]);
-    fprintf(stderr, "router: start_child(): Fork failed!\n");
-    return -1;
-  }
-  /* Parent */
+	} else if (pid < 0) { /* fork failed - yell and forget it! */
+	  close(tofd[0]);  close(tofd[1]);
+	  close(frmfd[0]); close(frmfd[1]);
+	  fprintf(stderr, "router: start_child(): Fork failed!\n");
+	  return -1;
+	}
+	/* Parent */
 
-  pipes_close_parent(tofd,frmfd);
+	MIBMtaEntry->rt.RouterProcesses     += 1;
+	MIBMtaEntry->rt.RouterProcessForks  += 1;
 
-  fd_nonblockingmode(tofd[1]);
-  fd_nonblockingmode(frmfd[0]);
+	pipes_close_parent(tofd,frmfd);
 
-  routerchilds[i].tochild   = tofd[1];
-  routerchilds[i].fromchild = frmfd[0];
-  routerchilds[i].childpid  = pid;
-  routerchilds[i].hungry    = 0;
-  routerchilds[i].childsize = 0;
-  routerchilds[i].childout  = 0;
-  return 0;
+	fd_nonblockingmode(tofd[1]);
+	fd_nonblockingmode(frmfd[0]);
+	
+	routerchilds[i].tochild   = tofd[1];
+	routerchilds[i].fromchild = frmfd[0];
+	routerchilds[i].childpid  = pid;
+	routerchilds[i].hungry    = 0;
+	routerchilds[i].childsize = 0;
+	routerchilds[i].childout  = 0;
+
+	return 0;
 }
 
 /*
@@ -271,18 +347,30 @@ int signum;
 
 	for (;;) {
 
+#ifdef HAVE_SYS_RESOURCE_H
+	  memset(&r, 0, sizeof(r));
+#endif
+
+#ifdef  HAVE_WAIT4
+#ifdef HAVE_SYS_RESOURCE_H
+	  pid = wait4(-1, &statloc, WNOHANG, &r);
+#else
+	  pid = wait4(-1, &statloc, WNOHANG, NULL);
+#endif
+#else
 #ifdef  HAVE_WAIT3
 #ifdef HAVE_SYS_RESOURCE_H
-	  pid = wait3(&statloc,WNOHANG,&r);
+	  pid = wait3(&statloc, WNOHANG, &r);
 #else
-	  pid = wait3(&statloc,WNOHANG,NULL);
+	  pid = wait3(&statloc, WNOHANG, NULL);
 #endif
 #else
 #ifdef	HAVE_WAITPID
-	  pid = waitpid(-1,&statloc,WNOHANG);
+	  pid = waitpid(-1, &statloc, WNOHANG);
 #else
 	  pid = wait(&statloc);
 #endif
+#endif
 #endif
 	  if (pid <= 0) break;
 
@@ -290,9 +378,10 @@ int signum;
 	    if (pid == routerchilds[i].childpid) {
 	      routerchilds[i].childpid = -pid;
 	      routerchilds[i].statloc = statloc;
-#if defined(HAVE_WAIT3) && defined(HAVE_SYS_RESOURCE_H)
+#if defined(HAVE_SYS_RESOURCE_H)
 	      routerchilds[i].r = r;
 #endif
+	      MIBMtaEntry->rt.RouterProcesses -= 1;
 	    }
 	}
 
@@ -326,17 +415,28 @@ static int reader_getc(rc)
     if (logfn) {
       loginit(SIGHUP); /* Reinit/rotate the log every at line .. */
       fprintf(stdout, "[%d] ROUTER CHILD PROCESS TERMINATED; wait() status = ", rc->childpid);
-      if (WIFSIGNALED(rc->statloc))
+      if (WIFSIGNALED(rc->statloc)) {
 	fprintf(stdout, "SIGNAL %d", WSIGNALSTATUS(rc->statloc));
-      else if (WIFEXITED(rc->statloc))
+	MIBMtaEntry->rt.RouterProcessFaults ++;
+      } else if (WIFEXITED(rc->statloc)) {
 	fprintf(stdout, "EXIT %d", WEXITSTATUS(rc->statloc));
-      else
+	if (WEXITSTATUS(rc->statloc)) /* non-zero */
+	  MIBMtaEntry->rt.RouterProcessFaults ++;
+      } else {
+	/* Uh ?? SIGNALED, and EXITED are handled above, what is this ? */
 	fprintf(stdout, "0x%04X ??", WEXITSTATUS(rc->statloc));
-#if defined(HAVE_WAIT3) && defined(HAVE_SYS_RESOURCE_H)
+	if (WEXITSTATUS(rc->statloc)) /* non-zero */
+	  MIBMtaEntry->rt.RouterProcessFaults ++;
+      }
+
+#if (defined(HAVE_WAIT3) || defined(HAVE_WAIT4))  && \
+    defined(HAVE_SYS_RESOURCE_H)
+
       fprintf(stdout, "; time = %ld.%06ld usr %ld.%06ld sys",
 	      (long)rc->r.ru_utime.tv_sec, (long)rc->r.ru_utime.tv_usec,
 	      (long)rc->r.ru_stime.tv_sec, (long)rc->r.ru_stime.tv_usec);
 #endif
+
       fprintf(stdout,"\n");
       fflush(stdout);
     }
@@ -366,11 +466,10 @@ static int reader_getc(rc)
 
 
 /* Single child reader.. */
-static int _parent_reader __((const int i));
-static int _parent_reader(i)
-     const int i;
+static int _parent_reader __((struct router_child *rc));
+static int _parent_reader(rc)
+     struct router_child *rc;
 {
-  struct router_child *rc = &routerchilds[i];
   int c;
 
   if (rc->fromchild < 0) return 0;
@@ -382,7 +481,7 @@ static int _parent_reader(i)
     if (c == EOF) {
       /* Because the socket/pipe is in NON-BLOCKING mode, we
 	 may drop here with an ERROR indication, which can be
-	 cleared and thing resume latter.. */
+	 cleared and thing resume later.. */
       if (errno)
 	break;
 
@@ -390,6 +489,8 @@ static int _parent_reader(i)
       if (rc->tochild >= 0)
 	close(rc->tochild);
       rc->tochild   = -1;
+      if (rc->fromchild >= 0)
+	close(rc->fromchild);
       rc->fromchild = -1;
       rc->hungry    = 0;
       rc->task_ino  = 0;
@@ -441,11 +542,10 @@ static int _parent_reader(i)
 }
 
 /* Single child writer.. */
-static void _parent_writer __((const int i));
-static void _parent_writer(i)
-     const int i;
+static void _parent_writer __((struct router_child *rc));
+static void _parent_writer(rc)
+     struct router_child *rc;
 {
-  struct router_child *rc = &routerchilds[i];
   int c, left;
 
   /* FD for writing ?? */
@@ -455,6 +555,10 @@ static void _parent_writer(i)
   for (;rc->childout < rc->childsize;) {
     left = rc->childsize - rc->childout;
     c = write(rc->tochild, rc->childline + rc->childout, left);
+    if (c < 0 && errno == EPIPE) {
+      pipes_shutdown_child(rc->tochild);
+      rc->tochild = -1;
+    }
     if (c <= 0) break;
     rc->childout += c;
   }
@@ -521,17 +625,24 @@ typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
 
 /* Return info re has some waiting been done */
 
-static int parent_reader()
+static int parent_reader(waittime)
+	int waittime;
 {
   fd_set rdset, wrset;
   int i, highfd, fd, rc;
   struct timeval tv;
 
+  if (notifysocket_reinit && notifysocket_reinit < now)
+    notifysock_init();
+
  redo_again:;
 
   _Z_FD_ZERO(rdset);
   _Z_FD_ZERO(wrset);
-  highfd = -1;
+
+  highfd = notifysocket;
+  if (notifysocket >= 0)
+    _Z_FD_SET(notifysocket, rdset);
 
   for (i = 0; i < MAXROUTERCHILDS; ++i) {
     /* Places to read from ?? */
@@ -550,9 +661,12 @@ static int parent_reader()
 	highfd = fd;
     }
   }
+#if 0
   if (highfd < 0) return 0; /* Nothing to do! */
+#endif
 
-  tv.tv_sec = tv.tv_usec = 1;
+  tv.tv_sec = waittime;
+  tv.tv_usec = 0;
   rc = select(highfd+1, &rdset, &wrset, NULL, &tv);
 
   if (rc == 0) return 1; /* Nothing to do, leave..
@@ -564,31 +678,36 @@ static int parent_reader()
       goto redo_again;
     /* Hmm.. Do it just blindly (will handle error situations) */
     for (i = 0; i < MAXROUTERCHILDS; ++i) {
-      _parent_writer(i);
-      _parent_reader(i);
+      _parent_writer(&routerchilds[i]);
+      _parent_reader(&routerchilds[i]);
     }
     return 0;  /* Urgh, an error.. */
   }
 
   /* Ok, select gave indication of *something* being ready for read */
+
+  if (notifysocket >= 0 && _Z_FD_ISSET(notifysocket, rdset))
+    notify_reader(notifysocket);
+
   for (i = 0; i < MAXROUTERCHILDS; ++i) {
     fd = routerchilds[i].tochild;
     if (fd >= 0 && _Z_FD_ISSET(fd, wrset))
-      _parent_writer(i);
+      _parent_writer(&routerchilds[i]);
     fd = routerchilds[i].fromchild;
     if (fd >= 0 && _Z_FD_ISSET(fd, rdset))
-      _parent_reader(i);
+      _parent_reader(&routerchilds[i]);
   }
   return 1; /* Did some productive job, don't sleep at the caller.. */
 }
 
 #else /* NO HAVE_SELECT */
-static int parent_reader()
+static int parent_reader(waittime)
+	int waittime;
 {
   int i;
   /* No select, but can do non-blocking -- we hope.. */
   for (i = 0; i < MAXROUTERCHILDS; ++i)
-    _parent_reader(i);
+    _parent_reader(&routerchilds[i], waittime);
   return 0;
 }
 #endif
@@ -660,8 +779,9 @@ static int parent_feed_child(fname,dir)
     s = strchr(fname,'-');
     if (s && isdigit(*fname)) {
       long thatpid = atoi(s+1);
-      if (thatpid > 1 &&
-	  kill(thatpid,0)==0) {
+      if ((thatpid > 1) &&
+	  (thatpid != MYPID) &&
+	  (kill(thatpid,0)==0)) {
 	/* Process of that PID does exist, and possibly even something
 	   we can kick.. (we should be *root* here anyway!) */
 	for (i = 0; i < MAXROUTERCHILDS; ++i) {
@@ -673,7 +793,7 @@ static int parent_feed_child(fname,dir)
       /* Hmm..  perhaps it is safe to feed to a subprocess */
     }
 
-    parent_reader();
+    parent_reader(0);
 
     rridx = 0;
     rc    = NULL;
@@ -688,7 +808,7 @@ static int parent_feed_child(fname,dir)
       if (rc->childpid == 0) {
 	start_child(rridx);
 	sleep(2); /* Allow a moment for child startup */
-	parent_reader();
+	parent_reader(0);
       }
 
       /* If we have a hungry child with all faculties intact.. */
@@ -709,7 +829,7 @@ static int parent_feed_child(fname,dir)
 #if 0 /* NO sync waiting here! */
   /* .. or if not, we wait here until it has been fed.. */
   while (rc->childout < rc->childsize)
-    parent_reader();
+    parent_reader(0);
 #endif
 
   return 1; /* Did feed successfully ?? */
@@ -803,6 +923,10 @@ static int dirqueuescan __((const char *dir,struct dirqueue *dq, int subdirs));
 struct dirqueue dirqb[ROUTERDIR_CNT];
 struct dirqueue *dirq[ROUTERDIR_CNT];
 
+/* Following is filled at  run_daemon(), and cleared also.. */
+static char     *routerdirs[ROUTERDIR_CNT];
+static char	*routerdirs2[ROUTERDIR_CNT];
+
 /*
  * Absorb any new files that showed up in our directory.
  */
@@ -832,10 +956,15 @@ dq_insert(DQ, ino, file, dir)
 	  return -1;
 	}
 
+#if 0
+	loginit(SIGHUP); /* Reinit/rotate the log every at line .. */
+	fprintf(stdout,"dqinsert: ino=%ld file='%s' dir='%s'\n",ino,file,dir);
+#endif
+
 	/* Is it already in the database ? */
 	if (sp_lookup((u_long)ino,dq->mesh) != NULL) {
 #if 0
-	  sfprintf(sfstderr,"scheduler: tried to dq_insert(ino=%ld, file='%s') already in queue\n",ino, file);
+	  fprintf(stderr,"daemonsub: tried to dq_insert(ino=%ld, file='%s') already in queue\n",ino, file);
 #endif
 	  return 1; /* It is! */
 	}
@@ -857,13 +986,11 @@ dq_insert(DQ, ino, file, dir)
 	if (dq->wrkspace <= dq->wrkcount) {
 	  /* Increase the space */
 	  dq->wrkspace = dq->wrkspace ? dq->wrkspace << 1 : 8;
-	  if (dq->stats == NULL)
-	    dq->stats = (struct dirstatname **)emalloc(sizeof(void*) *
-						       dq->wrkspace);
-	  else
-	    dq->stats = (struct dirstatname**)erealloc(dq->stats,
-						       sizeof(void*) *
-						       dq->wrkspace);
+
+	  /* malloc(size) == realloc(NULL,size) */
+	  dq->stats = (struct dirstatname**)erealloc(dq->stats,
+						     sizeof(void*) *
+						     dq->wrkspace);
 	}
 
 	dq->stats[dq->wrkcount] = dsn;
@@ -871,7 +998,17 @@ dq_insert(DQ, ino, file, dir)
 	dq->wrksum   += 1;
 	dq->sorted    = 0;
 
-	++MIBMtaEntry->mtaReceivedMessagesRt;
+	/* Account these into router job queue,
+	   counting them OUT of that queue is job for
+	   rfc822.c::sequencer()  ! 
+	*/
+
+	MIBMtaEntry->rt.ReceivedMessages += 1;
+	MIBMtaEntry->rt.StoredMessages   += 1;
+
+	i = (stbuf.st_size + stbuf.st_blksize -1)/1024;
+	MIBMtaEntry->rt.ReceivedVolume += i;
+	MIBMtaEntry->rt.StoredVolume   += i;
 
 	return 0;
 }
@@ -1058,6 +1195,98 @@ int syncweb(rc)
 	return wrkcnt;
 }
 
+/*
+ *  The  notify_reader()  receives messages from message submitters,
+ *  and uses them to find (most of the) new jobs.
+ *
+ *  The messages are of SPACE separated string:  'NEW dirname X/Y/filename'
+ *
+ */
+
+static void notify_reader(sock)
+	int sock;
+{
+	char buf[1000], *r, *p, *s;
+	int i, ok, cnt;
+	long ino;
+
+	/* Pick at most 10 in a row */
+	for (cnt=10; cnt!=0; --cnt) {
+
+	  i = recvfrom(sock, buf, sizeof(buf), 0, NULL, NULL);
+	  if (i <= 0 && errno != EINTR) return;
+	  if (i < 0) continue;
+
+	  if (i >= sizeof(buf)) i = sizeof(buf)-1;
+	  buf[i] = 0;
+
+	  /* The only supported notify is:
+	       "NEW router X/Y/file-name"
+	     message */
+
+#if 0 /* DEBUG IT! */
+	  if (logfn) {
+	    loginit(SIGHUP); /* Reinit/rotate the log every at line .. */
+	    fprintf(stdout, "NOTIFY got: i=%d '%s'\n", i, buf);
+	    fflush(stdout);
+	  }
+#endif
+
+	  if (strncmp(buf,"NEW ",4) != 0) continue;
+
+	  r = s = p = buf+4;
+	  while (*s != 0 && *s != ' ') ++s;
+	  if (*s == ' ') *s++ = 0;
+	  p = s;
+	  ok = 0;
+
+	  if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+	    ok = 1;
+	  } else  if ('A' <= s[0] && s[0] <= 'Z' && s[1] == '/') {
+	    s += 2;
+	    if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+	      ok = 1;
+	    } else  if ('A' <= s[0] && s[0] <= 'Z' && s[1] == '/') {
+	      s += 2;
+	      if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+		ok = 1;
+	      }
+	    }
+	  }
+	  if (!ok) continue;
+
+	  ino = atol(p);
+
+	  for (i = 0; i < ROUTERDIR_CNT; ++i) {
+	    if (routerdirs[i] && strcmp(routerdirs[i],r) == 0) {
+#if 0
+	      if (lstat(p,&stbuf) != 0) continue;
+	      if (!S_ISFILE(stbuf.st_mode)) continue;
+
+	      if (in_dirscanqueue(dirq[i],stbuf.st_ino)) continue;
+#endif
+#if 0 /* DEBUG IT! */
+	      if (logfn) {
+		loginit(SIGHUP); /* Reinit/rotate the log every at line .. */
+		fprintf(stdout, " ... routerdirs[%d] == '%s', ino=%ld, file='%s'\n",
+			i, routerdirs[i], ino, p);
+		fflush(stdout);
+	      }
+	      zsyslog((LOG_INFO, "notify dir='%s' ino=%ld file='%s'",
+		       r, ino, p));
+#endif
+
+	      dq_insert(dirq[i], ino, p, routerdirs2[i]);
+	      break;
+	    }
+	  }
+
+	}
+}
+
+
+
+
 /* "rd_doit()" at the feeding parent server */
 
 
@@ -1119,6 +1348,13 @@ rd_doit(filename, dirs)
 	   depending on actually doing something
 	   on a file */
 
+	int dir_fd_ = -1;
+	DIR *dp = NULL;
+
+	/* The 'filename' is always directoryless, whereas
+	   'dirs' points to the directory where it resides */
+
+
 #ifdef	USE_ALLOCA
 	char *buf;
 #else
@@ -1127,18 +1363,46 @@ rd_doit(filename, dirs)
 #endif
 	const char *av[3];
 	char *p;
-	int len;
+	int len, rc;
+	int lockfd = -1;
 	char pathbuf[512];
+	char dirhash[5];
 	char *sh_memlevel = getlevel(MEM_SHCMD);
 	int thatpid;
 	struct stat stbuf;
 
 	router_id = getpid();
+	dirhash[0] = 0;
 
 	*pathbuf = 0;
-	if (*dirs) {	/* If it is in alternate dir, move to primary one,
-			   and process there! */
-	  strcpy(pathbuf,dirs);
+#if defined(Z_DIRFD) && defined(HAVE_FCHDIR)
+	if (*dirs) {
+	  dp = opendir(".");
+	  if (dp) dir_fd_ = Z_DIRFD(dp);
+	  if (*dirs && dirs[0]  == '.' && dirs[1] == '.' && dirs[2] == '/') {
+	    char *p = pathbuf + 3;
+	    strcpy(pathbuf, dirs);
+	    while (*p && *p != '/') ++p;
+	    if (*p) *p++ = 0;
+	    chdir(pathbuf);
+	    strncpy(dirhash, p, sizeof(dirhash)-1);
+	    dirhash[sizeof(dirhash)-1] = 0;
+	  } else {
+	    strncpy(dirhash, dirs, sizeof(dirhash)-1);
+	    dirhash[sizeof(dirhash)-1] = 0;
+	  }
+
+	  *pathbuf = 0;
+	  if (dirhash[0]) {
+	    strcpy(pathbuf,dirhash);
+	    strcat(pathbuf,"/");
+	  }
+	}
+#endif
+	if (dirhash[0] && dir_fd_ < 0) {  /* If it is in alternate dir,
+					     move to primary one,
+					     and process there! */
+	  strcpy(pathbuf,dirhash);
 	  strcat(pathbuf,"/");
 	}
 	strcat(pathbuf,filename);
@@ -1157,7 +1421,7 @@ rd_doit(filename, dirs)
 #endif
 	  /* Probe it!
 	     Does the process exist ? */
-	  if (thatpid && kill(thatpid,0)==0 && thatpid != router_id) {
+	  if (thatpid && (kill(thatpid,0)==0) && (thatpid != router_id)) {
 	    /*
 	     * an already locked message file,
 	     * belonging to another process
@@ -1167,6 +1431,11 @@ rd_doit(filename, dirs)
 		      "** BUG **: %s%s not in primary router directory!\n",
 		      dirs,filename);
 	    }
+
+#if defined(HAVE_FCHDIR)
+	    if (dir_fd_ >= 0) fchdir(dir_fd_);
+#endif
+	    if (dp) closedir(dp);
 	    return 0;
 	    /*
 	     * This should not happen anywhere but at
@@ -1176,37 +1445,84 @@ rd_doit(filename, dirs)
 	     */
 	  }
 	}
+
+	lockfd = open(pathbuf, O_RDWR, 0);
+	if (lockfd >= 0) {
+#if defined(HAVE_LOCKF) && defined(F_LOCK)
+	  lseek(lockfd, (off_t)0, SEEK_SET); /* To the end of the file */
+	  if (lockf(lockfd, F_LOCK, 0) < 0) {
+	    /* return 0; FIXME: ??? ERRORS ?? HOW ???  */
+	  }
+#else
+#ifdef	HAVE_FLOCK
+	  if (flock(lockfd, LOCK_EX) < 0) {
+	    /* return 0; FIXME: ??? ERRORS ?? HOW ???  */
+	  }
+#else
+	  /* FCNTL() locking */
+	  struct flock ft; 
+	  memset( & ft, 0, sizeof(ft) );
+	  ft.l_type   = F_WRLCK;
+	  ft.l_whence = SEEK_SET;
+	  ft.l_start  = 0;
+	  ft.l_len    = 0;
+	  fcntl(lockfd, F_GETLK, &ft); /* XXX: ERROR PROCESSING! */
+#endif	/* HAVE_FLOCK */
+#endif	/* HAVE_LOCKF */
+	}
+
 	if (strncmp(filename,"core",4) != 0 &&
-	    (p == NULL || thatpid != router_id)) {
+	    ((p == NULL) || (thatpid != router_id))) {
 	  /* Not a core file, and ...
 	     not already in format of 'inode-pid' */
 	  /* If the pid did exist, we do not touch on that file,
 	     on the other hand, we need to rename the file now.. */
 #ifdef	USE_ALLOCA
-	  buf = (char*)alloca(len+16);
+	  buf = (char*)alloca(len+20);
 #else
 	  if (blen == 0) {
-	    blen = len+16;
-	    buf = (char *)malloc(len+16);
+	    blen = len+20;
+	    buf = (char *)malloc(len+20);
 	  }
-	  while (len + 12 > blen) {
+	  while (len + 18 > blen) {
 	    blen = 2 * blen;
 	    buf = (char *)realloc(buf, blen);
 	  }
 #endif
-	  /* Figure out its inode number */
-	  if (lstat(pathbuf,&stbuf) != 0) return 0; /* Failed ?  Well, skip it */
-	  if (!S_ISREG(stbuf.st_mode))   return 0; /* Not a regular file ??   */
+	  /* Figure out its inode number */	
+	  rc = lstat(pathbuf,&stbuf);
+	  if (rc != 0) {
+	    if (lockfd >= 0) close(lockfd);
+#if defined(HAVE_FCHDIR)
+	    if (dir_fd_ >= 0) fchdir(dir_fd_);
+#endif
+	    if (dp) closedir(dp);
+	    return 0; /* Failed ?  Well, skip it */
+	  }
 
-	  sprintf(buf, "%ld-%d", (long)stbuf.st_ino, router_id);
+	  if (!S_ISREG(stbuf.st_mode)) return 0; /* Not a regular file ?? */
 
-	  if (eqrename(pathbuf, buf) < 0)
+	  if (dirhash[0]) {
+	    sprintf(buf, "%s/%ld", dirhash, (long)stbuf.st_ino);
+	  } else {
+	    sprintf(buf, "%ld", (long)stbuf.st_ino);
+	  }
+
+	  if (strcmp(pathbuf, buf) != 0  &&
+	      eqrename(pathbuf, buf) < 0) {
+	    if (lockfd >= 0) close(lockfd);
+#if defined(HAVE_FCHDIR)
+	    if (dir_fd_ >= 0) fchdir(dir_fd_);
+#endif
+	    if (dp) closedir(dp);
 	    return 0;		/* something is wrong, erename() complains.
 				   (some other process picked it ?) */
+	  }
 	  filename = buf;
 	  /* message file is now "file-#" and belongs to this process */
 	}
 
+
 #ifdef	MALLOC_TRACE
 	mal_contents(stdout);
 #endif
@@ -1218,6 +1534,12 @@ rd_doit(filename, dirs)
 	s_apply(2, av); /* "process" filename (within  rd_doit() ) */
 	free_gensym();
 
+	if (lockfd >= 0) close(lockfd);
+#if defined(HAVE_FCHDIR)
+	if (dir_fd_ >= 0) fchdir(dir_fd_);
+#endif
+	if (dp) closedir(dp);
+
 	setlevel(MEM_SHCMD,sh_memlevel);
 
 #ifdef MALLOC_TRACE
@@ -1240,14 +1562,14 @@ run_stability(argc, argv)
 	case 2:
 		if (strcmp(argv[1], "on") == 0) {
 			real_stability = 1;
+			break;
 		} else if (strcmp(argv[1], "off") == 0) {
 			real_stability = 0;
-		} else {
-	default:
-			fprintf(stderr, "Usage: %s [ on | off ]\n", argv[0]);
-			return 1;
+			break;
 		}
-		break;
+	default:
+		fprintf(stderr, "Usage: %s [ on | off ]\n", argv[0]);
+		return EX_USAGE;
 	}
 	return 0;
 }
@@ -1258,18 +1580,30 @@ run_daemon(argc, argv)
 	const char *argv[];
 {
 	DIR *dirp[ROUTERDIR_CNT];  /* Lets say we have max 30 router dirs.. */
-	char *dirs[ROUTERDIR_CNT];
 	int i, ii;
-	char *s, *rd, *routerdirs = getzenv("ROUTERDIRS");
+	const char *s, *rd;
+	const char *routerdir_s = getzenv("ROUTERDIRS");
 	char pathbuf[256];
 	memtypes oval = stickymem;
 
 	time_t nextdirscan = 0;
 
+
+	MIBMtaEntry->sys.RouterMasterPID        =  getpid();
+	MIBMtaEntry->sys.RouterMasterStartTime  = time(NULL);
+	MIBMtaEntry->sys.RouterMasterStarts    += 1;
+
+
+	/* Zero the gauges at our startup.. */
+	MIBMtaEntry->rt.StoredMessages		= 0; /* in input queue */
+	MIBMtaEntry->rt.StoredVolume		= 0; /* in input queue */
+	MIBMtaEntry->rt.RouterProcesses       = 1; /* myself = 1 */
+	MIBMtaEntry->rt.RouterProcessForks    = 0;
+	MIBMtaEntry->rt.RouterProcessFaults   = 0;
+
 	if (nrouters > MAXROUTERCHILDS)
 	  nrouters = MAXROUTERCHILDS;
 
-
 	memset(routerchilds, 0, sizeof(routerchilds));
 
 	for (i = 0; i < MAXROUTERCHILDS; ++i)
@@ -1290,7 +1624,8 @@ run_daemon(argc, argv)
 	  dirq[i] = &dirqb[i];
 	  dirqb[i].mesh = sp_init();
 	  dirp[i] = NULL;
-	  dirs[i] = NULL;
+	  routerdirs[i] = NULL;
+	  routerdirs2[i] = NULL;
 	}
 	/* dirp[0] = opendir("."); */	/* assert dirp != NULL ... */
 
@@ -1306,26 +1641,27 @@ run_daemon(argc, argv)
 	ii = MAXROUTERCHILDS;
 
 	stickymem = MEM_MALLOC;
-	dirs[0] = strnsave(".",2);
-	if (routerdirs) {
+	routerdirs[0]  = strnsave("router",6);
+	routerdirs2[0] = strnsave(".",2);
+	if (routerdir_s) {
 	  /* Store up secondary router dirs! */
-	  rd = routerdirs;
+	  rd = routerdir_s;
 	  for (i = 1; i < ROUTERDIR_CNT && *rd; ) {
 
 	    s = strchr(rd,':');
-	    if (s)  *s = 0;
+	    if (s)  *(char*)s = 0;
 	    sprintf(pathbuf,"../%s",rd);
 	    /* strcat(pathbuf,"/"); */
 
-	    dirs[i] = strdup(pathbuf);
+	    routerdirs[i] = strdup(rd);
+	    routerdirs2[i] = strdup(pathbuf);
 
 	    if (ii > 2)
 	      routerchilds[ --ii ].dq = dirq[i];
 
 	    ++i;
 
-	    if (s)
-	      *s = ':';
+	    if (s) *(char*)s = ':';
 	    if (s)
 	      rd = s+1;
 	    else
@@ -1340,8 +1676,8 @@ run_daemon(argc, argv)
 	/* Do initial synchronous queue scan now */
 
 	for (i = 0; i < ROUTERDIR_CNT; ++i) {
-	  if (dirs[i])
-	    dirqueuescan(dirs[i], dirq[i], 1);
+	  if (routerdirs2[i])
+	    dirqueuescan(routerdirs2[i], dirq[i], 1);
 	}
 
 	for (i = 0; i < MAXROUTERCHILDS; ++i) {
@@ -1352,10 +1688,10 @@ run_daemon(argc, argv)
 	}
 
 	sleep(5); /* Wait a bit before continuting so that
-		     child processes have change to boot */
+		     child processes have a change to start */
 
 	/* Collect start reports (initial "#hungry\n" lines) */
-	parent_reader();
+	parent_reader(0);
 
 	for (; !mustexit ;) {
 
@@ -1364,10 +1700,9 @@ run_daemon(argc, argv)
 	      nextdirscan = now + 10;
 
 	      for (i = 0; i < ROUTERDIR_CNT; ++i) {
-		if (dirs[i])
-		  dirqueuescan(dirs[i], dirq[i], 1);
+		if (routerdirs2[i])
+		  dirqueuescan(routerdirs2[i], dirq[i], 1);
 	      }
-
 	    }
 
 	    for (i = 0; i < MAXROUTERCHILDS; ++i) {
@@ -1386,13 +1721,17 @@ run_daemon(argc, argv)
 	      }
 	    }
 
-	    if (! parent_reader())
-	      sleep(1);
+	    parent_reader(1);
 	}
 
-	for (i=0; i < ROUTERDIR_CNT; ++i)
-	  if (dirs[i])
-	    free(dirs[i]);
+	for (i=0; i < ROUTERDIR_CNT; ++i) {
+	  if (routerdirs2[i])
+	    free(routerdirs2[i]);
+	  routerdirs2[i] = NULL;
+	  if (routerdirs[i])
+	    free(routerdirs[i]);
+	  routerdirs[i] = NULL;
+	}
 
 	return 0;
 }
