@@ -2,7 +2,7 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	Modifications/maintance, Matti Aarnio, over years 1990-2001
+ *	Modifications/maintance, Matti Aarnio, over years 1990-2003
  *
  *	'longestmatch' driver kissg@sztaki.hu 970209
  */
@@ -28,6 +28,7 @@
 #endif
 #include "libdb/search.h"
 #include "splay.h"
+#include "sleepycatdb.h"
 
 #include "prototypes.h"
 
@@ -43,20 +44,21 @@ extern struct sptree *spt_databases; /* At conf.c */
 typedef enum { Nul, Boolean, Pathalias, Indirect, NonNull } postprocs;
 
 struct cache {
-	char		*key;
+	time_t		expiry;
 	unsigned long	keyhash;
+	char		*key;
 	struct cache	*next;
 	conscell	*value;
-	time_t		expiry;
 };
 
 #define DBFUNC(_fn_)  (* _fn_) __((search_info *))
-#define DBFUNCD(_fn_)  (* _fn_) __((conscell *DBFUNC(lookupfn), search_info *))
+#define DBFUNCD(_fn_) (* _fn_) __((conscell *DBFUNC(lookupfn), search_info *))
 #define DBFUNCV(_fn_) (* _fn_) __((search_info *, const char *))
 #define DBFUNCF(_fn_) (* _fn_) __((search_info *, FILE *))
 
 struct db_info {
 	const char	*file;			/* a file parameter */
+	const char	*cfgfile;		/* generalized config file */
 	const char	*subtype;		/* optional selector */
 	int		flags;			/* miscellaneous options */
 	int		cache_size;		/* default cache size */
@@ -74,6 +76,9 @@ struct db_info {
 	struct cache	*cache;			/* cache entry array */
 	struct cache	*cfirst;		/* LRU cache head entry */
 	struct cache	*cfree;			/* Chain of free entries */
+	void		*dbprivate;		/* DB specific private data;
+						   created by open, destroyed
+						   by close..  */
 };
 
 /* bits in the flags field */
@@ -82,13 +87,14 @@ struct db_info {
 #define	DB_MODCHECK	0x04
 #define DB_NEG_CACHE	0x08
 #define DB_PERCENTSUBST	0x10
+#define DB_DEFAULTKEY	0x20
 
 
 struct db_kind {
 	const char	*name;		/* database type identification */
 	struct db_info	config;		/* default configuration information */
 } db_kinds[] = {
-{ "incore",	{ NULL, NULL, 0, 0, 0, NULL, search_core, close_core,
+{ "incore",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_core, close_core,
 		  add_core,
 		  remove_core,
 		  print_core,
@@ -96,70 +102,70 @@ struct db_kind {
 		  owner_core,
 		  NULL,
 		  Nul,
-		  NULL } },
-{ "header",	{ NULL, NULL, 0, 0, 0, NULL, search_header, close_header,
+		  NULL, NULL } },
+{ "header",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_header, close_header,
 		  add_header, remove_header, print_header, count_header,
-		  owner_header, NULL, Nul, NULL } },
-{ "unordered",	{ NULL, NULL, 0, 10, 0, NULL, search_seq, close_seq,
+		  owner_header, NULL, Nul, NULL, NULL } },
+{ "unordered",	{ NULL, NULL, NULL, 0, 10, 0, NULL, search_seq, close_seq,
 		  add_seq, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul, NULL } },
+		  Nul, NULL, NULL } },
 #ifndef	HAVE_MMAP
-{ "ordered",	{ NULL, NULL, 0, 10, 0, NULL, search_bin, close_seq,
+{ "ordered",	{ NULL, NULL, NULL, 0, 10, 0, NULL, search_bin, close_seq,
 		  NULL, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul, NULL } },
+		  Nul, NULL, NULL } },
 #else /* HAVE_MMAP */ /* When using MMAP(), no cache is needed for ordered.. */
-{ "ordered",	{ NULL, NULL, 0, 0, 0, NULL, search_bin, close_seq,
+{ "ordered",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_bin, close_seq,
 		  NULL, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul, NULL } },
+		  Nul, NULL, NULL } },
 #endif
 
 #ifdef	HAVE_RESOLVER
-{ "hostsfile",	{ "/etc/hosts", NULL, 0, 0, 0, NULL, search_hosts, NULL,
-		  NULL, NULL, print_hosts, NULL, NULL, NULL, Nul, NULL } },
+{ "hostsfile",	{ "/etc/hosts", NULL, NULL, 0, 0, 0, NULL, search_hosts, NULL,
+		  NULL, NULL, print_hosts, NULL, NULL, NULL, Nul, NULL, NULL } },
 #endif	/* HAVE_RESOLVER */
 #ifdef	HAVE_RESOLVER
 #ifndef RESOLV_CONF
 # define RESOLV_CONF "/etc/resolv.conf"
 #endif
-{ "bind",	{ RESOLV_CONF, NULL, 0, 0, 0, NULL, search_res, NULL, NULL,
-		    NULL, NULL, NULL, NULL, NULL, Nul, NULL }},
+{ "bind",	{ RESOLV_CONF, NULL, NULL, 0, 0, 0, NULL, search_res, NULL, NULL,
+		    NULL, NULL, NULL, NULL, NULL, Nul, NULL, NULL }},
 #endif	/* HAVE_RESOLV */
-{ "selfmatch",	{ NULL, NULL, 0, 0, 0, NULL, search_selfmatch, NULL, NULL,NULL,
-		  print_selfmatch, count_selfmatch, NULL, NULL, Nul, NULL } },
-#ifdef	HAVE_NDBM_H
-{ "ndbm",	{ NULL, NULL, 0, 0, 0, NULL, search_ndbm, close_ndbm,
+{ "selfmatch",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_selfmatch, NULL, NULL,NULL,
+		  print_selfmatch, count_selfmatch, NULL, NULL, Nul, NULL, NULL } },
+#ifdef	HAVE_NDBM
+{ "ndbm",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_ndbm, close_ndbm,
 		  add_ndbm, remove_ndbm, print_ndbm, count_ndbm, owner_ndbm,
-		  modp_ndbm, Nul, NULL } },
+		  modp_ndbm, Nul, NULL, NULL } },
 #endif	/* HAVE_NDBM */
-#ifdef	HAVE_GDBM_H
-{ "gdbm",	{ NULL, NULL, 0, 0, 0, NULL, search_gdbm, close_gdbm,
+#ifdef	HAVE_GDBM
+{ "gdbm",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_gdbm, close_gdbm,
 		  add_gdbm, remove_gdbm, print_gdbm, count_gdbm, owner_gdbm,
-		  modp_gdbm, Nul, NULL } },
+		  modp_gdbm, Nul, NULL, NULL } },
 #endif	/* HAVE_GDBM */
 #ifdef	HAVE_DBM
-{ "dbm",	{ NULL, NULL, 0, 0, 0, NULL, search_dbm, close_dbm,
+{ "dbm",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_dbm, close_dbm,
 		  add_dbm, remove_dbm, print_dbm, count_dbm, owner_dbm,
-		  NULL, Nul, NULL } },
+		  NULL, Nul, NULL, NULL } },
 #endif	/* HAVE_DBM */
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-{ "btree",	{ NULL, NULL, 0, 0, 0, NULL, search_btree, close_btree,
+#ifdef HAVE_DB
+{ "btree",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_btree, close_btree,
 		  add_btree, remove_btree, print_btree, count_btree,
-		  owner_btree, modp_btree, Nul, NULL } },
-{ "bhash",	{ NULL, NULL, 0, 0, 0, NULL, search_bhash, close_bhash,
+		  owner_btree, modp_btree, Nul, NULL, NULL } },
+{ "bhash",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_bhash, close_bhash,
 		  add_bhash, remove_bhash, print_bhash, count_bhash,
-		  owner_bhash, modp_bhash, Nul, NULL } },
-#endif	/* HAVE_DB_H */
+		  owner_bhash, modp_bhash, Nul, NULL, NULL } },
+#endif	/* HAVE_DB */
 #ifdef	HAVE_YP
-{ "yp",		{ NULL, NULL, 0, 0, 0, NULL, search_yp, NULL, NULL,
-		  NULL, print_yp, NULL, owner_yp, NULL, Nul, NULL } },
+{ "yp",		{ NULL, NULL, NULL, 0, 0, 0, NULL, search_yp, NULL, NULL,
+		  NULL, print_yp, NULL, owner_yp, NULL, Nul, NULL, NULL } },
 #endif	/* HAVE_YP */
-#ifdef HAVE_LDAP
-{ "ldap",	{ NULL, NULL, 0, 0, 0, NULL, search_ldap, close_ldap,
-		  NULL, NULL, NULL, NULL, NULL, modp_ldap, Nul, NULL } },
+#ifdef USE_LDAP
+{ "ldap",	{ NULL, NULL, NULL, 0, 0, 0, NULL, search_ldap, close_ldap,
+		  NULL, NULL, NULL, NULL, NULL, modp_ldap, Nul, NULL, NULL } },
 #endif
-{ NULL, { NULL, NULL, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+{ NULL, { NULL, NULL, NULL, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
 	    /* proto_config is initialized from this entry */
-	    NULL, NULL, Nul, NULL }}
+	    NULL, NULL, Nul, NULL, NULL }}
 };
 
 
@@ -228,7 +234,8 @@ run_relation(argc, argv)
 	int c, errflg, set_cache_size, dbtest;
 	spkey_t symid;
 	memtypes oval;
-	char *cp, *dbtyp;
+	char *cp;
+	const char *dbtyp;
 
 	if (spt_files == NULL)          spt_files          = sp_init();
 	if (spt_files->symbols == NULL) spt_files->symbols = sp_init();
@@ -236,24 +243,30 @@ run_relation(argc, argv)
 	errflg = 0;
 	dbtyp = NULL;
 	set_cache_size = 0;
-	optind = 1;
+	zoptind = 1;
 	dbtest = 0;
-	proto_config
-		= db_kinds[sizeof db_kinds/(sizeof (struct db_kind))-1].config;
+	proto_config=db_kinds[sizeof(db_kinds)/(sizeof(db_kinds[0]))-1].config;
+
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "CbilmnNpud:f:s:L:t:Te:%");
+		c = zgetopt(argc,(char*const*)argv,":%C:bilmnNpud:f:s:L:t:Te:");
 		if (c == EOF)
 			break;
 		switch (c) {
 		case 'b':	/* boolean postprocessor */
 			proto_config.postproc = Boolean;
 			break;
+		case 'C': /* Generalized cfgfile definition argument */
+			oval = stickymem;
+			stickymem = MEM_PERM;
+			proto_config.cfgfile = strsave(zoptarg);
+			stickymem = oval;
+			break;
 		case 'd':	/* driver routine */
-			if (strcmp(optarg, "pathalias.nodot") == 0)
+			if (STREQ(zoptarg, "pathalias.nodot"))
 				proto_config.driver = find_nodot_domain;
-			else if (strcmp(optarg, "pathalias") == 0)
+			else if (STREQ(zoptarg, "pathalias"))
 				proto_config.driver = find_domain;
-			else if (strcmp(optarg, "longestmatch") == 0)
+			else if (STREQ(zoptarg, "longestmatch"))
 				proto_config.driver = find_longest_match;
 			else
 				++errflg;
@@ -261,7 +274,7 @@ run_relation(argc, argv)
 		case 'f':	/* file name */
 			oval = stickymem;
 			stickymem = MEM_PERM;
-			proto_config.file = strsave(optarg);
+			proto_config.file = strsave(zoptarg);
 			stickymem = oval;
 			break;
 		case 'i':	/* indirect reference postprocessor */
@@ -283,13 +296,13 @@ run_relation(argc, argv)
 			proto_config.postproc = Pathalias;
 			break;
 		case 's':	/* cache size */
-			proto_config.cache_size = atoi(optarg);
+			proto_config.cache_size = atoi(zoptarg);
 			set_cache_size = 1;
 			break;
 		case 't':	/* database type */
-			dbtyp = optarg;
-			if ((cp = strchr(optarg, ',')) != NULL
-			    || (cp = strchr(optarg, '/')) != NULL) {
+			dbtyp = zoptarg;
+			if ((cp = strchr(zoptarg, ',')) ||
+			    (cp = strchr(zoptarg, '/'))    ) {
 				*cp++ = '\0';
 				oval = stickymem;
 				stickymem = MEM_PERM;
@@ -302,7 +315,7 @@ run_relation(argc, argv)
 			dbtest = 1;
 			break;
 		case 'e':	/* expiry - cache data time to live */
-			proto_config.ttl = atol(optarg);
+			proto_config.ttl = atol(zoptarg);
 			break;
 		case 'u':	/* map all keys to uppercase */
 			proto_config.flags |= DB_MAPTOUPPER;
@@ -310,6 +323,9 @@ run_relation(argc, argv)
 		case '%':
 			proto_config.flags |= DB_PERCENTSUBST;
 			break;
+		case ':':
+			proto_config.flags |= DB_DEFAULTKEY;
+			break;
 		case 'F':	/* find routine */
 		case 'S':	/* search routine */
 		default:
@@ -317,22 +333,17 @@ run_relation(argc, argv)
 			break;
 		}
 	}
-	if (errflg || optind != argc - 1 || dbtyp == NULL) {
-		fprintf(stderr,
-		"Usage: %s -t dbtype[,subtype] [-f file -e# -s# -bilmnNpu -d driver] name\n",
-			argv[0]);
-		fprintf(stderr,
-		       "       %s -T -t dbtype dummyname\n", argv[0]);
-		fprintf(stderr,
-			"      dbtypes: ");
+	if (errflg || zoptind != argc - 1 || dbtyp == NULL) {
+		int first = 1;
+		fprintf(stderr, "Usage: %s -t dbtype[,subtype] [-f file -e# -s# -bilmnNpu -d driver] name\n", argv[0]);
+		fprintf(stderr, "       %s -T -t dbtype dummyname\n", argv[0]);
+		fprintf(stderr, "       dbtypes: ");
 		for (dbkp = &db_kinds[0];
-		     dbkp < &db_kinds[(sizeof db_kinds)/sizeof (struct db_kind)];
+		     dbkp < &db_kinds[sizeof(db_kinds)/sizeof(db_kinds[0])];
 		     ++dbkp) {
-		  if (dbkp->name != NULL) {
-		    if (dbkp == &db_kinds[0])
-		      fprintf(stderr,  "%s", dbkp->name);
-		    else
-		      fprintf(stderr, ",%s", dbkp->name);
+		  if (dbkp->name) {
+		    if (!first) fprintf(stderr, ","); first = 0;
+		    fprintf(stderr,  "%s", dbkp->name);
 		  }
 		}
 		fprintf(stderr,"\n");
@@ -345,13 +356,13 @@ run_relation(argc, argv)
 		return 2;
 	}
 	for (dbkp = &db_kinds[0];
-	     dbkp < &db_kinds[(sizeof db_kinds)/sizeof (struct db_kind)];
+	     dbkp < &db_kinds[sizeof(db_kinds)/sizeof(db_kinds[0])];
 	     ++dbkp) {
-		if (dbkp->name == NULL || strcmp(dbkp->name, dbtyp) == 0)
+		if (!dbkp->name  ||  STREQ(dbkp->name, dbtyp))
 			break;
 	}
 	/* the -1 in the following compensates for the terminating null entry */
-	if (dbkp >= &db_kinds[((sizeof db_kinds)/sizeof (struct db_kind))-1]) {
+	if (dbkp >= &db_kinds[(sizeof(db_kinds)/sizeof(db_kinds[0]))-1]) {
 	  if (!dbtest)
 	    fprintf(stderr,
 		    "relation: I don't know about the %s database type!\n",
@@ -361,24 +372,24 @@ run_relation(argc, argv)
 	if (dbtest)
 	  return 0;
 
-	symid = symbol(argv[optind]); /* Database name symbol  */
+	symid = symbol(argv[zoptind]); /* Database name symbol  */
 	if (sp_lookup(symid, spt_databases) != NULL) {
 		fprintf(stderr, "%s: %s is already a defined database!\n",
-				argv[0], argv[optind]);
+				argv[0], argv[zoptind]);
 		return 4;
 	}
 	if (spt_builtins != NULL && sp_lookup(symid, spt_builtins) != NULL) {
 		fprintf(stderr, "%s: %s is already a built in function!\n",
-				argv[0], argv[optind]);
+				argv[0], argv[zoptind]);
 		return 5;
 	}
 	if (spt_funclist != NULL && sp_lookup(symid, spt_funclist) != NULL) {
 		fprintf(stderr, "%s: %s is already a defined function!\n",
-				argv[0], argv[optind]);
+				argv[0], argv[zoptind]);
 		return 6;
 	}
 	if (proto_config.postproc == Indirect &&
-	    dbkp->config.owner != owner_seq) {
+	    dbkp->config.owner    != owner_seq) {
 		fprintf(stderr,
 			"%s: Indirect (-i) postprocessor can be used only with {un,}ordered databases!\n%s  Tried to use with: -t %s\n",
 			argv[0], argv[0], dbtyp);
@@ -386,19 +397,19 @@ run_relation(argc, argv)
 	}
 
 
-	if (dbkp->config.lookup == search_core
-	    || dbkp->config.lookup == search_header) {
+	if (dbkp->config.lookup == search_core ||
+	    dbkp->config.lookup == search_header) {
 		/* The database name is used as a key by the incore routines */
 		oval = stickymem;
 		stickymem = MEM_PERM;
-		proto_config.file = strsave(argv[optind]);
+		proto_config.file = strsave(argv[zoptind]);
 		stickymem = oval;
 		/* and the subtype is used to stash the splay tree */
 		proto_config.subtype = (char*) sp_init();
 	} else {
-		if (proto_config.file == NULL)
-			proto_config.file = dbkp->config.file;
-		if (proto_config.subtype == NULL)
+		if ((! proto_config.file) && (! proto_config.cfgfile))
+			proto_config.file    = dbkp->config.file;
+		if (! proto_config.subtype)
 			proto_config.subtype = dbkp->config.subtype;
 	}
 	if (proto_config.flags == 0)
@@ -439,6 +450,7 @@ run_relation(argc, argv)
 		dbip->cfree  = &dbip->cache[0];
 	} else
 		dbip->cache = NULL;	/* superfluous, but why not ... */
+	dbip->dbprivate = NULL; /* also superfluous .. */
 
 	sp_install(symid, dbip, 0, spt_databases);
 	register_cache_gc_markup_iterator();
@@ -494,8 +506,8 @@ iclistdbs(spl)
 
 	dbip = (struct db_info *)spl->data;
 	cp = NULL;
-	for (dbkp = &db_kinds[0];
-	     dbkp < &db_kinds[(sizeof db_kinds)/sizeof (struct db_kind)];
+	for (dbkp = &db_kinds[0] ;
+	     dbkp < &db_kinds[sizeof(db_kinds)/sizeof(db_kinds[0])] ;
 	     ++dbkp) {
 		if (dbkp->config.lookup == dbip->lookup) {
 			cp = dbkp->name;
@@ -504,8 +516,10 @@ iclistdbs(spl)
 	}
 	printf(" %s", cp);
 	i = strlen(cp);
-	if (dbip->lookup != search_core && dbip->lookup != search_header
-	    && dbip->postproc != Indirect && dbip->subtype != NULL) {
+	if (dbip->lookup   != search_core   &&
+	    dbip->lookup   != search_header &&
+	    dbip->postproc != Indirect      &&
+	    dbip->subtype  != NULL) {
 		printf(",%s", dbip->subtype);
 		i += 1 + strlen(dbip->subtype);
 	}
@@ -513,32 +527,53 @@ iclistdbs(spl)
 
 	printf("%*s", i, " ");
 	
-	for (i = 0, cachep = dbip->cfirst;
-	     cachep; cachep = cachep->next) ++i;
+	for (i = 0, cachep = dbip->cfirst ;
+	     cachep ; cachep = cachep->next) ++i;
 
 	printf("%4d/%-4d ", i, dbip->cache_size);
-	printf("%4d   ", (int)dbip->ttl);
+	printf("%4d ", (int)dbip->ttl);
 
 	i = 0;
 	if (dbip->flags & DB_MAPTOLOWER)
 		putchar('l'), ++i;
 	if (dbip->flags & DB_MAPTOUPPER)
-		putchar('U'), ++i;
+		putchar('u'), ++i;
+	if (dbip->flags & DB_MODCHECK)
+		putchar('m'), ++i;
+	if (dbip->flags & DB_NEG_CACHE)
+		putchar('N'), ++i;
+	if (dbip->flags & DB_PERCENTSUBST)
+		putchar('%'), ++i;
+	if (dbip->flags & DB_DEFAULTKEY)
+		putchar(':'), ++i;
 	switch (dbip->postproc) {
-	case NonNull: putchar('n'), ++i; break;
-	case Boolean: putchar('B'), ++i; break;
+	case NonNull:   putchar('n'), ++i; break;
+	case Boolean:   putchar('b'), ++i; break;
 	case Pathalias: putchar('p'), ++i; break;
-	case Indirect: putchar('@'), ++i; break;
+	case Indirect:  putchar('i'), ++i; break;
 	default: break;
 	}
+	if (dbip->driver == find_nodot_domain) {
+	  printf("{pa}");  i += 4;
+	} else if (dbip->driver == find_domain) {
+	  printf("{pa.}"); i += 5;
+	} else if (dbip->driver == find_longest_match) {
+	  printf("{lm}");  i += 4;
+	}
+	if (dbip->cfgfile) { putchar('C'); ++i; }
+	
 	if (i == 0)
 		putchar('-'), ++i;
-	i = (i > 2) ? 1 : 3 - i;
+	i = (i > 4) ? 1 : 5 - i;
 
 	printf("%*s", i, " ");
-	
-	if (dbip->lookup != search_core && dbip->lookup != search_header
-	    && dbip->file != NULL) {
+
+	if (dbip->cfgfile)
+	  printf("%s ", dbip->cfgfile);
+
+	if (dbip->lookup != search_core   &&
+	    dbip->lookup != search_header &&
+	    dbip->file   != NULL) {
 		printf("%s", dbip->file);
 		if (dbip->postproc == Indirect)
 			printf(" -> %s", dbip->subtype);
@@ -548,6 +583,37 @@ iclistdbs(spl)
 	return 0;
 }
 
+static int  dbs_atexit_set;
+
+static int
+dbs_atexit_close(spl)
+	struct spblk *spl;
+{
+	struct db_info *dbip;
+
+	dbip = (struct db_info *)spl->data;
+
+	if (dbip->close) {
+	  search_info si;
+	  memset(&si, 0, sizeof(si));
+	  si.file      =  dbip->file;
+	  si.cfgfile   =  dbip->cfgfile;
+	  si.subtype   =  dbip->subtype;
+	  si.ttl       =  dbip->ttl;
+	  si.dbprivate = &dbip->dbprivate;
+
+	  (*dbip->close)(&si, "atexit-close");
+	}
+
+	return 0;
+}
+
+static void dbs_atexit()
+{
+  sp_scan(dbs_atexit_close, (struct spblk *)NULL, spt_databases);
+}
+
+
 int
 run_db(argc, argv)
 	int argc;
@@ -558,10 +624,15 @@ run_db(argc, argv)
 	search_info si;
 	struct spblk *spl;
 
+	if (!dbs_atexit_set) {
+	  atexit(dbs_atexit);
+	  dbs_atexit_set = 1;
+	}
+
 	if (argc == 2 && (argv[1][0] == 'i' || argv[1][0] == 't')) {
 		/* print an index/toc of the databases */
 
-		printf("#DBname   Type{lookup,sub} cache{inuse/max} ttl Flgs File/param\n");
+		printf("#DBname Type{lookup,sub} cache{inuse/max} ttl Flgs File/param\n");
 
 		sp_scan(iclistdbs, (struct spblk *)NULL, spt_databases);
 		return 0;
@@ -610,10 +681,13 @@ run_db(argc, argv)
 		}
 	}
 
-	si.file = dbip->file;
-	si.key  = argv[3];
-	si.subtype = dbip->subtype;
-	si.ttl  = dbip->ttl;
+	memset(&si, 0, sizeof(si));
+	si.file      =  dbip->file;
+	si.cfgfile   =  dbip->cfgfile;
+	si.key       =   argv[3];
+	si.subtype   =  dbip->subtype;
+	si.ttl       =  dbip->ttl;
+	si.dbprivate = &dbip->dbprivate;
 
 	switch (argv[1][0]) {
 	case 'a':	/* add db key value */
@@ -707,21 +781,24 @@ run_db(argc, argv)
  */
 
 conscell *
-db(dbname, argv10)
-	const char *dbname, *argv10[];
+dblookup(dbname, argc, argv30)
+	const char *dbname, *argv30[];
+	const int argc;
 {
 	register int keylen;
 	conscell *l, *ll, *tmp;
 	struct spblk *spl;
 	struct db_info *dbip;
 	char *realkey;
-	const char *key = argv10[0];
+	const char *key = argv30[0];
 	search_info si;
 	struct cache *cache;
 	unsigned long khash;
 	char kbuf[BUFSIZ];	/* XX: */
 	int slen;
 	GCVARS3;
+	int zoptind;
+	int defaultkeys;
 
 	now = time(NULL);
 
@@ -741,7 +818,7 @@ db(dbname, argv10)
 		return NULL;
 	}
 	if (D_db)
-		fprintf(stderr, "%s(%s)\n", dbname, key);
+		fprintf(stderr, "%s(%s)", dbname, key);
 	/* apply flags */
 	realkey = NULL;
 
@@ -764,23 +841,57 @@ db(dbname, argv10)
 	} else
 	  khash = crc32(key);
 
-	si.file    = dbip->file;
-	si.key     = key;
-	si.subtype = dbip->subtype;
-	si.ttl     = dbip->ttl;
-	si.argv10  = argv10;
-	si.argv1   = NULL;
+	memset(&si, 0, sizeof(si));
+	si.file      =  dbip->file;
+	si.cfgfile   =  dbip->cfgfile;
+	si.key       =   key;
+	si.subtype   =  dbip->subtype;
+	si.ttl       =  dbip->ttl;
+	si.argv20    =   argv30;
+	/* si.argv0  =   NULL; */
+	si.flags     =  dbip->flags;
+	si.dbprivate = &dbip->dbprivate;
+	defaultkeys = 0;
+	/* si.defaultkey[0] = NULL; */
 
-	if ((dbip->flags & DB_MODCHECK)
-	    && dbip->modcheckp != NULL && (*dbip->modcheckp)(&si)) {
-		if (dbip->close != NULL) {
-			cacheflush(dbip);
-			(*dbip->close)(&si,"db modcheck");
-			if (dbip->postproc == Indirect) {
-				si.file = dbip->subtype;
-				(*dbip->close)(&si,"db modcheck indirect");
-				si.file = dbip->file;
-			}
+	zoptind = 1;
+	/* Recognize options:
+	   -: defaultval
+	   -- end of options
+	   Rest will be alike sendmail $@ "options"
+	*/
+	while (argv30[zoptind] && zoptind < argc && defaultkeys < 20-1) {
+	  if (STREQ("--",argv30[zoptind])) {
+	    ++zoptind;
+	    break;
+	  }
+	  if (STREQ("-:",argv30[zoptind]) && argv30[zoptind+1]) {
+	    ++zoptind;
+	    si.defaultkey[defaultkeys++] = argv30[zoptind];
+	    ++zoptind;
+	    continue;
+	  }
+	  /* ?? Other options ?? */
+	  break;
+	}
+	si.defaultkey[defaultkeys] = NULL;
+
+	if (D_db) {
+	  int i;
+	  for (i = 0; i < defaultkeys; ++i)
+	    fprintf(stderr, " -: %s", si.defaultkey[i]);
+	  fprintf(stderr, "\n");
+	}
+
+	if ((dbip->flags & DB_MODCHECK) &&
+	    dbip->modcheckp  &&  (*dbip->modcheckp)(&si) &&
+	    dbip->close) {
+		cacheflush(dbip);
+		(*dbip->close)(&si,"db modcheck");
+		if (dbip->postproc == Indirect) {
+			si.file = dbip->subtype;
+			(*dbip->close)(&si,"db modcheck indirect");
+			si.file = dbip->file;
 		}
 	}
 	/* look for the desired result in the cache first */
@@ -799,6 +910,7 @@ db(dbname, argv10)
 						"... expiring %s from cache\n",
 						cache->key);
 				if (cache->key) free(cache->key);
+				cache->keyhash = 0UL;
 				cache->key   = NULL;
 				cache->value = NULL; /* conscell GC does it */
 
@@ -820,7 +932,7 @@ db(dbname, argv10)
 			   match also strings in case sensitive manner. */
 
 			if (cache->keyhash == khash &&
-			    strcmp(cache->key, key) == 0) { /* CACHE HIT! */
+			    STREQ(cache->key, key)) { /* CACHE HIT! */
 
 				/* Move this entry to the head
 				   of the LRU list */
@@ -852,10 +964,8 @@ db(dbname, argv10)
 	  fflush(stderr);
 	}
 
-	if ((dbip->driver == NULL &&
-	     (l = (*dbip->lookup)(&si)) != NULL) ||
-	    (dbip->driver != NULL &&
-	     (l = (*dbip->driver)(dbip->lookup, &si)))) {
+	if ((!dbip->driver  && (l = (*dbip->lookup)(&si))) ||
+	    ( dbip->driver  && (l = (*dbip->driver)(dbip->lookup, &si)))) {
 
 		switch (dbip->postproc) {
 		case Boolean:
@@ -906,14 +1016,14 @@ db(dbname, argv10)
 				free(realkey);
 			}
 			UNGCPRO3;
-			if (si.argv1) free((void*)si.argv1);
-			si.argv1 = NULL;
+			if (si.argv0) free((void*)si.argv0);
+			si.argv0 = NULL;
 			return NULL;
 		}
 	}
 	if (!deferit && dbip->cache_size > 0) {
 		/* insert new cache entry at head of cache */
-		if (dbip->cfree == NULL) {
+		if (! dbip->cfree ) {
 
 			/* No free slots */
 
@@ -956,20 +1066,20 @@ db(dbname, argv10)
 				fprintf(stderr, "\n");
 		}
 		ll = l;
+		/* The "realkey" went into cache. */
 
-	} else if (dbip->cache_size > 0) {
+	} else {
 
-		free(realkey);
 		ll = l;
+		if (realkey) free(realkey);
 
-	} else
-		ll = l;
+	}
 
 	UNGCPRO3;
  post_subst:
 
-	if (ll && dbip->flags & DB_PERCENTSUBST) {
-	  char *buf, *b;
+	if (ll && (dbip->flags & DB_PERCENTSUBST)) {
+	  char *buf, *b, *e;
 	  const char *p, *s;
 	  int i;
 	  int do_subst = 0;
@@ -985,24 +1095,21 @@ db(dbname, argv10)
 	  for (;*p;++p) {
 	    if (*p == '%' && ('0' <= p[1] && p[1] <= '9')) {
 	      ++p;
-	      switch (*p) {
-	      case '0':
+	      s = NULL;
+	      i = (*p - '0');
+
+	      if (i == 0) {
 		s = si.key;
-		break;
-	      case '1':
-		s = si.argv1;
-		if (!s) s = si.argv10[1];
-		break;
-	      default:
-		i = (*p - '0');
-		if (si.argv1) --i;
-		s = si.argv10[i];
-		break;
+		if (si.argv0) s = si.argv0;
+	      } else {
+		--i;
+		if (zoptind+i < argc)
+		  s = si.argv20[zoptind+i];
 	      }
-	      if (s)
+	      if (s) {
 		slen += strlen(s);
-	      if (s)
 		do_subst = 1;
+	      }
 	      continue;
 	    }
 	    ++slen;
@@ -1015,29 +1122,44 @@ db(dbname, argv10)
 	    l = tmp = NULL;
 	    GCPRO3(l, ll, tmp);
 
-	    b = buf = malloc(slen + 3); /* probably 1 extra is enough .. */
+	    slen += 3;
+	    b = buf = malloc(slen); /* probably 1 extra is enough .. */
+	    e = b + slen-2;
 
 	    for (p = ll->cstring; *p; ++p) {
 	      if (*p == '%' && ('0' <= p[1] && p[1] <= '9')) {
 		++p;
-		switch (*p) {
-		case '0':
+		s = NULL;
+		i = (*p - '0');
+		if (i == 0) {
 		  s = si.key;
-		  break;
-		case '1':
-		  s = si.argv1;
-		  if (!s) s = si.argv10[1];
-		  break;
-		default:
-		  i = (*p - '1');
-		  if (si.argv1) --i;
-		  s = si.argv10[i];
-		  break;
+		  if (si.argv0) s = si.argv0;
+		} else {
+		  --i;
+		  if (zoptind+i < argc)
+		    s = si.argv20[zoptind+i];
+		}
+		while (s && *s) {
+		  if (b >= e) {
+		    i = b - buf;
+		    slen += 8;
+		    buf = realloc(buf, slen);
+		    if (!buf) goto malloc_failure;
+		    e = buf + slen-2;
+		    b = buf + i;
+		  }
+		  *b++ = *s++;
 		}
-		if (s)
-		  while (*s) *b++ = *s++;
 		continue;
 	      }
+	      if (b >= e) {
+		i = b - buf;
+		slen += 8;
+		buf = realloc(buf, slen);
+		if (!buf) goto malloc_failure;
+		e = buf + slen-2;
+		b = buf + i;
+	      }
 	      *b++ = *p;
 	    }
 	    *b = 0;
@@ -1045,9 +1167,10 @@ db(dbname, argv10)
 	    l = newstring(dupnstr(buf, b-buf), b-buf);
 	    free(buf);
 
+	  malloc_failure:
 
-	    if (si.argv1) free((void*)si.argv1);
-	    si.argv1 = NULL;
+	    if (si.argv0) free((void*)si.argv0);
+	    si.argv0 = NULL;
 
 	    ll = l;
 
@@ -1084,14 +1207,19 @@ cacheflush(dbip)
 {
 	struct cache *cache, *cnext;
 
-	if (dbip == NULL || dbip->cache_size == 0 || dbip->cache == NULL)
+	if (!dbip || (dbip->cache_size == 0) || (! dbip->cache) )
 		return;
 
 	/* flush cache */
 	
 	for (cache = dbip->cfirst; cache; cache = cnext) {
 		cnext = cache->next;
-		if (cache->key) free(cache->key);
+
+		if (cache->key)   free(cache->key);
+		cache->key   = NULL;
+		cache->value = NULL; /* conscell pointer; garbage
+					collector will free it.. */
+
 		cache->next = dbip->cfree;
 		dbip->cfree = cache;
 	}
@@ -1146,7 +1274,7 @@ dbtype(dbname)
 		return NULL;
 	}
 	for (dbkp = &db_kinds[0];
-	     dbkp < &db_kinds[(sizeof db_kinds)/sizeof (struct db_kind)];
+	     dbkp < &db_kinds[sizeof(db_kinds)/sizeof(db_kinds[0])];
 	     ++dbkp) {
 		if (dbkp->config.lookup == dbip->lookup)
 			return dbkp->name;
@@ -1208,8 +1336,8 @@ find_domain(lookupfn, sip)
 			sip->key = cp-1;
 			l = (*lookupfn)(sip);
 			if (l) {
-				if (sip->argv1) free((void*)sip->argv1);
-				sip->argv1 = dupnstr(buf, sip->key - buf);
+				if (sip->argv0) free((void*)sip->argv0);
+				sip->argv0 = dupnstr(buf, sip->key - buf);
 #ifdef PREDOT_TEST
 #ifndef HAVE_ALLOCA
 				free(buf);
@@ -1240,11 +1368,22 @@ find_domain(lookupfn, sip)
 #endif
 	/* Still failed ?  Try to look for "." */
 	sip->key = ".";
-	l = (*lookupfn)(sip);
+
+	if (sip->defaultkey[0]) {
+	    int i;
+	    for (i = 0 ; sip->defaultkey[i]; ++i) {
+		sip->key = sip->defaultkey[i];
+		l = (*lookupfn)(sip);
+		if (l) break;
+	    }
+	} else {
+	    l = (*lookupfn)(sip);
+	}
+
 	if (l) {
-		if (sip->argv1) free((void*)sip->argv1);
-		sip->argv1 = dupnstr(realkey, keylen);
-		return l;
+	    if (sip->argv0) free((void*)sip->argv0);
+	    sip->argv0 = dupnstr(realkey, keylen);
+	    return l;
 	}
 
 	return NULL;
@@ -1276,8 +1415,8 @@ find_nodot_domain(lookupfn, sip)
 			sip->key = cp;
 			l = (*lookupfn)(sip);
 			if (l) {
-				if (sip->argv1) free((void*)sip->argv1);
-				sip->argv1 = dupnstr(realkey,
+				if (sip->argv0) free((void*)sip->argv0);
+				sip->argv0 = dupnstr(realkey,
 						     sip->key - realkey);
 				return l;
 			}
@@ -1337,7 +1476,7 @@ find_longest_match(lookupfn, sip)
 			  ((oct3 & 255) <<  8) |
 			  ((oct4 & 255)));
 
-		sip->key=buf;
+		sip->key = buf;
 		for (prefix=32, h_mask=0xffffffffL;
 		    prefix>=0;  --prefix, h_mask<<=1) {
 			sprintf((char*)buf,"%u.%u.%u.%u/%d",
@@ -1368,8 +1507,8 @@ find_longest_match(lookupfn, sip)
 				sip->key = cp-1;
 				l = (*lookupfn)(sip);
 				if (l) {
-					if (sip->argv1) free((void*)sip->argv1);
-					sip->argv1 = dupnstr(realkey,
+					if (sip->argv0) free((void*)sip->argv0);
+					sip->argv0 = dupnstr(realkey,
 							     sip->key-realkey);
 					return l;
 				}
@@ -1377,10 +1516,22 @@ find_longest_match(lookupfn, sip)
 		}
 		/* Still failed ?  Try to look for "." */
 		sip->key = ".";
-		l = (*lookupfn)(sip);
+
+
+		if (sip->defaultkey[0]) {
+		    int i;
+		    for (i = 0 ; sip->defaultkey[i]; ++i) {
+			sip->key = sip->defaultkey[i];
+			l = (*lookupfn)(sip);
+			if (l) break;
+		    }
+		} else {
+		    l = (*lookupfn)(sip);
+		}
+
 		if (l) {
-			if (sip->argv1) free((void*)sip->argv1);
-			sip->argv1 = dupnstr(realkey, strlen(realkey));
+			if (sip->argv0) free((void*)sip->argv0);
+			sip->argv0 = dupnstr(realkey, strlen(realkey));
 			return l;
 		}
 	}
