@@ -21,6 +21,7 @@
 #include <pwd.h>			/* for run_homedir() */
 #include <grp.h>			/* for run_grpmems() */
 #include <errno.h>
+#include <sysexits.h>
 
 #include "zmsignal.h"
 #include "zsyslog.h"
@@ -44,6 +45,7 @@ extern conscell *s_value;
 #define ARGCV __((int argc, const char *argv[]))
 static int run_hostname    ARGCV;
 static int run_whataddress ARGCV;
+static int run_iserrormsg  ARGCV;
 static int run_erraddrlog  ARGCV;
 
 static conscell *run_dblookup __((conscell *avl, conscell *il));
@@ -61,8 +63,9 @@ extern int run_stability ARGCV;
 extern int run_process   ARGCV;
 
 static int run_grpmems   ARGCV;
+#if 0 /* dead code */
 static int run_praliases ARGCV;
-static int run_listaddresses ARGCV;
+#endif
 static int zap_DSN_notify ARGCV;
 static int post_zap_DSN_notify ARGCV;
 static int run_homedir   ARGCV;
@@ -102,6 +105,7 @@ struct shCmd fnctns[] = {
 {	"trace",	run_trace,	NULL,	NULL,	0	},
 {	"untrace",	run_trace,	NULL,	NULL,	0	},
 {	"hostname",	run_hostname,	NULL,	NULL,	0	},
+{	"iserrormsg",	run_iserrormsg,	NULL,	NULL,	0	},
 {	"sender",	run_whataddress,NULL,	NULL,	0	},
 {	"recipient",	run_whataddress,NULL,	NULL,	0	},
 {	"erraddron",	run_erraddrlog,	NULL,	NULL,	0	},
@@ -115,8 +119,9 @@ struct shCmd fnctns[] = {
 {	"process",	run_process,	NULL,	NULL,	0	},
 {	"rfc822",	run_rfc822,	NULL,	NULL,	0	},
 {	"groupmembers",	run_grpmems,	NULL,	NULL,	0	},
+#if 0
 {	"printaliases",	run_praliases,	NULL,	NULL,	0	},
-{	"listaddresses",run_listaddresses,NULL,	NULL,	SH_ARGV	},
+#endif
 {	"zapDSNnotify", zap_DSN_notify,	NULL,	NULL,	SH_ARGV	},
 {	"postzapDSNnotify", post_zap_DSN_notify, NULL, NULL, SH_ARGV },
 {	"listexpand",	NULL,	run_listexpand,	NULL,	SH_ARGV	},
@@ -168,22 +173,22 @@ static struct debugind {
 	const char	*name;
 	int		*indicator;
 } buggers[] = {
-	{	"rfc822",		&D_rfc822	},
-	{	"sequencer",		&D_sequencer	},
-	{	"rewrite",		&D_hdr_rewrite	},
-	{	"router",		&D_router	},
-	{	"functions",		&D_functions	},
-	{	"on",			&D_functions	},	/* dup */
-	{	"compare",		&D_compare	},
-	{	"matched",		&D_matched	},
 	{	"assign",		&D_assign	},
-	{	"regexp",		&D_regnarrate	},
-	{	"final",		&D_final	},
-	{	"db",			&D_db		},
 	{	"bind",			&D_bind		},
-	{	"resolv",		&D_resolv	},
-	{	"memory",		&D_alloc	},
+	{	"compare",		&D_compare	},
+	{	"db",			&D_db		},
 	{	"except",		0		},
+	{	"final",		&D_final	},
+	{	"functions",		&D_functions	},
+	{	"matched",		&D_matched	},
+	{	"memory",		&D_alloc	},
+	{	"on",			&D_functions	},	/* dup */
+	{	"regexp",		&D_regnarrate	},
+	{	"resolv",		&D_resolv	},
+	{	"rewrite",		&D_hdr_rewrite	},
+	{	"rfc822",		&D_rfc822	},
+	{	"router",		&D_router	},
+	{	"sequencer",		&D_sequencer	},
 	{	NULL,			0		}
 };
 
@@ -197,26 +202,27 @@ run_trace(argc, argv)
 	struct debugind *dbi;
 	int debug;
 	const char *prog;
+	int rc = 0;
 
 	if (argc == 1) {
 		fprintf(stderr, "Usage: %s all", argv[0]);
 		for (dbi = &buggers[0]; dbi->name != NULL; ++dbi)
 			fprintf(stderr, "|%s", dbi->name);
 		putc('\n', stderr);
-		return 1;
+		return EX_USAGE;
 	}
 	prog = argv[0];
 	debug = (strncmp(*argv, "un", 2) != 0);
 	while (--argc > 0) {
 		++argv;
-		if (strcmp(*argv, "off") == 0 || strcmp(*argv, "all") == 0) {
+		if (STREQ(*argv, "off")  ||  STREQ(*argv, "all")) {
 			for (dbi = &buggers[0]; dbi->name != NULL; ++dbi)
 			    if (dbi->indicator)
 				*(dbi->indicator) = (**argv == (debug?'a':'o'));
 			continue;
 		} else {
 			for (dbi = &buggers[0]; dbi->name != NULL; ++dbi) {
-				if (strcmp(*argv, dbi->name) == 0) {
+				if (STREQ(*argv, dbi->name)) {
 					if (dbi->indicator == NULL)
 					  debug = !debug; /* except */
 					else
@@ -225,11 +231,13 @@ run_trace(argc, argv)
 				}
 			}
 		}
-		if (dbi->name == NULL)
+		if (dbi->name == NULL) {
 			fprintf(stderr, "%s: unknown attribute: %s\n",
 					prog, *argv);
+			rc = EX_USAGE;
+		}
 	}
-	return 0;
+	return rc;
 }
 
 
@@ -307,6 +315,18 @@ run_whataddress(argc, argv)
 	return isRecpntAddr ? 0 : 1;
 }
 
+/*
+ * $(iserrormsg dummyargs)  returns a flag 
+ */
+
+static int
+run_iserrormsg(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	return !isErrorMsg;
+}
+
 /*
  * this is like accton(), but for logging errors in addresses.
  */
@@ -330,7 +350,7 @@ run_erraddrlog(argc, argv)
 		break;
 	default:
 		fprintf(stderr, "Usage: %s [ /path ]\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	return 0;
 }
@@ -348,21 +368,21 @@ run_dblookup(avl, il)
 	conscell *avl, *il; /* Inputs gc protected */
 {
 	conscell *l;
-	const char *argv10[10];
+	const char *argv30[30];
 	int i;
 
-	memset(argv10, 0, sizeof(argv10));
+	memset(argv30, 0, sizeof(argv30));
 
 	il = cdar(avl);
 	if (il == NULL || !STRING(il)) {
-		fprintf(stderr, "Usage: %s key [up_to_9_substitution_elements]\n", car(avl)->string);
+		fprintf(stderr, "Usage: %s key [up_to_19_substitution_elements_or_options]\n", car(avl)->string);
 		return NULL;
 	}
 	i = 0;
-	for (; il && i < 9 && STRING(il); il = cdr(il))
-	  argv10[i++] = il->string;
+	for (; il && i < 30-1 && STRING(il); il = cdr(il))
+	  argv30[i++] = il->string;
 
-	l = db(car(avl)->string, argv10);
+	l = dblookup(car(avl)->string, i, argv30);
 	if (l == NULL)
 		return NULL;
 	return l;
@@ -448,7 +468,7 @@ run_grpmems(argc, argv)
 
 	if (argc != 2) {
 		fprintf(stderr, "Usage: %s groupname\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	grp = zgetgrnam(argv[1]);
 	if (grp == NULL) {
@@ -465,6 +485,7 @@ static struct headerinfo aliashdr = {
 	"aliases", AddressList, Recipient, normal
 };
 
+#if 0 /* dead code, 'zmailer newdb' does these differently... */
 static int
 run_praliases(argc, argv)
 	int argc;
@@ -485,10 +506,10 @@ run_praliases(argc, argv)
 
 	verbose = 0;
 	indexfile = NULL;
-	optind = 1;
+	zoptind = 1;
 	errflg = 0;
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "vo:t");
+		c = zgetopt(argc, (char*const*)argv, "vo:t");
 		if (c == EOF)
 			break;
 		switch (c) {
@@ -496,7 +517,7 @@ run_praliases(argc, argv)
 			++verbose;
 			break;
 		case 'o':
-			indexfile = optarg;
+			indexfile = zoptarg;
 			break;
 		case 't':
 			tabsep = 1;
@@ -506,17 +527,17 @@ run_praliases(argc, argv)
 			break;
 		}
 	}
-	if (errflg || optind != argc - 1) {
+	if (errflg || zoptind != argc - 1) {
 		fprintf(stderr,
 			"Usage: %s [ -v ] [ -o indexoutputfile ] aliasfile\n",
 			argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 
 	e = (struct envelope *)tmalloc(sizeof (struct envelope));
-	if ((e->e_fp = fopen(argv[optind], "r")) == NULL) {
+	if ((e->e_fp = fopen(argv[zoptind], "r")) == NULL) {
 		c = errno;
-		fprintf(stderr, "%s: open(\"%s\"): ", argv[0], argv[optind]);
+		fprintf(stderr, "%s: open(\"%s\"): ", argv[0], argv[zoptind]);
 		errno = c;
 		perror("");
 		status = PERR_BADOPEN;
@@ -524,7 +545,7 @@ run_praliases(argc, argv)
 		setvbuf(e->e_fp, buf, _IOFBF, sizeof buf);
 		osiop = siofds[FILENO(e->e_fp)];
 		siofds[FILENO(e->e_fp)] = NULL;
-		e->e_file = argv[optind];
+		e->e_file = argv[zoptind];
 		status = makeLetter(e, 1);	/* Parse the aliases database
 						   as if all entries were of
 						   same syntax as "To:" et.al.
@@ -593,7 +614,7 @@ run_praliases(argc, argv)
 		size = ftell(stdout);
 		if (size - prevsize > maxsize)
 			maxsize = size - prevsize;
-		if (*cp == *(h->h_pname) && strcmp(cp, h->h_pname) == 0) {
+		if (*cp == *(h->h_pname) && STREQ(cp, h->h_pname)) {
 			fprintf(stderr, "%s: multiple definitions of '%s'.\n",
 					argv[0], h->h_pname);
 			cp = ":";
@@ -630,6 +651,7 @@ run_praliases(argc, argv)
 				argv[0]);
 	return errflg;
 }
+#endif /* ... dead code */
 
 #ifdef DEBUG_FOPEN
 void
@@ -718,7 +740,7 @@ static int zap_DSN_notify(argc, argv)
 	for (lc = l; lc && cdr(lc); pl = &cddr(lc),lc = *pl) {
 	  if (!STRING(lc))
 	    return 0; /* ?? */
-	  if (strcmp("DSN",lc->string)==0) {
+	  if (STREQ("DSN",lc->string)) {
 	    lc = cdr(lc);
 	    if (!lc || !STRING(lc))
 	      return 0;
@@ -887,7 +909,7 @@ static int post_zap_DSN_notify(argc, argv)
 	for (lc = l; lc && cdr(lc); pl = &cddr(lc),lc = *pl) {
 	  if (!STRING(lc))
 	    return 0; /* ?? */
-	  if (strcmp("DSN",lc->string)==0) {
+	  if (STREQ("DSN",lc->string)) {
 	    lc = cdr(lc);
 	    if (!lc || !STRING(lc))
 	      return 0;
@@ -1028,22 +1050,23 @@ run_listexpand(avl, il)
 	int   okaddresses = 0;
 	int   errcount = 0;
 	int   linecnt = 0;
+	int   euid = geteuid();
 	GCVARS3;
 
+
 	il = cdar(avl); /* The CDR (next) of the arg list.. */
 
 	errflag = 0;
 	erroraddress = NULL;
 	comment = "list";
-	optind = 1;
+	zoptind = 1;
 
 	while (il != NULL && STRING(il) && il->string[0] == '-' &&
 	       cdr(il) != NULL && STRING(cdr(il))) {
 	  switch( il->string[1] ) {
 	    case 'c':
 	      comment = (char*)cdr(il)->string;
-	      if (strchr(comment,'\n') != NULL ||
-		  strchr(comment,'\r') != NULL)
+	      if (strchr(comment,'\n') || strchr(comment,'\r'))
 		errflag = 1;
 	      break;
 	    case 'p':
@@ -1051,25 +1074,21 @@ run_listexpand(avl, il)
 	      break;
 	    case 'e':
 	      erroraddress = (char*)cdr(il)->string;
-	      if (strchr(erroraddress,'\n') != NULL ||
-		  strchr(erroraddress,'\r') != NULL)
+	      if (strchr(erroraddress,'\n') || strchr(erroraddress,'\r'))
 		errflag = 1;
 	      break;
 	    case 'E':
 	      if (errors_to != olderrors)
 		free(errors_to);
 	      errors_to = strdup((char*)cdr(il)->string);
-	      if (strchr(errors_to,'\n') != NULL ||
-		  strchr(errors_to,'\r') != NULL)
+	      if (strchr(errors_to,'\n') || strchr(errors_to,'\r'))
 		errflag = 1;
 	      break;
 	    case 'N':
 	      notary = (char *)cdr(il)->string;
-	      if (strchr(notary,'\n') != NULL ||
-		  strchr(notary,'\r') != NULL)
+	      if (strchr(notary,'\n') || strchr(notary,'\r'))
 		errflag = 1;
-	      if (strcmp(notary,"-")==0)
-		no_dsn = 1;
+	      if (STREQ(notary,"-")) no_dsn = 1;
 	      break;
 	    default:
 	      errflag = 1;
@@ -1087,7 +1106,7 @@ run_listexpand(avl, il)
 	if (errflag || cnt < 3 || cnt > 5 ||
 	    !STRING(il) || !STRING(cdr(il)) || !STRING(cddr(il)) ) {
 		fprintf(stderr,
-			"Usage: %s [ -e error-address ] [ -E errors-to-address ] [-p privilege] [ -c comment ] [ -N notarystring ] $attribute $localpart $origaddr [$plustail [$domain]]< /file/path \n",
+			"Usage: %s [ -e error-address ] [ -E errors-to-address ] [-p privilege] [ -c comment ] [ -N notarystring ] $attribute $localpart $origaddr [$plustail [$domain]] [ [<] /file/path ]\n",
 		car(avl)->string);
 		if (errors_to != olderrors)
 		  free(errors_to);
@@ -1131,7 +1150,7 @@ run_listexpand(avl, il)
 		errors_to = olderrors;
 		return NULL;
 	}
-	fd2 = dup(0);	/* Copy the file handle for latter returning
+	fd2 = dup(0);	/* Copy the file handle for later returning
 			   AFTER an fclose() has closed the primary
 			   copy.. */
 
@@ -1203,18 +1222,24 @@ run_listexpand(avl, il)
 		  ++errcount;
 		  if (hs.h_stamp == BadHeader) {
 		    if (erroraddress != NULL) {
-		      if (!iserrmessage()) {
+		      if (!isErrChannel && !isErrorMsg) {
 			if (mfp == NULL) {
-			  if ((mfp = mail_open(MSG_RFC822)) != NULL) {
+			  /* We are likely running under 'runas ..'
+			     Do reset now to ROOT, then to TRUSTED, ... */
+			  runasrootuser();
+			  runastrusteduser();
+			  mfp = mail_open(MSG_RFC822);
+			  runasrootuser();
+			  if (mfp != NULL) {
 			    osiop = siofds[FILENO(mfp)];
 			    siofds[FILENO(mfp)] = NULL;
 			    fprintf(mfp, "channel error\n");
+			    fprintf(mfp, "errormsg\n");
 			    fprintf(mfp, "to <%s>\n", erroraddress);
-			    fprintf(mfp, "to <postmaster>\n");
+			    fprintf(mfp, "to <postoffice>\n");
 			    fprintf(mfp, "env-end\n");
 			    fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
 			    fprintf(mfp, "To: %s\n", erroraddress);
-			    fprintf(mfp, "Cc: The Post Office <postmaster>\n");
 			    fprintf(mfp, "Subject: Error in %s\n", comment);
 			    fprintf(mfp, "Precedence: junk\n\n");
 			    /* Print the report: */
@@ -1226,8 +1251,10 @@ run_listexpand(avl, il)
 			  hdr_errprint(e, &hs, mfp, comment);
 			}
 		      }
+#if 0
 		      if (errcount == 1) /* At the first time only! */
 			printf("%s\n", erroraddress);
+#endif
 		    }
 		    fprintf(stderr,"Input file line number %d:\n",linecnt);
 		    hdr_errprint(e, &hs, stderr, comment);
@@ -1258,7 +1285,9 @@ run_listexpand(avl, il)
 	dup2(fd2,0);	/* Return the fd to descriptor 0..		*/
 	close(fd2);	/* .. and discard the backup copy..		*/
 
-	if (okaddresses == 0) {	/* if the file is empty, use error address */
+	if (okaddresses == 0 && !isErrChannel && !isErrorMsg) {
+		/* If the file is empty, use error address ...
+		   Except when this is already an error mesage, say nothing! */
 
 		if (erroraddress == NULL)
 			erroraddress = "postmaster";
@@ -1286,27 +1315,33 @@ run_listexpand(avl, il)
 		}
 
 		if (mfp == NULL) {
-		  if ((mfp = mail_open(MSG_RFC822)) != NULL) {
+		  /* We are likely running under 'runas ..'
+		     Do reset now to ROOT, then to TRUSTED, ... */
+		  runasrootuser();
+		  runastrusteduser();
+		  mfp = mail_open(MSG_RFC822);
+		  runasrootuser();
+		  if (mfp != NULL) {
 		    osiop = siofds[FILENO(mfp)];
 		    siofds[FILENO(mfp)] = NULL;
 		    fprintf(mfp, "channel error\n");
+		    fprintf(mfp, "errormsg\n");
 		    fprintf(mfp, "to <%s>\n", erroraddress);
-		    fprintf(mfp, "to <postmaster>\n");
+		    fprintf(mfp, "to <postoffice>\n");
 		    fprintf(mfp, "env-end\n");
 		    fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
 		    fprintf(mfp, "To: %s\n", erroraddress);
-		    fprintf(mfp, "Cc: The Post Office <postmaster>\n");
 		    fprintf(mfp, "Subject: Error in %s\n", comment);
 		    fprintf(mfp, "Precedence: junk\n\n");
 		    /* Print the report: */
 		    fprintf(mfp,"NO valid recipient addresses!\n");
-		    fprintf(mfp,"Verify source file protection/ownership/access-path\n");
-		    fprintf(mfp,"Current effective UID = %d\n", geteuid());
+		    fprintf(mfp,"Verify source file protection/ownership/access-path, and content.\n");
+		    fprintf(mfp,"Current effective UID = %d\n", euid);
 		  }
 		} else { /* mfp != NULL */
 		  fprintf(mfp,"\nNO valid recipient addresses!\n");
-		  fprintf(mfp,"Verify source file protection/ownership/access-path\n");
-		  fprintf(mfp,"Current effective UID = %d\n", geteuid());
+		  fprintf(mfp,"Verify source file protection/ownership/access-path, and content.\n");
+		  fprintf(mfp,"Current effective UID = %d\n", euid);
 		}
 	}
 
@@ -1505,6 +1540,7 @@ run_listexpand(avl, il)
 		   fwrite(buf,1,len+1,stdout); */
 	}
 
+#if 0
 	if (al == NULL) { /* ERROR! NO ADDRESSES! */
 	  int slen;
 	  al = conststring("error", 5);
@@ -1514,6 +1550,7 @@ run_listexpand(avl, il)
 	  al = ncons(al);
 	  al = ncons(al);
 	}
+#endif
 	UNGCPRO3;
 
 	if (errors_to != olderrors)
@@ -1524,289 +1561,6 @@ run_listexpand(avl, il)
 } /* end-of: run_listexpand() */
 
 
-static int
-run_listaddresses(argc, argv)
-	int argc;
-	const char *argv[];
-{
-	struct header hs;
-	struct envelope *e;
-	struct address *ap, *aroot = NULL, **atail = &aroot;
-	token822 *t;
-	struct addr *pp;
-	int c, n, errflag, stuff;
-	const char *comment, *erroraddress;
-	char *s, *s_end;
-	struct siobuf *osiop = NULL;
-	FILE *mfp = NULL, *fp;
-	char buf[4096];
-	int fd2;
-	int okaddresses = 0;
-	int errcount = 0, linecnt = 0;
-	char *old_errorsto = errors_to;
-
-	errflag = 0;
-	erroraddress = NULL;
-	comment = "list";
-	optind = 1;
-
-	while (1) {
-		c = getopt(argc, (char*const*)argv, "c:e:E:");
-		if (c == EOF)
-			break;
-		switch (c) {
-		case 'c':
-			comment = optarg;
-			break;
-		case 'e':
-			erroraddress = optarg;
-			break;
-		case 'E':
-			if (errors_to != old_errorsto)
-			  free(errors_to);
-			errors_to = (void*)strdup(optarg);
-			break;
-		default:
-			++errflag;
-			break;
-		}
-	}
-	if (errflag) {
-	  fprintf(stderr,
-		  "Usage: %s [ -e error-address ] [ -E errors-to-address ] [ -c comment ]\n",
-		  argv[0]);
-	  if (errors_to != old_errorsto)
-	    free(errors_to);
-	  errors_to = old_errorsto;
-	  return 1;
-	}
-	e = (struct envelope *)tmalloc(sizeof (struct envelope));
-	e->e_nowtime = now;
-	e->e_file = argv[0];
-	/* we only need sensible `e' if its a time stamp header,
-	   which it ain't, so "listaddress" is just fine for e_file .. */
-
-	hs.h_descriptor = &aliashdr;
-	hs.h_pname = argv[0];
-
-	initzline(4096);
-
-	/*
-	 * These hoops are so we can do this for both real stdin and
-	 * the fake I/O stuff in ../libsh.
-	 */
-	if ((fp = fdopen(0, "r")) == NULL) {
-		fprintf(stderr, "%s: fdopen failed\n", argv[0]);
-#ifdef DEBUG_FOPEN
-		report_fds(stderr);
-#endif
-		if (errors_to != old_errorsto)
-		  free(errors_to);
-		errors_to = old_errorsto;
-		return 1;
-	}
-	fd2 = dup(0);	/* Copy the file handle for latter returning
-			   AFTER an fclose() has closed the primary
-			   copy.. */
-
-	/* use a constant buffer to avoid memory leaks from stdio usage */
-	setvbuf(fp, buf, _IOFBF, sizeof buf);
-	c = 0;
-	while ((n = zgetline(fp)) > 0) {
-		++linecnt;
-		/*
-		 * For sendmail compatibility, addresses may not cross line
-		 * boundaries, and line boundary is just as good an address
-		 * separator as comma is, sigh.
-		 * Also lines beginning with '#' are comments, and blank
-		 * lines are (sortof) comments too.
-		 */
-		if (zlinebuf[n-1] == '\n')
-			--n;
-		if (zlinebuf[0] == '#')
-			continue;
-
-		stuff = 0;
-		s_end = &zlinebuf[n];
-		for (s = zlinebuf; s < s_end; ++s) {
-			if (isascii(*s) && !isspace(*s))
-				c = stuff = 1;
-			if (*s == ',')
-				c = 0;
-		}
-		if (!stuff)
-			continue;
-		if (c) {
-			zlinebuf[n] = ',';
-			++n;
-		}
-
-/* #ifdef SENDMAIL_COMPABILITY_KLUDGE */
-		/**
-		 * Additional sendmail compability kludge ++haa
-		 * If address starts with \, zap the \ character.
-		 * This is just to not to force people to edit their
-		 * .forwards :-(  "\haa" works as expected  :-)
-		 **/
-		if (zlinebuf[0] == '\\') zlinebuf[0] = ' ';
-		for (s = zlinebuf+1; s < s_end; ++s) {
-		  if (s[-1] == ',' && s[0]=='\\') s[0] = ' ';
-		}
-		/* end of sendmail compatibility kluge */
-/* #endif */
-		/* create h->h_lines */
-		/*
-		 * It is best to maintain a line at a time as tokens,
-		 * so errors will print out nicely.
-		 */
-
-		hs.h_lines = t = makeToken(zlinebuf, n);
-		t->t_type = Line;
-
-		/* fix up any trailing comma (more sendmail
-		   compatibility kluges) */
-		s = (char*)t->t_pname + TOKENLEN(t)-1;
-		if (c && (*s == ',' || *s == '\n'))
-		  *s = '\0';
-
-		hs.h_contents = hdr_scanparse(e, &hs, 1, 1);
-		hs.h_stamp = hdr_type(&hs);
-
-		if (hs.h_stamp == BadHeader || hs.h_contents.a == NULL) {
-		  ++errcount;
-		  if (hs.h_stamp == BadHeader) {
-		    if (erroraddress != NULL) {
-		      if (!iserrmessage()) {
-			if (mfp == NULL) {
-			  if ((mfp = mail_open(MSG_RFC822)) != NULL) {
-			    osiop = siofds[FILENO(mfp)];
-			    siofds[FILENO(mfp)] = NULL;
-			    fprintf(mfp, "channel error\n");
-			    fprintf(mfp, "to <%s>\n", erroraddress);
-			    fprintf(mfp, "to <postmaster>\n");
-			    fprintf(mfp, "env-end\n");
-			    fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
-			    fprintf(mfp, "To: %s\n", erroraddress);
-			    fprintf(mfp, "Cc: The Post Office <postmaster>\n");
-			    fprintf(mfp, "Subject: Error in %s\n", comment);
-			    fprintf(mfp, "Precedence: junk\n\n");
-			    /* Print the report: */
-			    fprintf(mfp,"Input file line number %d:\n",linecnt);
-			    hdr_errprint(e, &hs, mfp, comment);
-			  }
-			} else { /* mfp != NULL */
-			  fprintf(mfp,"Input file line number %d:\n",linecnt);
-			  hdr_errprint(e, &hs, mfp, comment);
-			}
-		      }
-		      if (errcount == 1) /* At the first time only! */
-			printf("%s\n", erroraddress);
-		    }
-		    fprintf(stderr,"Input file line number %d:\n",linecnt);
-		    hdr_errprint(e, &hs, stderr, comment);
-		  } else {		/* if (hs.h_contents.a == NULL) */
-#if 0
-		    if (errcount == 1) {
-		      /* Print only for the first intance.. */
-		      if (erroraddress != NULL)
-			printf("%s\n", erroraddress);
-		      fprintf(stderr, "%s: null input on line\n", argv[0]);
-		    }
-#endif
-		  }
-		  continue;
-		}
-
-		*atail = hs.h_contents.a;
-		while (*atail != NULL)
-		  atail = &((*atail)->a_next);
-
-		++okaddresses;
-	}
-
-	fclose(fp);	/* Now we discard the stdio buffers, but not the
-			   fd number 0!  Actually we use a copy of it..	*/
-	dup2(fd2,0);	/* Return the fd to descriptor 0..		*/
-	close(fd2);	/* .. and discard the backup copy..		*/
-
-	if (okaddresses == 0) {	/* if the file is empty, use error address */
-
-		if (erroraddress == NULL)
-			erroraddress = "postmaster";
-
-		hs.h_lines = t = makeToken(erroraddress, strlen(erroraddress));
-		t->t_type = Line;
-
-		/* fix up any trailing comma (more sendmail
-		   compatibility kluges) */
-		s = (char*)t->t_pname + TOKENLEN(t)-1;
-		if (c && (*s == ',' || *s == '\n'))
-		  *s = '\0';
-
-		hs.h_contents = hdr_scanparse(e, &hs, 1, 1);
-		hs.h_stamp = hdr_type(&hs);
-
-		if (hs.h_stamp == BadHeader || hs.h_contents.a == NULL) {
-		  /* OUTCH!  Even the "erroraddress" parameter is illegal! */
-		  fprintf(stderr,"%s: input parameters bad, empty input, even 'erroraddress' bad!\n", argv[0]);
-		} else {
-		  *atail = hs.h_contents.a;
-		  while (*atail != NULL)
-		    atail = &((*atail)->a_next);
-		}
-
-		if (mfp == NULL) {
-		  if ((mfp = mail_open(MSG_RFC822)) != NULL) {
-		    osiop = siofds[FILENO(mfp)];
-		    siofds[FILENO(mfp)] = NULL;
-		    fprintf(mfp, "channel error\n");
-		    fprintf(mfp, "to <%s>\n", erroraddress);
-		    fprintf(mfp, "to <postmaster>\n");
-		    fprintf(mfp, "env-end\n");
-		    fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
-		    fprintf(mfp, "To: %s\n", erroraddress);
-		    fprintf(mfp, "Cc: The Post Office <postmaster>\n");
-		    fprintf(mfp, "Subject: Error in %s\n", comment);
-		    fprintf(mfp, "Precedence: junk\n\n");
-		    /* Print the report: */
-		    fprintf(mfp,"NO valid recipient addresses!\n");
-		    fprintf(mfp,"Verify source file protection/ownership/access-path\n");
-		    fprintf(mfp,"Current effective UID = %d\n", geteuid());
-		  }
-		} else { /* mfp != NULL */
-		  fprintf(mfp,"\nNO valid recipient addresses!\n");
-		  fprintf(mfp,"Verify source file protection/ownership/access-path\n");
-		  fprintf(mfp,"Current effective UID = %d\n", geteuid());
-		}
-
-	}
-
-	if (mfp != NULL) {
-	  siofds[FILENO(mfp)] = osiop;
-	  mail_close(mfp);
-	}
-
-	for (ap = aroot; ap != NULL; ap = ap->a_next) {
-
-		for (pp = ap->a_tokens; pp != NULL; pp = pp->p_next)
-			if (pp->p_type == anAddress)
-				break;
-		if (pp == NULL)
-			continue;
-
-		pureAddress(stdout, pp);
-		/* printAddress(stdout, pp, 0); */
-		putc('\n', stdout);
-	}
-	/*
-	 * XX: If the alias expansion was a mail group,
-	 * we didn't print anything.
-	 */
-	if (errors_to != old_errorsto)
-	  free(errors_to);
-	errors_to = old_errorsto;
-	return 0;
-}
 
 static int
 run_homedir(argc, argv)
@@ -1819,7 +1573,7 @@ run_homedir(argc, argv)
 
 	if (argc != 2) {
 		fprintf(stderr, "Usage: %s name\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 
 	pw = zgetpwnam(argv[1]);
@@ -1852,7 +1606,7 @@ run_822date(argc, argv)
 	time_t dnow;
 
 	time(&dnow);
-	if (argc == 2 && strcmp(argv[1], "-s") == 0)
+	if (argc == 2 && STREQ(argv[1], "-s"))
 		printf("%ld\n", dnow);
 	else
 		printf("%s", rfc822date(&dnow));
@@ -1886,7 +1640,7 @@ run_filepriv(argc, argv)
 		fprintf(stderr, "Usage: %s [-M maxperm] pathname [ uid ]\n", argv0);
 		if (maxperm & (~0664))
 		  fprintf(stderr, "       maxperm must be 664 or stricter!\n");
-		return 1;
+		return EX_USAGE;
 	}
 	file = argv[1];
 	if (argc == 3 && isascii(argv[2][0]) && isdigit(argv[2][0])) {
@@ -2001,7 +1755,7 @@ run_runas(argc, argv)
 
 	if (argc < 3) {
 		fprintf(stderr, "Usage: %s user function [args...]\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	cp = argv[1];
 	if (*cp == '-')
@@ -2016,8 +1770,8 @@ run_runas(argc, argv)
 	if (!initeduid)
 		myuid = geteuid();
 	if (myuid == 0 && uid != 0) {
-		if (setreuid(0, uid) < 0) {
-			fprintf(stderr, "%s: setuid(%d): %s\n",
+		if (SETEUID(uid) < 0) {
+			fprintf(stderr, "%s: seteuid(%d): %s\n",
 				argv[0], uid, strerror(errno));
 			return 1;
 		}
@@ -2025,7 +1779,7 @@ run_runas(argc, argv)
 	/* must be builtin or defined function */
 	r = s_apply(argc - 2, argv + 2); /* within:  run_runas() */
 	if (myuid == 0 && uid != 0) {
-		if (setreuid(0, 0) < 0)
+		if (SETEUID(0) < 0)
 			abort(); /* user-identity change failed! */
 	}
 
@@ -2044,7 +1798,7 @@ run_cat(argc, argv)
 
 	if (argc < 2) {
 		fprintf(stderr, "Usage: %s [filenames ...]\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 
 	for ( ;argv[1] != NULL; ++argv) {
@@ -2075,7 +1829,7 @@ run_uid2login(argc, argv)
 {
 	if (argc != 2 || !isdigit(argv[1][0])) {
 		fprintf(stderr, "Usage: %s uid\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	printf("%s\n", uidpwnam(atoi(argv[1])));
 	return 0;
@@ -2088,7 +1842,7 @@ run_login2uid(argc, argv)
 {
 	if (argc != 2) {
 		fprintf(stderr, "Usage: %s login\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	printf("%d\n", login_to_uid(argv[1]));
 	return 0;
@@ -2105,7 +1859,7 @@ run_basename(argc, argv)
 	if (argc == 1) {
 		fprintf(stderr, "Usage: %s pathname suffix-to-strip\n",
 				argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 	cp = strrchr(argv[1], '/');
 	if (cp == NULL)
@@ -2113,7 +1867,7 @@ run_basename(argc, argv)
 	else
 		++cp;
 	if (argc > 2 && (len = strlen(cp) - strlen(argv[2])) > 0) {
-		if (strcmp(cp + len, argv[2]) == 0) {
+		if (STREQ(cp + len, argv[2])) {
 			while (len-- > 0)
 				putchar(*cp++);
 			putchar('\n');
@@ -2132,26 +1886,26 @@ run_syslog(argc, argv)
 	int c;
 	int prio = LOG_INFO;
 	int errflg = 0;
-	optind = 1;
+	zoptind = 1;
 
-	while ((c = getopt(argc, (char*const*)argv, "p:")) != EOF) {
+	while ((c = zgetopt(argc, (char*const*)argv, "p:")) != EOF) {
 		switch (c) {
 		case 'p':	/* priority */
-			if(!strcmp(optarg, "debug")) {
+			if(STREQ(zoptarg, "debug")) {
 				prio = LOG_DEBUG;
-			} else if(!strcmp(optarg, "info")) {
+			} else if(STREQ(zoptarg, "info")) {
 				prio = LOG_INFO;
-			} else if(!strcmp(optarg, "notice")) {
+			} else if(STREQ(zoptarg, "notice")) {
 				prio = LOG_NOTICE;
-			} else if(!strcmp(optarg, "warning")) {
+			} else if(STREQ(zoptarg, "warning")) {
 				prio = LOG_WARNING;
-			} else if(!strcmp(optarg, "err")) {
+			} else if(STREQ(zoptarg, "err")) {
 				prio = LOG_ERR;
-			} else if(!strcmp(optarg, "crit")) {
+			} else if(STREQ(zoptarg, "crit")) {
 				prio = LOG_CRIT;
-			} else if(!strcmp(optarg, "alert")) {
+			} else if(STREQ(zoptarg, "alert")) {
 				prio = LOG_ALERT;
-			} else if(!strcmp(optarg, "emerg")) {
+			} else if(STREQ(zoptarg, "emerg")) {
 				prio = LOG_EMERG;
 			} else {
 				++errflg;
@@ -2163,11 +1917,11 @@ run_syslog(argc, argv)
 		}
 	}
 
-	if (errflg || optind != argc - 1) {
+	if (errflg || zoptind != argc - 1) {
 		fprintf(stderr, "Usage: %s [-p prio] string\n", argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
-	zsyslog((prio, "%s", argv[optind]));
+	zsyslog((prio, "%s", argv[zoptind]));
 	return 0;
 }
 
@@ -2179,11 +1933,11 @@ run_recase(argc, argv)
 	char *cp;
 	int c, flag, errflg, action = 0;
 
-	optind = 1;
+	zoptind = 1;
 	errflg = 0;
 
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "ulp");
+		c = zgetopt(argc, (char*const*)argv, "ulp");
 		if (c == EOF)
 			break;
 		switch (c) {
@@ -2197,22 +1951,22 @@ run_recase(argc, argv)
 			break;
 		}
 	}
-	if (errflg || optind != argc - 1) {
-		fprintf(stderr, "Usage: %s [ -u | -l | -p ] string\n",
+	if (errflg || zoptind != argc - 1) {
+		fprintf(stderr, "Usage: %s [ -u | -l | -p ] -- string\n",
 				argv[0]);
-		return 1;
+		return EX_USAGE;
 	}
 
 	switch (action) {
 	case 'u':
-		strupper((char*)argv[optind]);
+		strupper((char*)argv[zoptind]);
 		break;
 	case 'l':
-		strlower((char*)argv[optind]);
+		strlower((char*)argv[zoptind]);
 		break;
 	case 'p':
 		flag = 1;
-		for (cp = (char*)argv[optind]; *cp != '\0'; ++cp) {
+		for (cp = (char*)argv[zoptind]; *cp != '\0'; ++cp) {
 			if (isascii(*cp) && isalnum(*cp)) {
 				if (flag && islower(*cp))
 					*cp = toupper(*cp);
@@ -2224,7 +1978,7 @@ run_recase(argc, argv)
 		}
 		break;
 	}
-	printf("%s\n", argv[optind]);
+	printf("%s\n", argv[zoptind]);
 	return 0;
 }
 
@@ -2287,7 +2041,7 @@ run_squirrel(argc, argv)
 		} else
 			flag = 1;
 		for (j = 0; j < (sizeof fyitable / sizeof fyitable[0]); ++j) {
-			if (cistrcmp(argv[i], fyitable[j].fyiname) == 0) {
+			if (CISTREQ(argv[i], fyitable[j].fyiname)) {
 				fyitable[j].fyisave = flag;
 				j = -1;
 				break;
@@ -2307,7 +2061,7 @@ run_squirrel(argc, argv)
 				fprintf(stderr, " %s", fyitable[j].fyiname);
 		}
 		fprintf(stderr, " ]\n");
-		return 1;
+		return EX_USAGE;
 	}
 	return 0;
 }
@@ -2326,7 +2080,7 @@ run_822syntax(argc, argv)
 	/*char buf[4096];*/
 
 	if (argc != 2)
-		return 2;
+		return EX_USAGE;
 	es.e_nowtime = now;
 	es.e_file = argv[0];
 	hs.h_descriptor = &addrhdr;
@@ -2337,7 +2091,7 @@ run_822syntax(argc, argv)
 	hs.h_contents = hdr_scanparse(&es, &hs, 1, 0);
 	hs.h_stamp = hdr_type(&hs);
 	if (hs.h_stamp == BadHeader) {
-		hdr_errprint(&es, &hs, stderr, "RFC822/976");
+		hdr_errprint(&es, &hs, stderr, "RFC822/976/2822");
 		return 1;
 	} else if (hs.h_contents.a == NULL)
 		return 1;
@@ -2362,31 +2116,31 @@ run_condquote_(argc, argv, condq)
 	/* We remove quotes when they are not needed, and add them when
 	   they really are needed! */
 
-	optind = 1;
+	zoptind = 1;
 	while (1) {
-	  c = getopt(argc, (char*const*)argv, "s:a:");
+	  c = zgetopt(argc, (char*const*)argv, "s:a:");
 	    
 	  if (c == EOF) break;
 	  switch (c) {
 	  case 's':
-	    spc = *optarg; /* First char only */
+	    spc = *zoptarg; /* First char only */
 	    break;
 	  case 'a':
-	    appstr = optarg;
+	    appstr = zoptarg;
 	    break;
 	  default:
 	    ++errflg;
 	    break;
 	  }
 	}
-	if (errflg || optind != argc - 1) {
+	if (errflg || zoptind != argc - 1) {
 	  fprintf(stderr,
 		  "Usage: %s [ -s SPCCHR ] [ -a APPENDSTR ] string\n",
 		  argv[0]);
-	  return 1;
+	  return EX_USAGE;
 	}
 
-	s = argv[optind];
+	s = argv[zoptind];
 
 	mustquote = rfc822_mustquote(s, spc);
 	/* A bitset:
