@@ -1,35 +1,20 @@
 /*
- *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
- *	This will be free software, but only when it is finished.
+ *      A component of ZMailer
  *
- *	Copyright 1996-2000 Matti Aarnio
+ *	Copyright 1996-2002 Matti Aarnio
  */
 
 /* LINTLIBRARY */
 
 #include "mailer.h"
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+
+#include "sleepycatdb.h"
+#ifdef HAVE_DB
+
 #ifdef HAVE_FCNTL_H
 # include <fcntl.h>
 #endif
 
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
-# include <db_185.h>
-#else
-#ifdef HAVE_DB2_DB_H
-# include <db2/db.h>
-#else
-#ifdef HAVE_DB_H
-# include <db.h>
-#else
-#ifdef HAVE_DB1_DB_H
-# include <db1/db.h>
-#endif
-#endif
-#endif
-#endif
-#endif
 
 #include <sys/file.h>
 #include "search.h"
@@ -38,10 +23,10 @@
 #include "libc.h"
 #include "libsh.h"
 
+#include <errno.h>
 extern int errno;
 extern int deferit;
 
-
 /*
  * Flush buffered information from this database, close any file descriptors.
  */
@@ -51,79 +36,128 @@ close_bhash(sip,comment)
 	search_info *sip;
 	const char *comment;
 {
-	DB *db;
-	struct spblk *spl = NULL;
-	spkey_t symid;
+	ZSleepyPrivate *prv;
 
-	if (sip->file == NULL)
+	if (*(sip->dbprivate) == NULL )
 		return;
-	symid = symbol_lookup_db(sip->file, spt_files->symbols);
-	if ((spkey_t)0 != symid)
-	  spl = sp_lookup(symid, spt_modcheck);
-	if (spl != NULL)
-	  sp_delete(spl, spt_modcheck);
-	spl = sp_lookup(symid, spt_files);
-	if (spl == NULL || (db = (DB *)spl->data) == NULL)
-		return;
-#ifdef HAVE_DB_OPEN2
-	(db->close)(db,0);
-#else
-	(db->close)(db);
-#endif
-	symbol_free_db(sip->file, spt_files->symbols);
-	sp_delete(spl, spt_files);
+
+	prv = *(sip->dbprivate);
+
+	SLEEPYCATDBCLOSE(prv->db);
+
+	zsleepyprivatefree(prv);
+
+	*(sip->dbprivate) = NULL;
 }
 
 
-static DB * open_bhash __((search_info *, int, const char *));
-static DB *
-open_bhash(sip, flag, comment)
+static ZSleepyPrivate * open_bhash __((search_info *, int, const char *));
+static ZSleepyPrivate *
+open_bhash(sip, roflag, comment)
 	search_info *sip;
-	int flag;
+	int roflag;
 	const char *comment;
 {
-	DB *db = NULL;
-	struct spblk *spl;
-	spkey_t symid;
 	int i;
+	ZSleepyPrivate **prvp = (ZSleepyPrivate **)sip->dbprivate;
+	DB *db = NULL;
+
+	if (sip->cfgfile) {
+		/* read the related configuration file, e.g.
+		   information about environment, etc.. */
+	}
 
 	if (sip->file == NULL)
 		return NULL;
 
-	symid = symbol_db(sip->file, spt_files->symbols);
-	spl = sp_lookup(symid, spt_files);
-	if (spl != NULL && flag != spl->mark)
-		close_bhash(sip,"open_bhash");
-	if (spl == NULL || (db = (DB *)spl->data) == NULL) {
-		for (i = 0; i < 3; ++i) {
-#ifdef HAVE_DB_OPEN2
-		  int err;
-		  db = NULL;
-		  err = db_open(sip->file, DB_HASH,
-				DB_NOMMAP|((flag == O_RDONLY) ? DB_RDONLY:DB_CREATE),
-				0644, NULL, NULL, &db);
-#else
-		  db = dbopen(sip->file, flag, 0, DB_HASH, NULL);
-#endif
-		  if (db != NULL)
-		    break;
-		  sleep(1); /* Open failed, retry after a moment */
-		}
-		if (db == NULL) {
-			++deferit;
-			v_set(DEFER, DEFER_IO_ERROR);
-			fprintf(stderr, "%s: cannot open %s!\n",
-					comment, sip->file);
-			return NULL;
-		}
-		if (spl == NULL)
-			sp_install(symid, (void *)db, flag, spt_files);
-		else {
-			spl->data = (void *)db;
-			spl->mark = flag;
-		}
+
+
+	if (*prvp && roflag != (*prvp)->roflag)
+ 		close_bhash(sip,"open_bhash");
+
+	if (*prvp) db = (*prvp)->db;
+
+	if (db == NULL) {
+
+	    char *phase = "";
+	    /* Three attempts to open it.. */
+	    for (i = 0; i < 3; ++i) {
+
+		int err;
+
+		*prvp = zsleepyprivateinit(sip->file, sip->cfgfile,
+					   DB_HASH);
+
+		if (!*prvp) break; /* URGH!! Out of memory! */
+
+	        err = zsleepyprivateopen(*prvp, roflag, 0644, &phase);
+		db = (*prvp)->db;
+
+		if (db)  break;
+
+		if (*prvp)
+		  zsleepyprivatefree(*prvp);
+		*prvp = NULL;
+
+		sleep(1); /* Open failed, retry after a moment */
+	    }
+
+	    /* Still failed ?? */
+	    if (db == NULL) {
+		++deferit;
+		v_set(DEFER, DEFER_IO_ERROR);
+		fprintf(stderr, "%s: cannot open%s %s!\n",
+			comment, phase, sip->file);
+
+		if (*prvp)
+		  zsleepyprivatefree(*prvp);
+
+		return NULL;
+	    }
 	}
-	return db;
+
+
+	/* Got it open ? */
+
+	if (db != NULL) {
+
+	    /* Prepare for  modp_bhash()  tests. */
+
+	    struct stat stbuf;
+	    int fd = -1, err = 0;
+
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
+	    err = (db->fd)(db, &fd);
+	    if (fstat(fd, &stbuf) < 0) {
+		fprintf(stderr, "open_bhash: cannot fstat(\"%s\"(%d))!  err=%d/%s (%s/%s)\n",
+			sip->file, fd, err, errno,
+			db_strerror(err), strerror(errno));
+
+		SLEEPYCATDBCLOSE(db);
+		if (*prvp)
+		  zsleepyprivatefree(*prvp);
+
+		return NULL;
+	    }
+#else
+	    fd = (db->fd)(db);
+	    if (fstat(fd, &stbuf) < 0) {
+		fprintf(stderr, "open_bhash: cannot fstat(\"%s\"/%d))!  err=%d (%s)\n",
+			sip->file, fd, errno, strerror(errno));
+
+		SLEEPYCATDBCLOSE(db);
+		if (*prvp)
+		  zsleepyprivatefree(*prvp);
+
+		return NULL;
+	    }
+#endif
+
+	    (*prvp)->mtime = stbuf.st_mtime;
+
+	}
+
+	return *prvp;
 }
 
 
@@ -135,6 +169,7 @@ conscell *
 search_bhash(sip)
 	search_info *sip;
 {
+	ZSleepyPrivate *prv;
 	DB *db;
 	DBT val, key;
 	int retry, rc;
@@ -143,15 +178,18 @@ search_bhash(sip)
 #if 0
 reopen:
 #endif
-	db = open_bhash(sip, O_RDONLY, "search_bhash");
-	if (db == NULL)
+	prv = open_bhash(sip, O_RDONLY, "search_bhash");
+	if (prv == NULL)
 	  return NULL; /* Huh! */
 
-	memset(&val, 0, sizeof(val));
+	db = prv->db;
+
 	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
 
 	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
+
 #ifdef DB_INIT_TXN
 	rc = (db->get)(db, NULL, &key, &val, 0);
 #else
@@ -186,13 +224,15 @@ add_bhash(sip, value)
 	DB *db;
 	DBT val, key;
 	int rc;
+	ZSleepyPrivate *prv;
 
-	db = open_bhash(sip, O_RDWR, "add_bhash");
-	if (db == NULL)
+	prv = open_bhash(sip, O_RDWR, "add_bhash");
+	if (prv == NULL)
 		return EOF;
+	db = prv->db;
 
-	memset(&val, 0, sizeof(val));
 	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
 
 	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
@@ -228,10 +268,12 @@ remove_bhash(sip)
 	DB *db;
 	DBT key;
 	int rc;
+	ZSleepyPrivate *prv;
 
-	db = open_bhash(sip, O_RDWR, "remove_bhash");
-	if (db == NULL)
+	prv = open_bhash(sip, O_RDWR, "remove_bhash");
+	if (prv == NULL)
 		return EOF;
+	db = prv->db;
 
 	memset(&key, 0, sizeof(key));
 
@@ -264,18 +306,22 @@ print_bhash(sip, outfp)
 	DB *db;
 	DBT key, val;
 	int rc;
-#ifdef HAVE_DB_OPEN2
-	DBC *curs;
+	ZSleepyPrivate *prv;
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
+	DBC *curs = NULL;
 
-	db = open_bhash(sip, O_RDONLY, "print_bhash");
-	if (db == NULL)
+	prv = open_bhash(sip, O_RDONLY, "print_bhash");
+	if (prv == NULL)
 		return;
+	db = prv->db;
 
 #ifdef HAVE_DB_CURSOR4
 	rc = (db->cursor)(db, NULL, &curs, 0);
 #else
 	rc = (db->cursor)(db, NULL, &curs);
 #endif
+	prv->cursor = curs;
+
 	memset(&val, 0, sizeof(val));
 	memset(&key, 0, sizeof(key));
 
@@ -295,11 +341,13 @@ print_bhash(sip, outfp)
 		rc = (curs->c_get)(curs, &key, &val, DB_NEXT);
 	}
 	(curs->c_close)(curs);
+	prv->cursor = NULL;
 #else
 
-	db = open_bhash(sip, O_RDONLY, "print_bhash");
-	if (db == NULL)
+	prv = open_bhash(sip, O_RDONLY, "print_bhash");
+	if (prv == NULL)
 		return;
+	db = prv->db;
 
 	memset(&val, 0, sizeof(val));
 	memset(&key, 0, sizeof(key));
@@ -335,18 +383,22 @@ count_bhash(sip, outfp)
 	DBT key, val;
 	int cnt = 0;
 	int rc;
-
-#ifdef HAVE_DB_OPEN2
+	ZSleepyPrivate * prv;
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
 	DBC *curs;
 
-	db = open_bhash(sip, O_RDONLY, "count_bhash");
+	prv = open_bhash(sip, O_RDONLY, "count_bhash");
 
-	if (db != NULL) {
+	if (prv && prv->db) {
+	  db = prv->db;
+
+	  curs = NULL;
 #ifdef HAVE_DB_CURSOR4
 	  rc = (db->cursor)(db, NULL, &curs, 0);
 #else
 	  rc = (db->cursor)(db, NULL, &curs);
 #endif
+	  prv->cursor = curs;
 
 	  memset(&val, 0, sizeof(val));
 	  memset(&key, 0, sizeof(key));
@@ -365,9 +417,11 @@ count_bhash(sip, outfp)
 	  }
 	}
 	(curs->c_close)(curs);
+	prv->cursor = NULL;
 #else
-	db = open_bhash(sip, O_RDONLY, "count_bhash");
-	if (db != NULL) {
+	prv = open_bhash(sip, O_RDONLY, "count_bhash");
+	if (prv != NULL) {
+	  db = prv->db;
 
 	  memset(&val, 0, sizeof(val));
 	  memset(&key, 0, sizeof(key));
@@ -377,10 +431,6 @@ count_bhash(sip, outfp)
 	    if (val.data == NULL) /* ???? When this would happen ? */
 	      continue;
 	    ++cnt;
-
-	    memset(&val, 0, sizeof(val));
-	    memset(&key, 0, sizeof(key));
-
 	    rc = (db->seq)(db, &key, &val, R_NEXT);
 	  }
 	}
@@ -402,20 +452,21 @@ owner_bhash(sip, outfp)
 {
 	DB *db;
 	struct stat stbuf;
-	int fd;
+	int fd = -1;
 
-	db = open_bhash(sip, O_RDONLY, "owner_bhash");
-	if (db == NULL)
+	ZSleepyPrivate *prv = open_bhash(sip, O_RDONLY, "owner_bhash");
+	if (!prv || !prv->db)
 		return;
+	db = prv->db;
 
-	/* There are more timing hazards, when the internal fd is not
+	/* There are timing hazards, when the internal fd is not
 	   available for probing.. */
-#ifdef HAVE_DB_OPEN2
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
 	(db->fd)(db, &fd);
 #else
 	fd = (db->fd)(db);
 #endif
-	if (fstat(fd, &stbuf) < 0) {
+	if (fd < 0 || fstat(fd, &stbuf) < 0) {
 		fprintf(stderr, "owner_bhash: cannot fstat(\"%s\")!\n",
 				sip->file);
 		return;
@@ -430,36 +481,38 @@ modp_bhash(sip)
 {
 	DB *db;
 	struct stat stbuf;
-	struct spblk *spl;
-	spkey_t symid;
-	int rval, fd;
+	int rval, fd = -1, err = 0;
+	int roflag = O_RDONLY;
 
-	db = open_bhash(sip, O_RDONLY, "owner_bhash");
-	if (db == NULL)
-		return 0;
+	ZSleepyPrivate *prv;
 
-#ifdef HAVE_DB_OPEN2
-	(db->fd)(db, &fd);
+	prv = open_bhash(sip, roflag, "owner_bhash"); /* if it isn't open.. */
+	if (!prv || !prv->db) return 0;
+	
+	roflag = prv->roflag;
+	db     = prv->db;
+
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
+	err = (db->fd)(db, &fd);
 #else
 	fd = (db->fd)(db);
 #endif
 	if (fstat(fd, &stbuf) < 0) {
-		fprintf(stderr, "modp_bhash: cannot fstat(\"%s\")!\n",
-				sip->file);
+		fprintf(stderr, "modp_bhash: cannot fstat(\"%s\"(%d))! err=%d\n",
+				sip->file, fd, err);
 		return 0;
 	}
 	if (stbuf.st_nlink == 0)
 		return 1;	/* Unlinked underneath of us! */
-	
-	symid = symbol_lookup_db(sip->file, spt_files->symbols);
-	spl = sp_lookup(symid, spt_modcheck);
-	if (spl != NULL) {
-		rval = ((long)stbuf.st_mtime != (long)spl->data ||
-			(long)stbuf.st_nlink != 1);
-	} else
-		rval = 0;
-	sp_install(symid, (void *)((long)stbuf.st_mtime),
-		   (long)stbuf.st_nlink, spt_modcheck);
+
+	if (roflag != O_RDONLY) return 0; /* We are a WRITER ??
+					     Of course it changes.. */
+
+	rval = (stbuf.st_mtime != prv->mtime || stbuf.st_nlink != 1);
+
+	prv->mtime = stbuf.st_mtime;
+
+
 	return rval;
 }
 #endif	/* HAVE_DB */
