@@ -5,16 +5,20 @@
  *	Port of NDBM version to GDBM by Matti Aarnio 1989-1997
  *
  *	Currently uses GDBM version 1.7.3 !
+ *	Supports also GDBM version 1.8.0!
+ */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 /* LINTLIBRARY */
 
 #include "mailer.h"
-#ifdef	HAVE_GDBM_H
+#ifdef	HAVE_GDBM
 #include <gdbm.h>
-#ifndef	GDBM_VERSION /* In a hope of GDBM 1.7.4 having this #define, and
-			the   gdbm_filefno() function.. */
-	/* Because GDBM 1.7.3 does not have gdbm_filefno() or some such,
+#ifndef	HAVE_GDBM_FDESC
+	/* Because GDBM 1.7.3 does not have gdbm_fdesc() or some such,
 	   we must do "Horrible Kludges"(TM) to find the fileno..	*/
 	/* THIS MUST TRACK THE GDBM INTERNAL DATA STRUCTURE ! Auurgh...! */
 struct __gdbmfoo {
@@ -27,11 +31,12 @@ struct __gdbmfoo {
 	int desc;       /* this is what we do need! */
 	/* the rest is irrelevant for us.. */
 };
-#define gdbm_filefno(db) (((struct __gdbmfoo *) (db))->desc)
+#define gdbm_fdesc(db) (((struct __gdbmfoo *) (db))->desc)
 #endif
 
 #include <fcntl.h>
 #include <sys/file.h>
+
 #include "libsh.h"
 #include "search.h"
 #include "io.h"
@@ -43,39 +48,56 @@ struct __gdbmfoo {
 extern int deferit;
 extern int nobody;	/* UID of NOBODY */
 
-static GDBM_FILE open_gdbm __((search_info *, int, const char *));
+typedef struct ZGdbmPrivate {
+  GDBM_FILE db;
+  int roflag;
+} ZGdbmPrivate;
 
-static GDBM_FILE
-open_gdbm(sip, flag, comment)
+
+static ZGdbmPrivate * open_gdbm __((search_info *, int, const char *));
+
+static ZGdbmPrivate *
+open_gdbm(sip, roflag, comment)
 	search_info *sip;
-	int flag;
+	int roflag;
 	const char *comment;
 {
-	GDBM_FILE db;
-	struct spblk *spl;
-	spkey_t symid;
-	int i;
+	int i, flag;
+	ZGdbmPrivate *prv;
 
 	if (sip->file == NULL)
 		return NULL;
-	if (spt_files == NULL)		spt_files = sp_init();
-	if (spt_files->symbols == NULL) spt_files->symbols = sp_init();
 
-	symid = symbol_db(sip->file, spt_files->symbols);
-	spl = sp_lookup(symid, spt_files);
+	prv = (ZGdbmPrivate*) sip->dbprivate;
 
-	if (spl != NULL && flag == O_RDWR && spl->mark != O_RDWR)
-		close_gdbm(sip);
-	if (flag == O_RDWR)	flag = GDBM_WRITER;
+	if (prv && roflag == O_RDWR && roflag != prv->roflag)
+	  close_gdbm(sip,"open_gdbm");
+
+	if (roflag == O_RDWR)	flag = GDBM_WRITER;
 	else			flag = GDBM_READER;
-	if (spl == NULL || (db = (GDBM_FILE)spl->data) == NULL) {
+
+	if (!prv || !prv->db) {
+
+		GDBM_FILE db = NULL;
+
+		prv = malloc(sizeof(*prv));
+		if (!prv) return NULL;
+		memset(prv, 0, sizeof(*prv));
+
 		for (i = 0; i < 3; ++i) {
-		  db = gdbm_open((char*)sip->file, 0, flag, 0, NULL);
+		  db = gdbm_open((void*)sip->file, 0, flag, 0, NULL);
 		  if (db != NULL)
 		    break;
+		  free(prv);
+		  prv = NULL;
 		  sleep(1);
 		}
-		
+
+		if (prv) {
+		  prv->db     = db;
+		  prv->roflag = roflag;
+		}
+
 		if (db == NULL) {
 			++deferit;
 			v_set(DEFER, DEFER_IO_ERROR);
@@ -83,12 +105,10 @@ open_gdbm(sip, flag, comment)
 					comment, sip->file);
 			return NULL;
 		}
-		if (spl == NULL)
-			sp_install(symid, (void *)db, flag, spt_files);
-		else
-			spl->data = (void *)db;
+
+		sip->dbprivate = (void*)prv;
 	}
-	return db;
+	return prv;
 }
 
 
@@ -103,31 +123,36 @@ search_gdbm(sip)
 	GDBM_FILE db;
 	datum val, key;
 	conscell *tmp;
-	struct spblk *spl = NULL;
-	int retry, i;
-	spkey_t symid;
+	int retry;
+	ZGdbmPrivate *prv;
 
 	retry = 0;
 
 reopen:
-	db = open_gdbm(sip, O_RDONLY, "search_gdbm");
+	prv = open_gdbm(sip, O_RDONLY, "search_gdbm");
 
-	if (db == NULL)
+	if (!prv || !prv->db)
 	  return NULL;
+	db = prv->db;
 
-	key.dptr  = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
+
+	key.dptr  = (void*)sip->key;
 	key.dsize = strlen(sip->key) + 1;
+
 	val = gdbm_fetch(db, key);
+
 	if (val.dptr == NULL) {
-		if (!retry && gdbm_errno) {
-			gdbm_close(db);
-			sp_install(symid, (u_char *)NULL, 0, spt_files);
-			++retry;
-			goto reopen;
-		}
-		return NULL;
+	  if (!retry && gdbm_errno != GDBM_NO_ERROR &&
+	      gdbm_errno != GDBM_EMPTY_DATABASE) {
+	    close_gdbm(sip,"search_gdbm");
+	    ++retry;
+	    goto reopen;
+	  }
+	  return NULL;
 	}
-	tmp = newstring(strnsave(val.dptr, val.dsize));
+	tmp = newstring(dupnstr(val.dptr, val.dsize), val.dsize);
 	free(val.dptr);
 	return tmp;
 }
@@ -137,8 +162,9 @@ reopen:
  */
 
 void
-close_gdbm(sip)
+close_gdbm(sip,comment)
 	search_info *sip;
+	const char *comment;
 {
 	GDBM_FILE db;
 	struct spblk *spl;
@@ -171,15 +197,22 @@ add_gdbm(sip, value)
 {
 	GDBM_FILE db;
 	datum val, key;
+	ZGdbmPrivate *prv;
 
-	db = open_gdbm(sip, O_RDWR, "add_gdbm");
-	if (db == NULL)
+	prv = open_gdbm(sip, O_RDWR, "add_gdbm");
+	if (!prv || !prv->db)
 		return EOF;
+	db = prv->db;
 
-	key.dptr  = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
+
+	key.dptr  = (void*)sip->key;
 	key.dsize = strlen(sip->key) + 1;
-	val.dptr  = (char*)value;
+
+	val.dptr  = (void*)value;
 	val.dsize = strlen(value) + 1;
+
 	if (gdbm_store(db, key, val, GDBM_REPLACE) < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -200,13 +233,18 @@ remove_gdbm(sip)
 {
 	GDBM_FILE db;
 	datum key;
+	ZGdbmPrivate *prv;
 
-	db = open_gdbm(sip, O_RDWR, "remove_gdbm");
-	if (db == NULL)
+	prv = open_gdbm(sip, O_RDWR, "remove_gdbm");
+	if (!prv || !prv->db)
 		return EOF;
+	db = prv->db;
 
-	key.dptr  = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+
+	key.dptr  = (void*)sip->key;
 	key.dsize = strlen(sip->key) + 1;
+
 	if (gdbm_delete(db, key) < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -229,23 +267,29 @@ print_gdbm(sip, outfp)
 {
 	GDBM_FILE db;
 	datum key, nextkey, val;
+	ZGdbmPrivate *prv;
 
-	db = open_gdbm(sip, O_RDONLY, "print_gdbm");
-	if (db == NULL)
+	prv = open_gdbm(sip, O_RDONLY, "print_gdbm");
+	if (!prv || !prv->db)
 		return;
+	db = prv->db;
 
-	for (key = gdbm_firstkey(db); key.dptr != NULL; nextkey = gdbm_nextkey(db, key)) {
+	memset(&key, 0, sizeof(key));
+	memset(&nextkey, 0, sizeof(nextkey));
+	memset(&val, 0, sizeof(val));
+
+	key = gdbm_firstkey(db);
+	while (key.dptr) {
 		val = gdbm_fetch(db, key);
-		if (val.dptr == NULL)
-			continue;
 		if (gdbm_errno)
 			break;
 		if (val.dptr == NULL || *val.dptr == '\0')
 			fprintf(outfp, "%s\n", key.dptr);
 		else
 			fprintf(outfp, "%s\t%s\n", key.dptr, val.dptr);
-		if (val.dptr != NULL)
+		if (val.dptr)
 			free(val.dptr);
+		nextkey = gdbm_nextkey(db, key);
 		free(key.dptr);
 		key = nextkey;
 	}
@@ -261,24 +305,24 @@ count_gdbm(sip, outfp)
 	search_info *sip;
 	FILE *outfp;
 {
-	GDBM_FILE db;
-	datum key, nextkey, val;
+	ZGdbmPrivate *prv;
 	int count = 0;
 
-	db = open_gdbm(sip, O_RDONLY, "count_gdbm");
-	if (db != NULL)
-	  for (key = gdbm_firstkey(db); key.dptr != NULL; nextkey = gdbm_nextkey(db, key)) {
-	    val = gdbm_fetch(db, key);
-	    if (val.dptr == NULL)
-	      continue;
-	    if (gdbm_errno)
-	      break;
+	prv = open_gdbm(sip, O_RDONLY, "count_gdbm");
+	if (prv && prv->db) {
+
+	  GDBM_FILE db = prv->db;
+	  datum key, nextkey;
+	  
+	  key = gdbm_firstkey(db);
+	  while (key.dptr) {
 	    ++count;
-	    if (val.dptr != NULL)
-	      free(val.dptr);
-	    free(key.dptr);
+	    nextkey = gdbm_nextkey(db, key);
+	    if (key.dptr) free(key.dptr);
+	    if (gdbm_errno) break;
 	    key = nextkey;
 	  }
+	}
 	fprintf(outfp,"%d\n",count);
 	fflush(outfp);
 }
@@ -298,14 +342,16 @@ owner_gdbm(sip, outfp)
 	GDBM_FILE db;
 	int	fno;
 	struct stat stbuf;
+	ZGdbmPrivate *prv;
 
-	db = open_gdbm(sip, O_RDONLY, "owner_gdbm");
-	if (db == NULL) {
+	prv = open_gdbm(sip, O_RDONLY, "owner_gdbm");
+	if (!prv || !prv->db) {
 	  fprintf(outfp, "%d\n", nobody);
 	  return;
 	}
+	db = prv->db;
 
-	fno = gdbm_filefno(db);
+	fno = gdbm_fdesc(db);
 	if (fstat(fno, &stbuf) < 0) {
 		fprintf(stderr, "owner_gdbm: cannot fstat(\"%s\")!\n",
 				sip->file);
@@ -326,12 +372,14 @@ modp_gdbm(sip)
 	struct spblk *spl;
 	spkey_t symid;
 	int rval;
+	ZGdbmPrivate *prv;
 
-	db = open_gdbm(sip, O_RDONLY, "modp_gdbm");
-	if (db == NULL)
+	prv = open_gdbm(sip, O_RDONLY, "modp_gdbm");
+	if (!prv || !prv->db)
 		return 0;
+	db = prv->db;
 
-	fno = gdbm_filefno(db);
+	fno = gdbm_fdesc(db);
 	if (fstat(fno, &stbuf) < 0) {
 		fprintf(stderr, "modp_gdbm: cannot fstat(\"%s\")!\n",
 				sip->file);
