@@ -2,6 +2,10 @@
  *	Copyright 1991 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2002
+ */
 
 /* LINTLIBRARY */
 
@@ -45,9 +49,6 @@ struct headerinfo mandatory_hdrs[] = {
 
 struct headerinfo optional_hdrs[] = {
 { "return-receipt-to",	AddressList,	Sender,		normal		},
-#if 0
-{ "return-path",	AMailboxList,	nilUserType,	normal		},
-#endif
 #if 1
 { "date",	 nilHeaderSemantics,	nilUserType,	normal		},
 { "resent-date", nilHeaderSemantics,	nilUserType,	Resent		},
@@ -58,18 +59,35 @@ struct headerinfo optional_hdrs[] = {
 { "encrypted",		Encrypted,	nilUserType,	normal		},
 { "errors-to",		AddressList,	Sender,		normal		},
 { "obsoletes",		MessageIDList,	nilUserType,	normal		},
+
 #if 0 /* Don't parse these .. */
 { "received",		Received,	nilUserType,	normal		},
 { "keywords",		PhraseList,	nilUserType,	normal		},
 { "references",		References,	nilUserType,	normal		},
 { "in-reply-to",	References,	nilUserType,	normal		},
 #endif
-{ "x-orcpt",	nilHeaderSemantics,	killUserType,	normal		},
-{ "x-envid",	nilHeaderSemantics,	killUserType,	normal		},
-{ "return-path", nilHeaderSemantics,	killUserType,	normal		},
-{ "resent-x-orcpt",	nilHeaderSemantics,	killUserType,	Resent	},
+
+{ "envelope-id",	nilHeaderSemantics,	killUserType,	normal	},
+{ "resent-envelope-id",	nilHeaderSemantics,	killUserType,	Resent	},
+
+{ "x-envid",		nilHeaderSemantics,	killUserType,	normal	},
 { "resent-x-envid",	nilHeaderSemantics,	killUserType,	Resent	},
+
+{ "x-orcpt",		nilHeaderSemantics,	killUserType,	normal	},
+{ "resent-x-orcpt",	nilHeaderSemantics,	killUserType,	Resent	},
+
+{ "original-recipient",	nilHeaderSemantics,	killUserType,	normal	},
+{ "resent-original-recipient",nilHeaderSemantics,killUserType,	Resent	},
+
+{ "x-envelope-to",	nilHeaderSemantics,	killUserType,	normal	},
+
+#if 0
+{ "return-path",	AMailboxList,	nilUserType,	normal		},
+#else
+{ "return-path", nilHeaderSemantics,	killUserType,	normal		},
+#endif
 { "resent-return-path", nilHeaderSemantics,	killUserType,	Resent	},
+
 #if 0 /* Special treatment.. */
 { "bcc",	nilHeaderSemantics,	killUserType,	normal		},
 { "Resent-bcc",	nilHeaderSemantics,	killUserType,	Resent		},
@@ -84,6 +102,7 @@ struct headerinfo envelope_hdrs[] = {
 { "env-end",	nilHeaderSemantics,	nilUserType,	eEnvEnd		},
 { "env-eof",	nilHeaderSemantics,	nilUserType,	eEnvEnd		},
 { "envid",	nilHeaderSemantics,	nilUserType,	eEnvid		},
+{ "errormsg",	nilHeaderSemantics,	nilUserType,	eErrorMsg	},
 { "external",	nilHeaderSemantics,	nilUserType,	eExternal	},
 { "from",		AMailboxList,	Sender,		eFrom		},
 { "fullname",		Phrase,		nilUserType,	eFullname	},
@@ -208,8 +227,7 @@ static struct sptree *
 open_header(sip)
 	search_info *sip;
 {
-	if (sip->subtype == NULL)
-	  return NULL;
+	if (! sip->subtype) return NULL;
 	return (struct sptree *)sip->subtype;
 }
 
