@@ -2,12 +2,17 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Lots of modifications (new guts, more or less..) by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2002
+ */
 
 /* LINTLIBRARY */
 
 #include "mailer.h"
 #ifdef	HAVE_RESOLVER
 #include <netdb.h>
+
 #include "search.h"
 #include "libz.h"
 #include "libc.h"
@@ -22,13 +27,16 @@ search_hosts(sip)
 	search_info *sip;
 {
 	struct hostent *hp;
-	conscell *tmp;
+	char *s;
+	int slen;
 
 	/* sethostent(1); */
 	hp = gethostbyname(sip->key);
 	if (hp == NULL)
 		return NULL;
-	return newstring(strsave(hp->h_name));
+	slen = strlen(hp->h_name);
+	s = dupnstr(hp->h_name, slen);
+	return newstring(s, slen);
 }
 
 /*
