@@ -72,10 +72,10 @@ extern long	dateParse __((struct tm *localtmptr, token822 *t));
 extern int	run_relation __((int argc, const char *argv[]));
 extern struct sptree *icdbspltree __((const char *name));
 extern int	run_db __((int argc, const char *argv[]));
-extern conscell	  *db     __((const char *dbname, const char *argv10[]));
-extern const char *dbfile __((const char *dbname));
-extern void	   dbfree __((void));
-extern const char *dbtype __((const char *dbname));
+extern conscell	  *dblookup __((const char *dbname, const int argc, const char *argv20[]));
+extern const char *dbfile   __((const char *dbname));
+extern void	   dbfree   __((void));
+extern const char *dbtype   __((const char *dbname));
 
 /* File: functions.c */
 extern int	funclevel;
@@ -106,12 +106,14 @@ extern void	optsave __((int type, struct envelope *e));
 extern char	*errors_to;
 
 extern int	run_rfc822 __((int argc, const char *argv[]));
-extern int	makeLetter __((struct envelope *e, int octothrp));
+extern int	makeLetter __((struct envelope *e));
 extern void	dumpInfo __((struct envelope *e));
 extern void	dumpHeaders __((struct header *h));
 extern void	dumpHeader __((struct header *h));
 extern int	isSenderAddr;
 extern int	isRecpntAddr;
+extern int	isErrorMsg;
+extern int	isErrChannel;
 extern void	squirrel __((struct envelope *e, const char *keyword, const char *text));
 extern struct header	*erraddress __((struct envelope *e));
 extern void	defer __((struct envelope *e, const char *why));
@@ -120,7 +122,6 @@ extern struct header	*mkTrace __((struct envelope *e, struct header *rcvdhdr));
 extern conscell	*pickaddress __((conscell *l));
 extern int	thesender __((struct envelope *e, struct address *a));
 extern conscell	*makequad __((void));
-extern int	iserrmessage __((void));
 extern int	sequencer __((struct envelope *e, const char *file));
 
 /* File: rfc822hdrs.c */
@@ -165,7 +166,7 @@ extern int   savefile;
 extern int   do_hdr_warning;
 
 /* File: rtsyslog.c */
-extern void rtsyslog __(( time_t msgmtime, long msgino, const char *from, const char *smtprelay, int size, int nrcpts, const char *msgid ));
+extern void rtsyslog __(( const char *spoolid, const time_t msgmtime, const char *from, const char *smtprelay, const int size, const int nrcpts, const char *msgid, const time_t starttime ));
 
 extern int	main __((int argc, const char *argv[]));
 extern int	login_to_uid __((const char *name));
