@@ -2,6 +2,10 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  */
+/*
+ *	Some modifications  by
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ */
 
 /*
  * This file contains most of the RFC822-specific message manipulation.
@@ -21,8 +25,15 @@
 #endif
 #include "zsyslog.h"
 
+#ifdef HAVE_SYS_UN_H
+#include <sys/socket.h>
+#include <sys/un.h>
+#endif
+
 #include "prototypes.h"
 #include "libsh.h"
+#include "zmsignal.h"
+#include "shmmib.h"
 
 #ifndef _IOFBF
 #define _IOFBF  0
@@ -79,14 +90,13 @@ makequad()
 	return l;
 }
 
-struct envelope *qate;
-
-int
-iserrmessage()
+static int
+iserrmessage(e)
+     struct envelope *e;
 {
-	return (qate      &&   qate->e_from_trusted   &&
-		(QCHANNEL(qate->e_from_trusted)->cstring == NULL ||
-		 CISTREQ(QCHANNEL(qate->e_from_trusted)->cstring, "error")));
+	return (e   &&   e->e_from_trusted   &&
+		(QCHANNEL(e->e_from_trusted)->cstring == NULL ||
+		 CISTREQ(QCHANNEL(e->e_from_trusted)->cstring, "error")));
 }
 
 
@@ -123,7 +133,8 @@ run_rfc822(argc, argv)
 #if 0
 	{
 	  int c;
-	  while ((c = getopt(argc, argv, "")) != EOF) {
+	  zoptind = 1;
+	  while ((c = zgetopt(argc, argv, "")) != EOF) {
 	    switch (c) {
 	    default:
 	      ++errflg;
@@ -131,7 +142,7 @@ run_rfc822(argc, argv)
 	    }
 	  }
 	}
-	if (errflg || optind != (argc - 1))
+	if (errflg || zoptind != (argc - 1))
 #endif
 	if (argc != 2)
 	  {
@@ -139,7 +150,7 @@ run_rfc822(argc, argv)
 		return PERR_USAGE;
 	}
 
-	/* file = argv[optind]; */
+	/* file = argv[zoptind]; */
 	file = argv[1];
 
 #ifdef	XMEM
@@ -152,7 +163,6 @@ run_rfc822(argc, argv)
 	/* XXX: If this tmalloc() fails, we crash! */
 	memset(e, 0, sizeof(*e)); /* Lots of pointers, etc here! */
 	GCPRO2(e->e_from_trusted, e->e_from_resolved);
-	qate = e;
 
 	if ((e->e_fp = fopen(file, "r")) == NULL) {
 	  fprintf(stderr, "router: cannot open %s\n", file);
@@ -168,7 +178,7 @@ run_rfc822(argc, argv)
 	  status = PERR_OK;
 	  setvbuf(e->e_fp, buf, _IOFBF, sizeof buf);
 	  e->e_file = file;
-	  status = makeLetter(e, 0);
+	  status = makeLetter(e);
 	}
 
 	if (status == PERR_OK)
@@ -181,10 +191,10 @@ run_rfc822(argc, argv)
 	case PERR_BADOPEN:
 		break;
 	case PERR_BADCONTINUATION:	/* fatal */
-		squirrel(e, "badenvelope", "continuation line prior to header");
+		squirrel(e, ":badenvelope", "continuation line prior to header");
 		break;
 	case PERR_BADSUBMIT:		/* fatal */
-		squirrel(e, "badsubmit", "unrecognized envelope information");
+		squirrel(e, ":badsubmit", "unrecognized envelope information");
 		break;
 	case PERR_LOOP:			/* fatal */
 		reject(e, "loopexceeded");
@@ -239,10 +249,12 @@ run_rfc822(argc, argv)
  * Read, store, and parse the message control information (envelope + header).
  */
 
+
+/* Live code, 'octothorp' eliminated (obsolete thing..) */
+
 int
-makeLetter(e, octothorp)
+makeLetter(e)
 	register struct envelope *e;
-	int octothorp;		/* does # at start of word start a comment? */
 {
 	register int	i;
 	register char	*cp;
@@ -256,48 +268,66 @@ makeLetter(e, octothorp)
 	e->e_msgOffset = 0;
 	e->e_nowtime = now = time(NULL);
 	if (efstat(FILENO(e->e_fp), &(e->e_statbuf)) < 0) {
-#ifdef	HAVE_ST_BLKSIZE
+#ifdef	HAVE_STRUCT_STAT_ST_BLKSIZE
 		e->e_statbuf.st_blksize = 0;
-#endif	/* !HAVE_ST_BLKSIZE */
+#endif	/* !HAVE_STRUCT_STAT_ST_BLKSIZE */
 		e->e_statbuf.st_mtime = e->e_nowtime;
 	}
 	e->e_localtime = *(localtime(&(e->e_statbuf.st_mtime)));
-#ifdef	HAVE_ST_BLKSIZE
+#ifdef	HAVE_STRUCT_STAT_ST_BLKSIZE
 	initzline((int)e->e_statbuf.st_blksize);
-#else	/* !HAVE_ST_BLKSIZE */
+#else	/* !HAVE_STRUCT_STAT_ST_BLKSIZE */
 	initzline(4096);
-#endif	/* !HAVE_ST_BLKSIZE */
+#endif	/* !HAVE_STRUCT_STAT_ST_BLKSIZE */
+
+
+	taspoolid(e->e_spoolid, e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino);
+
+
+	inheader = 0; /* 0: envelope, 1: RFC-822 */
+
+	/* Line of length 1 is LIKELY just "\n" */
+
+	while ((n = zgetline(e->e_fp)) > 0) {
 
-	inheader = 0;
-	while ((n = zgetline(e->e_fp)) > !octothorp) {
 		/* We do kludgy processing things in case the input
 		   does have a CRLF at the end of the line.. */
+
+		if (n == 1 && zlinebuf[0] == '\n') {
+			/* Separator in between RFC 822  headers, and body */
+			break;
+		}
+
 		if (n > 1 &&
 		    zlinebuf[n-2] == '\r' &&
 		    zlinebuf[n-1] == '\n') {
 			--n;
 			zlinebuf[n-1] = '\n';
-			if (n <= !octothorp) break;
+			if (n <= 1) break; /* Void line */
 		}
+
 		/* Ok, now we can proceed with the original agenda.. */
-		i = hdr_status(zlinebuf, zlinebuf, n, octothorp);
+		i = hdr_status(zlinebuf, zlinebuf, n, 0);
 		if (i > 0) {		/* a real message header */
-			if (octothorp && zlinebuf[0] == '#')
-				continue;
 			/* record start of headers for posterity */
 			if (!inheader)
 			  e->e_hdrOffset = zlineoffset(e->e_fp)-n;
+
 			/* cons up a new one at top of header list */
+
 			h = makeHeader(spt_headers, zlinebuf, i);
 			h->h_next = e->e_headers;
 			e->e_headers = h;
 			h->h_lines = makeToken(zlinebuf+i+1, n-i-2);
 			h->h_lines->t_type = Line;
 			++inheader;
+
 		} else if (i == 0) {	/* a continuation line */
 			if (inheader && zlinebuf[0] == ':') {
 				optsave(FYI_ILLHEADER, e);
+
 				/* cons up a new one at top of header list */
+
 				h = makeHeader(spt_headers,"X-Null-Field",12);
 				h->h_next = e->e_headers;
 				e->e_headers = h;
@@ -318,7 +348,7 @@ makeLetter(e, octothorp)
 					t = t->t_next;
 				t->t_next = makeToken(zlinebuf, n-1);
 				t->t_next->t_type = Line;
-			} else if (!octothorp) {
+			} else {
 				return PERR_BADCONTINUATION;
 			}
 		} else if (!inheader		/* envelope information */
@@ -326,8 +356,6 @@ makeLetter(e, octothorp)
 				|| *cp == '\n')) {
 			HeaderSemantics osem;
 
-			if (octothorp && zlinebuf[0] == '#')
-				continue;
 			/* cons up a new one at top of envelope header list */
 			h = makeHeader(spt_eheaders, zlinebuf, -i);
 			h->h_next = e->e_eHeaders;
@@ -344,14 +372,14 @@ makeLetter(e, octothorp)
 					progname, h->h_descriptor->class);
 				fwrite((char *)zlinebuf, sizeof (char), -i, stderr);
 				putc('\n', stderr);
-				h->h_contents = hdr_scanparse(e, h, octothorp, 0);
+				h->h_contents = hdr_scanparse(e, h, 0, 0);
 				h->h_stamp = hdr_type(h);
 				return PERR_BADSUBMIT;
 				/* break; */
 			case eFrom:		/* pity to break the elegance */
 				osem = h->h_descriptor->semantics;
 				h->h_descriptor->semantics = Mailbox;
-				h->h_contents = hdr_scanparse(e, h, octothorp, 0);
+				h->h_contents = hdr_scanparse(e, h, 0, 0);
 				h->h_stamp = hdr_type(h);
 				h->h_descriptor->semantics = osem;
 				break;
@@ -362,11 +390,11 @@ makeLetter(e, octothorp)
 				e->e_eHeaders = h->h_next;
 				break;
 			default:		/* a real envelope header */
-				h->h_contents = hdr_scanparse(e, h, octothorp, 0);
+				h->h_contents = hdr_scanparse(e, h, 0, 0);
 				h->h_stamp = hdr_type(h);
 				break;
 			}
-		} else if (!octothorp) {
+		} else {
 			/* Expected RFC-822 header, and got something else.
 			   We play as if it was the end of the headers, and
 			   start of the body.  Keep the first faulty input
@@ -388,7 +416,7 @@ makeLetter(e, octothorp)
 	for (ph = NULL, h = e->e_headers; h != NULL; h = nh) {
 		nh = h->h_next;
 		h->h_next = ph;
-		if (!octothorp && h->h_descriptor) {
+		if (h->h_descriptor) {
 			h->h_contents = hdr_scanparse(e, h, 0, 0);
 			h->h_stamp = hdr_type(h);
 			if (!hdr_nilp(h))	/* excise null-valued headers */
@@ -403,6 +431,7 @@ makeLetter(e, octothorp)
 	return PERR_OK;
 }
 
+
 void
 dumpInfo(e)
 	struct envelope *e;
@@ -439,7 +468,7 @@ dumpHeader(h)
 	switch (h->h_descriptor->semantics) {
 	case DateTime:
 		printf("\tUNIX time %ld, local time %s",
-			h->h_contents.d, ctime(&(h->h_contents.d)));
+		       (long)h->h_contents.d, ctime(&(h->h_contents.d)));
 		break;
 	case nilHeaderSemantics:
 		for (t = h->h_lines; t != NULL; t = t->t_next)
@@ -473,8 +502,8 @@ dumpHeader(h)
 		if ((t = h->h_contents.r->r_convert) != NULL)
 			printf("\tConvert: %s\n", formatToken(t));
 		printf("\tUNIX time %ld, local time %s",
-			h->h_contents.r->r_time,
-			ctime(&(h->h_contents.r->r_time)));
+		       (long) h->h_contents.r->r_time,
+		       ctime(&(h->h_contents.r->r_time)));
 		break;
 	default:
 		for (a = h->h_contents.a; a != NULL; a = a->a_next) {
@@ -495,8 +524,15 @@ dumpHeader(h)
  * can find out whether they are dealing with a sender or recipient address.
  * The variables are accessed through C coded config file functions.
  */
-int	isSenderAddr = 0;
-int	isRecpntAddr = 0;
+int	isSenderAddr;
+int	isRecpntAddr;
+
+/*
+ * Routing can sometimes want to have different view of the world when
+ * the message is _internally_ tagged with an 'errormsg' envelope header.
+ */
+int	isErrorMsg;
+int	isErrChannel;
 
 
 #define FindEnvelope(X)	\
@@ -619,16 +655,17 @@ erraddress(e)
 	if (h == NULL) {
 		/* everything else failed, so use the owner of the file */
 		if (!e->e_trusted)
-			h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 1);
+/*ROUTER*/		h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 1);
 #if	0
 		else
-			h = mkSender(e, POSTMASTER, 1);
+/*ROUTER*/		h = mkSender(e, POSTMASTER, 1);
 #endif
 	}
 	return h;
 }
 
 
+#if 0 /* DEAD CODE */
 /* Pick recipient address from the input line.
    EXTREMELY Simple minded parsing.. */
 static void pick_env_addr __((char *buf, FILE *mfp));
@@ -657,6 +694,7 @@ FILE *mfp;
 	  fprintf(mfp,"to <%s>\n",buf);
 	}
 }
+#endif /* DEAD CODE */
 
 static int nullhost __((conscell *cs));
 static int
@@ -665,7 +703,7 @@ nullhost(cs)
 {
 	const char *s = cs->cstring;
 
-	return (s == NULL || *s == '\0' || strcmp(s, "-") == 0);
+	return ( !s || *s == '\0' || STREQ(s, "-") );
 	/* actually we should also check for localhostness, but lets not
 	   get carried away... */
 }
@@ -688,9 +726,10 @@ reject(e, msgfile)
 	
 	h = erraddress(e);
 	if (h == NULL) {
-		squirrel(e, "noerraddr", "No one to return an error to!");
-		return;
+	  squirrel(e, "noerraddr", "No one to return an error to!");
+	  return;
 	}
+
 	/*
 	 * turn all (possibly more than one) the addresses in whatever h
 	 * is pointing at, into recipient addresses.
@@ -701,35 +740,40 @@ reject(e, msgfile)
 	path = (char*)alloca(3+strlen(mailshare)+strlen(FORMSDIR)+strlen(msgfile));
 #endif
 	sprintf(path, "%s/%s/%s", mailshare, FORMSDIR, msgfile);
+
 	fp = fopen(path, "r");
 	if (fp == NULL) {
-		perror(path);
-		squirrel(e, "norejform", "Couldn't open reject form file");
+	  perror(path);
+	  squirrel(e, "norejform", "Couldn't open reject form file");
 #ifndef	USE_ALLOCA
-		free(path);
+	  free(path);
 #endif
-		return;
+	  return;
 	}
 #ifndef	USE_ALLOCA
 	free(path);
 #endif
+
+	setvbuf(fp, vbuf, _IOFBF, sizeof vbuf);
+	fseek(fp,(off_t)0,0);
+
+
+	runastrusteduser();
 	mfp = mail_open(MSG_RFC822);
+	runasrootuser();
+
 	if (mfp == NULL) {
 	  squirrel(e, "mailcreatefail", "Couldn't open reject message file");
+	  fclose(fp);
 	  return;
 	}
-	setvbuf(fp, vbuf, _IOFBF, sizeof vbuf);
-	while (fgets(buf,sizeof(buf),fp)) {
-	  if (strncmp("ADR",buf,3)==0) {
-	    pick_env_addr(buf+4,mfp);
-	  } else if (strncmp("HDR",buf,3)==0 ||
-		     strncmp("SUB",buf,3)==0) {
-	    continue;
-	  } else
-	    break;
-	}
-	fputs("env-end\n",mfp);
-	fseek(fp,(off_t)0,0);
+
+	/* Mark this as an error message, but don't add the EXPLICITE
+	   envelope and RFC-822 header separator ("env-end") !         */
+
+	fputs("channel error\n", mfp);
+	fputs("errormsg\n",      mfp);
+	/* fputs("env-end\n",    mfp); */
 
 	/* who is it from? */
 	c_cp = h->h_pname;
@@ -920,7 +964,7 @@ mkSender(e, name, flag)
 		h = NULL;
 	if (h  &&  !flag  &&  !e->e_trusted) {
 		/* make sure the pretty login is valid. see thesender() */
-		l = router(h->h_contents.a, e->e_statbuf.st_uid, "sender", NULL);
+/*ROUTER*/	l = router(h->h_contents.a, e->e_statbuf.st_uid, "sender", NULL);
 		if (l) {
 			l = pickaddress(l);
 			flag = (QUSER(l) == NULL || !nullhost(QHOST(l)));
@@ -968,7 +1012,7 @@ mkSender(e, name, flag)
 	}
 	*ppp = NULL;
 	if (l == NULL && !e->e_trusted) {
-		l = router(sh->h_contents.a, e->e_statbuf.st_uid, "sender", NULL);
+/*ROUTER*/	l = router(sh->h_contents.a, e->e_statbuf.st_uid, "sender", NULL);
 		if (l)
 			e->e_from_trusted = pickaddress(l);
 	}
@@ -1033,26 +1077,23 @@ mkTrace(e, rcvdhdr)
 		th->h_contents.r->r_by = NULL;
 	/* via */
 	FindEnvelope(eVia);
-	if (h  &&  h->h_contents.a->a_tokens) {
+	if (h  &&  h->h_contents.a  &&  h->h_contents.a->a_tokens) {
 		th->h_contents.r->r_via = h->h_contents.a->a_tokens->p_tokens;
 	} else
 		th->h_contents.r->r_via = NULL;
 	/* with */
 	FindEnvelope(eWith);
-	if (h  &&  h->h_contents.a->a_tokens) {
+	if (h  &&  h->h_contents.a  &&  h->h_contents.a->a_tokens) {
 		th->h_contents.r->r_with = h->h_contents.a->a_tokens->p_tokens;
 	} else
 		th->h_contents.r->r_with = NULL;
-	/* id */
-	{
-	  char taspid[32];
-	  taspoolid(taspid, e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino);
 
-	  na = (struct addr *)tmalloc(sizeof (struct addr));
-	  na->p_tokens = makeToken(taspid, strlen(taspid));
-	  na->p_next = NULL;
-	  na->p_type = anAddress;	/* really a message id */
-	}
+	/* id */
+	na = (struct addr *)tmalloc(sizeof (struct addr));
+	na->p_tokens = makeToken(e->e_spoolid, strlen(e->e_spoolid));
+	na->p_next = NULL;
+	na->p_type = anAddress;	/* really a spool-id */
+
 	ap = (struct address *)tmalloc(sizeof (struct address));
 	ap->a_tokens = na;
 	ap->a_next = NULL;
@@ -1132,7 +1173,7 @@ thesender(e, a)
 {
 	conscell *l; /* Var life ends after return.. no GC protection */
 
-	l = router(a, e->e_statbuf.st_uid, "sender", NULL);
+/*ROUTER*/ l = router(a, e->e_statbuf.st_uid, "sender", NULL);
 	if (l == NULL)
 		return 0;
 	e->e_from_resolved = pickaddress(l);
@@ -1198,6 +1239,7 @@ sequencer(e, file)
 	struct address *a, *ap;
 	struct addr    *p = NULL;
 	char	       *ofpname, *path, *qpath;
+	const char     *pfile;
 	conscell       *l, *routed_addresses, *sender, *to;
 	conscell       *rwmchain;
 	struct rwmatrix *rwhead, *nsp, *rwp = NULL, *rcp = NULL;
@@ -1209,15 +1251,21 @@ sequencer(e, file)
 	char  vbuf[2048], verbosefile[1024];
 	const char     *envid;
 	const char     *notaryret;
-	int   nrcpts = 0;
+	int   onrcpts = 0;
+	int   inrcpts = 0;
 	const char     *fromaddr = "?from?";
 	const char     *msgidstr = "?msgid?";
 	const char     *smtprelay = NULL;
 	const char     *senderstr;
 	char subdirhash[8];
 	struct notary *DSN;
+	time_t start_now;
+	struct stat stbuf;
+	long infilesize_kb, taskfilesize_kb;
 	GCVARS5;
 
+	time(&start_now);
+
 	if (schedulersubdirhash < 0) {
 	  const char *s = getzenv("SCHEDULERDIRHASH");
 	  if (s && ((s[0] == '1' || s[0] == '2') && s[1] == 0))
@@ -1226,6 +1274,14 @@ sequencer(e, file)
 	    schedulersubdirhash = 0;
 	}
 
+	pfile = file;
+	if ('A' <= *pfile && *pfile <= 'Z') {
+	  ++pfile;
+	  if (*pfile == '/') ++pfile;
+	  if ('A' <= *pfile && *pfile <= 'Z') ++pfile;
+	  if (*pfile == '/') ++pfile;
+	}
+
 	if (schedulersubdirhash) {
 	  long ino = e->e_statbuf.st_ino;
 	  if (schedulersubdirhash > 1) {
@@ -1456,10 +1512,10 @@ sequencer(e, file)
 			h = copySender(e);
 		} else {
 			dprintf("Generate a sender based on owner of file\n");
-			h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 0);
+/*ROUTER*/		h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 0);
 		}
 		if (h == NULL)
-			h = mkSender(e, POSTMASTER, 1);
+/*ROUTER*/		h = mkSender(e, POSTMASTER, 1);
 		/* assert h != NULL */
 		h->h_next = e->e_eHeaders;
 		e->e_eHeaders = h;
@@ -1473,12 +1529,12 @@ sequencer(e, file)
 		dprintf("A sender was specified in the envelope\n");
 		if (!e->e_trusted) {
 			dprintf("Replace the sender based on owner of file\n");
-			h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 0);
+/*ROUTER*/		h = mkSender(e, uidpwnam(e->e_statbuf.st_uid), 0);
 		} else {
 			dprintf("Provide a full name for the originator\n");
 			/* ensure there is a fullnamemap entry */
 			login_to_uid(t->t_pname);
-			h = mkSender(e, t->t_pname, 0);
+/*ROUTER*/		h = mkSender(e, t->t_pname, 0);
 		}
 		h->h_next = e->e_eHeaders;
 		e->e_eHeaders = h;
@@ -1507,8 +1563,17 @@ sequencer(e, file)
 	dprintf("Originating channel determination\n");
 	def_uid = nobody;
 
+	isErrorMsg   = 0;
+	isErrChannel = 0;
+
 	if (e->e_trusted) {
 		/* The sender uid is known */
+
+		/* Is it perhaps an ERROR MESSAGE ? */
+		FindEnvelope(eErrorMsg);
+		if (h) isErrorMsg = 1;
+
+		/* Does it have CHANNEL ?  Is it "ERROR" ? */ 
 		FindEnvelope(eChannel);
 		if (h) {
 			dprintf("A channel was specified\n");
@@ -1529,6 +1594,9 @@ sequencer(e, file)
 				 */
 				optsave(FYI_NOCHANNEL, e);
 			}
+			if (iserrmessage(e))
+			  isErrChannel = 1;
+
 		}
 
 		h = rcvdhdr; /* FindEnvelopeLast(eRcvdFrom); */
@@ -1541,7 +1609,7 @@ sequencer(e, file)
 			cdr(QHOST(e->e_from_trusted)) = l;
 		}
 		FindEnvelope(eUser);
-		if (h  && h->h_contents.a->a_pname) {
+		if (h  && h->h_contents.a && h->h_contents.a->a_pname) {
 			/* a previous user was specified */
 			slen = strlen(h->h_contents.a->a_pname);
 			l = cdr(QHOST(e->e_from_trusted));
@@ -1564,29 +1632,48 @@ sequencer(e, file)
 			  if (h == NULL || h->h_stamp == BadHeader)
 			    FindHeader("from",!e->e_resent);
 			  if (h == NULL || h->h_stamp == BadHeader)
-			    h = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
+			    /* BAD input, must try to do something.. */
+/*ROUTER*/		    h = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
 			}
 			if (h == NULL)
 				abort(); /* Failed to make sender header */
 
-			/* This conscell lifetime is limited.. */
-			l = router(h->h_contents.a,
+			if (isErrChannel) {
+			    char *ss = (char*)tmalloc(10);
+			    strcpy(ss, "from <>");
+
+			    h = makeHeader(spt_eheaders, ss, 4);
+			    h->h_lines = makeToken(ss+5,2);
+			    h->h_lines->t_type = Line;
+			    h->h_contents = hdr_scanparse(e, h, 0, 0);
+			    /* And now we bend the result beyond any
+			       recognition.. */
+			    h->h_contents.a->a_tokens->p_type     =  anAddress;
+			    h->h_contents.a->a_tokens->p_tokens->t_len  = 2;
+			    h->h_contents.a->a_tokens->p_tokens->t_type = Atom;
+			    h->h_contents.a->a_tokens->p_tokens->t_next = NULL;
+			    h->h_stamp = hdr_type(h);
+			}
+
+			/* This conscell lifetime is limited, no GC protect. */
+/*ROUTER*/		l = router(h->h_contents.a,
 				   e->e_statbuf.st_uid, "sender", NULL);
+
 			if (l == NULL) {
 			  /* From: <>,  and no envelope 'from' .. */
-			  h = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
+/*ROUTER*/		  h = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
 			  if (h == NULL)
 			    abort(); /* Can't make Sender header ?? */
-			  l = router(h->h_contents.a,
+/*ROUTER*/		  l = router(h->h_contents.a,
 				     e->e_statbuf.st_uid, "sender", NULL);
 			}
 			if (l) {
-				/*
-				 * In case the router returns several addresses,
-				 * we pick one at random to use for sender priv
-				 * determination.
-				 */
-				e->e_from_resolved = pickaddress(l);
+			  /*
+			   * In case the router returns several addresses,
+			   * we pick one at random to use for sender priv
+			   * determination.
+			   */
+			  e->e_from_resolved = pickaddress(l);
 			}
 		}
 		if (nullhost(QHOST(e->e_from_resolved)) &&
@@ -1617,7 +1704,7 @@ sequencer(e, file)
 		}
 		if (h && h->h_contents.a) {
 			/* a Sender: was given */
-			if (!thesender(e, h->h_contents.a)) {
+/*ROUTER*/			if (!thesender(e, h->h_contents.a)) {
 				/* but it is fake, so correct it */
 				dprintf("The Sender: is not the sender\n");
 				set_pname(e, h, "Fake-Sender");
@@ -1625,12 +1712,12 @@ sequencer(e, file)
 				h = NULL;
 		} else if (nh  &&  nh->h_contents.a) {
 			/* only a From: was given */
-			if (!thesender(e, nh->h_contents.a)) {
+/*ROUTER*/		if (!thesender(e, nh->h_contents.a)) {
 			  /* but it is fake, so add a Sender: */
 			  dprintf("The From: is not the sender\n");
 			  if (h) {
 			    /* use our empty Sender: */
-			    ph = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
+/*ROUTER*/		    ph = mkSender(e,uidpwnam(e->e_statbuf.st_uid),0);
 			    h->h_contents.a = ph->h_contents.a;
 			    h = NULL;
 			  } else
@@ -1639,7 +1726,7 @@ sequencer(e, file)
 			  h = NULL;
 		}
 		if (h) {
-		  InsertHeader(h,mkSender(e, uidpwnam(e->e_statbuf.st_uid),0));
+/*ROUTER*/	  InsertHeader(h,mkSender(e, uidpwnam(e->e_statbuf.st_uid),0));
 		  set_pname(e, nh, "Sender");
 		}
 	}
@@ -1753,33 +1840,25 @@ sequencer(e, file)
 			dprintf(" none available!\n");
 			optsave(FYI_NOSENDER, e);
 		} else {
-			FindEnvelope(eChannel);
-			if (h) {
-			  if ((ap = h->h_contents.a) != NULL
-			      && (p = ap->a_tokens) != NULL
-			      && (p->p_tokens != NULL)) {
-			    t = p->p_tokens;
-			    if (TOKENLEN(t) == 5 &&
-				strncmp(t->t_pname,"error",5)==0) {
+			if (isErrChannel) {
 
-			      char *ss = (char*)tmalloc(10);
-			      strcpy(ss, "From: <>");
+			    char *ss = (char*)tmalloc(10);
+			    strcpy(ss, "From: <>");
 
-			      nh = makeHeader(spt_headers, ss, 4);
-			      nh->h_lines = makeToken(ss+5,3);
-			      nh->h_lines->t_type = Line;
-			      nh->h_contents = hdr_scanparse(e, nh, 0, 0);
-			      nh->h_stamp = hdr_type(nh);
-			      FindHeader("to",e->e_resent);
-			      if (h == NULL) {
-				for (h = e->e_headers; h != NULL; h = h->h_next)
-				  if (CISTREQ(h->h_pname, "subject"))
-				    break;
-			      }
-			      InsertHeader(h, nh);
-			      dprintf(" nope (added)!\n");
+			    nh = makeHeader(spt_headers, ss, 4);
+			    nh->h_lines = makeToken(ss+5,3);
+			    nh->h_lines->t_type = Line;
+			    nh->h_contents = hdr_scanparse(e, nh, 0, 0);
+			    nh->h_stamp = hdr_type(nh);
+			    FindHeader("to",e->e_resent);
+			    if (h == NULL) {
+			      for (h = e->e_headers; h != NULL; h = h->h_next)
+				if (CISTREQ(h->h_pname, "subject"))
+				  break;
 			    }
-			  }
+			    InsertHeader(h, nh);
+			    dprintf(" nope (added)!\n");
+
 			} else {
 
 			  /* We make a copy of eFrom envelope header
@@ -1847,7 +1926,7 @@ sequencer(e, file)
 	errors_to = NULL;
 	senderstr = NULL;
 	if (e->e_from_trusted) {
-	  if (CISTREQ("error",QCHANNEL(e->e_from_trusted)->cstring)) {
+	  if (isErrChannel) {
 	    senderstr = "<>"; /* From "BOX" -- from an error channel! */
 	  } else {
 	    senderstr = QUSER(e->e_from_trusted)->cstring;
@@ -1883,10 +1962,15 @@ sequencer(e, file)
 			DSN = NULL;
 		}
 		for (a = h->h_contents.a; a != NULL; a = a->a_next) {
-			l = router(a, def_uid, "recipient", senderstr);
+			++inrcpts;
+/*ROUTER*/		l = router(a, def_uid, "recipient", senderstr);
 			if (l == NULL)
 				continue;
-
+#if 0
+			if (!LIST(l) || !car(l) || !LIST(car(l))) {
+			FIXME: FIXME: react on badly structured result!
+			}
+#endif
 			if (routed_addresses == NULL) {
 				routed_addresses = ncons(car(l));
 			} else {
@@ -2045,6 +2129,7 @@ sequencer(e, file)
 	}
 
 	isSenderAddr = isRecpntAddr = 0;
+
 	dprintf("Make sure Date, From, To, are in the header\n");
 	FindHeader("date",e->e_resent);
 	if (h == NULL)
@@ -2078,13 +2163,14 @@ sequencer(e, file)
 
 	dprintf("Emit specification to the transport system\n");
 #ifdef	USE_ALLOCA
-	ofpname = (char *)alloca((u_int)(strlen(file)+9));
+	ofpname = (char *)alloca(30);
 #else
-	ofpname = (char *)emalloc((u_int)(strlen(file)+9));
+	ofpname = (char *)emalloc(30);
 #endif
 	/* Guaranteed unique within this machine */
-	sprintf(ofpname,".%s.%d", file, (int)getpid());
-	if ((ofp = fopen(ofpname, "w+")) == NULL) {
+	sprintf(ofpname,".%ld.%d", (long)time(NULL), (int)getpid());
+	ofp = fopen(ofpname, "w+");
+	if (ofp == NULL) {
 #ifndef USE_ALLOCA
 		free(ofpname);
 #endif
@@ -2098,6 +2184,9 @@ sequencer(e, file)
 		fchmod(FILENO(ofp), 0460);
 	}
 
+	fprintf(ofp, "%c%c0x%08lx\n",
+		_CF_FORMAT, _CFTAG_NORMAL, (long)_CF_FORMAT_KNOWN_SET);
+
 	FindEnvelope(eVerbose);
 	if (h  &&  h->h_contents.a  &&  h->h_contents.a->a_tokens) {
 		if (h->h_contents.a->a_tokens->p_tokens    &&
@@ -2107,12 +2196,12 @@ sequencer(e, file)
 			   h->h_contents.a->a_tokens->p_tokens,
 			   (token822 *)NULL, 0);
 		/*
-		 * We have to be careful how we open this file, since one might
-		 * imagine someone trying to append to /etc/passwd using this
-		 * stuff.  The only safe way is to open it with the permissions
-		 * of the owner of the message file.
+		 * We have to be careful how we open this file, since one
+		 * might imagine someone trying to append to /etc/passwd
+		 * using this stuff.  The only safe way is to open it
+		 * with the permissions of the owner of the message file.
 		 */
-		setreuid(0, e->e_statbuf.st_uid);
+		SETEUID(e->e_statbuf.st_uid);
 
 		vfp = fopen(verbosefile, "a");
 		if (vfp) {
@@ -2122,15 +2211,12 @@ sequencer(e, file)
 			fprintf(ofp, "%c%c%s\n", _CF_VERBOSE, _CFTAG_NORMAL,
 				verbosefile);
 		}
-		setreuid(0, 0);
+		SETEUID(0);
 	} else
 		vfp = NULL;
 
-	fprintf(ofp, "%c%c0x%08lx\n",
-		_CF_FORMAT, _CFTAG_NORMAL, (long)_CF_FORMAT_KNOWN_SET);
-
 	fprintf(ofp, "%c%c%s\n",
-		_CF_MESSAGEID, _CFTAG_NORMAL, file);
+		_CF_MESSAGEID, _CFTAG_NORMAL, pfile);
 	fprintf(ofp, "%c%c%d\n",
 		_CF_BODYOFFSET, _CFTAG_NORMAL, (int)(e->e_msgOffset));
 	if (e->e_messageid) {
@@ -2152,7 +2238,7 @@ sequencer(e, file)
 	 * to your disk space.
 	 */
 
-	if (!iserrmessage()) {
+	if (!isErrChannel) {
 #if 1
                 if (errors_to) {        /* [mea@utu.fi] Stupid, but workable.. */
                         putc(_CF_ERRORADDR, ofp);
@@ -2242,7 +2328,7 @@ sequencer(e, file)
 			ofperrors |= ferror(ofp);
 			if (ofperrors) break; /* Sigh.. */
 
-			if (!iserrmessage() && nsp->errto &&
+			if (!isErrChannel && nsp->errto &&
 			    nsp->errto->string) {
 				/* print envelope sender address */
 				putc(_CF_ERRORADDR, ofp);
@@ -2293,7 +2379,7 @@ sequencer(e, file)
 				}
 				
 				putc('\n', ofp);
-				++nrcpts;
+				++onrcpts;
 				/* DSN data output ! */
 				prdsndata(rcp->info, ofp, "recipient");
 				if (vfp) {
@@ -2356,11 +2442,11 @@ sequencer(e, file)
 	}
 	/* all is nirvana -- link the input file to somewhere safe */
 #ifdef	USE_ALLOCA
-	qpath = (char*)alloca(12+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)alloca(12+strlen(QUEUEDIR)+strlen(pfile));
 #else
-	qpath = (char*)emalloc(12+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)emalloc(12+strlen(QUEUEDIR)+strlen(pfile));
 #endif
-	sprintf(qpath, "../%s/%s%s", QUEUEDIR, subdirhash, file);
+	sprintf(qpath, "../%s/%s%s", QUEUEDIR, subdirhash, pfile);
 	fflush(ofp);
 #ifdef HAVE_FSYNC
 	while (fsync(FILENO(ofp)) < 0) {
@@ -2371,7 +2457,28 @@ sequencer(e, file)
 #endif
 	ofperrors |= ferror(ofp);
 
-	if ((fclose(ofp) != 0) || ofperrors || (erename(file, qpath) != 0)) {
+	taskfilesize_kb = 0;
+	if (!ofperrors) {
+	  memset(&stbuf, 0, sizeof(stbuf));
+	  fstat(FILENO(ofp), &stbuf); /* will always succeed.. */
+
+	  if (stbuf.st_blksize == 0) stbuf.st_blksize = 1024;
+	  taskfilesize_kb = (stbuf.st_size + stbuf.st_blksize -1) / 1024;
+	}
+
+
+	infilesize_kb = 0;
+	if (!ofperrors) {
+	  memset(&stbuf, 0, sizeof(stbuf));
+	  lstat(file, &stbuf); /* will always succeed.. */
+
+	  if (stbuf.st_blksize == 0) stbuf.st_blksize = 1024;
+	  infilesize_kb = (stbuf.st_size + stbuf.st_blksize -1) / 1024;
+	}
+
+
+	if ((fclose(ofp) != 0) || ofperrors ||
+	    (erename(file, qpath) != 0)) { /* ORIGINAL FILE PATH! */
 	  zunlink(qpath);
 	  zunlink(ofpname);
 #ifndef	USE_ALLOCA
@@ -2382,13 +2489,13 @@ sequencer(e, file)
 	}
 
 #ifndef USE_ALLOCA
-	path = (char*)emalloc(14+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)emalloc(14+strlen(TRANSPORTDIR)+strlen(pfile));
 #else
 	/* This actually reallocs more space from stack, but then it
 	   is just stack space and will disappear.. */
-	path = (char*)alloca(14+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)alloca(14+strlen(TRANSPORTDIR)+strlen(pfile));
 #endif
-	sprintf(path, "../%s/%s%s", TRANSPORTDIR, subdirhash, file);
+	sprintf(path, "../%s/%s%s", TRANSPORTDIR, subdirhash, pfile);
 #ifndef HAVE_RENAME
 	zunlink(path);	/* Should actually always fail.. */
 #endif
@@ -2416,9 +2523,62 @@ sequencer(e, file)
 		fclose(vfp);
 	}
 
-	rtsyslog(e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino,
+	rtsyslog(e->e_spoolid, e->e_statbuf.st_mtime,
 		 fromaddr, smtprelay, (int) e->e_statbuf.st_size,
-		 nrcpts, msgidstr);
+		 onrcpts, msgidstr, start_now);
+
+	MIBMtaEntry->rt.TransmittedMessages   += 1;
+	MIBMtaEntry->rt.ReceivedRecipients    += inrcpts;
+	MIBMtaEntry->rt.TransmittedRecipients += onrcpts;
+
+	MIBMtaEntry->rt.TransmittedVolume     += infilesize_kb;
+	MIBMtaEntry->rt.TransmittedVolume2    += taskfilesize_kb;
+
+	MIBMtaEntry->rt.StoredMessages        -= 1;
+	MIBMtaEntry->rt.StoredVolume          -= infilesize_kb;
+
+
+#ifdef AF_UNIX
+	do {
+
+	  const char *schedulernotify;
+	  char buf[1000];
+	  int notifysocket;
+	  struct sockaddr_un sad;
+#ifndef MSG_NOSIGNAL
+	  RETSIGTYPE (*oldsig)__((int));
+#endif
+
+	  schedulernotify = getzenv("SCHEDULERNOTIFY");
+	  if (!schedulernotify) break;
+
+	  memset(&sad, 0, sizeof(sad));
+	  sad.sun_family = AF_UNIX;
+	  strncpy(sad.sun_path, schedulernotify, sizeof(sad.sun_path));
+	  sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
+
+	  notifysocket = socket(PF_UNIX, SOCK_DGRAM, 0);
+	  if (notifysocket < 0) {
+	    perror("notifysocket: socket(PF_UNIX)");
+	    break;
+	  }
+	  fd_nonblockingmode(notifysocket);
+
+	  sprintf(buf, "NEW %s%.300s", subdirhash, pfile);
+
+	  /* We are running with SIGPIPE ignored.. */
+
+	  sendto(notifysocket, buf, strlen(buf),
+#ifdef MSG_NOSIGNAL
+		 MSG_NOSIGNAL|
+#endif
+		 MSG_DONTWAIT ,
+		 (struct sockaddr *)&sad, sizeof(sad));
+
+	  close(notifysocket);
+	} while (0);
+#endif /* AF_UNIX */
+
 
 #ifndef	USE_ALLOCA
 	free(ofpname);
@@ -2467,9 +2627,8 @@ prctladdr(info, fp, cfflag, comment)
 		&& LIST(x)) {	    /* ... and it is a list */
 	      /* Find the 'privilege' item */
 	      for (x = car(x); x != NULL; x = cddr(x)) {
-		if (STRING(x)
-		    && strcmp((char *)x->string,
-			      "privilege") == 0) {
+		if (STRING(x) &&
+		    STREQ((char *)x->string, "privilege")) {
 		  x = cdr(x);
 		  break;
 		}
@@ -2523,10 +2682,10 @@ prctladdr(info, fp, cfflag, comment)
 	    }
 	    if (cdr(l))
 	      putc(' ', fp);
-	  } else if (channel && strcmp(channel,"error")!=0)
+	  } else if (channel && !STREQ(channel,"error"))
 	    fprintf(stderr, "Malformed %s\n", comment);
 	}
-	if ((cfflag == _CF_SENDER) && channel && strcmp(channel,"error") == 0)
+	if ((cfflag == _CF_SENDER) && channel && STREQ(channel,"error"))
 		user = ""; /* error channel source address -> no user! */
 
 	if (!privilege) {
@@ -2602,7 +2761,7 @@ find_errto(info)
 	for (x = cadr(info); x != NULL; x = cddr(x)) {
 	  if (!STRING(x))
 	    return NULL; /* error in data */
-	  if (strcmp((char *)x->string,"ERR") == 0) {
+	  if (STREQ((char *)x->string,"ERR")) {
 	    x = cdr(x);
 	    break;
 	  }
