@@ -46,13 +46,52 @@ hdr_scanparse(e, h, commentflag, no_line_crossing)
 	const char	*cp, *ocp;
 	char		 c1, c2;
 	union misc	 retval;
-	token822	*tlist, **prev_tp, *scan_t, *nt;
+	token822	*tlist, **prev_tp, *scan_t, *nt, tt;
 	struct headerinfo *hd;
 
-	/* If we don't recognize the header, we can't parse it */
+	/* If we don't recognize the header, we can't parse it
+	   properly, but we can always do unstructured tokenization */
 	if ((hd = h->h_descriptor) == NULL || hd->hdr_name == NULL
 	    || hd->semantics == nilHeaderSemantics) {
+#if 1
+		tlist = NULL;
+		prev_tp = &tlist;
+		for (t = h->h_lines; t != NULL; t = t->t_next) {
+			if (t->t_type != Line)	/* sanity check */
+				continue;
+			cp = t->t_pname;
+			len = TOKENLEN(t);
+			while (len > 0) {
+			  ocp = cp;
+			  nt = t;
+			  scan_t = scan822utext(&cp, len, &nt);
+			  len -= cp - ocp;
+
+			  /* Append the scanner tokens to the list of tokens */
+
+			  *prev_tp = scan_t;
+			  while (scan_t != NULL) {
+				/*
+				 * Doing it in a loop allows the scanner to
+				 * return a token list instead of one token.
+				 */
+			    prev_tp = &(scan_t->t_next);
+			    scan_t = scan_t->t_next;
+			  }
+			}
+			tt.t_len = 1;
+			tt.t_pname = "\n";
+			tt.t_type = Space;
+			tt.t_next = NULL;
+			*prev_tp = copyToken(&tt);
+			prev_tp = &((*prev_tp)->t_next);
+		}
+		*prev_tp = NULL;
+
+		retval.t = tlist;
+#else
 		retval.t = h->h_lines;
+#endif
 		return retval;
 	}
 	/*
@@ -234,8 +273,8 @@ copySender(e)
 		cp = h->h_descriptor->hdr_name + e->e_resent;
 		for (cpp = prio_list; *cpp != NULL; ++cpp) {
 			/* printf("comparing '%s' and '%s'\n", *cpp, cp); */
-			if (**cpp == *cp && strcmp(*cpp, cp) == 0
-			    && (cpp - prio_list) < minval) {
+			if (**cpp == *cp  &&  STREQ(*cpp, cp) &&
+			    (cpp - prio_list) < minval) {
 				best = h;
 				minval = cpp - prio_list;
 			}
@@ -322,17 +361,19 @@ mkMessageId(e, unixtime)
 	  }
 	  *p = '\0';
 	  cp = tzbuf;
-	  sprintf(buf, "<%02d%s%d.%02d%02d%02d%.20s.%.20s+%d@%.90s>",
+	  sprintf(buf, "<%s/%02d%s%d.%02d%02d%02d%.20s.%.20s+%d@%.90s>",
+		  e->e_spoolid,
 		  ts->tm_year % 100, monthname[ts->tm_mon], ts->tm_mday,
 		  ts->tm_hour, ts->tm_min, ts->tm_sec,
 		  cp, e->e_file, ++genseq, myhostname);
 	}
 #else /* New way to do this .. more compressed one.. */
 	ts = gmtime(&unixtime);
-	sprintf(buf, "<%04d%02d%02d%02d%02d%02dZ%.20s+%d@%.90s>",
+	sprintf(buf, "<%s/%04d%02d%02d%02d%02d%02dZ+%d@%.90s>",
+		e->e_spoolid,
 		ts->tm_year + 1900, ts->tm_mon + 1, ts->tm_mday,
 		ts->tm_hour,        ts->tm_min,     ts->tm_sec,
-		e->e_file,          ++genseq,       myhostname);
+		++genseq,       myhostname);
 #endif
 	h->h_lines = makeToken(buf, strlen(buf));
 	h->h_lines->t_type = Line;
@@ -379,15 +420,12 @@ mkDate(isresent, unixtime)
 {
 	register struct header	*h;
 	char *cp, *s;
-	static struct headerinfo *dateDesc;
-	
+
 	/* There is really only one kind of dateDesc.. */
-	if (dateDesc == NULL) {
-		dateDesc = find_header(spt_headers,
-				       isresent ? "resent-date" : "date");
-		if (dateDesc == NULL)
-			return NULL;
-	}
+	static struct headerinfo dateDesc[] = {
+	  { "date",  DateTime,	nilUserType,  normal	}
+	};
+	
 	h = (struct header *)tmalloc(sizeof (struct header));
 	h->h_pname = isresent ? "Resent-Date" : "Date";
 	h->h_descriptor = dateDesc;
@@ -419,7 +457,7 @@ hdr_print(h, fp)
 	struct address *ap;
 	struct addr *pp;
 	HeaderSemantics sem;
-	int foldcol;
+	int foldcol, hadspc;
 	int canprefold = 0;
 
 	if (h == NULL)
@@ -430,19 +468,18 @@ hdr_print(h, fp)
 	if (h->h_stamp == BadHeader)
 	  sem = nilHeaderSemantics;
 
-	col = 1 + strlen(h->h_pname);
+	foldcol = col = 1 + strlen(h->h_pname);
 
 	fprintf(fp, "%s:", h->h_pname);
 
 	/* If it was pre-existing header, output all its pre-existing
 	   white-spaces now */
 
-	foldcol = 0; /* reusing for other purpose */
 	for (t = h->h_lines; t != NULL; t = t->t_next) {
 		const char *p = t->t_pname;
 		while (*p == ' ' || *p == '\t' ||
 		       *p == '\n' || *p == '\r') {
-		  putc(*p, fp);
+		  /* putc(*p, fp); */
 		  if (*p == ' ')
 		    ++col;
 		  else if (*p == '\t')
@@ -450,30 +487,39 @@ hdr_print(h, fp)
 		  else
 		    col = 0;
 		  ++p;
-		  ++foldcol;
 		}
 		if (*p != 0)
 		  break;
 		if (!t->t_next) /* Don't fold here ! */
 		  break;
 		col = 0;
-		putc('\n', fp);
-		foldcol = 0;
-	}
-	if (h->h_lines == 0) {
-		putc(' ', fp);
-		++col;
-		++foldcol;
+		/* putc('\n', fp); */
 	}
 
 	/* Fill to column 8 in all cases.. */
-	for ( ; col < 8; ++col)
-	  putc(' ', fp), ++foldcol;
+	if (col < 8) col = 8;
 
-	if (!foldcol) {
-	  /* Always at least one space! */
-	  putc(' ', fp);
-	  ++col;
+	/* always produces TABFULLY INDENTED headers, TA's
+	   may fold them to SPACEFULL if so desired.. */
+
+	hadspc = 0;
+	while (foldcol < col) {
+		int c2 = foldcol + 8 - (foldcol & 7);
+		if (c2 <= col) {
+		  putc('\t', fp);
+		  foldcol = c2;
+		} else {
+		  putc(' ', fp);
+		  foldcol += 1;
+		}
+		hadspc = 1;
+	}
+
+	/* Always at least one space! */
+	if (!hadspc) {
+		putc(' ', fp);
+		++col;
+		++foldcol;
 	}
 
 	foldcol = col;
@@ -481,23 +527,25 @@ hdr_print(h, fp)
 	switch (sem) {
 	case Received:
 		if (h->h_lines != NULL) {
-		  /* Write out the original lines, if possible */
-		  int first = 1;
-		  for (t = h->h_lines; t != NULL; t = t->t_next) {
-		    const char *p = t->t_pname;
-		    int len = (int)(TOKENLEN(t));
-		    if (first) {
-		      while ((len > 0) &&
-			     (*p == ' '||*p == '\t'||*p == '\n'||*p == '\r'))
-			++p, --len;
-		      if (!len) continue;
-		      first = 0;
+		    /* Write out the original lines, if possible */
+		    int first = 1;
+		    for (t = h->h_lines; t != NULL; t = t->t_next) {
+			const char *p = t->t_pname;
+			int len = (int)(TOKENLEN(t));
+			if (first) {
+			    while ((len > 0) &&
+				   (*p == ' '||*p == '\t'||
+				    *p == '\n'||*p == '\r'))
+				++p, --len;
+			    if (!len) continue;
+			    first = 0;
+			}
+			fwrite(p, sizeof (char), len, fp);
+			putc('\n', fp);
 		    }
-		    fwrite(p, sizeof (char), len, fp);
-		    putc('\n', fp);
-		  }
-		  break;
+		    break;
 		}
+		hadspc = 1;
 		if ((ap = h->h_contents.r->r_from) != NULL) {
 		    if (ap->a_tokens == NULL) {
 			if (ap->a_pname[0] != '(') {
@@ -520,21 +568,22 @@ hdr_print(h, fp)
 			if (pp)
 			  col = printLAddress(fp, pp, col, 8, 0);
 			else
-			  fprintf(fp, "(nil??)"), col += 7;
+			  fprintf(fp, "(nil?\?)"), col += 7;
 		    }
+		    hadspc = 0;
 		}
 		if ((ap = h->h_contents.r->r_by) != NULL) {
-			cl = printAddress(NULL, ap->a_tokens, col+4);
+			cl = printAddress(NULL, ap->a_tokens, col+3);
 			if (cl > hdr_width) {
 			  fprintf(fp, "\n\t");
 			  col = 8;
-			} else
+			} else if (!hadspc)
 			  putc(' ',fp), ++col;
 			fprintf(fp, "by ");
 			col = printLAddress(fp, ap->a_tokens, col+3, 8, 0);
 		}
 		if ((t = h->h_contents.r->r_via) != NULL) {
-			cl = fprintToken(NULL, t, col+5);
+			cl = fprintToken(NULL, t, col+4);
 			if (cl > hdr_width) {
 			  fprintf(fp, "\n\t");
 			  col = 8;
@@ -544,7 +593,7 @@ hdr_print(h, fp)
 			col = fprintToken(fp, t, col+4);
 		}
 		for (t = h->h_contents.r->r_with; t != NULL; t = t->t_next) {
-			cl = fprintToken(NULL, t, col+6);
+			cl = fprintToken(NULL, t, col+5);
 			if (cl > hdr_width) {
 			  fprintf(fp, "\n\t");
 			  col = 8;
@@ -554,18 +603,17 @@ hdr_print(h, fp)
 			col = fprintToken(fp, t, col+5);
 		}
 		if ((ap = h->h_contents.r->r_id) != NULL) {
-			cl = printAddress(NULL, ap->a_tokens, col+6);
+			cl = printAddress(NULL, ap->a_tokens, col+3);
 			if (cl > hdr_width) {
 			  fprintf(fp, "\n\t");
 			  col = 8;
 			} else
 			  putc(' ',fp), ++col;
-			fprintf(fp, "id <");
+			fprintf(fp, "id ");
 			col = printLAddress(fp, ap->a_tokens, col+5, 8, 0);
-			putc('>', fp); ++col;
 		}
 		if ((ap = h->h_contents.r->r_for) != NULL) {
-			cl = printAddress(NULL, ap->a_tokens, col+5);
+			cl = printAddress(NULL, ap->a_tokens, col+4);
 			if (cl > hdr_width) {
 			  fprintf(fp, "\n\t");
 			  col = 8;
@@ -588,6 +636,7 @@ hdr_print(h, fp)
 		    fprintf(fp, " %s\n", cp);
 		}
 		break;
+
 	case Address:
 	case Addresses:
 	case AddressList:
@@ -600,7 +649,9 @@ hdr_print(h, fp)
 	case RouteAddressInAngles:
 	case RouteAddress:
 	case UserAtDomain:
+
 		newline = 1;
+
 		for (ap = h->h_contents.a; ap != NULL; ap = ap->a_next) {
 
 		  addrlen = printAddress(NULL, ap->a_tokens, col);
@@ -637,7 +688,7 @@ hdr_print(h, fp)
 
 		    /* Explicite prefold */
 		    putc('\n', fp);
-		    for (col = 0; col + 8  < foldcol; col += 8)
+		    for (col = 0; col + 8 <= foldcol; col += 8)
 		      putc('\t', fp);
 		    for (;col < foldcol; ++col)
 		      putc(' ', fp);
@@ -656,17 +707,71 @@ hdr_print(h, fp)
 	case DateTime:
 		/* Use this ONLY if no original lines are available! */
 		if (h->h_contents.d > 0L && h->h_lines == NULL) {
-			char *cp, *s;
-			cp = rfc822date(&(h->h_contents.d));
-			s = cp + strlen(cp) -1;
-			if (*cp != 0 && *s == '\n') *s = 0;
-			fputs(cp, fp);
-			break;
+		  char *cp, *s;
+		  cp = rfc822date(&(h->h_contents.d));
+		  s = cp + strlen(cp) -1;
+		  if (*cp != 0 && *s == '\n') *s = 0;
+		  fputs(cp, fp);
+		  break;
 		} /* else we have the original line ... */
 		/* Fall thru ... */
+
 	case nilHeaderSemantics:
 	default:
 		/* Write out the original lines, if possible */
+#if 0
+
+	  /* Damn...  This doesn't work all the time.. */
+
+		if (h->h_contents.t != NULL && sem == nilHeaderSemantics) {
+		  int first = 1;
+		  int newcol;
+		  for (t = h->h_contents.t; t; t = t->t_next) {
+
+		    const char *p = t->t_pname;
+
+		    if (t->t_type == Line) {
+		      int len = (int)(TOKENLEN(t));
+		      if (first) {
+			while ((len > 0) &&
+			       (*p == ' '||*p == '\t'||*p == '\n'||*p == '\r'))
+			  ++p, --len;
+			if (!len) continue;
+			first = 0;
+		      }
+		      fwrite(p, sizeof (char), len, fp);
+		      putc('\n', fp);
+		      continue;
+		    }
+
+		    if (first) {
+		      if (t->t_type == Space && *p != '\n')
+			continue;
+		      first = 0;
+		    }
+		    newcol = fprintToken(NULL, t, 0);
+		    if ((t->t_type == Space && *p == '\n' && t->t_next) ||
+			((newcol + col > hdr_width) &&
+			 (newcol + foldcol <= hdr_width))) {
+		      putc('\n', fp);
+		      for (col = 0; col+8 <= foldcol; col += 8)
+			putc('\t', fp);
+		      for (;col < foldcol; ++col)
+			putc(' ', fp);
+		      if (col < 1)
+			putc(' ', fp), ++col;
+		      first = 1;
+		      if (t->t_type == Space)
+			continue;
+		    }
+		    first = 0;
+		    col = fprintToken(fp, t, col);
+		  }
+		  if (first) /* totally empty header! */
+		    putc('\n', fp);
+		}
+		else
+#endif
 		{
 		  int first = 1;
 		  for (t = h->h_lines; t != NULL; t = t->t_next) {
@@ -685,7 +790,9 @@ hdr_print(h, fp)
 		  if (first) /* totally empty header! */
 		    putc('\n', fp);
 		}
+
 		break;
+
 	}
 }
 
@@ -851,23 +958,23 @@ printAddress(fp, pp, col)
 static int ThisAddressLen(pp)
      register struct addr *pp;
 {
-  int len = 0;
-  token822 *t;
-  int hadSpecial = (pp->p_type == aSpecial);
+	int len = 0;
+	token822 *t;
+	int hadSpecial = (pp->p_type == aSpecial);
 
-  if (hadSpecial)
-    /* Ok, this one is aSpecial '<', next one is anAddress.
-       Scan as long as there are Addresses. */
-    ++len;
+	if (hadSpecial)
+	  /* Ok, this one is aSpecial '<', next one is anAddress.
+	     Scan as long as there are Addresses. */
+	  ++len;
 
-  for (pp = pp->p_next; pp && pp->p_type == anAddress; pp = pp->p_next)
-    for (t = pp->p_tokens; t != NULL; t = t->t_next)
-      len = fprintToken(NULL, t, len);
+	for (pp = pp->p_next; pp && pp->p_type == anAddress; pp = pp->p_next)
+	  for (t = pp->p_tokens; t != NULL; t = t->t_next)
+	    len = fprintToken(NULL, t, len);
 
-  if (hadSpecial) /* Just imply it.. */
-    len += 2;
+	if (hadSpecial) /* Just imply it.. */
+	  len += 2;
 
-  return len;
+	return len;
 }
 
 /* Return the column where the cursor is */
@@ -908,7 +1015,7 @@ printLAddress(fp, pp, col, foldcol, canprefold)
 		    if (alen + col > hdr_width) {
 		      /* Ok, fold now */
 		      putc('\n', fp);
-		      for (col = 0; col+8  < foldcol; col += 8)
+		      for (col = 0; col+8  <= foldcol; col += 8)
 			putc('\t', fp);
 		      for (;col < foldcol; ++col)
 			putc(' ', fp);
@@ -976,7 +1083,7 @@ printLAddress(fp, pp, col, foldcol, canprefold)
 			      && (fprintToken(NULL, t->t_next, col+2)
 				  >= hdr_width)))) {
 			  putc('\n', fp);
-			  for (col = 0; col+8  < foldcol; col += 8)
+			  for (col = 0; col+8 <= foldcol; col += 8)
 			    putc('\t', fp);
 			  for (;col < foldcol; ++col)
 			    putc(' ', fp);
