@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 
@@ -21,6 +21,7 @@
 #include <sys/file.h>
 #include "mail.h"
 #include "zsyslog.h"
+#include "zmsignal.h"
 #include "interpret.h"
 #include "splay.h"
 
@@ -28,12 +29,6 @@
 
 extern const char *postoffice; /* At libzmailer.a: mail.c */
 
-#if 0
-extern struct group *getgrnam __();
-extern struct passwd *getpwnam __();
-extern struct passwd *getpwuid __();
-#endif
-
 extern struct shCmd fnctns[];
 extern time_t time __((time_t *));
 
@@ -53,15 +48,11 @@ extern conscell *s_value;
 
 #include "shmmib.h"
 
-struct MIB_MtaEntry MIBMtaEntryLcl;
-struct MIB_MtaEntry *MIBMtaEntry = &MIBMtaEntryLcl;
-
 int	mustexit = 0;
 int	canexit = 0;
 int	router_id = 0;
 int	deferit;
 int	deferuid;
-int	origoptind;
 int	savefile = 0;
 const char * zshopts = "-O";
 int	nosyslog = 1;
@@ -78,8 +69,8 @@ main(argc, argv)
 	int c, errflg, daemonflg, killflg, interactiveflg, tac;
 	int version;
 	long offout, offerr;
-	char *config, *cp;
-	const char *tav[20], *av[3];
+	char *config;
+	const char *tav[20], *av[3], *cp;
 #ifdef	XMEM
 	FILE *fp;
 #endif	/* XMEM */
@@ -106,7 +97,7 @@ main(argc, argv)
 		progname = argv[0];
 	else
 		++progname;
-	origoptind = optind;	/* needed for reuse of getopt() */
+
 	logfn = config = NULL;
 	errflg = daemonflg = killflg = interactiveflg = version = 0;
 	tac = 0;
@@ -114,7 +105,7 @@ main(argc, argv)
 
 
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "m:n:dikf:o:t:L:P:r:sSVwW");
+		c = zgetopt(argc, (char*const*)argv, "m:n:dikf:o:t:L:P:r:sSVwWZ:");
 		if (c == EOF)
 			break;
 	  
@@ -126,7 +117,7 @@ main(argc, argv)
 #ifdef	XMEM
 			{ /* Rewritten to be portable...  Storing to
 			     fileno()" is not guaranteed to success.. */
-			  int fd = open(optarg, O_RDWR|O_CREAT|O_TRUNC,0644);
+			  int fd = open(zoptarg, O_RDWR|O_CREAT|O_TRUNC,0644);
 			  if (fd >= 0) {
 			    dup2(fd,30); /* we ASSUME have far less fd's
 					    in use.. */
@@ -141,23 +132,23 @@ main(argc, argv)
 #endif	/* XMEM */
 			break;
 		case 'n':
-			nrouters = atoi(optarg);
+			nrouters = atoi(zoptarg);
 			if (nrouters < 1)
 				nrouters = 1;
 			break;
 		case 'o':
-			zshopts = optarg;
+			zshopts = zoptarg;
 			break;
 		case 't':
 			if (tac < (sizeof tav)/(sizeof tav[0]))
-				tav[++tac] = optarg;
+				tav[++tac] = zoptarg;
 			else {
 				fprintf(stderr, "Too many trace options!\n");
-				fprintf(stderr, "Ignoring '%s'\n", optarg);
+				fprintf(stderr, "Ignoring '%s'\n", zoptarg);
 			}
 			break;
 		case 'f':	/* override default config file */
-			config = optarg;
+			config = (char*) zoptarg;
 			break;
 		case 'i':	/* first read config file, then read from tty */
 			interactiveflg = 1;
@@ -173,10 +164,10 @@ main(argc, argv)
 			nosyslog = 0;
 			break;
 		case 'L':	/* override default log file */
-			logfn = optarg;
+			logfn = zoptarg;
 			break;
 		case 'P':	/* override default postoffice */
-			postoffice = optarg;
+			postoffice = zoptarg;
 			break;
 		case 'V':
 			version = 1;
@@ -188,29 +179,33 @@ main(argc, argv)
 			do_hdr_warning = !do_hdr_warning;
 			break;
 		case 'r':
-			routerdirloops = atoi(optarg);
+			routerdirloops = atoi(zoptarg);
 			if (routerdirloops < 0)
 				routerdirloops = 0;
 			break;
+		case 'Z':
+			if (readzenv(zoptarg) == 0)
+			  ++errflg;
+			break;
 		case '?':
 		default:
-			errflg++;
+			++errflg;
 			break;
 		}
 	}
 
 	if (errflg || (interactiveflg && daemonflg)) {
 		fprintf(stderr,
-			"Usage: %s [ -dikV -n #routers -t traceflag -f configfile -L logfile -P postoffice]\n",
+			"Usage: %s [ -dikV -n #routers -t traceflag -f configfile -L logfile -P postoffice -Z zenvfile]\n",
 			progname);
 		exit(128+errflg);
 	}
+
 	time(&now);
 	mailshare = getzenv("MAILSHARE");
 	if (mailshare == NULL)
 		mailshare = MAILSHARE;
 	if (config == NULL) {
-		config = cf_suffix; /* XX: ???? delete this ? */
 		/* we don't need to remember this for long */
 		config = smalloc(MEM_TEMP, 3 + (u_int)(strlen(mailshare)
 					     + strlen(progname)
@@ -222,6 +217,46 @@ main(argc, argv)
 	    (postoffice = getzenv("POSTOFFICE")) == NULL)
 		postoffice = POSTOFFICE;
 
+
+	if (daemonflg || killflg) {
+
+	  /* Daemon attaches the SHM block, and may complain, but will not
+	     give up..  instead uses builtin fallback  */
+
+	  int r = Z_SHM_MIB_Attach (1);  /* R/W mode */
+
+	  if (r < 0) {
+	    /* Error processing -- magic set of constants: */
+	    switch (r) {
+	    case -1:
+	      /* fprintf(stderr, "No ZENV variable: SNMPSHAREDFILE\n"); */
+	      break;
+	    case -2:
+	      perror("Failed to open for exclusively creating of the SHMSHAREDFILE");
+	      break;
+	    case -3:
+	      perror("Failure during creation fill of SGMSHAREDFILE");
+	      break;
+	    case -4:
+	      perror("Failed to open the SHMSHAREDFILE at all");
+	      break;
+	    case -5:
+	      perror("The SHMSHAREDFILE isn't of proper size! ");
+	      break;
+	    case -6:
+	      perror("Failed to mmap() of SHMSHAREDFILE into memory");
+	      break;
+	    case -7:
+	      fprintf(stderr, "The SHMSHAREDFILE  has magic value mismatch!\n");
+	      break;
+	    default:
+	      break;
+	    }
+	    /* return; NO giving up! */
+	  }
+	}
+
+
 	if (killflg && !daemonflg) {
 		killprevious(-SIGTERM, pidfile);
 		exit(0);
@@ -229,16 +264,18 @@ main(argc, argv)
 
 	getnobody();
 
-	c = optind;	/* save optind since builtins can interfere with it */
+	c = zoptind;	/* save optind since builtins can interfere with it */
 
 	if (daemonflg && logfn == NULL) {
-		if ((cp = getzenv("LOGDIR")) != NULL)
+		if ((cp = (char *) getzenv("LOGDIR")) != NULL)
 			logdir = cp;
 		logfn = smalloc(MEM_PERM, 2 + (u_int)(strlen(logdir)
 					  + strlen(progname)));
 		sprintf((char*)logfn, "%s/%s", logdir, progname);
 	}
 
+
+
 	if (logfn != NULL) {
 		/* loginit is a signal handler, so can't pass log */
 		if (loginit(SIGHUP) < 0) /* do setlinebuf() there */
@@ -260,6 +297,10 @@ main(argc, argv)
 
 	stickymem = MEM_PERM;
 
+	/* We (and our children) run with SIGPIPE ignored.. */
+	SIGNAL_HANDLE(SIGPIPE, SIG_IGN);
+
+
 	initialize(config, argc - c, &argv[c]);
 
 	stickymem = MEM_TEMP;	/* this is the default allocation type */
@@ -270,6 +311,8 @@ main(argc, argv)
 	mal_leaktrace(1);
 #endif /* MALLOC_TRACE */
 
+
+
 	if (daemonflg) {
 		if (chdir(postoffice) < 0 || chdir(ROUTERDIR) < 0)
 		  fprintf(stderr, "%s: cannot chdir.\n", progname);
@@ -316,7 +359,7 @@ main(argc, argv)
 	if (c < argc) {
 	  savefile = 1;
 	  /*
-	   * we need to use a local variable (c) because optind is global
+	   * we need to use a local variable (c) because zoptind is global
 	   * and can (and will) be modified by the funcall()'s we do.
 	   */
 	  do {
@@ -363,11 +406,12 @@ initialize(configfile, argc, argv)
 	int argc;
 	const char *argv[];
 {
-	struct group *grp;
+	struct Zgroup *grp;
 	struct sptree_init *sptip;
 	int ac;
 	const char **cpp;
 	const char **av;
+	const char *zconfig;
 
 	av = (const char **)emalloc((5+argc)*(sizeof (char *)));
 	/* initialize shell */
@@ -416,12 +460,16 @@ initialize(configfile, argc, argv)
 		add_incoresp(*cpp, "", spt_goodguys);
 
 	if (files_group != NULL) {
-		if ((grp = getgrnam(files_group)) == NULL)
+		if ((grp = zgetgrnam(files_group)) == NULL)
 			files_gid = -1;
 		else
 			files_gid = grp->gr_gid;
 	}
 
+	zconfig = getzenv("ZCONFIG");
+	if (zconfig)
+	  v_set("ZCONFIG", zconfig);
+
 	/* source the router config file */
 	ac = 0;
 	av[ac++] = ".";
@@ -442,7 +490,7 @@ int
 login_to_uid(name)
 	const char	*name;
 {
-	struct passwd *pw;
+	struct Zpasswd *pw;
 	uid_t uid;
 	char buf[BUFSIZ];
 	char *cp;
@@ -453,7 +501,11 @@ login_to_uid(name)
 		memtypes oval = stickymem;
 
 		stickymem = MEM_MALLOC;
-		pw = getpwnam(name);
+
+		pw = zgetpwnam(name);
+		if (!pw)
+			pw = zgetpwnam(name);
+
 		if (pw == NULL) {
 			uid = nobody;
 		} else {
@@ -474,14 +526,14 @@ const char *
 uidpwnam(uid)
 	int	uid;
 {
-	struct passwd *pw;
+	struct Zpasswd *pw;
 	register const char *cp;
 	struct spblk *spl;
 	char buf[BUFSIZ];
 
 	spl = sp_lookup((u_long)uid, spt_uidmap);
 	if (spl == NULL) {
-		pw = getpwuid((uid_t)uid);
+		pw = zgetpwuid((uid_t)uid);
 		if (pw == NULL) {
 			/* memory shall be temporary in
 			   its nature for this data! */
@@ -605,10 +657,6 @@ logit(file, id, from, to)
 	if (id == NULL)
 		id = file;
 	baselen = strlen(file) + strlen(id) + 4;
-	printf("%.200s: file: %s %s...\n", id, file, from);
-	if (!nosyslog)
-	  zsyslog((LOG_INFO, "%.200s: file: %.150s %.250s...",
-		   id, file, from));
 	flen = strlen(from);
 	if (flen > 0)
 	  printf("%.200s: file: %s %s => %s\n", id, file, from, to);
