@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1996
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2002
  */
 
 /* stdout and stderr end up in qlogdir/progname when scheduler is a daemon */
@@ -32,6 +32,9 @@ int	nobody = -2;
 /* directory scanning interval in seconds */
 int	sweepinterval = 10;
 
+/* directory idle-scanning interval in seconds */
+int	idle_sweepinterval = 120;
+
 /* At most 30 new childs per second -- The 'R'-option can be used to increase this */
 int	forkrate_limit = 30;
 
@@ -40,8 +43,14 @@ int	default_full_content = 1;
 
 /* If we use ZMailer malloc library in debug mode, BAD-PTR looks like
    following value.. */
-void *BADPTR = (void*) 0x5555555555555555L;
+void *BADPTR = (void*) (unsigned long)0x5555555555555555L;
 
 /* Push reports regarding recipients out this often, unless
    parametrized to do differently */
-int global_report_interval = 900; /* 15min */
+int global_report_interval = 5*60; /* 5 min */
+
+/* Limit the time spent at one time at the cleanup scrubber */
+int expiry2_timelimit = 10; /* 10 seconds.. */
+
+/* How frequently the rough cleanup is run ? */
+int expiry2_sweepinterval = 900;  /* 15 minutes */
