@@ -4,12 +4,25 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 /*
- * This program must be installed suid to the uid the scheduler runs as
- * (usually root).  Unfortunately.
+ * Rayan 1988:
+ *  This program must be installed suid to the uid the scheduler runs as
+ *  (usually root).  Unfortunately.
+ *
+ * mea 1990:
+ *  This program can be run without suid-root -- depending on what one
+ *  needs of special features, e.g. if one aspires to see the verbose
+ *  queue printout along with exact message source and destination
+ *  addresses, message-ids, sizes, ...  Either run as root, or suid-root.
+ *
+ * mea 2001:
+ *  What has been true for 10+ years is still true, several things can
+ *  now be done without any sort of suid-privileges, others may need access
+ *  to the actual message files and need e.g. root powers.
+ *  Autentication for using 'MAILQv2' is orthogonal from suid:ing this.
  */
 
 
@@ -45,8 +58,10 @@
 # endif /* HAVE_NDIR_H */
 #endif /* HAVE_DIRENT_H */
 
+#ifdef HAVE_SYS_LOADAVG_H
+#include <sys/loadavg.h>
+#endif
 
-#if	defined(HAVE_SOCKET) && (defined(HAVE_RESOLVER) || defined(HAVE_YP))
 #include <netdb.h>
 #ifndef EAI_AGAIN
 # include "netdb6.h"
@@ -56,7 +71,7 @@
 #include <arpa/inet.h>
 #include <fcntl.h>
 #include <sys/file.h>
-#endif	/* HAVE_SOCKET */
+
 #ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
 #endif
@@ -73,9 +88,12 @@ int	D_alloc = 0;
 #include "libz.h"
 #include "libc.h"
 
-extern char *qoutputfile;
+#include "ta.h"
+
 extern int errno, pokedaemon();
 
+static void print_shm __((void));
+
 #ifndef strchr
 extern char *strchr(), *strrchr();
 #endif
@@ -101,7 +119,17 @@ char	path[MAXPATHLEN];
 #define  ISSPACE(cc) (cc == ' ' || cc == '\t')
 
 
-char *host = NULL;
+typedef struct threadtype {
+  const char *channel;
+  const char *host;
+  char *line;
+} threadtype;
+
+
+const char *host = NULL;
+
+const char *channel_opt = NULL;
+const char *host_opt    = NULL;
 
 int
 main(argc, argv)
@@ -111,32 +139,52 @@ main(argc, argv)
 	int fd, c, errflg, eval;
 	struct passwd *pw;
 #ifndef	AF_INET
-	char *rendezvous = NULL;
+	const char *rendezvous = NULL;
 	FILE *fp;
 	struct stat stbuf;
 	int r, pid, dsflag;
 #endif	/* AF_INET */
+	int prefer_4 = 0, prefer_6 = 0;
 	char *expn = NULL;
 
 	progname = argv[0];
 	verbose = debug = errflg = status = user = onlyuser = summary = 0;
+	
 	while (1) {
-	  c = getopt(argc, argv, "dip:r:stu:U:vVSQE:K:");
+	  c = getopt(argc, argv, "46c:dE:h:iK:Mp:Qr:sStu:U:vVZ:");
 	  if (c == EOF)
 	    break;
 	  switch (c) {
+	  case '4':
+	    prefer_4 = 1;
+	    prefer_6 = 0;
+	    break;
+	  case '6':
+	    prefer_4 = 0;
+	    prefer_6 = 1;
+	    break;
+	  case 'c':
+	    channel_opt = optarg;
+	    break;
 	  case 'd':
 	    ++debug;
 	    break;
 	  case 'E':
 	    expn = optarg;
 	    break;
+	  case 'h':
+	    host_opt = optarg;
+	    if (!channel_opt) channel_opt = "smtp";
+	    break;
 	  case 'i':
 	    user = getuid();
 	    onlyuser = 1;
 	    if (verbose == 0)
 	      ++verbose;
 	    break;
+	  case 'M':
+	    print_shm();
+	    break;
 #if defined(AF_INET) || defined(AF_UNIX)
 	  case 'p':
 	    port = optarg;
@@ -196,6 +244,10 @@ main(argc, argv)
 	    prversion("mailq");
 	    exit(0);
 	    break;
+	  case 'Z':
+	    if (readzenv(optarg) == 0)
+	      ++errflg;
+	    break;
 	  default:
 	    ++errflg;
 	    break;
@@ -216,9 +268,9 @@ main(argc, argv)
 	}
 	if (errflg) {
 #ifdef	AF_INET
-	  fprintf(stderr, "Usage: %s [-isSvt] [-p#] [host]\n", progname);
+	  fprintf(stderr, "Usage: %s [-46isSvt] [-Z zenvcfgfile] [-cchannel -hhost] [-p#] [host]\n", progname);
 #else  /* !AF_INET */
-	  fprintf(stderr, "Usage: %s [-isSvt]\n", progname);
+	  fprintf(stderr, "Usage: %s [-isSvt] [-Z zenvcfgfile] [-cchannel -hhost]\n", progname);
 #endif /* AF_INET */
 	  exit(EX_USAGE);
 	}
@@ -324,9 +376,13 @@ main(argc, argv)
 	    rc = _getaddrinfo_(host, "0", &req, &ai,
 			       (debug ? stderr : NULL));
 #endif
-#if defined(INET) && defined(INET6)
+#if defined(AF_INET6) && defined(INET6)
 	    {
 	      struct addrinfo *ai6;
+	      memset(&req, 0, sizeof(req));
+	      req.ai_socktype = SOCK_STREAM;
+	      req.ai_protocol = IPPROTO_TCP;
+	      req.ai_flags    = AI_CANONNAME;
 	      req.ai_family   = AF_INET6;
 	      ai6 = NULL;
 	      
@@ -340,12 +396,19 @@ main(argc, argv)
 		/* No IPv4, but have IPv6! */
 		ai = ai6;
 	      else if (ai && ai6) {
-		/* Catenate them, FIRST IPv6, then IPv4 things. */
 		struct addrinfo **aip;
-		aip = &ai6->ai_next;
-		while (*aip) aip = &(*aip)->ai_next;
-		*aip = ai;
-		ai = ai6;
+		if (prefer_4) {
+		  /* Catenate them, FIRST IPv4, then IPv6 things. */
+		  aip = &ai->ai_next;
+		  while (*aip) aip = &(*aip)->ai_next;
+		  *aip = ai6;
+		} else {
+		  /* Catenate them, FIRST IPv6, then IPv4 things. */
+		  aip = &ai6->ai_next;
+		  while (*aip) aip = &(*aip)->ai_next;
+		  *aip = ai;
+		  ai = ai6;
+		}
 	      }
 	    }
 #endif
@@ -355,7 +418,8 @@ main(argc, argv)
 	    }
 	    stashmyaddresses(NULL);
 	    if (ai && matchmyaddresses(ai) == 0) {
-	      fprintf(stdout, "[%s]\n", ai->ai_canonname);
+	      /* BSD systems can yield ai_canonname member NULL! */
+	      fprintf(stdout, "[%s]\n", ai->ai_canonname ? ai->ai_canonname : host);
 	      nonlocal = 1;
 	    } else
 	      nonlocal = 0;	/* "localhost" is per default a "local" */
@@ -395,8 +459,8 @@ main(argc, argv)
 	    if (rc < 0) {
 	      eval = errno;
 	      close(fd);
-	      fprintf(stderr, "%s: connect failed to %s",
-		      progname, ai->ai_canonname);
+	      fprintf(stderr, "%s: connect failed to %s\n",
+		      progname, ai->ai_canonname ? ai->ai_canonname : host);
 	      fd = -1;
 	      continue;
 	    }
@@ -404,7 +468,7 @@ main(argc, argv)
 	  if (fd >= 0)
 	    docat((char *)NULL, fd);
 	  else {
-	    fprintf(stderr, "%s: connect failed to %s",
+	    fprintf(stderr, "%s: connect failed to %s\n",
 		    progname, host);
 	  }
 	}
@@ -594,7 +658,7 @@ const char *dirpath;
 	dirp = opendir(dirpath);
 	if (dirp == NULL) {
 	  fprintf(stderr, "%s: opendir(%s): %s\n",
-		  progname, dpath, strerror(errno));
+		  progname, dirpath, strerror(errno));
 	  return -1;
 	}
 	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
@@ -651,15 +715,14 @@ checkrouter()
 	if (postoffice == NULL)
 	  return;
 	sprintf(path, "%s/%s", postoffice, ROUTERDIR);
-
 	n = countfiles(path);
-
-	fprintf(stdout,"%d entr%s in router queue: ", n, n != 1 ? "ies" : "y");
+	fprintf(stdout,"%d entr%s in %s/ directory: ", n, n != 1 ? "ies" : "y", path);
 
 	if (nonlocal)
 	  r = -2;
 	else
 	  r = isalive(PID_ROUTER, &pid, &fp);
+
 	switch (r) {
 	case EX_UNAVAILABLE:
 	  /* if the .router.pid file is younger than any core file,
@@ -691,7 +754,7 @@ checkrouter()
 	sprintf(path, "%s/%s", postoffice, DEFERREDDIR);
 	dirp = opendir(path);
 	if (dirp == NULL) {
-	  fprintf(stderr, "%s: opendir(%s): %s",
+	  fprintf(stderr, "%s: opendir(%s): %s\n",
 		  progname, path, strerror(errno));
 	  return;
 	}
@@ -716,7 +779,7 @@ checkrouter()
 void
 checkscheduler()
 {
-	int pid, n, r;
+	int pid, n, n2, r;
 	FILE *fp;
 
 	if (postoffice == NULL)
@@ -726,11 +789,9 @@ checkscheduler()
 	othern  = 0;
 
 	sprintf(path, "%s/%s", postoffice, TRANSPORTDIR);
-
 	n = countfiles(path);
-
-	fprintf(stdout,"%d message%s in transport queue: ",
-	       n, n != 1 ? "s" : "");
+	fprintf(stdout,"%d message%s in %s/ directory: ",
+	       n, n != 1 ? "s" : "", path);
 
 	if (nonlocal)
 	  r = -2;
@@ -761,6 +822,11 @@ checkscheduler()
 	  fprintf(stdout," (core exists)");
 	fprintf(stdout,"\n");
 
+	sprintf(path, "%s/%s", postoffice, QUEUEDIR);
+	n2 = countfiles(path);
+	if (n != n2)
+	  fprintf(stdout,"%d message%s in %s/ directory\n",
+		  n2, n2 != 1 ? "s" : "", path);
 }
 
 int
@@ -946,7 +1012,8 @@ if (debug)
 	      cfp->fd = -1;
 	      cfp->haderror = 0;
 	      cfp->head = NULL;
-	      cfp->nlines = 0;
+	      cfp->nlines        = 0;
+	      cfp->msgbodyoffset = 0;
 	      cfp->contents = NULL;
 	      cfp->logident = NULL;
 	      cfp->id = 0;
@@ -1092,9 +1159,9 @@ if (debug)
 
 static int r_i;
 
-extern int repscan __((struct spblk *));
+extern int repscan_v1 __((struct spblk *));
 int
-repscan(spl)
+repscan_v1(spl)
 	struct spblk *spl;
 {
 	register struct vertex *v, *vv;
@@ -1181,6 +1248,104 @@ repscan(spl)
 	return 0;
 }
 
+static struct ctlfile *readmq2cfp __((const char *fname));
+static struct ctlfile *readmq2cfp(fname)
+     const char *fname;
+{
+	struct ctlfile *cfp = NULL;
+	int i, fd, once;
+	struct stat stbuf;
+	char *s, *s0;
+
+	sprintf(path, "%s/%s/%s", postoffice, TRANSPORTDIR, fname);
+	if (lstat(path, &stbuf) != 0) return NULL;
+
+	cfp = malloc(sizeof(*cfp) + stbuf.st_size + 20);
+	if (!cfp) return NULL;
+
+	fd = open(path,O_RDONLY,0);
+	if (fd < 0) {
+	  /* whatever reason */
+	  free(cfp);
+	  return NULL;
+	}
+
+	s0 = (char *)(cfp+1);
+
+	i = read(fd, s0, stbuf.st_size);
+	close(fd);
+
+	memset(cfp, 0, sizeof(*cfp));
+	cfp->contents = s0;
+	cfp->nlines = stbuf.st_size; /* reuse the variable .. */
+
+	if (i != stbuf.st_size) {
+	  /* whatever reason.. */
+	  free(cfp);
+	  return NULL;
+	}
+
+	sprintf(path, "%s/%s/%s", postoffice, QUEUEDIR, fname);
+	if (lstat(path, &stbuf) == 0) {
+	  cfp->msgbodyoffset = stbuf.st_size;
+	  cfp->mtime         = stbuf.st_mtime;
+	}
+
+	s0[i] = 0;
+
+	once = 1;
+	for (s = s0; i > 0; ++s, --i) {
+
+	  char c;
+	  char *p;
+
+	  if (*s == '\n') {
+	    --i; ++s;
+	  }
+	  c = *s;
+	  once = 0;
+	  --i; ++s;
+	  --i; ++s;
+	  if (i > 0)
+	    p = memchr(s, '\n', i);
+	  else
+	    break;
+	  if (!p) break;
+	  switch(c) {
+	  case _CF_FORMAT:
+	    *p = 0;
+	    cfp->format = 0;
+	    sscanf(s, "%i", &cfp->format);
+	    i -= (p - s);
+	    s = p;
+	    break;
+	  case _CF_LOGIDENT:
+	    cfp->logident = s;
+	    *p = 0;
+	    i -= (p - s);
+	    s = p;
+	    break;
+	  case _CF_MSGHEADERS:
+	  case _CF_MIMESTRUCT:
+	    for (;i > 1; ++s, --i) {
+	      if (s[0] == '\n' && s[1] == '\n') {
+		*s = 0;
+		break;
+	      }
+	    }
+	    break;
+	  default:
+	    *p = 0;
+	    i -= (p - s);
+	    s = p;
+	    break;
+	  }
+	}
+
+	return cfp;
+}
+
+
 void query2 __((FILE *, FILE*));
 void query2(fpi, fpo)
 	FILE *fpi, *fpo;
@@ -1192,6 +1357,7 @@ void query2(fpi, fpo)
 	char *buf = NULL;
 	MD5_CTX CTX;
 	unsigned char digbuf[16];
+	struct ctlfile *cfp = NULL;
 
 	/* Authenticate the query - get challenge */
 	bufsize = 0;
@@ -1199,8 +1365,8 @@ void query2(fpi, fpo)
 	  return;
 
 	MD5Init(&CTX);
-	MD5Update(&CTX, challenge, strlen(challenge));
-	MD5Update(&CTX, v2password, strlen(v2password));
+	MD5Update(&CTX, (const void *)challenge,  strlen(challenge));
+	MD5Update(&CTX, (const void *)v2password, strlen(v2password));
 	MD5Final(digbuf, &CTX);
 	
 	fprintf(fpo, "AUTH %s ", v2username);
@@ -1223,12 +1389,23 @@ void query2(fpi, fpo)
 
 	if (schedq) {
 
-	  if (schedq > 2)
+	  switch (schedq) {
+	  case 4:
+	    strcpy(buf,"SHOW COUNTERS\n");
+	    break;
+	  case 3:
 	    strcpy(buf,"SHOW SNMP\n");
-	  else if (schedq > 1)
+	    break;
+	  case 2:
 	    strcpy(buf,"SHOW QUEUE SHORT\n");
-	  else
+	    break;
+	  case 1:
 	    strcpy(buf,"SHOW QUEUE THREADS\n");
+	    break;
+	  default:
+	    fprintf(stdout, "Bad -Q... command\n");
+	    return;
+	  }
 
 	  len = strlen(buf);
 
@@ -1258,54 +1435,76 @@ void query2(fpi, fpo)
 	    }
 
 	  }
+
+	  fprintf(fpo, "QUIT\n");
+	  fflush(fpo);
+
+	  close(FILENO(fpi));
+
 	} else {
 
 	  /* Non -Q* -mode processing */
 
 	  int linespace = 256;
-	  int linecnt = 0;
+	  int linecnt   = 0;
 	  char **lines = (char **) malloc(sizeof(char *) * linespace);
+	  int threadspace = 256;
+	  int threadcnt   = 0;
+	  threadtype *threads = (threadtype *) malloc(sizeof(threadtype) *
+						      threadspace);
 
-	  fprintf(fpo, "SHOW QUEUE THREADS2\n");
-	  fflush(fpo);
+	  if (channel_opt && host_opt) {
 
-	  bufsize = 0;
-	  if (GETLINE(buf, bufsize, bufspace, fpi))
-	    return;
+	    int i = strlen(channel_opt) + strlen(host_opt);
+	    lines[linecnt] = malloc(i+2);
+	    sprintf(lines[linecnt], "%s\t%s", channel_opt, host_opt);
+	    ++linecnt;
 
-	  if (*buf != '+') {
-	    fprintf(stdout,"Scheduler response: '%s'\n",buf);
-	    return;
-	  }
+	  } else {
+
+	    fprintf(fpo, "SHOW QUEUE THREADS2\n");
+	    fflush(fpo);
 
-	  for (;;) {
-	    char *b;
 	    bufsize = 0;
 	    if (GETLINE(buf, bufsize, bufspace, fpi))
-	      break;
-	    if (buf[0] == '.' && buf[1] == 0)
-	      break;
+	      return;
 
-	    if (linecnt+1 >= linespace) {
-	      linespace *= 2;
-	      lines = (char **)realloc((void**)lines,
-				       sizeof(char *) * linespace);
+	    if (*buf != '+') {
+	      fprintf(stdout,"Scheduler response: '%s'\n",buf);
+	      return;
 	    }
 
-	    /* Do leading dot duplication suppression */
-	    b = buf;
-	    if (*b == '.') {
-	      --bufsize;
-	      ++b;
+	    for (;;) {
+	      char *b;
+	      bufsize = 0;
+	      if (GETLINE(buf, bufsize, bufspace, fpi))
+		break;
+	      if (buf[0] == '.' && buf[1] == 0)
+		break;
+
+	      if (linecnt+1 >= linespace) {
+		linespace *= 2;
+		lines = (char **)realloc((void**)lines,
+					 sizeof(char *) * linespace);
+	      }
+
+	      /* Do leading dot duplication suppression */
+	      b = buf;
+	      if (*b == '.') {
+		--bufsize;
+		++b;
+	      }
+
+	      lines[linecnt] = malloc(bufsize+2);
+	      memcpy(lines[linecnt], b, bufsize+1);
+	      ++linecnt;
+
+	      /* fprintf(stdout,"%s\n", b); */
 	    }
-
-	    lines[linecnt] = malloc(bufsize+2);
-	    memcpy(lines[linecnt], b, bufsize+1);
-	    lines[++linecnt] = NULL;
-
-	    /* fprintf(stdout,"%s\n", b); */
 	  }
 
+	  lines[linecnt] = NULL;
+
 	  for (i = 0; lines[i] != NULL; ++i) {
 	    char *channel = lines[i];
 	    char *host    = strchr(channel, '\t');
@@ -1323,20 +1522,15 @@ void query2(fpi, fpo)
 
 	    bufsize = 0;
 	    if (GETLINE(buf, bufsize, bufspace, fpi))
-	      return;
+	      break; /* Response */
 
 	    if (*buf != '+') {
 	      fprintf(stdout,"Scheduler response: '%s'\n",buf);
-	      return;
+	      break;
 	    }
 
-
-	    printf("%s/%s:\n",channel, host);
-
 	    for (;;) {
-	      int j;
-	      char *split[10];
-	      char timebuf[30];
+
 	      bufsize = 0;
 	      if (GETLINE(buf, bufsize, bufspace, fpi))
 		break;
@@ -1350,57 +1544,198 @@ void query2(fpi, fpo)
 		++b;
 	      }
 
-	      /* Array elts:
-		 0) filepath under $POSTOFFICE/transport/
-		 1) error address in brackets
-		 2) recipient line offset within the control file
-		 3) message expiry time (time_t)
-		 4) next wakeup time (time_t)
-		 5) last feed time (time_t)
-		 6) count of attempts at the delivery
-		 7) "retry in NNN" or a pending on "channel"/"thread"
-		 8) possible diagnostic message from previous delivery attempt
-	      */
-
-	      for (j = 0; b && j < 9; ++j) {
-		split[j] = b;
-		b = strchr(b ? b : "", '\t');
-		if (b) *b++ = 0;
-	      }
-	      
-	      if (j != 9) {
-		fprintf(stderr,"Communication error! Malformed data entry!\n");
-		continue;
+	      if (threadcnt+2 >= threadspace) {
+		threadspace *= 2;
+		threads = (threadtype *)realloc((void*)threads,
+						sizeof(threadtype) *
+						threadspace);
 	      }
 
-	      printf("\t%s: (", split[0]);
+	      threads[threadcnt].channel = channel;
+	      threads[threadcnt].host    = host;
+	      threads[threadcnt].line    = malloc(bufsize + 2);
+	      memcpy(threads[threadcnt].line, b, bufsize+1);
+	      ++threadcnt;
+	    }
+	    
+	  }
 
-	      *timebuf = 0;
-	      saytime((long)(atol(split[3]) - now), timebuf, 1);
+	  threads[threadcnt].channel = NULL;
+	  threads[threadcnt].host    = NULL;
+	  threads[threadcnt].line    = NULL;
 
-	      printf("%s tries, expires in %s", split[6], timebuf);
+	  fprintf(fpo, "QUIT\n");
+	  fflush(fpo);
 
-	      printf(") %s\n", split[8]);
+	  close(FILENO(fpi));
 
-	      if (verbose) {
-		/* XXX: Pick message file control data to present old-style
-		   verbose report */
-		/* XXX: printf("\t  id\t%s", ... );
-		   XXX: printf("...\n"); */
-		printf("\t  from\t%s\n", *split[1] ? split[1] : "<>");
-		/* XXX: TO line ... */
-	      }
+	  for (i = 0; threads[i].line != NULL; ++i) {
+	    static const char *channel = NULL;
+	    static const char *host    = NULL;
+
+	    int j;
+	    char *split[11], *s, *ocp, *b;
+	    char timebuf[30];
+
+	    if (channel != threads[i].channel ||
+		host    != threads[i].host) {
+
+	      channel = threads[i].channel;
+	      host    = threads[i].host;
+
+	      printf("%s/%s:\n",channel, host);
 
 	    }
 
-	  }
+	    b = threads[i].line;
 
-	}
+
+	    /* Array elts:
+	       0) filepath under $POSTOFFICE/transport/
+	       1) number WITHIN a group of recipients
+	       2) error address in brackets
+	       3) recipient line offset within the control file
+	       4) message expiry time (time_t)
+	       5) next wakeup time (time_t)
+	       6) last feed time (time_t)
+	       7) count of attempts at the delivery
+	       8) "retry in NNN" or a pending on "channel"/"thread"
+	       9) possible diagnostic message from previous delivery attempt
+	    */
+
+	    for (j = 0; b && j < 10; ++j) {
+	      split[j] = b;
+	      if (j == 1) {
+		/* The 'number within group' got added here
+		   after the rest of the interface was working. */
+		if (!('0' <= *b && *b <= '9')) {
+		  split[1] = "0";
+		  ++j;
+		  split[j] = b;
+		}
+	      }
+	      b = strchr(b, '\t');
+	      if (b) *b++ = 0;
+	    }
+	      
+	    if (j != 10) {
+	      fprintf(stderr,"Communication error! Malformed data entry!\n");
+	      continue;
+	    }
+
+	    j = atoi(split[1]);
+
+	    if (j == 0) {
+
+	      printf("\t%s: (", split[0]);
+
+	      if (!verbose) {
+		/* First recipient in the group */
+		printf("%s tries, ", split[7]);
+
+		*timebuf = 0;
+		saytime((long)(atol(split[4]) - now), timebuf, 1);
+
+		printf("expires in %s)", timebuf);
+
+		printf(" %s\n", split[9]);
+	      } /* !verbose */
+	    }
+
+
+	    if (verbose) {
+	      if (j == 0) {
+		if (cfp) free(cfp);
+		cfp = readmq2cfp(split[0]);
+	      }
+	      if (cfp) {
+		if (j == 0) {
+		  /* First recipient in the group */
+
+		  *timebuf = 0;
+		  saytime((long)(now - cfp->mtime), timebuf, 1);
+
+		  printf("%s tries, age %s, ", split[7], timebuf);
+
+		  *timebuf = 0;
+		  saytime((long)(atol(split[4]) - now), timebuf, 1);
+
+		  printf("expires in %s, %ld+%ld bytes)", timebuf,
+			 (long)cfp->nlines, (long)cfp->msgbodyoffset);
+
+		  printf("\n");
+
+		  if (cfp->logident)
+		    printf("\t  id\t%s\n", cfp->logident);
+
+		  printf("\t  from\t%s\n", *split[2] ? split[2] : "<>");
+		}
+
+		s = cfp->contents + atoi(split[3]) +2;
+		if (s > (cfp->contents + cfp->nlines)) {
+		  printf("\t\tto-ptr bad; split[3]='%s'\n",split[3]);
+		  continue; /* BAD! */
+		}
+
+		if (*s == ' ' || (*s >= '0' && *s <= '9'))
+		  s += _CFTAG_RCPTPIDSIZE;
+
+		if ((cfp->format & _CF_FORMAT_DELAY1) || *s == ' ' ||
+		    (*s >= '0' && *s <= '9')) {
+		  /* Newer DELAY data slot - _CFTAG_RCPTDELAYSIZE bytes */
+		  s += _CFTAG_RCPTDELAYSIZE;
+		}
+
+		s = skip821address(s); /* skip channel */
+		while (*s == ' ' || *s == '\t') ++s;
+		s = skip821address(s); /* skip host */
+		while (*s == ' ' || *s == '\t') ++s;
+
+		ocp = s;
+		s = skip821address(s); /* skip user */
+		*s++ = 0;
+		fprintf(stdout,"\t  to\t%s",ocp);
+
+		/* XXX: ORCPT data! */
+		
+
+		fprintf(stdout,"\n");
+
+	      } else /* not have cfp */ {
+		/* Can't show 'message-id', nor 'to' addresses,
+		   but have 'from'! */
+		if (j == 0) {
+		  /* First recipient in the group */
+		  printf("\t  from\t%s\n", *split[2] ? split[2] : "<>");
+		}
+	      }
+
+	      /* remember to show the diagnostics */
+
+	      /* Show all CR separated sub-lines as their OWN 'diag' lines! */
+
+	      s = split[9];
+	      while (*s == '\r') ++s;
+	      while (*s) {
+		printf("\t  diag\t");
+		for (;*s && *s != '\r'; ++s) putchar(*s);
+		putchar('\n');
+		while (*s == '\r') ++s;
+	      }
+
+	    } /* verbose */
+
+	  } /* all recipients towards each host */
+	  /* all channel/host pairs */	  
+
+	} /* No -Q processing */
+
+	free(cfp);
 }
 
 void
 report(fpi,fpo)
-	FILE *fpi, *fpo;
+     FILE *fpi, *fpo;
 {
 	int rc = parse(fpi);
 	if (rc == 0)
@@ -1409,6 +1744,7 @@ report(fpi,fpo)
 	  query2(fpi,fpo);
 	  return;
 	}
+
 	if (schedq) {
 	  /* Old-style processing */
 	  int prevc = -1;
@@ -1432,7 +1768,7 @@ report(fpi,fpo)
 	}
 
 	r_i = 0;
-	sp_scan(repscan, (struct spblk *)NULL, spt_ids[L_CHANNEL]);
+	sp_scan(repscan_v1, (struct spblk *)NULL, spt_ids[L_CHANNEL]);
 	if (!r_i) {
 	  if (onlyuser)
 	    fprintf(stdout,"No user messages found\n");
@@ -1446,7 +1782,7 @@ report(fpi,fpo)
 
 void
 printaddrs(v)
-	struct vertex *v;
+     struct vertex *v;
 {
 	register char *cp;
 	int	i, fd;
@@ -1475,7 +1811,7 @@ printaddrs(v)
 	  errno = 0;
 	  if (read(fd, v->cfp->contents, stbuf.st_size) < stbuf.st_size){
 	    fprintf(stdout,"\t\tread(%d): %s\n", (int)stbuf.st_size,
-		   errno == 0 ? "failed" : strerror(errno));
+		    errno == 0 ? "failed" : strerror(errno));
 	    close(fd);
 	    return;
 	  }
@@ -1487,6 +1823,14 @@ printaddrs(v)
 	      if (*++cp == _CF_SENDER)
 		break;
 	      switch (*cp) {
+	      case _CF_FORMAT:
+		++cp;
+		v->cfp->format = 0;
+		sscanf(cp,"%i",&v->cfp->format);
+		if (v->cfp->format & (~_CF_FORMAT_KNOWN_SET))
+		  fprintf(stdout, "Unsupported SCHEDULER file format flags seen: 0x%x at file '%s'",
+			  v->cfp->format, path);
+		break;
 	      case _CF_LOGIDENT:
 		v->cfp->logident = cp + 2;
 		break;
@@ -1544,28 +1888,72 @@ printaddrs(v)
 	  cp = v->cfp->contents + v->index[i] + 2;
 	  if (*cp == ' ' || (*cp >= '0' && *cp <= '9'))
 	    cp += _CFTAG_RCPTPIDSIZE;
-	  while (!ISSPACE(*cp)) ++cp;
-	  while ( ISSPACE(*cp)) ++cp;
-	  while (!ISSPACE(*cp)) ++cp;
-	  while ( ISSPACE(*cp)) ++cp;
-	  ocp = cp;
-	  while (*cp != '\0' && *cp != '\n')
-	    ++cp;
-	  if (*cp-- == '\n') {
-	    while (!ISSPACE(*cp)) --cp;
-	    while ( ISSPACE(*cp)) --cp;
-	    if (cp >= ocp)
-	      *++cp = '\0';
-	    else {
-	      cp = ocp;
-	      while (*cp != '\0' && *cp != '\n')
-		++cp;
-	      *cp = '\0';
-	    }
+
+	  if ((v->cfp->format & _CF_FORMAT_DELAY1) || *cp == ' ' ||
+	      (*cp >= '0' && *cp <= '9')) {
+	    /* Newer DELAY data slot - _CFTAG_RCPTDELAYSIZE bytes */
+	    cp += _CFTAG_RCPTDELAYSIZE;
 	  }
+
+	  cp = skip821address(cp); /* skip channel */
+	  while (*cp == ' ' || *cp == '\t') ++cp;
+	  cp = skip821address(cp); /* skip host */
+	  while (*cp == ' ' || *cp == '\t') ++cp;
+
+	  ocp = cp;
+	  cp = skip821address(cp); /* skip user */
+	  *cp++ = 0;
 	  fprintf(stdout,"\t");
 	  if (i == 0)
 	    fprintf(stdout,"  to");
 	  fprintf(stdout,"\t%s\n",ocp);
 	}
 }
+
+
+static void print_shm __((void))
+{
+  int r, i;
+
+  r = Z_SHM_MIB_Attach (0); /* Attach read-only! */
+
+  if (r < 0) {
+    /* Error processing -- magic set of constants: */
+    switch (r) {
+    case -1:
+      fprintf(stderr, "No ZENV variable: SNMPSHAREDFILE\n");
+      break;
+    case -2:
+      perror("Failed to open for exclusively creating of the SHMSHAREDFILE");
+      break;
+    case -3:
+      perror("Failure during creation fill of SGMSHAREDFILE");
+      break;
+    case -4:
+      perror("Failed to open the SHMSHAREDFILE at all");
+      break;
+    case -5:
+      perror("The SHMSHAREDFILE isn't of proper size! ");
+      break;
+    case -6:
+      perror("Failed to mmap() of SHMSHAREDFILE into memory");
+      break;
+    case -7:
+      fprintf(stderr, "The SHMSHAREDFILE  has magic value mismatch!\n");
+      break;
+    default:
+      break;
+    }
+    return;
+  }
+
+  r = (Z_SHM_MIB_is_attached() > 0); /* Attached and WRITABLE ? */
+
+
+#define sfprintf fprintf
+#define fp       stdout
+
+#include "mailq.inc"  /* shared stuff with  mq2.c  module */
+
+	exit(0);
+}
