@@ -1,7 +1,7 @@
 /*
  *	ZMailer 2.99.53+ Scheduler "mailq2" routines
  *
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1999
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1999-2003
  *
  */
 
@@ -22,6 +22,9 @@
 # include <sys/select.h>
 #endif
 
+#ifdef HAVE_SYS_LOADAVG_H
+#include <sys/loadavg.h>
+#endif
 
 #if	defined(BSD4_3) || defined(sun)
 #include <sys/file.h>
@@ -32,6 +35,14 @@
 #include <sys/socket.h>
 #include <sys/time.h>
 
+#include "zmsignal.h"
+
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+#include "libc.h"
+
+
 #ifndef	NFDBITS
 /*
  * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
@@ -109,6 +120,9 @@ static void mq2_discard(mq)
     free(mq->challenge);
 
   free(mq);
+
+  --mq2count;
+  MIBMtaEntry->sc.MQ2sockParallel --;
 }
 
 /* EXTERNAL */
@@ -199,13 +213,14 @@ int mq2_wflush(mq)
     } else {
       /* Error ??? */
       if (errno == EAGAIN || errno == EINTR)
-	break; /* Back latter .. */
+	break; /* Back later .. */
       /* Err... what ?? */
 
       if (verbose)
 	sfprintf(sfstderr, " -- failure; errno = %d\n", errno);
 
       mq2_discard(mq);
+      MIBMtaEntry->sc.MQ2sockWriteFails ++;
 
       return -1;
     }
@@ -302,6 +317,7 @@ static void mq2_read(mq)
 
   if (mq->fd < 0) {
     mq2_discard(mq);
+    MIBMtaEntry->sc.MQ2sockReadFails ++;
     return;
     /* Zap! */
   }
@@ -313,7 +329,12 @@ static void mq2_read(mq)
 
   if (mq->inbufsize+80 >= mq->inbufspace) {
     mq->inbufspace *= 2;
-    mq->inbuf = erealloc(mq->inbuf, mq->inbufspace);
+    /* Abort if line size is too large.. */
+    if (mq->inbufspace > 17000) {
+      free(mq->inbuf);
+      mq->inbuf = NULL;
+    } else
+      mq->inbuf = erealloc(mq->inbuf, mq->inbufspace);
   }
 
   if (mq->inbuf == NULL) {
@@ -326,6 +347,7 @@ static void mq2_read(mq)
 
   if (i == 0) {
     mq2_discard(mq);
+    MIBMtaEntry->sc.MQ2sockReadEOF ++;
     return; /* ZAP! */
   }
   if (i > 0) {
@@ -333,7 +355,7 @@ static void mq2_read(mq)
     mq->inbufsize += i;
   } else {
     if (errno == EINTR || errno == EAGAIN) {
-      /* Ok, come back latter */
+      /* Ok, come back later */
     } else {
       mq2_discard(mq); /* ZAP! */
     }
@@ -356,6 +378,7 @@ void mq2_register(fd, addr)
      Usockaddr *addr;
 {
   struct mailq *mq;
+  void *mq2test;
 
   static int cnt = 0;
   char buf[200];
@@ -372,16 +395,17 @@ void mq2_register(fd, addr)
     return;
   }
   memset(mq, 0, sizeof(*mq));
+
+  ++mq2count;
+  MIBMtaEntry->sc.MQ2sockParallel ++;
   
   mq->fd = fd;
-  mq->apopteosis = now + max_mq_life;
+  mq->apoptosis = now + max_mq_life;
   mq->qaddr = *addr;
 
   mq->nextmailq = mq2root;
   mq2root = mq;
 
-  fd_nonblockingmode(fd);
-
   /* 
      Scheduler writes following to the interface socket:
 
@@ -390,19 +414,37 @@ void mq2_register(fd, addr)
   */
 
   mq2_puts(mq,"version zmailer 2.0\n");
+
+  mq2test = mq2_authuser(mq, NULL);
+
+  if (! mq2test) {
+
+    fd_blockingmode(fd);
+
+    mq2_puts(mq, "550 NO ACCESS FOR YOU\n");
+    mq2_wflush(mq);
+  mq2_abort:;
+    MIBMtaEntry->sc.MQ2sockAuthRej ++;
+    mq2_discard(mq);
+
+  } else {
   
-  gettimeofday(&tv,NULL);
+    fd_nonblockingmode(fd);
 
-  sprintf(buf,"MAILQ-V2-CHALLENGE: %ld.%ld.%d",
-	  (long)tv.tv_sec, (long)tv.tv_usec, ++cnt);
+    gettimeofday(&tv,NULL);
 
-  mq->challenge = strdup(buf);
+    sprintf(buf,"220 MAILQ-V2-CHALLENGE: %ld.%ld.%d",
+	    (long)tv.tv_sec, (long)tv.tv_usec, ++cnt);
 
-  mq2_puts(mq, buf);
-  mq2_puts(mq, "\n");
-  mq2_wflush(mq);
+    mq->challenge = strdup(buf); /* This MAY yield NULL */
+    if (!mq->challenge) goto mq2_abort;
 
-  mq->auth = 0;
+    mq2_puts(mq, buf);
+    mq2_puts(mq, "\n");
+    mq2_wflush(mq);
+
+    mq->auth = 0;
+  }
 }
 
 /* EXTERNAL */
@@ -434,6 +476,8 @@ void mq2_areinsets(rdmaskp, wrmaskp)
 {
     struct mailq *mq;
 
+    timed_log_reinit();
+
     /* The mq-queue may change while we are going it over,
        thus we *must not* try to do write and read things
        at the same time! and be *very* carefull at following
@@ -447,8 +491,10 @@ void mq2_areinsets(rdmaskp, wrmaskp)
       }
 
       /* Time of forced death ? */
-      if (now > mq->apopteosis)
+      if (now > mq->apoptosis) {
+	MIBMtaEntry->sc.MQ2sockTimedOut ++;
 	mq2_discard(mq);
+      }
 
       mq = mq2;
     }
@@ -553,7 +599,7 @@ static int mq2cmd_etrn(mq,s)
 
   /* 's' points to the first arg, 't' points to string after
      separating white-space has been skipped. */
-  sfprintf(sfstdout,"%s MQ2 ETRN: %s\n", timestring(), s);
+  sfprintf(sfstdout,"%s MQ2 ETRN: %s %s\n", timestring(), s, t);
   rc = turnme(s);
   if (rc)
     mq2_puts(mq, "+OK; an ETRN started something.\n");
@@ -563,6 +609,92 @@ static int mq2cmd_etrn(mq,s)
   return 0;
 }
 
+
+/* INTERNAL */
+static void mq2_show_snmp(mq)
+     struct mailq *mq;
+{
+  int r, i;
+  Sfio_t *fp;
+  struct mq2discipline mq2d;
+
+  fp = sfnew(NULL, NULL, 0, 0, SF_LINE|SF_WRITE);
+  if (!fp) {
+    mq2_puts(mq, " *** FAILURE: sfnew(NULL, NULL, 0, -1, SF_LINE|SF_STRING|SF_WRITE);\n");
+    return;
+  }
+
+  memset(&mq2d, 0, sizeof(mq2d));
+  mq2d.mq = mq;
+  mq2d.D.writef  = mq2_sfwrite;
+
+  sfdisc(fp, &mq2d.D);
+
+
+  r = (Z_SHM_MIB_is_attached() > 0); /* Attached and WRITABLE ? */
+
+
+#include "mailq.inc" /* Shared stuff with  mailq.c  program */
+
+
+  zsfsetfd(fp, -1);
+  sfclose(fp);
+
+
+}
+
+
+
+/* INTERNAL */
+static int mq2cmd_reroute(mq,s)
+     struct mailq *mq;
+     char *s;
+{
+  char *t = s, *u;
+
+  while (*t && (*t != ' ') && (*t != '\t')) ++t;
+  if (*t) *t++ = '\000';
+  while (*t == ' ' || *t == '\t') ++t;
+  u = t;
+  while (*u && (*u != ' ') && (*u != '\t')) ++u;
+  if (*u) *u++ = '\000';
+  while (*u == ' ' || *u == '\t') ++u;
+
+  /* 's' points to the first arg, 't' points to string after
+     separating white-space has been skipped. */
+
+  return -1;
+}
+
+
+/* INTERNAL */
+static int mq2cmd_kill(mq,s)
+     struct mailq *mq;
+     char *s;
+{
+  char *t = s;
+
+  if (!(mq->auth & MQ2MODE_KILL)) {
+    mq2_puts(mq, "-No KILL allowed for you.\n");
+    return 0;
+  }
+  while (*t && (*t != ' ') && (*t != '\t')) ++t;
+  if (*t) *t++ = '\000';
+  while (*t == ' ' || *t == '\t') ++t;
+
+  /* 's' points to the first arg, 't' points to string after
+     separating white-space has been skipped. */
+  if (strcmp(s, "MSG") == 0) {
+	sfprintf(sfstdout,"%s MQ2 KILL MSG: %s %s\n", timestring(), t);
+	deletemsg(t, NULL);
+	mq2_puts(mq, "+OK; KILLed something.\n\n");
+  } else {
+        mq2_puts(mq, "-Unknown request.\n\n");
+  }
+
+  return 0;
+}
+
 /* INTERNAL */
 static int mq2cmd_show(mq,s)
      struct mailq *mq;
@@ -581,9 +713,25 @@ static int mq2cmd_show(mq,s)
   /* 's' points to the first arg, 't' points to string after
      separating white-space has been skipped. */
 
+  /* This really should be "SHOW SNMP", but that command
+     was reserved for other use... */
+  if (strcmp(s,"COUNTERS") == 0) {
+
+    MIBMtaEntry->sc.MQ2sockCommandShowCounters ++;
+
+    mq2_puts(mq, "+OK until LF.LF\n");
+    mq2_show_snmp(mq);
+    mq2_puts_(mq, ".\n");
+    return 0;
+  }
+
   if (strcmp(s,"SNMP") == 0) {
+
     if (! (MQ2MODE_SNMP & mq->auth)) /* If not allowed operation, exit! */
       return -1;
+
+    MIBMtaEntry->sc.MQ2sockCommandShowQueueVeryShort ++;
+
     mq2_puts(mq, "+OK until LF.LF\n");
     mq2_thread_report(mq, MQ2MODE_SNMP, NULL, NULL);
     mq2_puts_(mq, ".\n");
@@ -596,6 +744,9 @@ static int mq2cmd_show(mq,s)
 
       if (! (MQ2MODE_QQ & mq->auth)) /* If not allowed operation, exit! */
 	return -1;
+
+      MIBMtaEntry->sc.MQ2sockCommandShowQueueShort ++;
+
       mq2_puts(mq, "+OK until LF.LF\n");
       mq2_thread_report(mq, MQ2MODE_QQ, NULL, NULL);
       mq2_puts_(mq, ".\n");
@@ -605,6 +756,9 @@ static int mq2cmd_show(mq,s)
 
       if (! (MQ2MODE_FULL & mq->auth)) /* If not allowed operation, exit! */
 	return -1;
+
+      MIBMtaEntry->sc.MQ2sockCommandShowQueueThreads2 ++;
+
       mq2_puts(mq, "+OK until LF.LF\n");
       mq2_thread_report(mq, MQ2MODE_FULL2, NULL, NULL);
       mq2_puts_(mq, ".\n");
@@ -614,6 +768,9 @@ static int mq2cmd_show(mq,s)
 
       if (! (MQ2MODE_FULL & mq->auth)) /* If not allowed operation, exit! */
 	return -1;
+
+      MIBMtaEntry->sc.MQ2sockCommandShowQueueThreads2 ++;
+
       mq2_puts(mq, "+OK until LF.LF\n");
       mq2_thread_report(mq, MQ2MODE_FULL, NULL, NULL);
       mq2_puts_(mq, ".\n");
@@ -632,6 +789,8 @@ static int mq2cmd_show(mq,s)
     if (! (MQ2MODE_FULL & mq->auth)) /* If not allowed operation, exit! */
       return -1;
 
+    MIBMtaEntry->sc.MQ2sockCommandShowThread ++;
+
     mq2_puts(mq, "+OK until LF.LF\n");
     mq2_thread_report(mq, MQ2MODE_FULL, channel, host);
     mq2_puts_(mq, ".\n");
@@ -650,6 +809,8 @@ static void mq2interpret(mq,s)
 {
   char *t = s;
 
+  MIBMtaEntry->sc.MQ2sockCommands ++;
+
   while (*t && (*t != ' ') && (*t != '\t')) ++t;
   if (*t) *t++ = '\000';
   while (*t == ' ' || *t == '\t') ++t;
@@ -660,17 +821,22 @@ static void mq2interpret(mq,s)
   if (cistrcmp(s,"QUIT")==0 || cistrcmp(s,"EXIT") == 0) {
     mq2_puts(mq, "+Bye bye\n");
     mq2_wflush(mq);
-    close(mq->fd);
-    mq->fd = -1;
+    mq2_discard(mq);
+    MIBMtaEntry->sc.MQ2sockCommandQUIT ++;
+    return;
   }
 
   if (mq->auth == 0 && strcmp(s,"AUTH") == 0) {
+    MIBMtaEntry->sc.MQ2sockCommandAUTH ++;
     mq2auth(mq,t);
+    if (! mq->auth)
+      MIBMtaEntry->sc.MQ2sockAuthRej ++;
     return;
   }
 
   if (!mq->auth) {
     mq2_puts(mq,"-BAD; USER MUST AUTHENTICATE\n");
+    MIBMtaEntry->sc.MQ2sockCommandsRej ++;
     return;
   }
 
@@ -678,11 +844,22 @@ static void mq2interpret(mq,s)
     if (mq2cmd_show(mq,t) == 0)
       return;
   }
+  if (strcmp(s,"KILL") == 0) {
+    if (mq2cmd_kill(mq,t) == 0)
+      return;
+  }
+  if (strcmp(s,"REROUTE") == 0) {
+    if (mq2cmd_reroute(mq,t) == 0)
+      return;
+  }
   if (strcmp(s,"ETRN") == 0) {
+    MIBMtaEntry->sc.MQ2sockCommandETRN ++;
     mq2cmd_etrn(mq,t);
     return;
   }
 
+  MIBMtaEntry->sc.MQ2sockCommandsRej ++;
+
   mq2_puts(mq, "-MAILQ2 Unknown command, or refused by access control; VERB='");
   mq2_puts(mq, s);
   mq2_puts(mq, "' REST='");
