@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2001
  *
  *	We produce RFC 1894 format reports with everything it contains.
  */
@@ -77,14 +77,15 @@ decodeXtext(fp,xtext)
 	}
 }
 
-static void scnotaryreport __((Sfio_t *, struct not *, int *, int));
+static void scnotaryreport __((Sfio_t *, struct not *, int *, int, int));
 
 static void	/* There is notaryreport() on transporters also.. */
-scnotaryreport(errfp,notary,haserrsp,notifyrespectflg)
+scnotaryreport(errfp,notary,haserrsp,notifyrespectflg,headstyle)
 	Sfio_t *errfp;
 	struct not *notary;
 	int *haserrsp;
 	int notifyrespectflg;
+	int headstyle;
 {
 	char *rcpt, *action, *status, *diagstr, *wtt;
 	char *taid;
@@ -160,26 +161,49 @@ scnotaryreport(errfp,notary,haserrsp,notifyrespectflg)
 	} else
 	  typetag = type_local;
 	if (notary->orcpt) {
-	  sfprintf(errfp, "Original-Recipient: ");
+	  if (headstyle)
+	    sfprintf(errfp, "  Original Recipient:\n      ");
+	  else
+	    sfprintf(errfp, "Original-Recipient: ");
 	  decodeXtext(errfp,notary->orcpt);
 	  sfputc(errfp,'\n');
 	}
-	sfprintf(errfp, "Final-Recipient: %s;%s\n", typetag, rcpt);
-	sfprintf(errfp, "Action: %s\n", action);
+	if (headstyle) {
+	  sfprintf(errfp, "  Final Recipient:\n      %s;%s\n", typetag, rcpt);
+	} else {
+	  sfprintf(errfp, "Final-Recipient: %s;%s\n", typetag, rcpt);
+	  sfprintf(errfp, "Action: %s\n", action);
+	}
 	if (status) {
 	  if (*status == '4' || *status == '5')
 	    *haserrsp = 1;
-	  sfprintf(errfp, "Status: %s\n", status);
+	  if (headstyle)
+	    sfprintf(errfp, "  Status:\n      %s\n", status);
+	  else
+	    sfprintf(errfp, "Status: %s\n", status);
+	}
+	if (wtt && wtt[0] != 0) {
+	  if (headstyle)
+	    sfprintf(errfp, "  Remote MTA:\n      %s\n", wtt);
+	  else
+	    sfprintf(errfp, "Remote-MTA: %s\n", wtt);
+	}
+	if (notary->tstamp != 0) {
+	  if (headstyle)
+	    sfprintf(errfp, "  Last Attempt Date:\n      %s",
+		     rfc822date(&notary->tstamp));
+	  else
+	    sfprintf(errfp, "Last-Attempt-Date: %s",
+		     rfc822date(&notary->tstamp));
+	}
+	if (taid && headstyle)
+	  sfprintf(errfp, "  X-ZTAID:\n      %s\n", taid);
+	if (headstyle)
+	  sfprintf(errfp, "  Diagnostic Code:\n      %s\n", diagstr);
+	else {
+	  sfprintf(errfp, "Diagnostic-Code: %s\n", diagstr);
+	  sfprintf(errfp, "\n");
 	}
-	sfprintf(errfp, "Diagnostic-Code: %s\n", diagstr);
-	if (wtt && wtt[0] != 0)
-	  sfprintf(errfp, "Remote-MTA: %s\n", wtt);
-	if (notary->tstamp != 0)
-	  sfprintf(errfp, "Last-Attempt-Date: %s",
-		   rfc822date(&notary->tstamp));
-	if (taid)
-	  sfprintf(errfp, "X-ZTAID: %s\n", taid);
-	sfprintf(errfp, "\n");
 
 	action[-1] = '\001';
 	if (status) status[-1] = '\001';
@@ -203,17 +227,14 @@ msgerror(vp, offset, message)
 
 	if (vp->notary) notary = vp->notary;
 
-	if (vp->cfp->dirind > 0) {
-	  sprintf(path, "%.300s/%.100s",
-		  cfpdirname(vp->cfp->dirind), vp->cfp->mid);
-	} else {
-	  strncpy(path,vp->cfp->mid,400);
-	  path[400] = 0;
-	}
+	sprintf(path, "%s%.400s", cfpdirname(vp->cfp->dirind), vp->cfp->mid);
 
 	/* exclusive access required, but we're the only scheduler... */
 	fp = sfopen(NULL, path, "a");
 	if (fp == NULL) {
+
+	  timed_log_reinit();
+
 	  sfprintf(sfstderr,
 		   "Cannot open control file %s to deposit", vp->cfp->mid);
 	  sfprintf(sfstderr,
@@ -228,7 +249,11 @@ msgerror(vp, offset, message)
 		 time(NULL), notary, message);
 	sfsync(fp);
 #ifdef HAVE_FSYNC
-	fsync(sffileno(fp));
+	while (fsync(sffileno(fp)) < 0) {
+	  if (errno == EINTR || errno == EAGAIN)
+	    continue;
+	  break;
+	}
 #endif
 	sfclose(fp);
 }
@@ -305,16 +330,16 @@ writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary, actionset)
 	      strchr(eaddr,'=') == NULL /*Smells of X.400*/)))
 	  eaddr = "postmaster"; /* Paranoid, eh ? */
 	sfprintf(errfp, "channel error\n");
+	sfprintf(errfp, "errormsg\n"); /* SCHEDULER WROTE THIS! */
 
 	if (!*no_error_reportp)
 	  printenvaddr(errfp, eaddr);
-	sprintf(path, "%s/%s/%s", mailshare, FORMSDIR,
-		deliveryform ? deliveryform : "delivery");
+	sprintf(path, "%s/%s/%s", mailshare, FORMSDIR, deliveryform);
 	fp = sfopen(NULL, path, "r");
 	if (fp != NULL) {
 	  int inhdr = 1, hadsubj =0;
 	  buf[sizeof(buf)-1] = 0;
-	  while (cfgets(buf,sizeof(buf)-1,fp) >= 0) {
+	  while (csfgets(buf,sizeof(buf)-1,fp) >= 0) {
 	    if (strncmp(buf,"HDR",3)==0) {
 	      continue;
 	    } else if (strncmp(buf,"ADR",3)==0) {
@@ -333,7 +358,7 @@ writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary, actionset)
 	    sfprintf(errfp, "To: %s\n", eaddr);
 	  else if (*no_error_reportp < 0)
 	    sfprintf(errfp, "To: dummy:; (error trapped source)\n");
-	  while (cfgets(buf,sizeof(buf)-1,fp) >= 0) {
+	  while (csfgets(buf,sizeof(buf)-1,fp) >= 0) {
 	    if (strncmp(buf,"HDR",3)==0) {
 	      sfprintf(errfp, "%s", buf+4);
 	    } else if (strncmp(buf,"ADR",3)==0) {
@@ -419,7 +444,7 @@ writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary, actionset)
 	  sfprintf(errfp, "This report is classified as 'Multiple-fault', because\n");
 	  sfprintf(errfp, "the error report template file (%s) was not found.\n\n",path);
 	  sfprintf(errfp, "Please report this to this system's postmaster.\n\n");
-	  sfprintf(errfp, "Here are report messages regarding email you (propably) sent:\n\n");
+	  sfprintf(errfp, "Here are report messages regarding email you (probably) sent:\n\n");
 	}
 }
 
@@ -431,9 +456,9 @@ reporterrs(cfpi, delayreports)
 	const int delayreports;
 {
 	int i, n, wroteheader, byteidx, headeridx = -1, drptidx, fd;
-	int *lp;
+	int *lp, ignored = 0;
 	time_t tstamp;
-	char *midbuf, *cp, *cp2, *action, *eaddr;
+	char *cp, *cp2, *action, *eaddr;
 	char *deliveryform;
 	Sfio_t *errfp, *fp;
 	char *notary;
@@ -441,54 +466,53 @@ reporterrs(cfpi, delayreports)
 	char *envid;
 	int notarycnt = 0;
 	int notaryspc = 0;
-	struct ctlfile *cfp;
+	struct ctlfile *cfp = NULL;
 	char *rcpntpointer;
 	int no_error_report = 0;
 	int has_errors = 0;
 	int lastoffset;
-	char path[MAXPATHLEN], mpath[MAXPATHLEN];
 	int mypid;
 	long format;
 	char boundarystr[400];
-	char spoolid[30];
+	char rptspoolid[30];
+	time_t mtime;
+	long ino;
 	int actionsets[5]; /* 0:DELIVERED, 1:FAILED, 2:RELAYED, 3:DELAYED,
 			      4:EXPANDED */
 	ACTSETENUM thisaction;
+	char path[MAXPATHLEN], mpath[MAXPATHLEN];
 
-	if (cfpi->haderror == 0)
-		return;
-	if (cfpi->erroraddr == NULL)
-		return;
+	if (cfpi->haderror == 0) {
+	  if (verbose > 1)
+	    sfprintf(sfstdout, "reporterrs: No errors! bailing out!\n");
+	  return;
+	}
+	if (cfpi->erroraddr == NULL) {
+	  if (verbose)
+	    sfprintf(sfstdout, "reporterrs: No error address! bailing out!\n");
+	  return;
+	}
 	no_error_report = cfpi->iserrmesg;
 
-	taspoolid(spoolid, cfpi->mtime, cfpi->id);
-
 	eaddr        = cfpi->erroraddr;
-	deliveryform = cfpi->deliveryform;
+	deliveryform = cfpi->deliveryform ? cfpi->deliveryform : "delivery";
 	envid        = cfpi->envid;
-	/* re-read the control file to get some unstashed information */
-	midbuf       = cfpi->mid;
 
 	/* exclusive access required, but we're the only scheduler... */
-	if (cfpi->dirind > 0)
-	  sprintf(mpath,"%s/%s", cfpdirname(cfpi->dirind), cfpi->mid);
-	else
-	  strcpy(mpath,cfpi->mid);
+	sprintf(mpath,"%s%.400s", cfpdirname(cfpi->dirind), cfpi->mid);
 
 	fd = open(mpath, O_RDWR, 0);
 	if (fd < 0 ||
 	    (cfp = slurp(fd, cfpi->id)) == NULL) {
 	  sfprintf(sfstderr,
-		   "%s: unexpected absence of control file %s for error processing!\n",
-		   progname, mpath);
+		   "%s: unexpected absence of control file %s for error processing (%s)!\n",
+		   progname, mpath, cfpi->spoolid ? cfpi->spoolid : "-");
 	  if (fd >= 0)
 	    close(fd);
 	  return;
 	}
 	lp = &cfp->offset[0];
 	mypid = getpid();
-	if (cfp->mid) free(cfp->mid);
-	cfp->mid = midbuf; /* This is the original one! */
 	wroteheader = 0;
 	lastoffset = cfp->offset[cfp->nlines-1];
 	format = 0L;
@@ -588,8 +612,9 @@ reporterrs(cfpi, delayreports)
 	  if (format & _CF_FORMAT_DELAY1)
 	    rcpntpointer += _CFTAG_RCPTDELAYSIZE;
 
-	  if (do_syslog)
-	    zsyslog((LOG_INFO, "%s: <%s>: %s", cfp->mid, rcpntpointer, cp));
+	  if (do_syslog > 1)
+	    zsyslog((LOG_INFO, "%s: <%s>: %s",
+		     cfpi->spoolid ? cfpi->spoolid:"-", rcpntpointer, cp));
 
 	  if (notary != NULL && *notary != 0) {
 	    if (!notaries) {
@@ -663,7 +688,7 @@ reporterrs(cfpi, delayreports)
 		if (*d) *d++ = 0;
 		continue;
 	      }
-	      /* Report conditional testing in the latter phases */
+	      /* Report conditional testing in the later phases */
 	    }
 	    while (*cp == ' ' || *cp == '\t') ++cp;
 	    notaries[notarycnt].message = cp;
@@ -674,20 +699,30 @@ reporterrs(cfpi, delayreports)
 	    case ACTSET_RELAYED:
 	      if (notaries[notarycnt].notifyflgs & NOT_SUCCESS)
 		actionsets[thisaction] += 1;
-	      else thisaction = ACTSET_NONE;
+	      else {
+		thisaction = ACTSET_NONE;
+		++ignored;
+	      }
 	      break;
 	    case ACTSET_DELAYED:
 	      if (notaries[notarycnt].notifyflgs & NOT_DELAY)
 		actionsets[thisaction] += 1;
-	      else thisaction = ACTSET_NONE;
+	      else {
+		thisaction = ACTSET_NONE;
+		++ignored;
+	      }
 	      break;
 	    case ACTSET_FAILED:
 	      if (notaries[notarycnt].notifyflgs & NOT_FAILURE)
 		actionsets[thisaction] += 1;
-	      else thisaction = ACTSET_NONE;
+	      else {
+		thisaction = ACTSET_NONE;
+		++ignored;
+	      }
 	      break;
 	    default:
 	      thisaction = ACTSET_NONE;
+	      ++ignored;
 	      break;
 	    }
 	    notaries[notarycnt].thisaction = thisaction;
@@ -706,9 +741,17 @@ reporterrs(cfpi, delayreports)
 
 	if (notaries == NULL) {
 	  /* Oops!  No recipients on which to report anything ?! */
+
+	  if (verbose)
+	    sfprintf(sfstdout, "No reports! bailing out!\n");
+
 	  close(cfp->fd);
-	  cfp->mid = NULL; /* we don't want to loose the original one! */
 	  free_cfp_memory(cfp);
+
+	  if (do_syslog)
+	    zsyslog((LOG_INFO, "%s: Abnormal: No notaries to report; 1",
+		     cfpi->spoolid ? cfpi->spoolid:"-"));
+
 	  return;
 	}
 
@@ -728,6 +771,15 @@ reporterrs(cfpi, delayreports)
 	  free((void*)notaries);
 	  close(cfp->fd);
 	  free_cfp_memory(cfp);
+
+	  if (verbose > 1)
+	    sfprintf(sfstdout, "reporterrs: No reports! bailing out!\n");
+
+	  if (do_syslog)
+	    zsyslog((LOG_INFO, "%s: Nothing to report, ignorecount: %d",
+		     cfpi->spoolid ? cfpi->spoolid:"-", ignored));
+
+
 	  return;
 	}
 
@@ -744,6 +796,11 @@ reporterrs(cfpi, delayreports)
 	  free((void*)notaries);
 	  close(cfp->fd);
 	  free_cfp_memory(cfp);
+
+	  if (do_syslog)
+	    zsyslog((LOG_INFO, "%s: ABNORMAL: Failed to open report file!",
+		     cfpi->spoolid ? cfpi->spoolid:"-"));
+
 	  return;
 	}
 
@@ -762,6 +819,25 @@ reporterrs(cfpi, delayreports)
 	writeheader(errfp, eaddr, &no_error_report, deliveryform, boundarystr,
 		    actionsets);
 
+
+	if (mydomain() != NULL) {
+	  sfprintf(errfp, "Reporting-MTA: dns; %s\n", mydomain() );
+	} else {
+	  sfprintf(errfp, "Reporting-MTA: x-local-hostname; -unknown-\n");
+	}
+	if (envid != NULL) {
+	  sfprintf(errfp, "Original-Envelope-Id: ");
+	  decodeXtext(errfp,envid);
+	  sfputc(errfp, '\n');
+	}
+	/* rfc822date() returns a string with trailing newline! */
+	sfprintf(errfp, "Arrival-Date: %s", rfc822date(&cfpi->mtime));
+	if (cfpi->spoolid)
+	  sfprintf(errfp, "Local-Spool-ID: %s\n", cfpi->spoolid);
+	sfprintf(errfp, "\n\n");
+
+
+
 	n = 0;
 	for (i = 0; i < notarycnt; ++i) {
 	  /* Scan to the start of the message text */
@@ -792,35 +868,27 @@ reporterrs(cfpi, delayreports)
 	    --n;
 	    break;
 	  }
-	  if (notaries[i].orcpt) {
-	    sfprintf(errfp, "  Original Recipient:\n    ");
-	    decodeXtext(errfp,notaries[i].orcpt);
-	    sfputc(errfp,'\n');
-	  } else if (notaries[i].inrcpt) {
-	    sfprintf(errfp, "  Arrived Recipient:\n    ");
+	  if (notaries[i].inrcpt) {
+	    sfprintf(errfp, "  Arrived Recipient:\n      ");
 	    decodeXtext(errfp, notaries[i].inrcpt);
 	    sfputc(errfp,'\n');
 	  }
-	  sfprintf(errfp, "  Control data:\n");
-	  sfprintf(errfp, "    %s\n", notaries[i].rcpntp);
-	  sfprintf(errfp, "  Diagnostic texts:\n");
+	  scnotaryreport(errfp, &notaries[i], &has_errors, no_error_report, 1);
+
+	  sfprintf(errfp, "  Control data:\n      %s\n", notaries[i].rcpntp);
+	  sfprintf(errfp, "  Diagnostic texts:\n      ");
 	  ccp = notaries[i].message;
-	  if (*ccp != '\r')
-	    sfprintf(errfp, "    ");
-	  if (strchr(ccp, '\r')) {
-	    sfprintf(errfp, "...\\\n    ");
-	    s = ccp;
-	    if (*s == '\r') ++s; /* Skip possible first '\r' */
-	    for (; *s != '\0'; ++s) {
-	      if (*s == '\r') {
-		sfprintf(errfp,"\n    ");
-	      } else {
-		sfputc(errfp, *s);
-	      }
+	  s = ccp;
+	  while (*s == '\r') ++s; /* Skip possible first '\r' */
+	  for (; *s != '\0'; ++s) {
+	    if (*s == '\r') {
+	      sfprintf(errfp,"\n     ");
+	    } else {
+	      sfputc(errfp, *s);
 	    }
-	    sfputc(errfp, '\n');
-	  } else
-	    sfprintf(errfp, "%s\n", ccp);
+	  }
+	  sfputc(errfp, '\n');
+
 	}
 
 	sfprintf(errfp,"\n\
@@ -848,24 +916,21 @@ sent you this report, please include this information in your question!\n\
 	}
 	/* rfc822date() returns a string with trailing newline! */
 	sfprintf(errfp, "Arrival-Date: %s", rfc822date(&cfpi->mtime));
-	sfprintf(errfp, "Local-Spool-ID: %s\n", spoolid);
+	if (cfpi->spoolid)
+	  sfprintf(errfp, "Local-Spool-ID: %s\n", cfpi->spoolid);
 	sfprintf(errfp, "\n");
 
 	/* Now scan 'em all again for IETF-NOTARY */
 	for (i = 0; i < notarycnt; ++i)
-	  scnotaryreport(errfp, &notaries[i],&has_errors,no_error_report);
+	  scnotaryreport(errfp, &notaries[i],&has_errors,no_error_report, 0);
 
 	sfprintf(errfp,"\n\
 Following is copy of the message headers. Original message content may\n\
 be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 
 	/* path to the message body */
-	if (cfpi->dirind > 0)
-	  sprintf(path, "../%s/%s/%s",
-		  QUEUEDIR, cfpdirname(cfpi->dirind), cfpi->mid);
-	else
-	  sprintf(path, "../%s/%s",
-		  QUEUEDIR, cfpi->mid);
+	sprintf(path, "../%s/%s%.300s",
+		QUEUEDIR, cfpdirname(cfpi->dirind), cfpi->mid);
 
 	fp = sfopen(NULL, path, "r");
 	if (fp != NULL) {
@@ -882,14 +947,14 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	  } else {
 	    /* Scan the input, and drop off the ZMailer
 	       envelope headers */
-	    while (cfgets(buf,sizeof(buf),fp) >= 0) {
+	    while (csfgets(buf,sizeof(buf),fp) >= 0) {
 	      const char *s = buf;
 	      while (*s && *s != ':' && *s != ' ' && *s != '\t') ++s;
 	      if (*s == ':') break;
 	      *buf = 0;
 	    }
 	    /* We leave the first scan-phase with  buf[]  containing some
-	       valid RFC-822 -style header, propably "Received:" */
+	       valid RFC-822 -style header, probably "Received:" */
 	    if (*buf)
 	      sfprintf(errfp, "%s", buf);
 	    else {
@@ -916,12 +981,13 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	}
 	/* rfc822date() returns a string with trailing newline! */
 	sfprintf(errfp, "Arrival-Date: %s", rfc822date(&cfpi->mtime));
-	sfprintf(errfp, "Local-Spool-ID: %s\n", spoolid);
+	if (cfpi->spoolid)
+	  sfprintf(errfp, "Local-Spool-ID: %s\n", cfpi->spoolid);
 	sfprintf(errfp, "\n");
 
 	/* Now scan 'em all again for IETF-NOTARY */
 	for (i = 0; i < notarycnt; ++i) {
-	  scnotaryreport(errfp, &notaries[i],&has_errors,no_error_report);
+	  scnotaryreport(errfp, &notaries[i],&has_errors,no_error_report, 0);
 	  if (notaries[i].not != NULL)
 	    free(notaries[i].not);
 	}
@@ -946,14 +1012,14 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	    /* Scan the input, and drop off the Zmailer
 	       envelope headers */
 	    sfseek(fp, (Sfoff_t)0, SEEK_SET);
-	    while (cfgets(buf,sizeof(buf),fp) >= 0) {
+	    while (csfgets(buf,sizeof(buf),fp) >= 0) {
 	      const char *s = buf;
 	      while (*s && *s != ':' && *s != ' ' && *s != '\t') ++s;
 	      if (*s == ':') break;
 	      *buf = 0;
 	    }
 	    /* We leave the first scan-phase with  buf[]  containing some
-	       valid RFC-822 -style header, propably "Received:" */
+	       valid RFC-822 -style header, probably "Received:" */
 	    if (*buf)
 	      sfprintf(errfp, "%s", buf);
 	    else {
@@ -978,13 +1044,23 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	/* And cap the tail with paired MIME boundary.. */
 	sfprintf(errfp, "--%s--\n", boundarystr);
 
-	if (no_error_report > 0)
-	  sfmail_close_alternate(errfp,POSTMANDIR,":error-on-error");
-	else
-	  sfmail_close(errfp);	/* XX: check for error */
+	ino = 0; mtime = 0;
 	close(cfp->fd);
-	cfp->mid = NULL; /* we don't want to loose the original one! */
 	free_cfp_memory(cfp);
+
+	if (no_error_report > 0) {
+	  sfmail_close_alternate_async(errfp,POSTMANDIR,":error-on-error", msgwriteasync);
+	  sprintf(rptspoolid, "POSTMAN :error-on-error"); /* < 30 chr ! */
+	} else {
+	  _sfmail_close_async(errfp, &ino, &mtime, msgwriteasync);	/* XX: check for error */
+	  taspoolid(rptspoolid, mtime, ino);
+	}
+
+	if (do_syslog)
+	  zsyslog((LOG_INFO, "%s: Created '%s' report on spoolid: %s",
+		   cfpi->spoolid ? cfpi->spoolid:"-",
+		   deliveryform, rptspoolid));
+
 }
 
 
@@ -1012,13 +1088,7 @@ msgdelayed(vp, offset, notary, message)
 	  if (s) *s++ = 0;
 	}
 
-	if (vp->cfp->dirind > 0) {
-	  sprintf(path, "%.300s/%.100s",
-		  cfpdirname(vp->cfp->dirind), vp->cfp->mid);
-	} else {
-	  strncpy(path,vp->cfp->mid,400);
-	  path[400] = 0;
-	}
+	sprintf(path, "%s%.300s", cfpdirname(vp->cfp->dirind), vp->cfp->mid);
 
 	/* exclusive access required, but we're the only scheduler... */
 	fp = sfopen(NULL, path, "a");
@@ -1052,7 +1122,11 @@ msgdelayed(vp, offset, notary, message)
 	sfprintf(fp, "\t%s\n", message);
 	sfsync(fp);
 #ifdef HAVE_FSYNC
-	fsync(sffileno(fp));
+	while (fsync(sffileno(fp)) < 0) {
+	  if (errno == EINTR || errno == EAGAIN)
+	    continue;
+	  break;
+	}
 #endif
 	sfclose(fp);
 }
@@ -1138,7 +1212,7 @@ struct spblk *spl;
 	  nvp = vp->next[L_CTLFILE];
 
 	  /* We essentially report accumulated reports every
-	     ``global_report_interval'' seconds (2h by default) */
+	     ``global_report_interval'' seconds (2 minutes by default) */
 
 	  if (vp->nextrprttime <= now)
 	    doreport = 1;
