@@ -1,6 +1,6 @@
 /*	Prototypes of ZMailer Scheduler component routines	*/
 /*
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-1999
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-2002
  */
 
 #ifndef __
@@ -41,6 +41,7 @@ extern u_int ranny __((u_int m));
 /* agenda.c */
 extern time_t qipcretry;
 extern int doagenda __((void));
+extern int doexpiry2 __((void));
 extern int turnme __((const char *));
 
 /* conf.c */
@@ -52,10 +53,12 @@ extern const char *replhost;
 extern const char *replchannel;
 extern int  nobody;
 extern int  sweepinterval;
+extern int  idle_sweepinterval;
 extern const char *mq2authfile;
-extern const char *mailqsock;
 extern const void *BADPTR;
 extern int   global_report_interval;
+extern int   expiry2_timelimit;
+extern int   expiry2_sweepinterval;
 
 /* msgerror.c */
 extern void msgerror __((struct vertex *vp, long offset, const char *message));
@@ -78,6 +81,7 @@ extern void   defaultconfigentry __((struct config_entry *ce, struct config_entr
 extern void   vtxprint __((struct vertex *vp));
 extern struct config_entry *readconfig __((const char *file));
 extern struct config_entry *rereadconfig __((struct config_entry *head, const char *file));
+extern char  *zenvexpand __(( char *line));
 
 /* resources.c */
 extern int  resources_query_nofiles  __((void));
@@ -91,11 +95,15 @@ extern const char *progname;
 extern const char *rendezvous;
 extern const char *pidfile;
 extern const char *mailshare;
-extern const char *log;
+extern const char *logfn;
+extern int  msgwriteasync;
 extern int  hungry_childs;
 extern int  global_maxkids;
 extern int  verbose;
 extern int  querysocket;
+extern int  notifysocket;
+extern char *mailqsock;
+extern char *notifysock;
 extern int  do_syslog;
 extern struct ctlfile *slurp __((int fd, long ino));
 extern void free_cfp_memory __((struct ctlfile *cfp));
@@ -112,6 +120,7 @@ extern time_t mytime __((time_t *));
 extern int dq_insert __((void*, long, const char*, int));
 extern int in_dirscanqueue __((void *, long));
 extern const char *cfpdirname __((int));
+extern void timed_log_reinit __((void));
 
 extern struct MIB_MtaEntry *MIBMtaEntry;
 
@@ -134,6 +143,8 @@ extern int   idleprocs;
 extern void  web_detangle __((struct vertex *vp, int ok));
 extern void  reschedule __((struct vertex *vp, int factor, int index));
 extern int   thread_reschedule __((struct thread *, time_t, int index));
+extern int   thread_expire2 __((struct thread *thr, time_t timelimit, int killall, const char *msgstr));
+extern int   thread_count_files __((void));
 
 /* transport.c */
 extern struct procinfo *cpids;
@@ -148,6 +159,7 @@ extern void queryipcinit __((void));
 #if defined(USE_BINMKDIR) || defined(USE_BINRMDIR)
 extern int system __((char*));
 #endif
+extern void receive_notify __((int fd));
 
 /* update.c */
 extern void update __((int, char *));
@@ -181,3 +193,5 @@ extern int  mq2_active __((void));
 
 /* mq2auth.c */
 extern void mq2auth __((struct mailq *, char *));
+struct mq2pw; /* forward definition */
+extern struct mq2pw * mq2_authuser __((struct mailq *mq, char *user));
