@@ -4,10 +4,10 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2001
  */
 
-#include "hostenv.h"
+#include "mailer.h"
 #include <sfio.h>
 #include <ctype.h>
 #include <pwd.h>
@@ -21,8 +21,13 @@
 #endif
 
 #include "libz.h"
+#include "libc.h"
 
-static void celink __((struct config_entry *, struct config_entry **, struct config_entry **));
+#define SKIPSPACE(Y) while (*Y == ' ' || *Y == '\t' || *Y == '\n') ++Y
+#define SKIPTEXT(Y)  while (*Y && *Y != ' ' && *Y != '\t' && *Y != '\n') ++Y
+#define SKIPDIGIT(Y) while ('0' <= *Y && *Y <= '9') ++Y
+
+static void celink __((struct config_entry *, struct config_entry **, struct config_entry **, int copy));
 static int readtoken __((Sfio_t *fp, char *buf, int buflen, int *linenump));
 static int paramparse __((char *line));
 
@@ -31,6 +36,7 @@ static int paramparse __((char *line));
 static int rc_command		RCKEYARGS;
 static int rc_expform		RCKEYARGS;
 static int rc_expiry		RCKEYARGS;
+static int rc_expiry2		RCKEYARGS;
 static int rc_group		RCKEYARGS;
 static int rc_interval		RCKEYARGS;
 static int rc_maxchannel	RCKEYARGS;
@@ -55,48 +61,46 @@ static int rc_sysnice		RCKEYARGS;
 
 extern int errno;
 
-extern struct group *getgrnam();
-extern struct passwd *getpwnam();
-
 struct config_entry *default_entry = NULL;
 struct config_entry *rrcf_head     = NULL;
 
 /* where the  MAILQv2  authentication dataset file is ? */
-char * mq2authfile = NULL;
+const char * mq2authfile = NULL;
 
 
 static struct rckeyword {
 	const char	*name;
 	int		(*parsef)();
 } rckeys[] = {
-{	"interval",		rc_interval	},	/* time */
+{	"ageorder",		rc_ageorder	},	/* boolean */
+{	"bychannel",		rc_bychannel	},	/* boolean */
+{	"command",		rc_command	},	/* array of strings */
+{	"deliveryform",		rc_deliveryform	},	/* string */
+{	"expiry",		rc_expiry	},	/* time */
+{	"expiry2",		rc_expiry2	},	/* time */
+{	"expiryform",		rc_expform	},	/* string */
+{	"group",		rc_group	},	/* number */
 {	"idlemax",		rc_idlemax	},	/* time */
+{	"interval",		rc_interval	},	/* time */
 {	"maxchannel",		rc_maxchannel	},	/* number */
 {	"maxchannels",		rc_maxchannel	},	/* number */
-{	"expiry",		rc_expiry	},	/* time */
-{	"command",		rc_command	},	/* array of strings */
-{	"expiryform",		rc_expform	},	/* string */
-{	"deliveryform",		rc_deliveryform	},	/* string */
-{	"retries",		rc_retries	},	/* array of numbers */
-{	"reporttimes",		rc_reporttimes	},	/* array of numbers */
 {	"maxring",		rc_maxring	},	/* number */
 {	"maxrings",		rc_maxring	},	/* number */
 {	"maxta",		rc_maxta	},	/* number */
 {	"maxthr",		rc_maxthr	},	/* number */
 {	"maxtransport",		rc_maxta	},	/* number */
 {	"maxtransports",	rc_maxta	},	/* number */
-{	"user",			rc_user		},	/* number */
-{	"group",		rc_group	},	/* number */
-{	"skew",			rc_skew		},	/* number */
-{	"bychannel",		rc_bychannel	},	/* boolean */
-{	"ageorder",		rc_ageorder	},	/* boolean */
-{	"queueonly",		rc_queueonly	},	/* boolean */
-{	"wakeuprestartonly",	rc_wakeuprestartonly },	/* boolean */
+{	"nice",			rc_nice		},	/* number */
 {	"overfeed",		rc_overfeed	},	/* number */
 {	"priority",		rc_priority	},	/* number */
-{	"nice",			rc_nice		},	/* number */
-{	"syspriority",		rc_syspriority	},	/* number */
+{	"queueonly",		rc_queueonly	},	/* boolean */
+{	"reporttimes",		rc_reporttimes	},	/* array of numbers */
+{	"retries",		rc_retries	},	/* array of numbers */
+{	"skew",			rc_skew		},	/* number */
 {	"sysnice",		rc_sysnice	},	/* number */
+{	"syspriority",		rc_syspriority	},	/* number */
+{	"user",			rc_user		},	/* number */
+{	"wakeuprestartonly",	rc_wakeuprestartonly },	/* boolean */
 {	NULL,			0		}
 };
 
@@ -116,6 +120,7 @@ defaultconfigentry(ce,defaults)
 	  ce->interval		= defaults->interval;
 	  ce->idlemax		= defaults->idlemax;
 	  ce->expiry		= defaults->expiry;
+	  ce->expiry2		= defaults->expiry2;
 	  ce->expiryform	= defaults->expiryform;
 	  ce->uid		= defaults->uid;
 	  ce->gid		= defaults->gid;
@@ -139,11 +144,14 @@ defaultconfigentry(ce,defaults)
 	} else if (defaults == NULL) {
 	  /* Compile these defaults in.. Only for the "*" / "* / *" entry.. */
 	  ce->next	= NULL;
+#if 0
 	  ce->mark	= 0;
+#endif
 
 	  ce->interval	= -1;
 	  ce->idlemax   = -1;
 	  ce->expiry	= -1;
+	  ce->expiry2	= -1;
 	  ce->expiryform = NULL;
 	  ce->uid	= -1;
 	  ce->gid	= -1;
@@ -179,12 +187,13 @@ vtxprint(vp)
 		   vp->orig[L_HOST]->name);
 	else
 	  sfprintf(sfstdout, "%s/%s", ISS(ce->channel), ISS(ce->host));
-	sfprintf(sfstdout," %p\n",		ce);
+	sfprintf(sfstdout," %p  mark %d\n",	ce, ce->mark);
 	sfprintf(sfstdout,"\tinterval %d\n",	(int)ce->interval);
 	sfprintf(sfstdout,"\tidlemax %d\n",	ce->idlemax);
-	sfprintf(sfstdout,"\texpiry %d\n",		(int)ce->expiry);
+	sfprintf(sfstdout,"\texpiry %d\n",	(int)ce->expiry);
+	sfprintf(sfstdout,"\texpiry2 %d\n",	(int)ce->expiry2);
 	sfprintf(sfstdout,"\texpiryform %s\n",	ISS(ce->expiryform));
-	sfprintf(sfstdout,"\tdeliveryform %s\n",	ISS(ce->deliveryform));
+	sfprintf(sfstdout,"\tdeliveryform %s\n", ISS(ce->deliveryform));
 	sfprintf(sfstdout,"\tuid %d\n",		ce->uid);
 	sfprintf(sfstdout,"\tgid %d\n",		ce->gid);
 	sfprintf(sfstdout,"\tcommand %s\n",	ISS(ce->command));
@@ -199,11 +208,11 @@ vtxprint(vp)
 							 " WAKEUPRESTARTONLY");
 	}
 	sfprintf(sfstdout,"\n");
-	sfprintf(sfstdout,"\tmaxkids %d\n",	ce->maxkids);
+	sfprintf(sfstdout,"\tmaxkids %d\n",		ce->maxkids);
 	sfprintf(sfstdout,"\tmaxkidChannel %d\n",	ce->maxkidChannel);
 	sfprintf(sfstdout,"\tmaxkidThread  %d\n",	ce->maxkidThread);
 	sfprintf(sfstdout,"\tmaxkidThreads %d\n",	ce->maxkidThreads);
-	sfprintf(sfstdout,"\toverfeed %d\n",	ce->overfeed);
+	sfprintf(sfstdout,"\toverfeed %d\n",		ce->overfeed);
 
 	if (ce->priority >= 80)
 	  sfprintf(sfstdout,"\tpriority %d\n",	ce->priority - 100);
@@ -239,9 +248,10 @@ vtxprint(vp)
 }
 
 static void
-celink(ce, headp, tailp)
+celink(ce, headp, tailp, copy)
 	struct config_entry *ce;
 	struct config_entry **headp, **tailp;
+	int copy;
 {
 	if (ce == default_entry && *headp != NULL && *tailp != NULL)
 	  return; /* XX: ?? */
@@ -251,13 +261,20 @@ celink(ce, headp, tailp)
 	else {
 	  (*tailp)->next = ce;
 	  (*tailp) = ce;
+
+	  if (!copy) return;
+
 	  for (ce = (*headp); ce != (*tailp); ce = ce->next) {
-	    if (ce->mark == 0)
-	      continue;
+
+	    if (verbose)
+	      sfprintf(sfstdout,"celink() ce = %p  mark=%d\n", ce, ce->mark);
+
+	    if (ce->mark == 0) continue;
 	    ce->mark = 0;
 	    ce->interval	= (*tailp)->interval;
 	    ce->idlemax		= (*tailp)->idlemax;
 	    ce->expiry		= (*tailp)->expiry;
+	    ce->expiry2		= (*tailp)->expiry2;
 	    ce->expiryform	= (*tailp)->expiryform;
 	    ce->uid		= (*tailp)->uid;
 	    ce->gid		= (*tailp)->gid;
@@ -287,6 +304,7 @@ readconfig(file)
 	struct vertex v;
 	Sfio_t *fp;
 	int linenum = 0;
+	int attrs = 0;
 
 	ce = head = tail = NULL;
 	errflag = 0;
@@ -311,8 +329,13 @@ readconfig(file)
 	    }
 
 	    if (ce != NULL)
-	      celink(ce, &head, &tail);
+	      celink(ce, &head, &tail, attrs);
+	    attrs = 0;
 	    ce = (struct config_entry *)emalloc(sizeof (struct config_entry));
+	    memset((void*)ce, 0, sizeof(ce));
+	    if (verbose) sfprintf(sfstdout,"CE= %p mark=1\n", ce);
+
+	    defaultconfigentry(ce,NULL);
 	    ce->mark = 1;
 	    if ((s = strchr(line, '/')) != NULL) {
 	      *s = 0;
@@ -344,7 +367,7 @@ readconfig(file)
 	      while (p >= line && (*p == ' ' || *p == '\t'))
 		*p-- = '\0';
 	      a = cp+1;
-	      while (*a == ' ' || *a == '\t') ++a;
+	      SKIPSPACE(a);
 	      if (*a == '"') {
 		++a;
 		cp = a;
@@ -358,8 +381,11 @@ readconfig(file)
 	      }
 	    }
 
-	    if (ce)
+	    if (ce && ce->mark) {
+	      if (verbose)
+		sfprintf(sfstdout," reading entry, ce = %p, mark=0\n", ce);
 	      ce->mark = 0;
+	    }
 
 	    for (rckp = &rckeys[0]; rckp->name != NULL ; ++rckp)
 	      if (cistrcmp(rckp->name, line) == 0) {
@@ -372,6 +398,9 @@ readconfig(file)
 		      progname, line, file, linenum);
 	      ++errflag;
 	    }
+
+	    if (!errflag) attrs = 1;
+
 	  } else {
 	    sfprintf(sfstderr, "%s: illegal syntax at %s:%d\n",
 		    progname, file, linenum);
@@ -379,7 +408,7 @@ readconfig(file)
 	  }
 	}
 	if (ce != NULL)
-	  celink(ce, &head, &tail);
+	  celink(ce, &head, &tail, 1);
 	sfclose(fp);
 	if (verbose) {
 	  struct threadgroup tg;
@@ -406,14 +435,13 @@ readtoken(fp, buf, buflen, linenump)
 
 redo_readtoken:
 	if (lp == NULL) {
-	  if (cfgets(line, sizeof line, fp) < 0)
+	  if (csfgets(line, sizeof line, fp) < 0)
 	    return -1;
 	  *linenump += 1;
 	  lp = line;
 	}
 	/* Skip initial white-space */
-	while (*lp != '\0' && isspace(0xFF & *lp))
-	  ++lp;
+	SKIPSPACE(lp);
 	/* Now it is one of: a token, a comment start, or end of line */
 	if (*lp == '\0' || *lp == '#') {
 	  /* Comment/EOL */
@@ -422,12 +450,12 @@ redo_readtoken:
 	}
 	/* Now we scan for the token + possible value */
 	elp = lp;
-	while (*elp && !isspace(0xFF & *elp) && *elp != '=' && *elp != '#')
+	while (*elp && *elp != ' ' && *elp != '\t' && *elp != '\n' && *elp != '=' && *elp != '#')
 	  ++elp;
 	if (isspace(0xFF & *elp)) {
 	  /* Allow spaces after the token and before '=' */
 	  char *p = elp;
-	  while (*p && isspace(0xFF & *p)) ++p;
+	  SKIPSPACE(p);
 	  if (*p == '=')
 	    elp = p;
 	}
@@ -435,12 +463,12 @@ redo_readtoken:
 	if (*elp == '=') {
 	  /* Allow spaces between '=', and value */
 	  ++elp;
-	  while (*elp && isspace(0xFF & *elp)) ++elp;
+	  SKIPSPACE(elp);
 	  if (*elp == '"') {
 	    ++elp;
 	    while (*elp != '"' && *elp != '\0') {
 	      if (*elp == '\\' && *(elp+1) == '\n') {
-		if (cfgets(elp, sizeof line - (elp - line), fp) < 0) {
+		if (csfgets(elp, sizeof line - (elp - line), fp) < 0) {
 		  sfprintf(sfstderr,
 			  "%s: bad continuation line\n",
 			  progname);
@@ -457,8 +485,7 @@ redo_readtoken:
 	    }
 	    ++elp;
 	  } else {
-	    while (*elp && !isspace(0xFF & *elp) && *elp != '"')
-	      ++elp;
+	    SKIPTEXT(elp);
 	  }
 	}
 	strncpy(buf, lp, elp-lp);
@@ -538,17 +565,14 @@ static int rc_command(key, arg, ce)
 
 	ce->command = strsave(arg);
 	j = 0;
-	for (cp = ce->command; *cp != '\0' && isascii(*cp);) {
+	for (cp = ce->command; *cp;) {
 	  argv[j++] = cp;
 	  if (j >= (sizeof argv)/(sizeof argv[0]))
 	    break;
-	  while (*cp != '\0' && !isspace(0xFF & *cp))
-	    ++cp;
-	  if (*cp == '\0')
-	    break;
+	  SKIPTEXT(cp);
+	  if (*cp == '\0')  break;
 	  *cp++ = '\0';
-	  while (*cp != '\0' && isspace(0xFF & *cp))
-	    ++cp;
+	  SKIPSPACE(cp);
 	}
 	argv[j++] = NULL;
 	if (j > 0) {
@@ -585,16 +609,24 @@ static int rc_expiry(key, arg, ce)
 	return 0;
 }
 
+static int rc_expiry2(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	ce->expiry2 = parse_interval(arg,NULL);
+	return 0;
+}
+
 static int rc_group(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	struct group *gr;
+	struct Zgroup *gr;
 
 	if (isascii(*arg) && isdigit(*arg))
 	  ce->gid = atoi(arg);
-	else if ((gr = getgrnam(arg)) == NULL) {
-	  sfprintf(sfstderr, "%s: unknown group: %s\n", progname, arg);
+	else if ((gr = zgetgrnam(arg)) == NULL) {
+	  sfprintf(sfstderr, "%s: unknown group: '%s'\n", progname, arg);
 	  return 1;
 	} else
 	  ce->gid = gr->gr_gid;
@@ -662,7 +694,7 @@ static int rc_syspriority(key, arg, ce)
 	  sfprintf(sfstderr, "%s: Bad UNIX priority value, acceptable in range: -20..19; input=\"%s\"\n", progname, arg);
 	  return 1;
 	}
-#ifdef HAVE_SETPRIORITY
+#if defined(HAVE_SETPRIORITY) && defined(HAVE_SYS_RESOURCE_H)
 	/* PRIO_PROCESS depends likely of  HAVE_SYS_RESOURCE_H */
 	setpriority(PRIO_PROCESS, 0, i);
 #endif
@@ -734,13 +766,11 @@ static int rc_retries(key, arg, ce)
 
 	j = 0;
 	for (cp = arg; *cp != '\0'; ++cp) {
-	  while (*cp != '\0' && isspace(0xFF & *cp))
-	    ++cp;
+	  SKIPSPACE(cp);
 	  if (*cp == '\0')
 	    break;
 	  d = cp++;
-	  while (*cp != '\0' && !isspace(0xFF & *cp))
-	    ++cp;
+	  SKIPTEXT(cp);
 	  c = *cp;
 	  *cp = '\0';
 	  i = atoi(d);
@@ -780,13 +810,11 @@ static int rc_reporttimes(key, arg, ce)
 
 	j = 0;
 	for (cp = arg; *cp != '\0'; ++cp) {
-	  while (*cp != '\0' && isspace(0xFF & *cp))
-	    ++cp;
+	  SKIPSPACE(cp);
 	  if (*cp == '\0')
 	    break;
 	  d = cp++;
-	  while (*cp != '\0' && !isspace(0xFF & *cp))
-	    ++cp;
+	  SKIPTEXT(cp);
 	  c = *cp;
 	  *cp = '\0';
 	  i = parse_interval(d, NULL);
@@ -815,11 +843,11 @@ static int rc_user(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	struct passwd *pw;
+	struct Zpasswd *pw;
 
 	if (isascii(*arg) && isdigit(*arg))
 	  ce->uid = atoi(arg);
-	else if ((pw = getpwnam(arg)) == NULL) {
+	else if ((pw = zgetpwnam(arg)) == NULL) {
 	  sfprintf(sfstderr, "%s: unknown user: %s\n", progname, arg);
 	  return 1;
 	} else
@@ -882,6 +910,59 @@ static int rc_wakeuprestartonly(key, arg, ce)
 
 extern int mailqmode;
 
+char *zenvexpand(line)
+     char *line;
+{
+	char *s = line;
+	char *v;
+	const char *e;
+	char end_c;
+	int spotlen;
+	int e_len;
+
+	if (!line) return NULL;
+
+	for (;*s;++s) {
+	  if (*s == '$') {
+	    if (s[1] == '{') {
+	      v = s;
+	      for (;*s && *s != '}'; ++s) ;
+	      end_c = *s;
+	      *s = 0;
+	      spotlen = s - v + 1;
+	      e = getzenv(v+2);
+	      *s = end_c;
+	      if (!e) continue; /* No such zenv-variable :-/ */
+	      e_len = strlen(e);
+	      if (e_len > spotlen) {
+		/* Must expand a bit */
+		char *n = malloc(strlen(line)+e_len-spotlen+2);
+		int p = v - line;
+		if (!n) continue; /* alloc failure */
+		if (p > 0)
+		  memcpy(n, line, p);
+		s = n + p;
+		memcpy(s, e, e_len);
+		s += e_len;
+		p += spotlen;
+		strcpy(s, line + p); /* Tail */
+		free(line);
+		line = n;
+	      } else {
+		/* New data has same or smaller size */
+		memcpy(v, e, e_len);
+		if (e_len < spotlen) /* Smaller size */
+		  strcpy(v + e_len, v + spotlen);
+		v += e_len;
+		s = v;
+	      }
+	    }
+	  }
+	}
+	return line;
+}
+
+
 static int paramparse(line)
 	char *line;
 {
@@ -893,7 +974,7 @@ static int paramparse(line)
 	  while (p >= line && (*p == ' ' || *p == '\t'))
 	    *p-- = '\0';
 	  a = s+1;
-	  while (*a == ' ' || *a == '\t') ++a;
+	  SKIPSPACE(a);
 	  if (*a == '"') {
 	    ++a;
 	    s = a;
@@ -909,8 +990,8 @@ static int paramparse(line)
 
 	if (cistrcmp(line,"authfile")==0 && a) {
 	  if (mq2authfile)
-	    free(mq2authfile);
-	  mq2authfile = strsave(a);
+	    free((void*)mq2authfile);
+	  mq2authfile = zenvexpand(strsave(a));
 
 	  if (mq2authfile && access(mq2authfile,R_OK)==0)
 	    mailqmode = 2;
@@ -920,8 +1001,20 @@ static int paramparse(line)
 
 	if (cistrcmp(line,"mailqsock")==0 && a) {
 	  if (mailqsock)
-	    free(mailqsock);
-	  mailqsock = strsave(a);
+	    free((void*)mailqsock);
+	  mailqsock = zenvexpand(strsave(a));
+	  return 0;
+	}
+
+	if (cistrcmp(line,"msgwriteasync")==0 && a) {
+	  msgwriteasync = atoi(s);
+	  return 0;
+	}
+
+	if (cistrcmp(line,"notifysock")==0 && a) {
+	  if (notifysock)
+	    free((void*)notifysock);
+	  notifysock = zenvexpand(strsave(a));
 	  return 0;
 	}
 
