@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 /*
@@ -13,6 +13,7 @@
 
 #include <stdio.h>
 #include <sfio.h>
+
 #include <sys/param.h>
 #include "hostenv.h"
 #include <ctype.h>
@@ -23,6 +24,7 @@
 #include <string.h>
 #include "ta.h"
 #include "sysexits.h"
+#include "libc.h"
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -101,6 +103,7 @@ time_t	sched_starttime;
 int	do_syslog = 0;
 int	verbose = 0;
 int	querysocket = -1;	/* fd of TCP socket to listen for queries */
+int	msgwriteasync = 0;
 int	D_alloc = 0;
 int	hungry_childs = 0;
 int	global_wrkcnt = 0;
@@ -122,16 +125,23 @@ char * procselect = NULL;	/* Non-null defines  channel/host specifier
 char *  procselhost = NULL;	/* Just spliced out 'host'-part of the above */
 extern int forkrate_limit;	/* How many forks per second ? */
 int	mailqmode = 1;		/* ZMailer v1.0 mode on mailq */
-const char *  mailqsock = NULL;
+char *  mailqsock;
+char *  notifysock;
+
+int    interim_report_interval = 5*60; /* 5 minutes */
+
+static int vtxprep_skip;
+static int vtxprep_skip_any;
+static int vtxprep_skip_lock;
+static time_t next_dirscan;
+static time_t next_idlecleanup;
+static time_t next_interim_report_time;
+static time_t next_expiry2_scan;
+static struct sptree *dirscan_mesh;
+
+static int newents_limit = 40000;
+static int newents_timelimit = 5;
 
-static int vtxprep_skip      = 0;
-static int vtxprep_skip_any  = 0;
-static int vtxprep_skip_lock = 0;
-static time_t next_dirscan     = 0;
-static time_t next_idlecleanup = 0;
-static time_t next_interim_report_time = 0;
-static struct sptree *dirscan_mesh = NULL;
-static int newents_limit = 400;
 extern int default_full_content; /* on conf.c */
 
 #include "memtypes.h"
@@ -146,7 +156,6 @@ static int  globmatch   __((const char *, const char*));
 static void vtxdo   __((struct vertex *, struct config_entry *, const char *));
 
 extern void  cfp_mksubdirs __((const char *, const char*));
-extern const char *cfpdirname __((int));
 
 static RETSIGTYPE sig_exit   __((int sig));
 static RETSIGTYPE sig_quit   __((int sig));
@@ -156,11 +165,6 @@ static RETSIGTYPE sig_readcf __((int sig));
 
 extern char *strerror __((int err));
 
-static struct MIB_MtaEntry MIBMtaEntryLocal = {0,};
-struct MIB_MtaEntry *MIBMtaEntry = &MIBMtaEntryLocal;
-
-
-static int    timeserver_pid = 0;
 extern time_t mytime          __((time_t *));
 static void   init_timeserver __((void));
 
@@ -199,8 +203,8 @@ int sig;
 	  sfset(sfstderr, SF_LINE, 1);
 	}
 	if (statusfn != NULL && statuslog != NULL) {
-	  sfsync(statuslog);
-	  if (sfopen(statuslog, statusfn, "a") != statuslog) {
+	  sfclose(statuslog);
+	  if (sfopen(NULL, statusfn, "a") != statuslog) {
 	    sfprintf(sfstderr,"%s: cannot open statuslog: %s, errno=%d\n", progname, statusfn, errno);
 	    return -1;
 	  }
@@ -213,6 +217,42 @@ int sig;
 	return 0;
 }
 
+/*
+ *  A self-timed log re-init caller to make sure, that
+ *  no log file is kept open for more than 30 seconds
+ *  in one go.  This allows (within 30 seconds) the files
+ *  to be rotates at first to new names, then after that
+ *  30 second delay, to be further processed - e.g. compressed.
+ *
+ */
+
+void timed_log_reinit()
+{
+  static time_t next_log_reinit;
+
+  mytime(&now);
+
+  if (next_log_reinit < now) {
+    struct stat stbuf;
+
+    next_log_reinit = now + 25;
+    loginitsched(SIGHUP);
+
+    if (logfn != NULL &&
+	lstat(logfn,&stbuf) == 0 &&
+	S_ISREG(stbuf.st_mode)) {
+
+      long fsspace;
+      fsspace = free_fd_statfs(sffileno(sfstdout));
+      MIBMtaEntry->sys.LogFreeSpace = fsspace;
+      fsspace = used_fd_statfs(sffileno(sfstdout));
+      MIBMtaEntry->sys.LogUsedSpace = fsspace;
+    }
+  }
+}
+
+
+
 struct dirstatname {
 	struct stat st;
 	long ino;
@@ -248,11 +288,11 @@ struct ctlfile *cfp;
 	   and can be written to.  If the file does not
 	   exist, no logging shall happen! */
 
-	setreuid(0, cfp->uid);
+	SETEUID(cfp->uid);
 	fd = open(cfp->vfpfn, O_WRONLY|O_APPEND, 0);
-	setreuid(0, 0);
+	SETEUID(0);
 	if (fd < 0) return NULL; /* Can't open it! */
-	vfp = sfnew(NULL, NULL, 0, fd, SF_WRITE|SF_APPEND|SF_LINE);
+	vfp = sfnew(NULL, NULL, 0, fd, SF_WRITE|SF_APPENDWR|SF_LINE);
 	if (!vfp) return NULL; /* Failure to open */
 
 	sfseek(vfp, (Sfoff_t)0, SEEK_END);
@@ -284,17 +324,19 @@ struct ctlfile *cfp;
 	/* Throw all way, if no vertices.. */
 	if (!cfp->head) {
 	  /* This should *not* be happening.. */
+	  sfprintf(sfstderr,
+		   "%s: SHOULD NOT HAPPEN: cfp->head == NULL; spoolid: %s\n",
+		   progname, cfp->spoolid);
 	  unctlfile(cfp, 1);
 	  return;
 	}
 
 	for (vp = cfp->head; vp != NULL; vp = nvp) {
-	  /* Stash the next pointer, because the CFP content
-	     may become trashed at the end of the run. */
+	  /* Stash the next pointer, because the VP content
+	     WILL become trashed at the end of the loop. */
 	  nvp = vp->next[L_CTLFILE];
 
-	  MIBMtaEntry->mtaStoredRecipients     -= vp->ngroup;
-	  MIBMtaEntry->mtaReceivedRecipientsSc -= vp->ngroup;
+	  MIBMtaEntry->sc.StoredRecipientsSc   -= vp->ngroup;
 	  vp->ngroup = 0;
 	  unvertex(vp,1,1); /* Don't unlink()! Just free()! */
 	}
@@ -337,21 +379,26 @@ struct ctlfile *cfp;
 {
 	if (cfp->contents)	free(cfp->contents);
 	if (cfp->vfpfn)		free(cfp->vfpfn);
+	if (cfp->spoolid)	free(cfp->spoolid);
 	if (cfp->mid)		free(cfp->mid);
 	if (cfp->erroraddr)	free(cfp->erroraddr);
 	if (cfp->logident)	free(cfp->logident);
 	if (cfp->envid)		free(cfp->envid);
 
-memset(cfp, 0x55, sizeof(*cfp));
+	/* memset(cfp, 0x55, sizeof(*cfp)); */
 
 	free((char *)cfp);
+
+	--global_wrkcnt;
+	--MIBMtaEntry->sc.StoredMessagesSc;
 }
 
 
 struct dirqueue dirqb = { 0, };
 struct dirqueue *dirq = &dirqb;
 
-const char **ArgvSave;
+char *ArgvSave;
+const char *EOArgvSave;
 
 extern int main __((int, const char **));
 
@@ -485,16 +532,6 @@ main(argc, argv)
 	__free_hook = my_free_hook;
 #endif
 
-	syslogflg = getzenv("SYSLOGFLG");
-	if (syslogflg == NULL)
-	  syslogflg = "";
-	t = syslogflg;
-	for ( ; *t ; ++t ) {
-	  if (*t == 'c' || *t == 'C')
-	    break;
-	}
-	do_syslog = (*t != '\0');
-
 #ifdef HAVE_SETGROUPS
 	/* We null supplementary groups list entirely */
 	setgroups(0, NULL);
@@ -503,7 +540,8 @@ main(argc, argv)
 	freeze = 0;
 	/* setlinebuf(stderr);  -- no need for this ? */
 
-	ArgvSave = argv;
+	ArgvSave   = (char *) argv[0];
+	EOArgvSave = argv[argc-1] + strlen(argv[argc-1]) + 1;
 
 	mytime(&sched_starttime);
 
@@ -552,7 +590,7 @@ main(argc, argv)
 	verbose = errflg = version = 0;
 	for (;;) {
 		c = getopt(argc, (char*const*)argv,
-			   "divE:f:Fl:HL:M:nN:p:P:q:QR:SVW");
+			   "divE:f:Fl:HL:M:nN:p:P:q:QR:SVWZ:");
 		if (c == EOF)
 		  break;
 		switch (c) {
@@ -560,9 +598,11 @@ main(argc, argv)
 			config = optarg;
 			break;
 		case 'E':
-			newents_limit = atoi(optarg);
-			if (newents_limit < 10)
-			  newents_limit = 10;
+		        sscanf(optarg, "%d,%d", &newents_limit, &newents_timelimit);
+
+			if (newents_limit < 10)      newents_limit = 10;
+			if (newents_timelimit < 2)   newents_timelimit = 2;
+			if (newents_timelimit > 15)  newents_timelimit = 15;
 			break;
 		case 'F':
 			freeze = 1;
@@ -646,6 +686,10 @@ main(argc, argv)
 		case 'S':
 			syncstart = 1;
 			break;
+		case 'Z':
+			if (readzenv(optarg) == 0)
+			  ++errflg;
+			break;
 		case '?':
 		default:
 			errflg++;
@@ -655,17 +699,27 @@ main(argc, argv)
 
 	if (errflg) {
 	  sfprintf(sfstderr,
-		   "Usage: %s [-dHisvV -M (1|2) -f configfile -L logfile -P postoffice -Q rendezvous]\n",
+		   "Usage: %s [-dHisvV -M (1|2) -f configfile -L logfile -P postoffice -Q rendezvous -Z zenvfile]\n",
 		   progname);
 	  exit(128+errflg);
 	}
+
+	syslogflg = getzenv("SYSLOGFLG");
+	if (syslogflg == NULL)
+	  syslogflg = "";
+	t = syslogflg;
+	for ( ; *t ; ++t ) {
+	  if (*t == 'c' || *t == 'C')
+	    break;
+	}
+	do_syslog = (*t != '\0');
+
+
 	cp = getzenv("SCHEDULERDIRHASH");
 	if (cp){
 	  if ((cp[0] == '1' || cp[0] == '2') &&
 	      cp[1] == 0)
 	    hashlevels = cp[0] - '0';
-	  else
-	    hashlevels = 0;
 	}
 
 	mailshare = getzenv("MAILSHARE");
@@ -678,6 +732,45 @@ main(argc, argv)
 	  logfn = emalloc(2 + (u_int)(strlen(qlogdir) + strlen(progname)));
 	  sprintf((char*)logfn, "%s/%s", qlogdir, progname);
 	}
+
+
+	/* If we are a daemon, or doing verbose foobar, attach the SHM MIB now */
+	if (daemonflg || verbose) {
+
+	  int r = Z_SHM_MIB_Attach (1);
+
+	  if (r < 0) {
+	    /* Error processing -- magic set of constants: */
+	    switch (r) {
+	    case -1:
+	      /* fprintf(stderr, "No ZENV variable: SNMPSHAREDFILE\n"); */
+	      break;
+	    case -2:
+	      perror("Failed to open for exclusively creating of the SHMSHAREDFILE");
+	      break;
+	    case -3:
+	      perror("Failure during creation fill of SGMSHAREDFILE");
+	      break;
+	    case -4:
+	      perror("Failed to open the SHMSHAREDFILE at all");
+	      break;
+	    case -5:
+	      perror("The SHMSHAREDFILE isn't of proper size! ");
+	      break;
+	    case -6:
+	      perror("Failed to mmap() of SHMSHAREDFILE into memory");
+	      break;
+	    case -7:
+	      fprintf(stderr, "The SHMSHAREDFILE  has magic value mismatch!\n");
+	      break;
+	    default:
+	      break;
+	    }
+	    /* return; NO giving up! */
+	  }
+	}
+
+
 	if (logfn != NULL) {
 	  /* loginit is a signal handler, so can't pass log */
 	  if (loginitsched(SIGHUP) < 0) /* do setlinebuf() there */
@@ -727,6 +820,8 @@ main(argc, argv)
 	  /* NOTREACHED */
 	}
 
+
+
 	if (postoffice == NULL && (postoffice = getzenv("POSTOFFICE")) == NULL)
 	  postoffice = POSTOFFICE;
 
@@ -751,6 +846,26 @@ main(argc, argv)
 	  sfprintf(sfstdout, "%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
 		   progname, Version, (int)getpid(), (char *)rfc822date(&now));
 	}
+
+
+	/* Now we are either interactive, or daemon, lets attach monitoring
+	   memory block.. and fill it in.  */
+
+	MIBMtaEntry->sys.SchedulerMasterPID        = getpid();
+	MIBMtaEntry->sys.SchedulerMasterStartTime  = time(NULL);
+	MIBMtaEntry->sys.SchedulerMasterStarts    += 1;
+
+	/* Zero the gauges at our startup.. */
+	MIBMtaEntry->sc.StoredMessagesSc		= 0;
+	MIBMtaEntry->sc.StoredThreadsSc		= 0;
+	MIBMtaEntry->sc.StoredVerticesSc		= 0;
+	MIBMtaEntry->sc.StoredRecipientsSc		= 0;
+	MIBMtaEntry->sc.StoredVolumeSc		= 0;
+	MIBMtaEntry->sc.TransportAgentProcessesSc	= 0;
+	MIBMtaEntry->sc.TransportAgentsActiveSc	= 0;
+	MIBMtaEntry->sc.TransportAgentsIdleSc		= 0;
+
+
 	/* Actually we want this to act as daemon,
 	   even when not in daemon mode.. */
 	if (killprevious(SIGTERM, pidfile) != 0) {
@@ -764,6 +879,20 @@ main(argc, argv)
 	  spt_mesh[i]->symbols = sp_init();
 	}
 	zopenlog("scheduler", LOG_PID, LOG_MAIL);
+
+	mustexit = gotalarm = dumpq = rereadcf = 0;
+	canexit = 0;
+	SIGNAL_IGNORE(SIGPIPE);
+	SIGNAL_HANDLE(SIGALRM, sig_alarm);	/* process agenda */
+	SIGNAL_HANDLE(SIGUSR2, sig_iot);	/* dump queue info */
+
+	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* split */
+	SIGNAL_HANDLE(SIGQUIT, sig_quit);	/* Slow shutdown */
+
+	/* call it to create the timeserver -- if possible */
+	init_timeserver(); /* Will take around 3-4 secs.. */
+
+
 	if (optind < argc) {
 	  /* process the specified control files only */
 	  for (; optind < argc; ++optind) {
@@ -783,14 +912,6 @@ main(argc, argv)
 	  killpidfile(pidfile);
 	  exit(0);
 	}
-	mustexit = gotalarm = dumpq = rereadcf = 0;
-	canexit = 0;
-	SIGNAL_IGNORE(SIGPIPE);
-	SIGNAL_HANDLE(SIGALRM, sig_alarm);	/* process agenda */
-	SIGNAL_HANDLE(SIGUSR2, sig_iot);	/* dump queue info */
-
-	/* call it to create the timeserver -- if possible */
-	init_timeserver(); /* Will take around 3-4 secs.. */
 
 	queryipcinit();
 
@@ -799,10 +920,7 @@ main(argc, argv)
 	vtxprep_skip_lock = 0;
 	syncweb(dirq);
 
-
 	canexit = 1;
-	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* split */
-	SIGNAL_HANDLE(SIGQUIT, sig_quit);	/* Slow shutdown */
 #ifdef	MALLOC_TRACE
 	mal_leaktrace(1);
 #endif	/* MALLOC_TRACE */
@@ -833,33 +951,48 @@ main(argc, argv)
 
 	  canexit = 0;
 
+	  /* SAH Hmm... next_dirscan is NOT reset unless we find < 150
+	   * new files in dirqueuescan - meaning this will be true.
+	   * So a 'busy' system will potentially be running dirqueuescan()
+	   * EVERY pass thru the loop... This could be bad
+	   */
+
 	  if (now >= next_dirscan) {
 	    /* Directory scan time for new jobs ..    */
 	    /* Do it recursively every now and then,
 	       so that if we forget some jobs, they will
 	       become relearned soon enough.          */
-	    if (dirqueuescan(".", dirq, 
-#if 1 /* Do recursively every time */
-			     1
-#else
-			     (now >= next_idlecleanup)
-#endif
-			     ) < 150) {
-	      /* If we have more things to scan, don't quit yet! */
-	      next_dirscan = now + sweepinterval;  /* 10 seconds interval */
-	      if (dirq->wrksum > 100)
-		next_dirscan = now + 60; /* Lots of work, pick new jobs
-					    less frequently */
-	    }
+	    int wrk;
+	    i = dirqueuescan(".", dirq, (now >= next_idlecleanup));
+	    mytime(&now);
+	    wrk = dirq->wrksum;
+	    wrk >>= 5; /* Divide by 32 -- just presume 32 msgs/sec.. */
+	    if (wrk > 10) wrk = 10; /* But limit to 10... */
+	    next_dirscan = now + sweepinterval + wrk; /* 10 .. 20 second
+							 sweep interval */
 	  }
 
 	  if (now >= next_idlecleanup) {
-	    next_idlecleanup = now + 20; /* 20 second interval */
 	    idle_cleanup();
+
+	    if (next_idlecleanup == 0)  next_idlecleanup = now;
+
+	    /* At regular intervals... */
+	    next_idlecleanup += idle_sweepinterval;
+	  }
+
+	  if (now >= next_expiry2_scan) {
+	    if (next_expiry2_scan == 0)  next_expiry2_scan = now;
+	    i = doexpiry2();
+	    if (i == 0) 
+	      /* At regular intervals... */
+	      next_expiry2_scan += expiry2_sweepinterval;
 	  }
 
 	  if (now >= next_interim_report_time) {
-	    next_interim_report_time = now + 120; /* every 2 minutes */
+	    if (next_interim_report_time == 0) next_interim_report_time = now;
+	    /* At regular intervals... */
+	    next_interim_report_time += interim_report_interval;
 	    interim_report_run();
 	  }
 
@@ -896,7 +1029,7 @@ main(argc, argv)
 	      /* If more things in queue, submit them up to 200 pieces, or
 		 until spent two seconds at it! */
 	      time_t now0 = now + 2;
-	      for (i=0; i < 200 && now > now0; ++i) {
+	      for (i=0; i < 200 && now < now0; ++i) {
 		if (syncweb(dirq) < 0)
 		  break; /* Out of queue! */
 		mytime(&now);
@@ -1056,7 +1189,7 @@ dq_insert(DQ, ino, file, delay)
 	  dq->wrkcount2 += 1;
 	  dq->wrksum    += 1;
 	}
-	++MIBMtaEntry->mtaReceivedMessagesSc;
+	++MIBMtaEntry->sc.ReceivedMessagesSc;
 	return 0;
 }
 
@@ -1103,6 +1236,15 @@ static int dirqueuescan(dir, dq, subdirs)
 	struct stat stbuf;
 	char file[MAXNAMLEN+1];
 	int newents = 0;
+	static time_t timelimit;
+	time_t then;
+
+	mytime( &now );
+	then = now;
+
+	if (dir && dir[0] == '.' && dir[1] == 0) {
+	  timelimit = now + newents_timelimit;
+	}
 
 #if 0
 	static time_t modtime = 0;
@@ -1133,7 +1275,14 @@ static int dirqueuescan(dir, dq, subdirs)
 
 	  /* Here we are presuming the time retrieval is essentially
 	     _free_ in form of shared memory segment.. */
-	  queryipccheck();
+	  mytime( &now );
+	  if (now >  then) {
+	    then = now;
+	    queryipccheck();
+	  }
+
+	  if (!syncstart && now > timelimit)
+	    break; /* At most  newents_timelimit  seconds in this scanner... */
 
 	  if (!syncstart && newents > newents_limit)
 	    break; /* At most NNN per one go */
@@ -1235,6 +1384,7 @@ int syncweb(dq)
 	  dq->wrkcount2 = 0;
 	  dq->wrkspace  = dq->wrkspace2;
 	  dq->wrkspace2 = 0;
+	  wrkidx = dq->wrkcount - 1; /* SAH ReSync wrkidx                    */
 	  dq->sorted = 0;
 	}
 
@@ -1295,8 +1445,12 @@ int syncweb(dq)
 
 	  /* Can open ? */
 	  if ((fd = eopen(file, O_RDWR, 0)) < 0) {
+	    /* Don't unlink here, ... */
+	    /* errno == EMFILE  --> out of FDs!  URGH! */
+#if 0
 	    if (getuid() == 0)
 	      eunlink(file,"sch-syncweb");	/* hrmpf! */
+#endif
 	  } else {
 	    /* Ok, schedule! */
 	    if (schedule(fd, file, ino, 0) != NULL) {
@@ -1387,7 +1541,7 @@ static int sync_cfps(oldcfp, newcfp, proc)
 	      for (j = i+1; j < ovp->ngroup; ++j)
 		ovp->index[j-1] = ovp->index[j];
 	      ovp->ngroup -= 1;
-	      MIBMtaEntry->mtaStoredRecipients -= 1;
+	      MIBMtaEntry->sc.StoredRecipientsSc -= 1;
 	    next_i:;
 	    }
 	    if (ovp->ngroup <= 0)
@@ -1425,7 +1579,7 @@ static int sync_cfps(oldcfp, newcfp, proc)
 	    while (ovp && ovp != vp) {
 	      novp = ovp->next[L_CTLFILE];
 
-	      MIBMtaEntry->mtaStoredRecipients -= vp->ngroup;
+	      MIBMtaEntry->sc.StoredRecipientsSc -= vp->ngroup;
 	      ovp->ngroup = 0;
 	      unvertex(ovp,-1,1); /* Don't unlink()! free() *just* ovp! */
 
@@ -1449,14 +1603,13 @@ static int sync_cfps(oldcfp, newcfp, proc)
 	while (ovp) {
 	  novp = ovp->next[L_CTLFILE];
 
-	  MIBMtaEntry->mtaStoredRecipients -= ovp->ngroup;
+	  MIBMtaEntry->sc.StoredRecipientsSc -= ovp->ngroup;
 	  ovp->ngroup = 0;
 	  unvertex(ovp,-1,1); /* Don't unlink()! free() *just* ovp! */
 	  /* Leaves CFP unharmed */
 
 	  ovp = novp;
 	}
-	
 
 	if (verbose)
 	  sfprintf(sfstdout," -- Synced %d OLD vertices\n", ovp_count);
@@ -1506,17 +1659,18 @@ void resync_file(proc, file)
 	}
 
 	oldcfp = (struct ctlfile *)(spl->data);
-	oldcfp->id = 0; /* Don't scramble spt_mesh[] latter below */
+	oldcfp->id = 0; /* Don't scramble spt_mesh[] later below */
 
 	if (spl != NULL)
 	  sp_delete(spl, spt_mesh[L_CTLFILE]);
 	spl = NULL;
 
-	oldcfp->resynccount += 1;
+	++ oldcfp->resynccount;
 
-	sfprintf(sfstdout, "Resyncing file \"%s\" (ino=%d pid=%d of=%d ho='%s') reqcnt=%d\n",
+	sfprintf(sfstdout, "Resyncing file \"%s\" (ino=%d pid=%d of=%d ho='%s') reqcnt=%d global-work-count=%ld\n",
 		 file, (int) ino, proc->pid, proc->overfed,
-		 (proc->ho ? proc->ho->name : "<NULL>"), oldcfp->resynccount);
+		 (proc->ho ? proc->ho->name : "<NULL>"),
+		 oldcfp->resynccount, global_wrkcnt);
 	/* sfprintf(sfstdout, " .. in processing db\n"); */
 
 	if (oldcfp->resynccount > MAXRESYNCS) {
@@ -1561,24 +1715,98 @@ void resync_file(proc, file)
 
 	  if (oldcfp->head == NULL) {
 	    cfp_free(oldcfp,spl);
-	    if (verbose)
-	      sfprintf(sfstdout," .. LOST in resync ?!\n");
+	    sfprintf(sfstdout," .. LOST in resync ?!  (wrkcnt %ld)\n", global_wrkcnt);
 	  } else
-	    if (verbose)
-	      sfprintf(sfstdout," .. resynced!\n");
+	    sfprintf(sfstdout," .. resynced!  (wrkcnt %ld)\n", global_wrkcnt);
 
 	} else {
 
-	  if (verbose)
-	    sfprintf(sfstdout," .. NOT resynced!\n");
-
 	  /* Sigh.. Throw everything away :-( */
 	  oldcfp->id = ino;
 	  cfp_free(oldcfp, NULL);
+
+	  sfprintf(sfstdout," .. NOT resynced!  (wrkcnt %ld)\n", global_wrkcnt);
 	}
 }
 
 
+
+void receive_notify(fd)
+	int	fd;
+{
+	char buf[1000], *s, *file;
+	int i, ok, cnt;
+	u_long ino;
+
+	/* 10 receives at the time AT MOST */
+	for (cnt=10;cnt;--cnt) {
+
+	  i = recvfrom(fd, buf, sizeof(buf), 0, NULL, NULL);
+	  if (i <= 0 && errno != EINTR) return;
+	  if (i < 0) continue;
+
+	  if (i >= sizeof(buf)) i = sizeof(buf)-1;
+	  buf[i] = 0;
+
+	  /* The only supported notify is:
+	       "NEW X/Y/file-name"
+	     message */
+
+
+	  if (strncmp(buf,"NEW ",4) != 0) continue;
+
+	  file = s = buf+4;
+	  ok = 0;
+
+	  if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+	    ok = 1;
+	  } else  if ('A' <= s[0] && s[0] <= 'Z' && s[1] == '/') {
+	    s += 2;
+	    if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+	      ok = 1;
+	    } else  if ('A' <= s[0] && s[0] <= 'Z' && s[1] == '/') {
+	      s += 2;
+	      if ('1' <= s[0] && s[0] <= '9' && strchr(s, '/') == NULL) {
+		ok = 1;
+	      }
+	    }
+	  }
+
+	  ino = atol(s);
+
+#if 0
+	  sfprintf(sfstderr,"SCH-NOTIFY: len=%d '%s' file='%s', ino=%ld ok=%d\n",
+		   i, buf, file, ino, ok);
+	  zsyslog((LOG_INFO, "Got notify: '%s'", buf));
+#endif
+
+	  if (!ok) continue;
+
+#if 0
+	  if (lstat(file,&stbuf) != 0) continue;
+	  if (!S_ISFILE(stbuf.st_mode)) continue;
+
+	  if (in_dirscanqueue(dirq,stbuf.st_ino)) continue;
+#endif
+
+	  /* We may have this file in processing state...  */
+	  {
+	    struct spblk *spl;
+	    spl = sp_lookup(ino, spt_mesh[L_CTLFILE]);
+	    if (spl != NULL) {
+	      /* Already in processing, don't touch.. */
+	      printf("File: '%s' active (not previously locked)\n", buf+4);
+	      continue;
+	    }
+	  }
+
+	  dq_insert(dirq, ino, buf+4, -1);
+	}
+} 
+
+
+
+
 /*
  * The schedule() function is in charge of reading the control file from
  * the scheduler directory, and creating all the appropriate links to the
@@ -1609,9 +1837,6 @@ static struct ctlfile *schedule(fd, file, ino, reread)
 	  return NULL;
 	}
 
-	++MIBMtaEntry->mtaStoredMessages;
-	++global_wrkcnt;
-
 	if (cfp->head == NULL) {
 	  unctlfile(cfp, 0); /* Delete the file.
 				(decrements those counters above!) */
@@ -1649,7 +1874,7 @@ slurp(fd, ino)
 	register char *s;
 	register int i;
 	char *contents;
-	int *offset, *ip, *lp;
+	int *offset;
 	int offsetspace;
 	struct stat stbuf;
 	struct ctlfile *cfp;
@@ -1675,15 +1900,17 @@ slurp(fd, ino)
 	cfp = (struct ctlfile *)emalloc(sizeof(struct ctlfile));
 	memset((void*)cfp, 0, sizeof(struct ctlfile));
 
-	cfp->fd = fd;
-	cfp->dirind = -1; /* Not known -- or top-level */
-	cfp->uid = stbuf.st_uid;
+	cfp->fd       = fd;
+	cfp->dirind   = -1; /* Not known -- or top-level */
+	cfp->uid      = stbuf.st_uid;
 	cfp->envctime = stbuf.st_ctime;
 	cfp->contents = contents;
+
 	/* 
-	   cfp->vfpfn = NULL;
-	   cfp->head = NULL;
-	   cfp->mark = V_NONE;
+	   cfp->vfpfn    = NULL;
+	   cfp->spoolid  = NULL;
+	   cfp->head     = NULL;
+	   cfp->mark     = V_NONE;
 	   cfp->haderror = 0;
 	   cfp->mid      = NULL;
 	   cfp->envid    = NULL;
@@ -1741,13 +1968,22 @@ slurp(fd, ino)
 	cfp = (struct ctlfile *)erealloc((void*)cfp,
 					 (u_int) (sizeof(struct ctlfile) +
 						  i * (sizeof offset[0])));
-	lp = &(cfp->offset[0]);
-	ip = &(offset[0]);
-	/* copy over the offsets */
-	while (--i >= 0)
-	  *lp++ = *ip++;
-	cfp->id = ino;
-	/* cfp->mid = NULL; */
+	if (cfp) {
+	  /* copy over the offsets */
+	  memcpy(&(cfp->offset[0]), offset, sizeof(offset[0]) * i);
+
+	  cfp->id = ino;
+	  /* cfp->mid = NULL; */
+
+	  /* INC there in every case! */
+	  ++global_wrkcnt;
+	  ++MIBMtaEntry->sc.StoredMessagesSc;
+
+	} else {
+	  /* realloc() failed.. */
+	  free(contents);
+	}
+
 	free(offset);	/* release the block */
 	return cfp;
 }
@@ -1858,6 +2094,14 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	offspc = 16;
 	offarr = (struct offsort *)emalloc(offspc * sizeof(struct offsort));
 
+	if (!offarr) {
+	  /* malloc() failure.. */
+	  sfprintf(sfstderr,"malloc failed, discarding job: '%s'\n", file);
+	  cfp_free(cfp, NULL);
+	  return NULL;
+	}
+
+
 	mypid = getpid();
 	opcnt = 0;
 	svn = 0;
@@ -1923,16 +2167,16 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	    if (*cp == _CFTAG_NOTOK) {
 #if 0
 	      ++cfp->rcpnts_failed;
-	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
-	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
+	      ++MIBMtaEntry->sc.ReceivedRecipientsSc,
+	      ++MIBMtaEntry->sc.TransmittedRecipientsSc;
 #endif
 	      prevrcpt = -1;
 	    } else if (*cp != _CFTAG_OK) {
 	      ++cfp->rcpnts_work;
 	    } else { /* _CFTAG_OK */
 #if 0
-	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
-	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
+	      ++MIBMtaEntry->sc.ReceivedRecipientsSc,
+	      ++MIBMtaEntry->sc.TransmittedRecipientsSc;
 #endif
 	    }
 	  }
@@ -1979,6 +2223,12 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 		offarr = (struct offsort *)erealloc(offarr,
 						    sizeof(struct offsort) *
 						    offspc);
+		if (!offarr) {
+		  /* realloc() failure.. */
+		  sfprintf(sfstderr,"realloc() failed, discarding job: '%s'\n", file);
+		  cfp_free(cfp, NULL);
+		  return NULL;
+		}
 	      }
 	      offarr[opcnt].offset = *lp + 2;
 	      strlower(cp);
@@ -2005,10 +2255,6 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      prevrcpt = opcnt;
 	      ++opcnt;
 
-	      /* Account for all yet to be delivered recipients */
-	      ++MIBMtaEntry->mtaStoredRecipients;
-	      ++MIBMtaEntry->mtaReceivedRecipientsSc;
-
 	      break;
 	    case _CF_RCPTNOTARY:
 	      /* IETF-NOTARY-DRPT+NOTIFY DATA! */
@@ -2118,7 +2364,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	    /* We have a desire to use subdirs, now the magic of hashing
 	       into subdirs...  inode number ? */
 
-	    int hash;
+	    int hash, rc;
 
 	    if (hashlevels > 1) {
 	      hash = atol(cfp->mid) % (26*26);
@@ -2129,34 +2375,46 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	    cfp->dirind = hash;
 	    /* Ok, we have the hash values, now move the file
 	       to match with our hashes.. */
-	    sprintf(path, "%s/%s", cfpdirname(hash), fpath);
-	    if (rename(fpath, path) != 0) {
+	    sprintf(path, "%s%s", cfpdirname(hash), fpath);
+	    
+	    while ((rc = rename(fpath, path)) != 0) {
+	      if (errno != EINTR) break;
+	    }
+	    if (rc != 0) {
 	      /* Failed, why ? */
 	      if (errno != ENOENT) {
 		/* For any other than 'no such (target) directory' */
 		cfp->dirind = -1;
 	      } else {
 		cfp_mksubdirs("",cfpdirname(hash));
-		if (rename(fpath, path) != 0)
+		while ((rc = rename(fpath, path)) != 0) {
+		  if (errno != EINTR) break;
+		}
+		if (rc != 0)
 		  cfp->dirind = -1; /* Failed for any reason */
 		else
 		  strcpy(fpath, path);
 	      }
 	    } else
-		strcpy(fpath, path);
+	      strcpy(fpath, path);
 
 	    if (cfp->dirind >= 0) {
 	      /* Successfully renamed the transport file to a subdir,
 		 now do the same to the queue directory! */
 	      sprintf(path,  "../%s/%s", QUEUEDIR, cfp->mid);
-	      sprintf(path2, "../%s/%s/%s",
+	      sprintf(path2, "../%s/%s%s",
 		      QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
 
-	      if (rename(path,path2) != 0) {
+	      while ((rc = rename(path, path2)) != 0) {
+		if (errno != EINTR) break;
+	      }
+	      if (rc != 0) {
 		if (errno == ENOENT) {
 		  /* No dirs ?? */
 		  cfp_mksubdirs(QUEUEDIR,cfpdirname(hash));
-		  rename(path, path2);
+		  while ((rc = rename(path, path2)) != 0) {
+		    if (errno != EINTR) break;
+		  }
 		}
 	      }
 	      /* If failed, it will be reported below */
@@ -2164,10 +2422,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  }
 	}
 
-
 	if (cfp->mid != NULL) {
 	  if (cfp->dirind >= 0)
-	    sprintf(mfpath, "../%s/%s/%s",
+	    sprintf(mfpath, "../%s/%s%s",
 		    QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
 	  else
 	    sprintf(mfpath, "../%s/%s", QUEUEDIR, cfp->mid);
@@ -2191,6 +2448,20 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 					another machine, and run things
 					in there, and still have same
 					expiration times.. */
+
+	/* Reuse the buffer ...
+	   Generate same spoolid string that all other subsystems also
+	   report to syslog. */
+	taspoolid(path2, cfp->mtime, cfp->id);
+	cfp->spoolid = strsave(path2);
+	if (!cfp->spoolid) {
+	  /* malloc() failure.. */
+	  sfprintf(sfstderr,"malloc() failed, discarding job: '%s'\n", file);
+	  cfp_free(cfp, NULL);
+	  free(offarr);
+	  return NULL;
+	}
+
 	cfp->fd = -1;
 	/* sort it to get channels and channel/hosts together */
 	bcp = cfp->contents;
@@ -2258,6 +2529,15 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      u_int alloc_size = (u_int) (sizeof (struct vertex) +
 					  (i - svn - 1) * sizeof (int));
 	      vp = (struct vertex *)emalloc(alloc_size);
+	      if (!vp) {
+		/* malloc() failure.. */
+		sfprintf(sfstderr,"malloc() failed, discarding job: '%s'\n", file);
+		cfp_free(cfp, NULL);
+		free(offarr);
+		return NULL;
+	      }
+	      MIBMtaEntry->sc.StoredVerticesSc += 1;
+
 	      memset((char*)vp, 0, alloc_size);
 	      vp->cfp             = cfp;
 	      vp->next[L_CTLFILE] = NULL;
@@ -2274,6 +2554,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      vp->notary          = NULL;
 #endif
 	      vp->ngroup       = i - svn;
+	      MIBMtaEntry->sc.StoredRecipientsSc   += (i - svn);
+	      MIBMtaEntry->sc.ReceivedRecipientsSc += (i - svn);
+
 	      /* vp->sender       = strsave(offarr[svn].sender); */
 	      vp->wakeup       = offarr[svn].wakeup;
 	      vp->headeroffset = offarr[svn].headeroffset; /*They are similar*/
@@ -2309,6 +2592,15 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  u_int alloc_size = (u_int) (sizeof (struct vertex) +
 				      (i - svn - 1) * sizeof (int));
 	  vp = (struct vertex *)emalloc(alloc_size);
+	  if (!vp) {
+	    /* malloc() failure.. */
+	    sfprintf(sfstderr,"malloc() failed, discarding job: '%s'\n", file);
+	    cfp_free(cfp, NULL);
+	    free(offarr);
+	    return NULL;
+	  }
+	  MIBMtaEntry->sc.StoredVerticesSc += 1;
+
 	  memset((void*)vp, 0, alloc_size);
 	  vp->cfp = cfp;
 	  vp->next[L_CTLFILE] = NULL;
@@ -2323,6 +2615,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  vp->notary       = NULL;
 #endif
 	  vp->ngroup = i - svn;
+	  MIBMtaEntry->sc.StoredRecipientsSc   += (i - svn);
+	  MIBMtaEntry->sc.ReceivedRecipientsSc += (i - svn);
+
 	  /* vp->sender = strsave(offarr[snv].sender); */
 	  vp->wakeup       = offarr[svn].wakeup;
 	  vp->headeroffset = offarr[svn].headeroffset; /* Just any of them will do */
@@ -2367,9 +2662,10 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 
 	{
 	  Sfio_t *vfp = vfp_open(cfp);
-	  if (vfp != NULL && cfp->mid != NULL)
-	    sfprintf(vfp, "scheduler processing %s\n", cfp->mid);
-	  if (vfp) sfclose(vfp);
+	  if (vfp) {
+	    if (cfp->mid) sfprintf(vfp, "scheduler processing %s\n", cfp->mid);
+	    sfclose(vfp);
+	  }
 	}
 
 	return cfp;
@@ -2417,6 +2713,7 @@ static void ce_fillin(thg,cep)
 	if (cep->interval != -1) ce->interval = cep->interval;
 	if (cep->idlemax  != -1) ce->idlemax  = cep->idlemax;
 	if (cep->expiry   != -1) ce->expiry   = cep->expiry;
+	if (cep->expiry2  != -1) ce->expiry2  = cep->expiry2;
 	ce->expiryform     = cep->expiryform;
 	if (cep->uid      != -1) ce->uid = cep->uid;
 	if (cep->gid      != -1) ce->gid = cep->gid;
@@ -2480,8 +2777,9 @@ static void vtxdo(vp, cehdr, path)
 	  }
 	}
 	if (n == 0) {
-	  sfprintf(sfstderr, "%s: no pattern matched %s/%s address\n",
-		   progname, vp->orig[L_CHANNEL]->name,vp->orig[L_HOST]->name);
+	  sfprintf(sfstderr, "%s: no pattern matched %s/%s address %s%s\n",
+		   progname, vp->orig[L_CHANNEL]->name,vp->orig[L_HOST]->name,
+		   path ? "file=":"", path);
 	  /* XX: memory leak here? */
 	  return;
 	}
@@ -2495,6 +2793,14 @@ static void vtxdo(vp, cehdr, path)
 	else
 	  vp->ce_expiry = 0;
 
+	if (tp->expiry2 > 0 && tp->expiry > 0)
+	  vp->ce_expiry2 = tp->expiry2 + vp->ce_expiry;
+	else
+	  vp->ce_expiry2 = /* vp->ce_expiry */ 0;
+
+	if (vp->ce_expiry2 < vp->ce_expiry)
+	  vp->ce_expiry2 = vp->ce_expiry + 24*3600;
+
 	if (tp->reporttimes[0])
 	  vp->nextrprttime = now + tp->reporttimes[0];
 	else
@@ -2576,7 +2882,7 @@ static int globmatch(pattern, string)
 	      int i = 0, c;
 	      while ((c = *p++) != 0) {
 		/* Scan for special chars in pattern.. */
-		if (c == '*' || c == '[' || c == '{' || c == '\\' || c == '?') {
+		if (c == '*' || c == '[' || c == '{' || c == '\\' || c == '?'){
 		  i = 1; /* Found! */
 		  break;
 		}
@@ -2692,65 +2998,80 @@ int sig;
 	alarm(1);
 }
 
+static int    timeserver_pid = 0;
+
+static void timer_health_check __((void))
+{
+  /* XXX: TODO! TIMER HEALTH CHECK */
+  if (timeserver_pid > 0) {
+  }
+  if (running_on_alarm) {
+  }
+}
+
 
 #if defined(HAVE_MMAP)
-struct timeserver {
-	int	pid;
-#ifdef HAVE_SELECT
-	struct timeval tv;
-#else
-	time_t	time_sec;
-#endif
-#define MAPSIZE 16*1024 /* Should work at all systems ? */
-};
+
 struct timeserver *timeserver_segment = NULL;
 static void init_timeserver()
 {
 	int ppid;
 
+	if ( Z_SHM_MIB_is_attached() ) {
+
+	  timeserver_segment = & MIBMtaEntry->ts;
+
+	} else { /* SHARED MIB MEMORY BLOCK IS NOT ATTACHED */
+
+#define MAPSIZE 16*1024 /* Just in case other methods didn't work.. e.g.
+			   that MIB-block.. */
+
 #if !defined(MAP_ANONYMOUS) || defined(__linux__)
-	/* must have a file ? (SunOS 4.1, et.al.) */
-	int fd = -1;
-	char blk[1024];
-	int i;
-	Sfio_t *fp = sftmp(0); /* Create the backing file fairly reliably. */
+	  /* must have a file ? (SunOS 4.1, et.al.) */
+	  int fd = -1;
+	  char blk[1024];
+	  int i;
+	  Sfio_t *fp = sftmp(0); /* Create the backing file fairly reliably. */
 
-	if (fp)
-	  fd = sffileno(fp);
-	if (fd < 0) {
-	  perror("init_timeserver() didn't create tempfile ??");
-	  return; /* Brr! */ 
-	}
-
-	memset(blk,0,sizeof(blk));
-	for (i=0; i < MAPSIZE; i += sizeof(blk))
-	  sfwrite(fp, blk, sizeof(blk));
-	sfsync(fp);
-	sfseek(fp, 0, 0);
+	  if (fp)
+	    fd = sffileno(fp);
+	  if (fd < 0) {
+	    perror("init_timeserver() didn't create tempfile ??");
+	    return; /* Brr! */ 
+	  }
+	  
+	  memset(blk,0,sizeof(blk));
+	  for (i=0; i < MAPSIZE; i += sizeof(blk))
+	    sfwrite(fp, blk, sizeof(blk));
+	  sfsync(fp);
+	  sfseek(fp, 0, 0);
 
 #ifndef MAP_FILE
 # define MAP_FILE 0 /* SunOS 4.1 does not have this */
 #endif
-	timeserver_segment = (void*)mmap(NULL, MAPSIZE, PROT_READ|PROT_WRITE,
-					 MAP_FILE|MAP_SHARED, fd, 0);
-	sfclose(fp);
+	  timeserver_segment = (void*)mmap(NULL, MAPSIZE, PROT_READ|PROT_WRITE,
+					   MAP_FILE|MAP_SHARED, fd, 0);
+	  sfclose(fp);
 	
-#else
+#else /* MAP_ANONYMOUS and not Linux */
+
 #ifndef MAP_VARIABLE
 # define MAP_VARIABLE 0 /* Some system have MAP_ANONYMOUS, but
 			   no MAP_VARIABLE.. */
 #endif
-	/* This MAP_ANONYMOUS does work at DEC OSF/1 ... */
-	timeserver_segment = (void*)mmap(NULL, MAPSIZE, PROT_READ|PROT_WRITE,
-					 MAP_VARIABLE|MAP_ANONYMOUS|MAP_SHARED,
-					 -1, 0);
+	  /* This MAP_ANONYMOUS does work at DEC OSF/1 ... */
+	  timeserver_segment = (void*)mmap(NULL, MAPSIZE, PROT_READ|PROT_WRITE,
+					   MAP_VARIABLE|MAP_ANONYMOUS|MAP_SHARED,
+					   -1, 0);
 #endif
-	if (-1L == (long)timeserver_segment
-	    || timeserver_segment == NULL) {
-	  perror("mmap() of timeserver segment gave");
-	  timeserver_segment = NULL;
-	  return; /* Brr.. */
-	}
+	  if (-1L == (long)timeserver_segment
+	      || timeserver_segment == NULL) {
+	    perror("mmap() of timeserver segment gave");
+	    timeserver_segment = NULL;
+	    return; /* Brr.. */
+	  }
+
+	} /* Have global shared server segment, or not ... */
 
 	ppid = fork();
 	if (ppid > 0) {
@@ -2777,38 +3098,77 @@ static void init_timeserver()
 
 	  return;
 	}
+
 	if (ppid < 0) return; /* Error ?? brr.. */
 
-	ArgvSave[1] = NULL;
-	strcpy((char*)ArgvSave[0],"TimeServer");
 
-	ppid = getppid(); /* who is our parent ? */
+	/* ========  TIME SERVER CHILD ======= */
 
-#ifdef HAVE_SELECT
-	gettimeofday(&timeserver_segment->tv, NULL);
+
+	{
+
+	  int space_check_count = 0;
+
+	  MIBMtaEntry->sc.schedulerTimeserverStartTime = time(NULL);
+	  MIBMtaEntry->sc.schedulerTimeserverStarts ++;
+
+
+#ifdef HAVE_SETPROCTITLE
+	  setproctitle("[TimeServer]");
 #else
-	time(&timeserver_segment->time_sec);
+	  strncpy(ArgvSave,"[TimeServer]", EOArgvSave - ArgvSave);
 #endif
 
-	for(;;) {
+	  if (ArgvSave < EOArgvSave)
+	    ArgvSave[EOArgvSave-ArgvSave-1] = 0;
+
+	  ppid = getppid(); /* who is our parent ? */
+
 #ifdef HAVE_SELECT
-	  struct timeval tv;
-	  int rc;
-
-	  tv.tv_sec = 0;
-	  tv.tv_usec = 300000;
-
 	  gettimeofday(&timeserver_segment->tv, NULL);
-
-	  rc = select(0,NULL,NULL,NULL,&tv);
 #else
 	  time(&timeserver_segment->time_sec);
-	  sleep(1);
 #endif
-	  /* Is the parent still alive ?? */
-	  if (kill(ppid, 0) != 0)
-	    break; /* No ?? Out! */
-	}
+
+	  for(;;) {
+#ifdef HAVE_SELECT
+	    struct timeval tv;
+	    int rc;
+
+	    tv.tv_sec = 0;
+	    tv.tv_usec = 300000;
+
+	    gettimeofday(&timeserver_segment->tv, NULL);
+
+	    rc = select(0,NULL,NULL,NULL,&tv);
+#else
+	    time(&timeserver_segment->time_sec);
+	    sleep(1);
+#endif
+	    /* Is the parent still alive ?? */
+	    if (kill(ppid, 0) != 0)
+	      break; /* No ?? Out! */
+
+	    /* Now check and fill in the filesystem free space
+	       gauges */
+	    if (--space_check_count  < 0) {
+	      long spc = Z_SHM_FileSysFreeSpace();
+
+	      MIBMtaEntry->sys.SpoolFreeSpace = spc;
+	      /* Log FS space is monitored by the self-timed
+		 log reinit code. */
+
+#ifdef HAVE_SELECT
+	      space_check_count = 100; /* ticks about 3 times a second */
+#else
+	      space_check_count =  30; /* ticks once a second.. */
+#endif
+	    }
+
+	  }
+
+	} /* time-server child block */
+
 	_exit(0);
 }
 #else
@@ -2822,6 +3182,12 @@ static void init_timeserver()
 time_t mytime(timep)
 time_t *timep;
 {
+	static int healthcheck = 10000;
+
+	if (--healthcheck <= 0) {
+	  healthcheck = 10000;
+	  timer_health_check();
+	}
 #if defined(HAVE_MMAP)
 	if (timeserver_pid) {
 	  time_t t;
@@ -2860,11 +3226,11 @@ int hash;
 	static char dirbuf[8];
 
 	if (hash < 0)
-	  return ".";
+	  return "";
 	if (hash < 256) {
-	  sprintf(dirbuf, "%c", hash);
+	  sprintf(dirbuf, "%c/", hash);
 	} else {
-	  sprintf(dirbuf, "%c/%c", (hash >> 8) & 0xff, hash & 0xff);
+	  sprintf(dirbuf, "%c/%c/", (hash >> 8) & 0xff, hash & 0xff);
 	}
 	return dirbuf;
 }
@@ -2878,18 +3244,22 @@ const char *topdir, *newpath;
 {
 	char path[256];
 	int omask = umask(022);
+	char *s;
 
 	if (*topdir != 0)
 	  sprintf(path, "../%s/%s", topdir, newpath);
 	else
 	  strcpy(path, newpath);
 
-	if (mkdir(path,0755) != 0) {
+	s = strrchr(path,'/');
+	if (s[1] == 0) *s = 0; /* Trailing slash elimination */
+
+	if (mkdir(path,02755) != 0) {
 	  char *s = strrchr(path,'/');
 	  if (s) *s = 0;
-	  mkdir(path,0755);
+	  mkdir(path,02755);
 	  if (s) *s = '/';
-	  mkdir(path,0755);
+	  mkdir(path,02755);
 	}
 	umask(omask);
 }
