@@ -5,7 +5,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2002
  */
 
 #define USE_SIGREAPER /* DO Use SIGCLD-driven reaper.. */
@@ -29,6 +29,7 @@ struct config_entry {
 	char	*host;		/* host part of pattern			     */
 	time_t	interval;	/* how often to start up these things	     */
 	time_t	expiry;		/* bounce the message after this long in q's */
+	time_t	expiry2;	/* bounce the message after this long in q's */
 	char	*expiryform;	/* use this form as the error message	     */
 			/* Following two need to be SIGNED variables	     */
 	int	uid;		/* what uid to run the transport agent under */
@@ -81,6 +82,7 @@ struct ctlfile {
 	uid_t	uid;		/* the owner of the control file (= msg file)*/
 	time_t	mtime;		/* original msg file's mtime (~ arrival)     */
 	time_t	envctime;	/* when the transport file was created	     */
+	char*   spoolid;	/* buffer for msg file spoolid storage	     */
 	int	haderror;	/* some errors/diagnostics need processing   */
 	int	resynccount;	/* How many resync requests ?		     */
 #define MAXRESYNCS 30
@@ -101,6 +103,9 @@ struct ctlfile {
 	int	rcpnts_work;	/* .. yet to deliver ?			     */
 	int	mark;		/* flag used by selector() to pass filenames */
 	int	msgbodyoffset;	/* size of original headers to skip on errrpt*/
+	int	msgbodysize;	/* header size + body size, in kB	     */
+	int	msgheadsize;	/* header size (from within transport file)  */
+	int	msgfilesizekb;	/* Sum of both, round up to nearest kB, div  */
 	int	format;		/* Message format version -- _CF_FORMAT data */
 	int	offset[1];	/* array of nlines byte offsets into the file*/
 };
@@ -229,6 +234,7 @@ struct vertex {
 #define NOT_TRACE   020 /* RFC 2852 */
 	int		ce_pending;	/* pending on what ?		     */
 	time_t		ce_expiry;	/* when this vertex expires ?        */
+	time_t		ce_expiry2;	/* when this vertex expires ? w/o attempts */
 	int		attempts;	/* count of number of TA invocations */
 	int		retryindex;	/* cur index into ce->retries array  */
 	time_t		wakeup;		/* time to wake up and run this      */
@@ -277,7 +283,7 @@ struct mailq {
 	Usockaddr	qaddr;
 #endif /* HAVE_SOCKET */
 
-	time_t		apopteosis;
+	time_t		apoptosis;
 
 	char		*challenge;
 
@@ -303,3 +309,4 @@ struct mailq {
 #define MQ2MODE_FULL2	0x0008
 #define MQ2MODE_ETRN	0x0010
 #define MQ2MODE_KILL	0x0020
+#define MQ2MODE_REROUTE 0x0040
