@@ -1,7 +1,7 @@
 /*
  *	ZMailer 2.99.16+ Scheduler "threads" routines
  *
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-2000
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-2003
  *
  *
  *	These "threads" are for book-keeping of information
@@ -168,7 +168,8 @@ struct thread *thr;
 	thr->prevthg->nextthg = thr->nextthg;
 	thr->nextthg->prevthg = thr->prevthg;
 
-	thg->threads -= 1;
+	thg->threads                      -= 1;
+	MIBMtaEntry->sc.StoredThreadsSc -= 1;
 
 	if (thg->thread == thr)
 	  thg->thread = thr->nextthg;	/* pick other */
@@ -229,6 +230,7 @@ struct thread *thr;
 	}
 
 	thg->threads += 1;
+	MIBMtaEntry->sc.StoredThreadsSc += 1;
 }
 
 
@@ -274,6 +276,7 @@ struct config_entry *cep;
 	thr->thvertices   = vtx;
 	thr->lastthvertex = vtx;
 	thr->jobs       = 1;
+
 	vtx->thread     = thr;
 	thr->channel    = strsave(vtx->orig[L_CHANNEL]->name);
 	thr->host       = strsave(vtx->orig[L_HOST   ]->name);
@@ -643,19 +646,18 @@ memset(wp, 0x55, sizeof(*wp));
 /*
  * unthread(vtx) -- detach this vertex from its thread
  */
-static void unthread __((struct vertex *vtx));
-static void unthread(vtx)
-struct vertex *vtx;
+static void unthread __((struct thread *thr, struct vertex *vtx));
+static void unthread(thr, vtx)
+     struct thread *thr;
+     struct vertex *vtx;
 {
-	struct thread *thr = vtx->thread;
-
 	if (vtx->previtem != NULL)
 	  vtx->previtem->nextitem = vtx->nextitem;
 	if (vtx->nextitem != NULL)
 	  vtx->nextitem->previtem = vtx->previtem;
 
 	if (thr) {
-	  thr->jobs -= 1;
+	  thr->jobs                        -= 1;
 
 	  if (thr->nextfeed     == vtx)
 	    thr->nextfeed       = thr->nextfeed->nextitem;
@@ -684,10 +686,9 @@ web_detangle(vp, ok)
 
 	struct thread *thr = vp->thread;
 
-	if (thr && thr->nextfeed == vp)
-	  thr->nextfeed = vp->nextitem; /* Oh, next one then.. */
+	/* unthread() will also unpick the  nextfeed link.. */
 
-	unthread(vp);
+	unthread(thr, vp);
 
 	/* The thread can now be EMPTY! */
 
@@ -853,6 +854,7 @@ thread_start(thr, queueonly_too)
 	    proc = proc->pnext;
 	    while (proc) {
 	      sfprintf(sfstderr, " %p", proc);
+	      proc = proc->pnext;
 	    }
 	    sfprintf(sfstderr, " }\n");
 	  }
@@ -890,7 +892,13 @@ thread_start(thr, queueonly_too)
 	  proc->pnext = proc->pprev   = NULL;
 
 	  thg->idlecnt -= 1;
+
 	  --idleprocs;
+
+	  /* Move to ACTIVE state */
+	  MIBMtaEntry->sc.TransportAgentsActiveSc += 1;
+	  MIBMtaEntry->sc.TransportAgentsIdleSc   -= 1;
+
 	  
 	  /* It may be that while we idled it, it died at the idle queue.. */
 	  if (proc->pid <= 0 || proc->tofd < 0) {
@@ -1068,19 +1076,72 @@ pick_next_vertex(proc)
 	return (vtx != NULL);
 }
 
+/*
+ *  The  thread_expire2()  will handle exceedingly old things
+ *  with ages in excess of expire+expire2 (seconds) in queue
+ *  even if no successfull delivery attempt has been made.
+ *
+ *  Return the kill-count.
+ *
+ *  Side-effect warning: 
+ *         Afterwards the THR may point to nonexistent object!
+ */
+
+
+int
+thread_expire2(thr, timelimit, killall, msgstr)
+     struct thread *thr;
+     time_t timelimit;
+     int killall;	  /* later uses in mind.. now dummy parameter */
+     const char *msgstr;  /* ... likewise. */
+{
+	int killcount = 0;
+	struct vertex       *vtx  = thr->thvertices;
+	struct vertex	*nextvtx;
+
+	for ( ;vtx; vtx = nextvtx) {
+	  int expire_this = 0;
+
+	  nextvtx = vtx->nextitem;
+
+	  /* Time to expire ? */
+	  if (vtx->ce_expiry > 0 && vtx->ce_expiry <= now &&
+	      vtx->attempts  > 0) {
+	    expire_this = 1;
+	  }
+	  if (vtx->ce_expiry2 > 0 && vtx->ce_expiry2 <= now) {
+	    expire_this = 1;
+	  }
+	  if (expire_this) {
+	    /* ... and now expire it! */
+	    /* this MAY invalidate also the THREAD object! */
+
+	    expire(vtx, -1); /* ... them all. */
+	    ++killcount;
+
+	    mytime(&now);
+	    if (now > timelimit) break;
+	  }
+	}
+
+	return killcount;
+}
+
+
 /*
  * The  thread_reschedule()  updates threads time-chain to match the
- * new value of wakeup for the  doagenda()  to latter use.
+ * new value of wakeup for the  doagenda()  to later use.
  * Return 0 for DESTROYED thread, 1 for EXISTING thread.
  */
 
 int
 thread_reschedule(thr, retrytime, index)
-struct thread *thr;
-int index;
-time_t retrytime;
+     struct thread *thr;
+     int index;
+     time_t retrytime;
 {
 	struct vertex *vtx = thr->thvertices;
+	struct vertex *nvtx;
 	time_t wakeup = 0;
 	int skew;
 
@@ -1095,7 +1156,7 @@ time_t retrytime;
 	}
 
 	/* If there are multiple kids working still, DON'T reschedule! */
-	if (thr->thrkids > 0 || !vtx) return 1;
+	if ((thr->thrkids > 0)  ||  (vtx == NULL)) return 1;
 
 	/* find out when to retry */
 	mytime(&now);
@@ -1162,33 +1223,40 @@ time_t retrytime;
 	  retrytime = wakeup;
 
 	/* Reschedule ALL vertices on this thread */
-	while (vtx) {
+	for ( ;vtx; vtx = nvtx) {
+	  int expire_this = 0;
+
+	  nvtx = vtx->nextitem;
 
 	  /* Time to expire ? */
 	  if (vtx->ce_expiry > 0 && vtx->ce_expiry <= now &&
 	      vtx->attempts  > 0) {
-	    struct vertex *nvtx = vtx->nextitem;
-
+	    expire_this = 1;
+	  }
+	  if (vtx->ce_expiry2 > 0 && vtx->ce_expiry2 <= now) {
+	    expire_this = 1;
+	  }
+	  if (expire_this) {
 	    /* ... and now expire it! */
 	    /* this MAY invalidate also the THREAD object! */
 
 	    if (thr->jobs > 1) {
-	      expire(vtx,index);
+	      expire(vtx, index);
 	    } else {
-	      expire(vtx,index);
+	      expire(vtx, index);
 	      thr = NULL; /* The THR-pointed object is now invalid */
 	    }
-	    vtx = nvtx;
 	    continue;
 	  }
 
+	  /* Didn't expire, so time to tune the wakeup ... */
+
 	  if (vtx->wakeup < retrytime)
 	    vtx->wakeup = retrytime;
 	  if (wakeup > vtx->wakeup || wakeup == 0)
 	    wakeup = vtx->wakeup;
-
-	  vtx = vtx->nextitem;
 	}
+
 	if (thr != NULL)
 	  thr->wakeup = wakeup;
 
@@ -1371,7 +1439,7 @@ idle_cleanup()
 
 		zsyslog((LOG_ERR, "ZMailer scheduler kludge shutdown of TA channel (info for debug only); %s/%s/%d HA=%ds",
 			 thr->channel, thr->host, thr->thgrp->withhost,
-			 now - p->hungertime));
+			 (int)(now - p->hungertime)));
 	      }
 	    }
 	  }
@@ -1657,6 +1725,7 @@ void thread_report(fp,mqmode)
 	}
 
 	if (mqmode & (MQ2MODE_FULL | MQ2MODE_QQ | MQ2MODE_SNMP)) {
+	  long files;
 	  *timebuf = 0;
 	  saytime((long)(now - sched_starttime), timebuf, 1);
 	  sfprintf(fp,"Kids: %d  Idle: %2d  Msgs: %3d  Thrds: %3d  Rcpnts: %4d  Uptime: ",
@@ -1666,17 +1735,21 @@ void thread_report(fp,mqmode)
 	  else
 	    sfprintf(fp, "%s\n",timebuf);
 
-	  sfprintf(fp, "Msgs in %lu out %lu stored %lu ",
-		   (u_long)MIBMtaEntry->mtaReceivedMessagesSc,
-		   (u_long)MIBMtaEntry->mtaTransmittedMessagesSc,
-		   (u_long)MIBMtaEntry->mtaStoredMessages);
+	  sfprintf(fp, "Msgs in %lu out %lu stored %ld ",
+		   (u_long)MIBMtaEntry->sc.ReceivedMessagesSc,
+		   (u_long)MIBMtaEntry->sc.TransmittedMessagesSc,
+		   (long)MIBMtaEntry->sc.StoredMessagesSc);
 
-	  sfprintf(fp, "Rcpnts in %lu out %lu stored %lu",
-		   (u_long)MIBMtaEntry->mtaReceivedRecipientsSc,
-		   (u_long)MIBMtaEntry->mtaTransmittedRecipientsSc,
-		   (u_long)MIBMtaEntry->mtaStoredRecipients);
+	  files = thread_count_files();
+	  if ((long)MIBMtaEntry->sc.StoredMessagesSc != files)
+	    sfprintf(fp, "(%ld) ", files);
 
-	  if (rcptsum != MIBMtaEntry->mtaStoredRecipients)
+	  sfprintf(fp, "Rcpnts in %lu out %lu stored %ld",
+		   (u_long)MIBMtaEntry->sc.ReceivedRecipientsSc,
+		   (u_long)MIBMtaEntry->sc.TransmittedRecipientsSc,
+		   (long)MIBMtaEntry->sc.StoredRecipientsSc);
+
+	  if (rcptsum != MIBMtaEntry->sc.StoredRecipientsSc)
 	    sfprintf(fp, " (%d)", rcptsum);
 
 	  sfprintf(fp, "\n");
@@ -1727,10 +1800,10 @@ void thread_detail_report(fp,mqmode,channel,host)
 	    struct ctlfile *cfp = vp->cfp;
 	    for (i = 0; i < vp->ngroup; ++i) {
 	      /* Spoolfile */
-	      if (vp->cfp->dirind >= 0)
-		sfprintf(fp, "%s/%s", cfpdirname(cfp->dirind), cfp->mid);
-	      else
-		sfprintf(fp, "%s", cfp->mid);
+	      sfprintf(fp, "%s%s", cfpdirname(cfp->dirind), cfp->mid);
+
+	      /* How manyth in a group ? */
+	      sfprintf(fp, "\t%d", i);
 	      /* Sender index -- or sender address */
 	      sfprintf(fp, "\t%s", cfp->erroraddr);
 	      /* Recipient offset */
@@ -1803,3 +1876,22 @@ int thread_count_recipients()
 	  }
 	return jobtotal;
 }
+
+
+static int thread_files_count;
+
+static int spl_thread_cnt_files __((struct spblk *spl));
+static int spl_thread_cnt_files(spl)
+struct spblk *spl;
+{
+	++thread_files_count;
+	return 0;
+}
+
+
+int thread_count_files __((void))
+{
+	thread_files_count = 0;
+	sp_scan(spl_thread_cnt_files, NULL, spt_mesh[L_CTLFILE]);
+	return  thread_files_count;
+}
