@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 
@@ -38,6 +38,8 @@
 #include "ta.h"
 #include "libz.h"
 
+#include "libc.h"
+
 extern int forkrate_limit;
 extern int freeze;
 extern int mailqmode;
@@ -59,18 +61,10 @@ extern void *dirq;
 
 extern FILE *vfp_open __((struct ctlfile *));
 
-#ifdef  HAVE_WAITPID
+#ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
 # include <sys/wait.h>
-#else
-# ifdef HAVE_WAIT3
-#  include <sys/wait.h> /* Has BSD wait3() */
-# else
-#  ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
-#   include <sys/wait.h>
-#  else /* Not POSIX.1 compatible, lets fake it.. */
+#else /* Not POSIX.1 compatible, lets fake it.. */
 extern int wait();
-#  endif
-# endif
 #endif
 
 #ifndef WEXITSTATUS
@@ -84,9 +78,12 @@ struct procinfo *cpids = NULL;
 
 #define	MAXFILESPERTRANSPORT	1000
 
-int	numkids = 0;
-int	readsockcnt = 0; /* Count how many childs to read there are;
-			    this for the SLOW Shutdown */
+int	numkids;
+int	readsockcnt;		/* Count how many childs to read there are;
+				   this for the SLOW Shutdown */
+
+int	notifysocket = -1;	/* fd of UDP/AF_UNIX socket to listen for notifies */
+
 
 static void cmdbufalloc __((int, char **, int *));
 
@@ -121,6 +118,7 @@ static int
 flush_child(proc)
      struct procinfo *proc;
 {
+
 	if (proc->pid < 0 && proc->tofd >= 0) {
 	  pipes_shutdown_child(proc->tofd);
 	  proc->tofd = -1;
@@ -182,7 +180,7 @@ flush_child(proc)
 	    proc->feedtime = now;
 	  } else
 	    break; /* Zero or negative.. let the writing
-		      proceed latter... */
+		      proceed later... */
 	  proc->cmdlen -= rc;
 	}
 	if (proc->cmdlen) return 1; /* Incomplete write */
@@ -254,29 +252,19 @@ feed_child(proc)
 	  cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
 	  strcpy(cmdbuf,"\n");
 
-	} else if (vtx->cfp->dirind > 0) {
+	} else {
 
 	  const char *d = cfpdirname(vtx->cfp->dirind);
 	  if (proc->thg->withhost) { /* cmd-line was with host */
-	    cmdlen = 2 + strlen(d) + strlen(vtx->cfp->mid);
+	    cmdlen = 1 + strlen(d) + strlen(vtx->cfp->mid);
 	    cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
-	    sprintf(cmdbuf, "%s/%s\n", d, vtx->cfp->mid);
+	    sprintf(cmdbuf, "%s%s\n", d, vtx->cfp->mid);
 	  } else {
-	    cmdlen = 3+strlen(d)+strlen(vtx->cfp->mid)+strlen(proc->ho->name);
+	    cmdlen = 2+strlen(d)+strlen(vtx->cfp->mid)+strlen(proc->ho->name);
 	    cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
-	    sprintf(cmdbuf, "%s/%s\t%s\n", d, vtx->cfp->mid, proc->ho->name);
+	    sprintf(cmdbuf, "%s%s\t%s\n", d, vtx->cfp->mid, proc->ho->name);
 	  }
-	} else {
 
-	  if (proc->thg->withhost) { /* cmd-line was with host */
-	    cmdlen = 1 + strlen(vtx->cfp->mid);
-	    cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
-	    sprintf(cmdbuf, "%s\n", vtx->cfp->mid);
-	  } else {
-	    cmdlen = 2+strlen(vtx->cfp->mid)+strlen(proc->ho->name);
-	    cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
-	    sprintf(cmdbuf, "%s\t%s\n", vtx->cfp->mid, proc->ho->name);
-	  }
 	}
 
 	if ((proc->cmdlen + cmdlen) >= proc->cmdspc)
@@ -423,7 +411,7 @@ ta_hungry(proc)
 
 	    }
 	    /* As long as we have things to feed (or have fed anything!),
-	       we stay in the STUFFING state.  This was if we latter get
+	       we stay in the STUFFING state.  This was if we later get
 	       new workitems,  thread_linkin() can simply put them into
 	       the 'nextfeed' pointer, and we feed them right away. */
 	    return;
@@ -444,7 +432,9 @@ ta_hungry(proc)
 	     This TA may have other things to poke at! */
 
 	  thr0 = proc->pthread;
-	  /* proc->pthread = NULL; */ /* NO KILL YET! */
+
+	  /* proc->pthread = NULL; */
+	  /* ^^^  NO KILL YET! pick_next_thread() needs this! */
 
 	  /* Disconnect the previous thread from the proc. */
 
@@ -554,8 +544,12 @@ ta_hungry(proc)
 	  proc->thg->idleproc = proc;
 
 	  proc->thg->idlecnt += 1;
+
 	  ++idleprocs;
-	  
+	  MIBMtaEntry->sc.TransportAgentsActiveSc -= 1;
+	  MIBMtaEntry->sc.TransportAgentsIdleSc   += 1;
+
+
 	  /* Failed to pick anything, reschedule the old thread.
 	     Possibly killed the thread.. */
 	  if (thr0) thread_reschedule(thr0, 0, -1);
@@ -595,7 +589,9 @@ start_child(vhead, chwp, howp)
 	struct web *chwp, *howp;
 {
 #define MAXARGC 40
-	char *av[1+MAXARGC], *ev[1+MAXARGC], *s, *os, *cp, *ocp;
+	char * av[1+MAXARGC];
+	char * ev[1+MAXARGC];
+	char *os, *cp, *ocp, *s;
 	char buf[MAXPATHLEN*4];
 	char buf2[MAXPATHLEN];
 
@@ -659,7 +655,7 @@ start_child(vhead, chwp, howp)
 		  } else if (strcmp(ocp,"channel")==0) {
 		    strcpy(s,chwp->name);
 		  } else {
-		    char *t = getzenv(ocp);
+		    const char *t = getzenv(ocp);
 		    if (t)
 		      strcpy(s, t);
 		  }
@@ -688,7 +684,7 @@ start_child(vhead, chwp, howp)
 	  } else if (avi == 0 && av[0][0] != '/') {
 	    /* Must add ${MAILBIN}/ta/ to be the prefix.. */
 
-	    static char *mailbin = NULL;
+	    static const char *mailbin = NULL;
 
 	    if (!mailbin) mailbin = getzenv("MAILBIN");
 	    if (!mailbin) mailbin = MAILBIN;
@@ -707,9 +703,15 @@ start_child(vhead, chwp, howp)
 	  if (evi >= MAXARGC) evi = MAXARGC;
 	}
 	av[avi] = NULL;
-	if ((s = getenv("TZ")))       ev[evi++] = s-3; /* Pass the TZ      */
-	if ((s = getzenv("PATH")))    ev[evi++] = s-5; /* Pass the PATH    */
-	if ((s = getzenv("ZCONFIG"))) ev[evi++] = s-8; /* Pass the ZCONFIG */
+	{
+	  const char *t;
+	  if ((t = getenv("TZ")))
+	    ev[evi++] = (char*) t-3; /* Pass the TZ      */
+	  if ((t = getzenv("PATH")))
+	    ev[evi++] = (char*) t-5; /* Pass the PATH    */
+	  if ((t = getzenv("ZCONFIG")))
+	    ev[evi++] = (char*) t-8; /* Pass the ZCONFIG */
+	}
 	ev[evi] = NULL;
 
 	/* fork off the appropriate command with the appropriate stdin */
@@ -764,7 +766,7 @@ static int runcommand(argv, env, vhead, chwp, howp)
 	  for (i = 3; i < scheduler_nofiles; ++i)
 	    close(i);
 
-#ifdef HAVE_SETPRIORITY
+#if defined(HAVE_SETPRIORITY) && defined(HAVE_SYS_RESOURCE_H)
 	  if (prio >= 80) { /* MAGIC LIMIT VALUE FOR ABSOLUTE SET! */
 	    setpriority(PRIO_PROCESS, 0, i - 100);
 	  } else
@@ -846,7 +848,13 @@ static void stashprocess(pid, fromfd, tofd, chwp, howp, vhead, argv)
 	proc->thg           = vhead->thread->thgrp;
 	proc->thg->transporters += 1;
 	proc->pthread->thrkids  += 1;
+
 	++numkids;
+	MIBMtaEntry->sc.TransportAgentProcessesSc += 1;
+	MIBMtaEntry->sc.TransportAgentForksSc     += 1;
+	MIBMtaEntry->sc.TransportAgentsActiveSc   += 1;
+
+
 	proc->tofd          = tofd;
 	proc->pnext         = proc->pthread->proc;
 	proc->pthread->proc = proc;
@@ -918,10 +926,10 @@ static void reclaim(fromfd, tofd)
 {
 	struct procinfo *proc = &cpids[fromfd];
 
-if (verbose)
-  sfprintf(sfstderr,"reclaim(%d,%d) pid=%d, reaped=%d, chan=%s, host=%s\n",
-	   fromfd, tofd, (int)proc->pid, proc->reaped,
-	   proc->ch->name, proc->ho->name);
+	if (verbose)
+	  sfprintf(sfstderr,"reclaim(%d,%d) pid=%d, reaped=%d, chan=%s, host=%s\n",
+		   fromfd, tofd, (int)proc->pid, proc->reaped,
+		   proc->ch->name, proc->ho->name);
 
 	if (proc->reaped &&
 	    ((WIFEXITED(proc->waitstat) && WEXITSTATUS(proc->waitstat)) ||
@@ -933,6 +941,7 @@ if (verbose)
 		   "reclaim(%d,%d) pid=%d, reaped=%d, chan=%s, host=%s ",
 		   fromfd, tofd, (int)proc->pid, proc->reaped,
 		   proc->ch->name, proc->ho->name);
+
 	  if (WIFEXITED(proc->waitstat))
 	    sfprintf(sfstderr,"EXIT=%d\n", WEXITSTATUS(proc->waitstat));
 	  else if (WIFSIGNALED(proc->waitstat))
@@ -992,6 +1001,10 @@ if (verbose)
 	    proc->thg->idlecnt -= 1;
 	    --idleprocs;
 
+	    /* Virtually move to ACTIVE state for count off below */
+	    MIBMtaEntry->sc.TransportAgentsActiveSc += 1;
+	    MIBMtaEntry->sc.TransportAgentsIdleSc   -= 1;
+
 	    /* Remove this entry from the chains */
 	    if (proc == proc->thg->idleproc)
 	      proc->thg->idleproc = proc->pnext;
@@ -1009,6 +1022,9 @@ if (verbose)
 	  delete_threadgroup(proc->thg);
 	  proc->thg    = NULL;
 	}
+
+	MIBMtaEntry->sc.TransportAgentsActiveSc   -= 1;
+	MIBMtaEntry->sc.TransportAgentProcessesSc -= 1;
 	--numkids;
 	proc->thg    = NULL;
 }
@@ -1081,6 +1097,8 @@ time_t timeout;
 	struct timeval tv;
 	struct procinfo *proc = cpids;
 
+	timed_log_reinit();
+
 	if (in_select) {
 	  sfprintf(sfstderr,"**** recursed into mux()! ***\n");
 	  return 0;
@@ -1120,6 +1138,11 @@ time_t timeout;
 	  if (maxf < querysocket)
 	    maxf = querysocket;
 	}
+	if (notifysocket >= 0) {
+	  _Z_FD_SET(notifysocket, rdmask);
+	  if (maxf < notifysocket)
+	    maxf = notifysocket;
+	}
 
 	/* Although we don't react on the results of these MQ2 fd's,
 	   getting them to break timeouts is important for MAILQv2
@@ -1132,14 +1155,18 @@ time_t timeout;
 	  return -1;
 
 	++maxf;
-	/*sfprintf(sfstderr, "about to select on %x [%d]\n",
-			  mask.fds_bits[0], maxf); */
+	/* sfprintf(sfstderr, "about to select on %x [%d]\n",
+	   mask.fds_bits[0], maxf); */
 
 	in_select = 1;
 
 	n = select(maxf, &rdmask, &wrmask, NULL, &tv);
+
 	if (n < 0) {
 	  int err = errno;
+
+	  timed_log_reinit();
+
 	  /* sfprintf(sfstderr, "got an interrupt (%d)\n", errno); */
 	  in_select = 0;
 	  if (errno == EINTR || errno == EAGAIN)
@@ -1162,19 +1189,29 @@ time_t timeout;
 	} else if (n == 0) {
 	  /* sfprintf(sfstderr, "abnormal 0 return from select!\n"); */
 	  /* -- just a timeout -- fast or long */
+	  timed_log_reinit();
 	  in_select = 0;
 	  return 1;
 	} else {
 	  /*sfprintf(sfstderr, "got %d ready (%x)\n", n, rdmask.fds_bits[0]);*/
 
 	  /* In case we really should react.. */
-	  queryipccheck();
+	  if (querysocket >= 0 &&
+	      _Z_FD_ISSET(querysocket, rdmask))
+	    queryipccheck();
+	  
+	  if (notifysocket >= 0 &&
+	      _Z_FD_ISSET(notifysocket, rdmask))
+	    receive_notify(notifysocket);
 
 	  if (cpids != NULL) {
 
 	    for (proc = cpids, i = 0; i < maxf; ++i, ++proc) {
+	      
+	      timed_log_reinit();
 
-	      if (proc->pid != 0 && _Z_FD_ISSET(i, rdmask)) {
+	      if (proc->pid < 0 ||
+		  (proc->pid > 0 && _Z_FD_ISSET(i, rdmask))) {
 		_Z_FD_CLR(i, rdmask);
 		/*sfprintf(sfstderr,"that is fd %d\n",i);*/
 		/* do non-blocking reads from this fd */
@@ -1213,7 +1250,8 @@ queryipccheck()
 	struct timeval tv;
 	int maxfd;
 
-	mytime(&now);
+	timed_log_reinit(); /* internal: mytime(&now); */
+
 	if (!mq2_active() && (lastqueryipccheck == now)) return;
 	lastqueryipccheck = now;
 
@@ -1226,7 +1264,7 @@ queryipccheck()
 	   */
 	}
 
-	if (querysocket >= 0) {
+	if (querysocket >= 0 || notifysocket >= 0) {
 
 	  maxfd = querysocket;
 
@@ -1235,7 +1273,13 @@ queryipccheck()
 
 	  _Z_FD_ZERO(rdmask);
 	  _Z_FD_ZERO(wrmask);
-	  _Z_FD_SET(querysocket, rdmask);
+	  if (querysocket >= 0)
+	    _Z_FD_SET(querysocket, rdmask);
+	  if (notifysocket >= 0) {
+	    _Z_FD_SET(notifysocket, rdmask);
+	    if (notifysocket > maxfd)
+	      maxfd = notifysocket;
+	  }
 
 	  if (mailqmode == 2) {
 	    maxfd = mq2add_to_mask(&rdmask, &wrmask, maxfd);
@@ -1245,8 +1289,12 @@ queryipccheck()
 	  } else 
 	    n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
 
+	  if (n > 0 && notifysocket >= 0 &&
+	      _Z_FD_ISSET(notifysocket, rdmask)) {
+	    receive_notify(notifysocket);
+	  }
 
-	  if (n > 0 &&
+	  if (n > 0 && querysocket >= 0 &&
 	      _Z_FD_ISSET(querysocket, rdmask)) {
 	    Usockaddr raddr;
 	    int raddrlen = sizeof(raddr);
@@ -1254,29 +1302,59 @@ queryipccheck()
 	    n = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
 	    if (n >= 0) {
 	      if (mailqmode == 1) {
-		int pid = fork();
+		int pid;
+
+		MIBMtaEntry->sc.MQ1sockConnects ++;
+		MIBMtaEntry->sc.MQ1sockParallel ++;
+
+		pid = fork();
 		if (pid == 0) {
 #if defined(F_SETFD)
 		  fcntl(n, F_SETFD, 1); /* close-on-exec */
 #endif
+#ifdef USE_TCPWRAPPER
 #ifdef HAVE_TCPD_H /* TCP-Wrapper code */
 		  if (wantconn(n, "mailq") == 0) {
-		    char *msg = "refusing 'mailq' query from your whereabouts\r\n";
+		    char *msg = "500 TCP-WRAPPER refusing 'mailq' query from your whereabouts\r\n";
 		    int   len = strlen(msg);
 		    write(n,msg,len);
+		    MIBMtaEntry->sc.MQ1sockParallel --;
+		    MIBMtaEntry->sc.MQ1sockTcpWrapRej ++;
 		    _exit(0);
 		  }
+#endif
 #endif
 		  qprint(n);
+		  close(n);
+		  MIBMtaEntry->sc.MQ1sockParallel --;
 		  /* Silence memory debuggers about this child's
 		     activities by doing exec() on the process.. */
 		  /* execl("/bin/false","false",NULL); */
 		  _exit(0); /* _exit() should be silent, too.. */
 		}
+		if (pid < 0)
+		  MIBMtaEntry->sc.MQ1sockParallel --;
 		close(n);
 	      } else {
 		/* mailqmode == 2 */
-		mq2_register(n, &raddr);
+
+		MIBMtaEntry->sc.MQ2sockConnects ++;
+
+#if 0  /* NOT IN MAILQ-V2 MODE ! */
+#ifdef USE_TCPWRAPPER
+#ifdef HAVE_TCPD_H /* TCP-Wrapper code */
+		if (wantconn(n, "mailq") == 0) {
+		  char *msg = "500 TCP-WRAPPER refusing 'mailq' query from your whereabouts\r\n";
+		  int   len = strlen(msg);
+		  write(n,msg,len);
+		  MIBMtaEntry->sc.MQ2sockTcpWrapRej ++;
+		  close(n);
+		}
+		else
+#endif
+#endif
+#endif
+		  mq2_register(n, &raddr);
 	      }
 	    }
 	  }
@@ -1289,136 +1367,246 @@ queryipcinit()
 	int modecode = 1; /* Modes: 1=TCP, 2=UNIX, default=TCP */
 	char *modedata = NULL;
 
-	if (querysocket >= 0)
-		return;
+	while (notifysocket < 0) {
 
-	if (mailqsock) {
-	  if (cistrncmp(mailqsock,"UNIX:",5)==0) {
-	    modedata = mailqsock+5;
-	    modecode = 2;
-	  } else if (*mailqsock == '/') {
-	    /* If it begins with '/', it is AF_UNIX socket */
-	    modedata = mailqsock;
-	    modecode = 2;
-	  } else if (cistrncmp(mailqsock,"TCP:",4)==0) {
-	    modedata = mailqsock+4;
-	    modecode = 1;
-	  } else {
-	    /* The default mode is TCP/IP socket */
-	    modedata = mailqsock;
-	    modecode = 1;
-	  }
-	}
-
-#ifdef  AF_UNIX
-	if (modecode == 2) {
-	  struct sockaddr_un sad;
-	  int on = 1;
-
-	  mytime(&now);
-	  qipcretry = now + 5;
-
-	  sad.sun_family = AF_UNIX;
-	  strncpy(sad.sun_path, modedata, sizeof(sad.sun_path));
-	  sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
-
-	  if ((querysocket = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
-	    perror("querysocket: socket(PF_UNIX)");
-	    return;
+	  if (!notifysock) {
+	    notifysock = (char *)getzenv("SCHEDULERNOTIFY");
+	    if (notifysock) {
+	      notifysock = zenvexpand(notifysock);
+	    } else {
+	      notifysock = strsave("UNIX:${POSTOFFICE}/.scheduler.notify");
+	      notifysock = zenvexpand(notifysock);
+	    }
 	  }
 
-	  setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
-
-	  /* In case that one already exists.. */
-	  unlink(sad.sun_path);
-
-	  if (bind(querysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
-	    perror("bind:mailq socket");
-	    close(querysocket);
-	    querysocket = -1;
-	    return;
-	  }
-#if defined(F_SETFD)
-	  fcntl(querysocket, F_SETFD, 1); /* close-on-exec */
-#endif
-
-	  if (listen(querysocket, 5) < 0) {
-	    perror("listen:mailq socket");
-	    close(querysocket);
-	    querysocket = -1;
-	    return;
-	  }
-	  qipcretry = 0;
-	}
-#endif
-#ifdef	AF_INET
-	if (modecode == 1) {
-	  struct servent *serv;
-	  Usockaddr ua;
-	  int on = 1;
-	  int port = 174; /* MAGIC knowledge */
-
-	  mytime(&now);
-	  if (!modedata || !*modedata || sscanf(modedata,"%d",&port) != 1) {
-	    if ((serv = getservbyname(modedata ? modedata : "mailq", "tcp")) == NULL) {
-	      sfprintf(sfstderr, "No 'mailq' tcp service defined!\n");
-	    } else
-	      port = ntohs(serv->s_port);
-	  }
-	  qipcretry = now + 5;
-	  memset(&ua, 0, sizeof(ua));
-	  ua.v4.sin_port        = htons(port);
-	  ua.v4.sin_family      = AF_INET;
-	  ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
-	  querysocket = -1;
-#ifdef INET6
-	  querysocket = socket(PF_INET6, SOCK_STREAM, 0);
-#endif
-	  if (querysocket < 0)
-	    querysocket = socket(PF_INET, SOCK_STREAM, 0);
-#ifdef INET6
-	  else {
-	    memset(&ua, 0, sizeof(ua));
-	    ua.v6.sin6_port   = htons(port);
-	    ua.v6.sin6_family = AF_INET6;
-	  }
-#endif
-	  if (querysocket < 0) {
-	    perror("querysocket: socket(PF_INET)");
-	    return;
-	  }
-	  setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
-
-#ifdef INET6
-	  if (ua.v6.sin6_family == AF_INET6) {
-	    if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v6) < 0) {
-	      perror("bind:mailq ipv6 socket");
-	      close(querysocket);
-	      querysocket = -1;
-	      return;
+	  if (notifysock) {
+	    if (cistrncmp(notifysock,"UNIX:",5)==0) {
+	      modedata = (char *)notifysock+5;
+	      modecode = 2;
+	    } else if (*notifysock == '/') {
+	      /* If it begins with '/', it is AF_UNIX socket */
+	      modedata = (char *)notifysock;
+	      modecode = 2;
+	    } else if (cistrncmp(notifysock,"UDP:",4)==0) {
+	      modedata = (char *)notifysock+4;
+	      modecode = 1;
+	    } else {
+	      /* The default mode is UDP/IP socket */
+	      modedata = (char *)notifysock;
+	      modecode = 1;
 	    }
 	  } else
-#endif
-	    if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v4) < 0) {
-	      perror("bind:mailq ipv4 socket");
-	      close(querysocket);
-	      querysocket = -1;
-	      return;
+	    modecode = 0;
+
+#ifdef  AF_UNIX
+	  if (modecode == 2) {
+	    struct sockaddr_un sad;
+	    int on = 1, oldumask;
+
+	    memset(&sad, 0, sizeof(sad));
+	    sad.sun_family = AF_UNIX;
+	    strncpy(sad.sun_path, modedata, sizeof(sad.sun_path));
+	    sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
+
+	    if ((notifysocket = socket(PF_UNIX, SOCK_DGRAM, 0)) < 0) {
+	      perror("notifysocket: socket(PF_UNIX)");
+	      break;
 	    }
 
+	    setsockopt(notifysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+	    /* In case that one already exists.. */
+	    unlink(sad.sun_path);
+
+	    oldumask = umask(0577);
+
+	    if (bind(notifysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
+	      perror("bind:UNIX notify socket");
+	      umask(oldumask);
+	      close(notifysocket);
+	      notifysocket = -1;
+	      break;
+	    }
+	    umask(oldumask);
+
+	    fcntl(notifysocket, F_SETFL,
+		  fcntl(notifysocket, F_GETFL, 0)|O_NONBLOCK);
+
 #if defined(F_SETFD)
-	  fcntl(querysocket, F_SETFD, 1); /* close-on-exec */
+	    fcntl(notifysocket, F_SETFD, 1); /* close-on-exec */
+#endif
+	  }
+#endif /* AF_UNIX */
+	  break;
+	}
+
+	while (querysocket < 0) {
+
+	  modedata = NULL;
+	  modecode = 0;
+	  if (mailqsock) {
+	    if (cistrncmp(mailqsock,"UNIX:",5)==0) {
+	      modedata = (char *)mailqsock+5;
+	      modecode = 2;
+	    } else if (*mailqsock == '/') {
+	      /* If it begins with '/', it is AF_UNIX socket */
+	      modedata = (char *)mailqsock;
+	      modecode = 2;
+	    } else if (cistrncmp(mailqsock,"TCP:",4)==0) {
+	      modedata = (char *)mailqsock+4;
+	      modecode = 1;
+	    } else {
+	      /* The default mode is TCP/IP socket */
+	      modedata = (char *)mailqsock;
+	      modecode = 1;
+	    }
+	  } else {
+	    /* The default mode is TCP/IP socket */
+	    modedata = (char *)mailqsock;
+	    modecode = 1;
+	  }
+
+#ifdef  AF_UNIX
+	  if (modecode == 2) {
+	    struct sockaddr_un sad;
+	    int on = 1, oldumask;
+
+	    memset(&sad, 0, sizeof(sad));
+	    sad.sun_family = AF_UNIX;
+	    strncpy(sad.sun_path, modedata, sizeof(sad.sun_path));
+	    sad.sun_path[ sizeof(sad.sun_path)-1 ] = 0;
+
+	    if ((querysocket = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
+	      perror("querysocket: socket(PF_UNIX)");
+	      break;
+	    }
+
+	    setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+	    /* In case that one already exists.. */
+	    unlink(sad.sun_path);
+
+	    oldumask = umask(0111);
+
+	    if (bind(querysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
+	      perror("bind:UNIX mailq socket");
+	      umask(oldumask);
+	      close(querysocket);
+	      querysocket = -1;
+	      break;
+	    }
+	    umask(oldumask);
+
+	    fd_nonblockingmode(querysocket);
+
+#if defined(F_SETFD)
+	    fcntl(querysocket, F_SETFD, 1); /* close-on-exec */
 #endif
 
-	  if (listen(querysocket, 5) < 0) {
-	    perror("listen:mailq socket");
-	    close(querysocket);
+	    if (listen(querysocket, 5) < 0) {
+	      perror("listen:UNIX mailq socket");
+	      close(querysocket);
+	      querysocket = -1;
+	      break;
+	    }
+	  }
+#endif
+#ifdef	AF_INET
+	  if (modecode == 1) {
+	    struct servent *serv;
+	    Usockaddr ua;
+	    int on = 1;
+	    int port = 174;
+	    char *modedata2 = NULL;
+
+	    if (modedata) {
+	      modedata2 = strchr(modedata,'@');
+	      if (modedata2) *modedata2++ = 0;
+	    }
+
+	    if (!modedata || !*modedata || sscanf(modedata,"%d",&port) != 1) {
+	      serv = getservbyname(modedata ? modedata : "mailq", "tcp");
+	      if (serv == NULL) {
+		sfprintf(sfstderr, "No 'mailq' tcp service defined!\n");
+		port = 174; /* magic knowledge */
+	      } else
+		port = ntohs(serv->s_port);
+	    }
+	    if (modedata2)  modedata2[-1] = '@';
+
+	    memset(&ua, 0, sizeof(ua));
+	    if (zgetbindaddr(NULL, AF_INET, &ua))
+		ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
+	    ua.v4.sin_port        = htons(port);
+	    ua.v4.sin_family      = AF_INET;
 	    querysocket = -1;
-	    return;
+#ifdef INET6
+	    querysocket = socket(PF_INET6, SOCK_STREAM, 0);
+#endif
+	    if (querysocket < 0)
+	      querysocket = socket(PF_INET, SOCK_STREAM, 0);
+#ifdef INET6
+	    else {
+	      memset(&ua, 0, sizeof(ua));
+	      if (zgetbindaddr(NULL, AF_INET6, &ua))
+		  ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
+	      ua.v6.sin6_port   = htons(port);
+	      ua.v6.sin6_family = AF_INET6;
+	    }
+#endif
+	    if (querysocket < 0) {
+#ifdef INET6
+	      perror("querysocket: socket(PF_INET6/PF_INET)");
+#else
+	      perror("querysocket: socket(PF_INET)");
+#endif
+	      break;
+	    }
+	    setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+#ifdef INET6
+	    if (ua.v6.sin6_family == AF_INET6) {
+	      if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v6) < 0) {
+		perror("bind:TCP6 mailq socket");
+		close(querysocket);
+		querysocket = -1;
+		break;
+	      }
+	    } else
+#endif
+	      if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v4) < 0) {
+		perror("bind:TCP4 mailq socket");
+		close(querysocket);
+		querysocket = -1;
+		break;
+	      }
+
+	    fcntl(querysocket, F_SETFL,
+		  fcntl(querysocket, F_GETFL, 0)|O_NONBLOCK);
+
+
+#if defined(F_SETFD)
+	    fcntl(querysocket, F_SETFD, 1); /* close-on-exec */
+#endif
+
+	    if (listen(querysocket, 5) < 0) {
+	      perror("listen:TCP mailq socket");
+	      close(querysocket);
+	      querysocket = -1;
+	      break;
+	    }
 	  }
-	  qipcretry = 0;
-	}
 #endif	/* AF_INET */
+	  break;
+	}
+
+	if (querysocket >= 0 && notifysocket >= 0)
+	  qipcretry = 0; /* Successfull init done. */
+	else {
+	  mytime(&now); 
+	  qipcretry = now + 5; /* Will do retry soon.. */
+
+	}
 }
 
 #else	/* !HAVE_SELECT */
@@ -1644,13 +1832,17 @@ int signum;
 	for (;;) {
 
 #ifdef	HAVE_WAITPID
-	  pid = waitpid(-1,&statloc,WNOHANG);
+	  pid = waitpid(-1, &statloc, WNOHANG);
+#else
+#ifdef  HAVE_WAIT4
+	  pid = wait4(-1, &statloc, WNOHANG, NULL);
 #else
 #ifdef  HAVE_WAIT3
-	  pid = wait3(&statloc,WNOHANG,NULL);
+	  pid = wait3(&statloc, WNOHANG, NULL);
 #else
 	  pid = wait(&statloc);
 #endif
+#endif
 #endif
 	  if (pid <= 0) break;
 
