@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-2000.
+ *      Copyright 1991-2003.
  */
 
 /*
@@ -17,6 +17,9 @@
 #define SKIPTEXT(Y)  while (*Y && *Y != ' ' && *Y != '\t') ++Y
 #define SKIPDIGIT(Y) while ('0' <= *Y && *Y <= '9') ++Y
 
+static int called_getbindaddr = 0;
+static Usockaddr bindaddr;
+
 static void dollarexpand __((unsigned char *s0, int space));
 static void dollarexpand(s0, space)
      unsigned char *s0;
@@ -116,10 +119,11 @@ static void dollarexpand(s0, space)
 }
        
 
-static void cfparam __((char *, int));
-static void cfparam(str,size)
+static void cfparam __((char *, int, const char *, int));
+static void cfparam(str, size, cfgfilename, linenum)
      char *str;
-     int size;
+     int size, linenum;
+     const char *cfgfilename;
 {
     char *name, *param1, *param2, *param3;
     char *str0 = str;
@@ -155,6 +159,12 @@ static void cfparam(str,size)
 	hdr220lines[i+1] = NULL;
 	return;
     }
+    if (cistrcmp(name, "sasl-mechanisms") == 0) {
+      param2 = strchr(str, '\n');
+      if (param2) *param2 = 0;
+      SASL_Auth_Mechanisms = strdup(str);
+      return;
+    }
 
     /* Do '$' expansions on the string */
     dollarexpand((unsigned char *)str, size - (str - str0));
@@ -208,54 +218,16 @@ static void cfparam(str,size)
       if (bindport != 0 && bindport != 0xFFFFU)
 	bindport_set = 1;
     } else if (cistrcmp(name, "BindAddress") == 0 && param1) {
-      memset(&bindaddr, 0, sizeof(bindaddr));
-      bindaddr_set = 1;
-#if defined(AF_INET6) && defined(INET6)
-      if (cistrncmp(param1,"[ipv6 ",6) == 0 ||
-	  cistrncmp(param1,"[ipv6:",6) == 0 ||
-	  cistrncmp(param1,"[ipv6.",6) == 0) {
-	char *s = strchr(param1,']');
-	if (s) *s = 0;
-	if (inet_pton(AF_INET6, param1+6, &bindaddr.v6.sin6_addr) < 1) {
-	  /* False IPv6 number literal */
-	  /* ... then we don't set the IP address... */
-	  bindaddr_set = 0;
-	}
-	bindaddr.v6.sin6_family = AF_INET6;
-      } else
-#endif
-	if (*param1 == '[') {
-	  char *s = strchr(param1,']');
-	  if (s) *s = 0;
-	  if (inet_pton(AF_INET, param1+1, &bindaddr.v4.sin_addr) < 1) {
-	    /* False IP(v4) number literal */
-	    /* ... then we don't set the IP address... */
-	    bindaddr_set = 0;
-	  }
-	  bindaddr.v4.sin_family = AF_INET;
-	} else {
-	  if (CISTREQN(param1,"iface:",6)) {
-#if defined(AF_INET6) && defined(INET6)
-	    bindaddr.v6.sin6_family = AF_INET6;
-	    if (zgetifaddress(AF_INET6, param1+6,
-			      (struct sockaddr *)&bindaddr.v6.sin6_addr))
-	      /* Didn't get IPv6 interface address of given name.. */
-#endif
-	      {
-		if (zgetifaddress(AF_INET, param1+6,
-				  (struct sockaddr *)&bindaddr.v4.sin_addr)) {
-		  /* No recognized interface! */
-		  bindaddr_set = 0;
-		} else
-		  /* Got IPv4 type interface address */
-		  bindaddr.v4.sin_family = AF_INET;
-	      }
-	  } else {
-	    /* XXX: TODO: Try to see if this is an interface name, and pick
-	       IPv4 and/or IPv6 addresses for that interface. */
-	    bindaddr_set = 0;
-	  }
-	}
+      called_getbindaddr=1;
+      if (!zgetbindaddr(param1, use_ipv6, &bindaddr)) {
+	bindaddrs_count += 1;
+	bindaddrs = realloc( bindaddrs,
+			     sizeof(bindaddr) * (bindaddrs_count +1) );
+	if (!bindaddrs)
+	  bindaddrs_count = 0;
+	else
+	  bindaddrs[ bindaddrs_count-1 ] = bindaddr;
+      }
     }
 
     /* SMTP Protocol limit & policy tune options */
@@ -264,9 +236,9 @@ static void cfparam(str,size)
 	sscanf(param1, "%ld", &maxsize);
     } else if (cistrcmp(name, "min-availspace") == 0 && param1) {
 	if (sscanf(param1, "%ld", &minimum_availspace) == 1) {
-	  minimum_availspace *= 1024;
-	  if (minimum_availspace < 1000000)
-	    minimum_availspace = 1000000;
+	  /* Minimum of minimum is 1000 kB ! */
+	  if (minimum_availspace < 1000)
+	    minimum_availspace = 1000;
 	}
     } else if (cistrcmp(name, "RcptLimitCnt") == 0 && param1) {
 	sscanf(param1, "%d", &rcptlimitcnt);
@@ -297,9 +269,10 @@ static void cfparam(str,size)
 	use_tcpwrapper = 1;
     }
 
-    else if (cistrcmp(name, "tarpit") == 0 && param2 /* 2 params */) {
+    else if (cistrcmp(name, "tarpit") == 0 && param3 /* 3 params */) {
 	sscanf(param1,"%d",&tarpit_initial);
 	sscanf(param2,"%d",&tarpit_exponent);
+	sscanf(param3,"%d",&tarpit_toplimit);
     }
 
     else if (cistrcmp(name, "deliverby") == 0) {
@@ -319,6 +292,9 @@ static void cfparam(str,size)
       if (access(param1, X_OK) == 0)
 	contentfilter = strdup(param1);
     }
+    else if (cistrcmp(name, "debug-contentfilter") == 0) {
+      debug_content_filter = 1;
+    }
 
     /* A few facility enablers: (default: off) */
 
@@ -334,6 +310,8 @@ static void cfparam(str,size)
       auth_ok = 1;
     } else if (cistrcmp(name, "auth-login-also-without-tls") == 0) {
       auth_login_without_tls = 1;
+    } else if (cistrcmp(name, "smtp-auth-sasl") == 0) {
+      do_sasl = 1;
     } else if (cistrcmp(name, "msa-mode") == 0) {
       msa_mode = 1;
     } else if (cistrcmp(name, "smtp-auth-pipe") == 0 && param1) {
@@ -374,34 +352,69 @@ static void cfparam(str,size)
       etrn_ok = 0;
     } else if (cistrcmp(name, "no-multiline-replies") == 0) {
       multilinereplies = 0;
+    } else if (cistrcmp(name, "force-rcpt-notify-never") == 0) {
+      force_rcpt_notify_never = 1;
     }
 
+#ifdef HAVE_OPENSSL
+
     /* TLSv1/SSLv* options */
 
     else if (cistrcmp(name, "use-tls") == 0)
       starttls_ok = 1;		/* Default: OFF */
 
-    else if (cistrcmp(name, "listen-ssmtp") == 0) {
+    else if (cistrcmp(name, "listen-ssmtp") == 0)
       ssmtp_listen = 1;		/* Default: OFF */
 
+    else if (cistrcmp(name, "outlook-tls-bug") == 0) {
+      detect_incorrect_tls_use = 1;	/* Default: OFF */
+
     } else if (cistrcmp(name, "tls-cert-file") == 0 && param1) {
-      if (tls_cert_file) free(tls_cert_file);
+      if (tls_cert_file) free((void*)tls_cert_file);
       tls_cert_file = strdup(param1);
       if (!tls_key_file)	/* default the other */
 	tls_key_file = strdup(param1);
 
     } else if (cistrcmp(name, "tls-key-file")  == 0 && param1) {
-      if (tls_key_file) free(tls_key_file);
+      if (tls_key_file) free((void*)tls_key_file);
       tls_key_file = strdup(param1);
       if (!tls_cert_file)	/* default the other */
 	tls_cert_file = strdup(param1);
 
+    } else if (cistrcmp(name, "tls-dcert-file") == 0 && param1) {
+      if (tls_dcert_file) free((void*)tls_dcert_file);
+      tls_dcert_file = strdup(param1);
+      if (!tls_dkey_file)	/* default the other */
+	tls_dkey_file = strdup(param1);
+
+    } else if (cistrcmp(name, "tls-dkey-file")  == 0 && param1) {
+      if (tls_dkey_file) free((void*)tls_dkey_file);
+      tls_dkey_file = strdup(param1);
+      if (!tls_dcert_file)	/* default the other */
+	tls_dcert_file = strdup(param1);
+
+    } else if (cistrcmp(name, "tls-dh1024")  == 0 && param1) {
+      if (tls_dh1024_param) free((void*)tls_dh1024_param);
+      tls_dh1024_param = strdup(param1);
+
+    } else if (cistrcmp(name, "tls-dh512")  == 0 && param1) {
+      if (tls_dh512_param) free((void*)tls_dh512_param);
+      tls_dh512_param = strdup(param1);
+
+    } else if (cistrcmp(name, "tls-random-source")  == 0 && param1) {
+      if (tls_random_source) free((void*)tls_random_source);
+      tls_random_source = strdup(param1);
+
+    } else if (cistrcmp(name, "tls-cipher-list")  == 0 && param1) {
+      if (tls_cipherlist) free((void*)tls_cipherlist);
+      tls_cipherlist = strdup(param1);
+
     } else if (cistrcmp(name, "tls-CAfile")    == 0 && param1) {
-      if (tls_CAfile) free(tls_CAfile);
+      if (tls_CAfile) free((void*)tls_CAfile);
       tls_CAfile = strdup(param1);
 
     } else if (cistrcmp(name, "tls-CApath")    == 0 && param1) {
-      if (tls_CApath) free(tls_CApath);
+      if (tls_CApath) free((void*)tls_CApath);
       tls_CApath = strdup(param1);
 
     } else if (cistrcmp(name, "tls-loglevel")  == 0 && param1) {
@@ -420,17 +433,16 @@ static void cfparam(str,size)
       sscanf(param1,"%d", & tls_req_cert);
 
     } else if (cistrcmp(name, "tls-use-scache") == 0) {
-#ifdef HAVE_OPENSSL
       tls_use_scache = 1;
-#endif /* - HAVE_OPENSSL */
 
     } else if (cistrcmp(name, "tls-scache-timeout") == 0 && param1) {
-#ifdef HAVE_OPENSSL
       sscanf(param1,"%d", & tls_scache_timeout);
-#endif /* - HAVE_OPENSSL */
+
     } else if (cistrcmp(name, "lmtp-mode") == 0) {
       lmtp_mode = 1;
     }
+#endif /* - HAVE_OPENSSL */
+
 
     /* Cluster-wide ETRN support for load-balanced smtp relay use */
     else if (cistrcmp(name, "etrn-cluster") == 0 && param3 /* 3 params */) {
@@ -445,6 +457,7 @@ static void cfparam(str,size)
 
     else {
       /* XX: report error for unrecognized PARAM keyword ?? */
+      type(NULL,0,NULL, "Cfgfile '%s' line %d has bad PARAM keyword/missing parameters: '%s'", cfgfilename, linenum, name);
     }
 }
 
@@ -455,12 +468,14 @@ readcffile(name)
     FILE *fp;
     struct smtpconf scf, *head, *tail = NULL;
     char c, *cp, buf[1024], *s, *s0;
+    int linenum = 0;
 
     if ((fp = fopen(name, "r")) == NULL)
 	return NULL;
     head = NULL;
     buf[sizeof(buf) - 1] = 0;
     while (fgets(buf, sizeof buf, fp) != NULL) {
+	++linenum;
 	c = buf[0];
 	if (c == '#' || (isascii(c) && isspace(c)))
 	    continue;
@@ -475,7 +490,7 @@ readcffile(name)
 	cp = buf;
 	SKIPSPACE(cp);
 	if (strncmp(cp, "PARAM", 5) == 0) {
-	    cfparam(cp, sizeof(buf) -(cp-buf));
+	    cfparam(cp, sizeof(buf) -(cp-buf), name, linenum);
 	    continue;
 	}
 	scf.flags = "";
@@ -515,6 +530,18 @@ readcffile(name)
 	configuration_ok = 1; /* At least something! */
     }
     fclose(fp);
+    if (!called_getbindaddr) {
+	bindaddr_set = !zgetbindaddr(NULL, use_ipv6, &bindaddr);
+	if (bindaddr_set) {
+	  bindaddrs_count += 1;
+	  bindaddrs = realloc( bindaddrs,
+			       sizeof(Usockaddr) * (bindaddrs_count +1) );
+	  if (!bindaddrs)
+	    bindaddrs_count = 0;
+	  else
+	    bindaddrs[ bindaddrs_count-1 ] = bindaddr;
+	}
+    }
     return head;
 }
 
