@@ -1,55 +1,16 @@
 /*
  *   mx_client_verify() -- subroutine for ZMailer smtpserver
  *
- *   By Matti Aarnio <mea@nic.funet.fi> 1997-1999
+ *   By Matti Aarnio <mea@nic.funet.fi> 1997-1999,2002-2003
  */
 
 #include "smtpserver.h"
-
-#if	defined(TRY_AGAIN) && defined(HAVE_RESOLVER)
-#define	BIND		/* Want BIND (named) nameserver support enabled */
-#endif	/* TRY_AGAIN */
-#ifdef	BIND
-#ifdef NOERROR
-#undef NOERROR		/* Several SysV-streams using systems have NOERROR,
-			   which is not the same as  <arpa/nameser.h> has! */
-#endif
-#include <arpa/nameser.h>
-#include <resolv.h>
-
-#ifndef	BIND_VER
-#ifdef	GETLONG
-/* 4.7.3 introduced the {GET,PUT}{LONG,SHORT} macros in nameser.h */
-#define	BIND_VER	473
-#else	/* !GETLONG */
-#define	BIND_VER	472
-#endif	/* GETLONG */
-#endif	/* !BIND_VER */
-#endif	/* BIND */
-
-#ifndef	T_TXT
-# define T_TXT 16	/* Text strings */
-#endif
-
-extern int h_errno;
-extern int res_mkquery(), res_send(), dn_skipname(), dn_expand();
+#include "zresolv.h"
 
 extern int use_ipv6;
 
-#if	defined(BIND_VER) && (BIND_VER >= 473)
-typedef u_char msgdata;
-#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
-typedef char msgdata;
-#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
-
-
 static int dnsmxlookup __((const char*, int, int, int));
 
-typedef union {
-	HEADER qb1;
-	char qb2[8000];
-} querybuf;
-
 extern int debug;
 static char * txt_buf = NULL;
 
@@ -65,7 +26,7 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	int qlen, n, i, qdcount, ancount, nscount, arcount, maxpref, class;
 	u_short type;
 	int saw_cname = 0, had_mx_record = 0;
-	/* int ttl; */
+	int ttl;
 	struct addrinfo req, *ai;
 #define MAXMX 128
 	char *mx[MAXMX];
@@ -151,22 +112,25 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  if (n < 0)
 	    break;
 	  cp += n;
-	  type = _getshort(cp);
-	  cp += 2;
-	  /* class = _getshort(cp); */
-	  cp += 2;
-	  /* ttl = _getlong(cp); */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+
+	  NS_GET16(type,  cp); /* type  */
+	  NS_GET16(class, cp); /* class */
+	  NS_GET32(ttl,   cp); /* ttl   */
+	  NS_GET16(n,     cp); /* dlen  */
+
+	  if (cp + n > eom) {
+	    /* BAD data.. */
+	    break;
+	  }
+
 	  if (type == T_CNAME) {
 	    cp += dn_expand((msgdata *)&answer, eom, cp,
 			    (void*)realname, sizeof realname);
 	    saw_cname = 1;
 	    continue;
-	  } else if (type != qtype)  {
+	  }
+
+	  if (type != qtype)  {
 	    /* Not looked for .. */
 	    cp += n;
 	    continue;
@@ -195,13 +159,24 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  } /* ===== END OF MX DATA PROCESING ========= */
 
 	  if (type == T_TXT) {
-	    int len = (*cp++) & 0xFF; /* 0..255 chars */
+	    int len = (*cp) & 0xFF; /* 0..255 chars */
+
+	    /* Mal-formed inputs are possible overflowing the buffer.. */
+	    if (len > (eom - cp))
+	      len = (eom - cp);
+	    if (len > n - 1)
+	      len = n - 1;
+
 	    if (txt_buf != NULL)
 	      free(txt_buf);
 	    txt_buf = emalloc(len+1);
+	    ++cp;
 	    memcpy(txt_buf, cp, len);
 	    txt_buf[len] = '\0';
-	    for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
+	    for (i = 0; i < mxcount; ++i) {
+	      if (mx[i]) free(mx[i]);
+	      mx[i] = NULL;
+	    }
 	    return 1; /* OK! */
 	  }
 
@@ -238,16 +213,12 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    break;
 	  cp += n;
 	  if (cp+10 > eom) { cp = eom; break; }
-	  /* type = _getshort(cp); */
-	  cp += 2;
-	  /* class = _getshort(cp); */
-	  cp += 2;
-	  /* mx[nmx].expiry = now + _getlong(cp); */ /* TTL */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones, I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+
+	  NS_GET16(type,  cp); /* type  - short */
+	  NS_GET16(class, cp); /* class - short */
+	  cp += NS_INT32SZ;    /* ttl   - long  */
+	  NS_GET16(n, cp);     /* dlen  - short */
+
 	  cp += n; /* We simply skip this data.. */
 	  if (cp <= eom)
 	    --nscount;
@@ -268,16 +239,11 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  if (n < 0) { cp = eom; break; }
 	  cp += n;
 	  if (cp+10 > eom) { cp = eom; break; }
-	  type = _getshort(cp);
-	  cp += 2;
-	  class = _getshort(cp);
-	  cp += 2;
-	  /* mx[nmx].expiry = now + _getlong(cp); */ /* TTL */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones, I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+
+	  NS_GET16(type,  cp); /* type  - short */
+	  NS_GET16(class, cp); /* class - short */
+	  cp += NS_INT32SZ;    /* ttl   - long  */
+	  NS_GET16(n, cp);     /* dlen  - short */
 
 	  if (cp + n > eom) { cp = eom; break; }
 
@@ -334,24 +300,23 @@ dnsmxlookup(host, depth, mxmode, qtype)
 
 #if 1
 		if (debug) {
-		  struct sockaddr * sa = (struct sockaddr*) &usa;
-		  if (sa->sa_family == AF_INET) {
-		    inet_ntop(AF_INET, (void*) & ((struct sockaddr_in *)sa)->sin_addr, (char *)buf, sizeof(buf));
+		  if (usa.v4.sin_family == AF_INET) {
+		    inet_ntop(AF_INET, (void*) & usa.v4.sin_addr, (char *)buf, sizeof(buf));
 		    printf("000-  matching %s AR address IPv4:[%s]\n", mx[i], buf);
 		  }
 #if defined(AF_INET6) && defined(INET6)
-		  else if (sa->sa_family == AF_INET6) {
-		    inet_ntop(AF_INET6, (void*) & ((struct sockaddr_in6 *)sa)->sin6_addr, buf, sizeof(buf));
+		  else if (usa.v6.sin6_family == AF_INET6) {
+		    inet_ntop(AF_INET6, (void*) & usa.v6.sin6_addr, buf, sizeof(buf));
 		    printf("000-  matching %s AR address IPv6:[%s]\n", mx[i], buf);
 		  }
 #endif
 		  else
 		    printf("000- matching unknown %s AR address family address; AF=%d\n",
-			   mx[i], sa->sa_family);
+			   mx[i], usa.v4.sin_family);
 		}
 #endif
 
-		i = matchmyaddress( (struct sockaddr *)&usa );
+		i = matchmyaddress( &usa );
 		if (i == 1) {
 		  if (debug)
 		    printf("000-   AR ADDRESS MATCH!\n");
@@ -407,21 +372,18 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  req.ai_socktype = SOCK_STREAM;
 	  req.ai_protocol = IPPROTO_TCP;
 	  req.ai_flags    = AI_CANONNAME;
+	  /*  ai_family  set above. */
 	  ai = NULL;
 	  /* This resolves CNAME, it should not happen in case
 	     of MX server, though..    */
-#define GETADDRINFODEBUG 0
-#if !GETADDRINFODEBUG
-	  i = getaddrinfo(mx[n], "0", &req, &ai);
-#else
+#ifdef HAVE__GETADDRINFO_
 	  i = _getaddrinfo_(mx[n], "0", &req, &ai,
 			    debug ? stdout : NULL);
+#else
+	  i = getaddrinfo(mx[n], "0", &req, &ai);
+#endif
 	  if (debug)
 	    printf("000-  getaddrinfo('%s','0') -> r=%d, ai=%p\n",mx[n],i,ai);
-#endif
-
-	  if (debug)
-	    printf("000-  getaddrinfo('%s') yields %d\n", mx[n], i);
 	    
 	  if (i != 0)
 	    continue;		/* Well well.. spurious! */
@@ -439,24 +401,25 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    ++k;
 #if 1
 	    if (debug) {
-	      struct sockaddr * sa = ai2->ai_addr;
+	      Usockaddr * usa = (Usockaddr *) ai2->ai_addr;
 	      char buf[60];
-	      if (sa->sa_family == AF_INET) {
-		inet_ntop(AF_INET, (void*) & ((struct sockaddr_in *)sa)->sin_addr, buf, sizeof(buf));
+
+	      if (usa->v4.sin_family == AF_INET) {
+		inet_ntop(AF_INET, (void*) & usa->v4.sin_addr, buf, sizeof(buf));
 		printf("000-  matching %s address IPv4:[%s]\n", mx[n], buf);
 	      }
 #if defined(AF_INET6) && defined(INET6)
-	      else if (sa->sa_family == AF_INET6) {
-		inet_ntop(AF_INET6, (void*) & ((struct sockaddr_in6 *)sa)->sin6_addr, buf, sizeof(buf));
+	      else if (usa->v6.sin6_family == AF_INET6) {
+		inet_ntop(AF_INET6, (void*) & usa->v6.sin6_addr, buf, sizeof(buf));
 		printf("000-  matching %s address IPv6:[%s]\n", mx[n], buf);
 	      }
 #endif
 	      else
 		printf("000- matching %s unknown address family address; AF=%d\n",
-		       mx[n], sa->sa_family);
+		       mx[n], usa->v4.sin_family);
 	    }
 #endif
-	    rc = matchmyaddress(ai2->ai_addr);
+	    rc = matchmyaddress((Usockaddr *)ai2->ai_addr);
 	    if (rc == 1) {
 	      if (debug)
 		printf("000-   ADDRESS MATCH!\n");
@@ -490,21 +453,67 @@ perhaps_address_record:
 	  req.ai_socktype = SOCK_STREAM;
 	  req.ai_protocol = IPPROTO_TCP;
 	  req.ai_flags    = AI_CANONNAME;
-	  req.ai_family   = 0; /* Both OK (IPv4/IPv6) */
+	  req.ai_family   = PF_INET;
 	  ai = NULL;
 
+
 	  /* This resolves CNAME, it should not happen in case
 	     of MX server, though..    */
-#if !GETADDRINFODEBUG
-	  i = getaddrinfo((const char*)host, "0", &req, &ai);
-#else
+#ifdef HAVE__GETADDRINFO_
+	  if (debug)
+	    printf("000-  perhaps A?\n");
 	  i = _getaddrinfo_((const char*)host, "0", &req, &ai, debug ? stdout : NULL);
+#else
+	  i = getaddrinfo((const char*)host, "0", &req, &ai);
 #endif
 	  if (debug)
-	    printf("000-   perhaps A? getaddrinfo('%s','0') -> r=%d, ai=%p\n",host,i,ai);
-	  if (i != 0) /* Found nothing! */
+	    printf("000-   getaddrinfo('%s','0') (PF_INET) -> r=%d (%s), ai=%p\n",host,i,gai_strerror(i),ai);
+
+#if defined(AF_INET6) && defined(INET6)
+	  if (use_ipv6) {
+
+	    /* Want, but not have AAAA, ask for it. */
+
+	    int n2;
+	    struct addrinfo *ai2 = NULL;
+
+	    memset(&req, 0, sizeof(req));
+	    req.ai_socktype = SOCK_STREAM;
+	    req.ai_protocol = IPPROTO_TCP;
+	    req.ai_flags    = AI_CANONNAME;
+	    req.ai_family   = PF_INET6;
+
+	  /* This resolves CNAME, it should not happen in case
+	     of MX server, though..    */
+#ifdef HAVE__GETADDRINFO_
+	    n2 = _getaddrinfo_((const char *)host, "0", &req, &ai2,
+			       debug ? stdout : NULL);
+#else
+	    n2 = getaddrinfo((const char *)host, "0", &req, &ai2);
+#endif
+	    if (debug)
+	      printf("000-   getaddrinfo('%s','0') (PF_INET6) -> r=%d (%s), ai=%p\n",host,n2,gai_strerror(n2),ai2);
+
+
+	    if (i != 0 && n2 == 0) {
+	      /* IPv6 address, no IPv4 (or error..) */
+	      i = n2;
+	      ai = ai2; ai2 = NULL;
+	    }
+	    if (ai2 && ai) {
+	      /* BOTH ?!  Catenate them! */
+	      struct addrinfo **aip;
+	      aip = &ai->ai_next;
+	      while (*aip) aip = &((*aip)->ai_next);
+	      *aip = ai2;
+	    }
+	  }
+#endif
+
+	  if (i)
 	    return i;
 
+
 	  i = matchmyaddresses(ai);
 #if 1
 	  /* With this we can refuse to accept any message with
@@ -625,7 +634,7 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
      char **msgp;
 {
 	char hbuf[2000], *s, *suf;
-	/* int hspc; */
+	int hspc;
 	struct hostent *hp;
 	int has_ok = 0;
 
@@ -644,18 +653,23 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	}
 
 	suf = hbuf + strlen(hbuf);
-	/* hspc = sizeof(hbuf)-strlen(hbuf)-1; */
+	hspc = sizeof(hbuf) - strlen(hbuf) - 2;
 
 	while (*rbldomain) {
 	  /* "rbldomain" is possibly a COLON-demarked set of
 	     domain names:  rbl.maps.vix.com:dul.maps.vix.com
 	     which isn't so easy to read, but ... */
+	  /* The 2000 char buffer should be way oversized
+	     for this routine's needs..  And it is managerial
+	     input at the policy database, which has the "unpredictable"
+	     size...  */
 	  s = strchr(rbldomain, ':');
 	  if (s) *s = 0;
 	  if (strcmp(rbldomain,"+") == 0)
-	    strcpy (suf, "rbl.maps.vix.com");
+	    strncpy (suf, "rbl.maps.vix.com", hspc);
 	  else
-	    strcpy (suf, rbldomain);
+	    strncpy (suf, rbldomain, hspc);
+	  suf[hspc] = '\0';
 
 	  if (s) {
 	    *s = ':';
@@ -664,6 +678,17 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	    rbldomain += strlen(rbldomain);
 	  }
 
+	  /* Add explicite DOT into the tail of the lookup object.
+	     That way the lookup should never use resolver's  SEARCH
+	     suffix set. */
+
+	  s = suf + strlen(suf) - 1;
+	  if (*s != '.') {
+	    *(++s) = '.';
+	    *(++s) = 0;
+	  }
+
+
 	  if (debug)
 	    printf("000- looking up DNS A object: %s\n", hbuf);
 
@@ -682,11 +707,12 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	      has_ok = 1;
 	      continue; /* Isn't  127.0.0.* */
 	    }
+#if 0
 	    if (strcmp("127.0.0.4",abuf) == 0) {
 	      /* ORBS NETBLOCK */
 	      if (has_ok) continue;
 	    }
-
+#endif
 	    /* Ok, then lookup for the TXT entry too! */
 	    if (debug)
 	      printf("000- looking up DNS TXT object: %s\n", hbuf);
@@ -695,11 +721,22 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	      if (*msgp != NULL)
 		free(*msgp);
 	      *msgp = strdup(txt_buf);
+	      s = *msgp;
+	      if (s) {
+		for ( ;*s; ++s) {
+		  int c = ((*s) & 0xFF);
+		  /* Characters not printable in ISO-8859-*
+		     are masked with space. */
+		  if (c < ' ' || (c >= 127 && c < 128+32) || c == 255)
+		    *s = ' ';
+		}
+	      }
+	      type(NULL,0,NULL,"Found DNS TXT object: %s\n", *msgp ? *msgp : "<nil>");
 	    }
 	    return -1;
 	  }
 	  /* Didn't find A record */
-	  type(NULL,0,NULL, "Looking up DNS A object: %s", hbuf);
+	  type(NULL,0,NULL, "Didn't find DNS A object: %s", hbuf);
 	}
 
 	return 0;
