@@ -1,6 +1,6 @@
 /*
  *  relaytest.c -- module for ZMailer's smtpserver
- *  By Matti Aarnio <mea@nic.funet.fi> 1997-2000
+ *  By Matti Aarnio <mea@nic.funet.fi> 1997-2003
  *
  */
 
@@ -19,30 +19,15 @@
 #include <fcntl.h>
 #include <errno.h>
 
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
-# include <db_185.h>
-#else
-#ifdef HAVE_DB2_DB_H
-# include <db2/db.h>
-#else
-#ifdef HAVE_DB_H
-# include <db.h>
-#else
-#ifdef HAVE_DB1_DB_H
-# include <db1/db.h>
-#endif
-#endif
-#endif
-#endif
-#endif
 
-#ifdef HAVE_NDBM_H
+#include "sleepycatdb.h"
+
+#ifdef HAVE_NDBM
 #define datum Ndatum
 #include <ndbm.h>
 #undef datum
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 #define datum Gdatum
 #include <gdbm.h>
 #undef datum
@@ -74,6 +59,15 @@
 #include "policytest.h"
 
 
+/* We are not including  "smtpserver.h",  thus have to do local prototype.. */
+#if defined(HAVE_STDARG_H) && defined(HAVE_VPRINTF)
+extern void type __((void *, int code, const char *status, const char *fmt,...));
+#else
+extern void type __(( /* void *SS, int code, const char *status, const char *fmt, ... */ ));
+#endif
+
+
+
 extern int debug;
 extern int percent_accept;
 
@@ -119,9 +113,9 @@ const char *key;
     return buf;
 }
 
-static char *showattr __((const char *key));
+static char *showattr __((const unsigned char *key));
 static char *showattr(key)
-const char *key;
+const unsigned char *key;
 {
     static char buf[500];
     sprintf(buf,"%d/%s/'%s'", key[0], KA(key[1]), key+2);
@@ -136,17 +130,25 @@ static int valueeq(value,str)
     return (strcmp(value,str) == 0);
 }
 
-static char *showresults __((char *values[]));
-static char *showresults(values)
-char *values[];
+static char *showresults __((struct policystate *state));
+static char *showresults(state)
+struct policystate *state;
 {
     static char buf[2000];
     int i;
+    char **values=state->values;
 
     buf[0] = '\0';
     for (i = P_A_FirstAttr; i <= P_A_LastAttr; ++i) {
 	sprintf(buf+strlen(buf),"%s ",KA(i));
-	sprintf(buf+strlen(buf),"%s ",values[i] ? values[i] : ".");
+	if (i == P_A_InboundSizeLimit )
+	    sprintf(buf+strlen(buf),"%li ",state->maxinsize);
+	else if (i == P_A_OutboundSizeLimit )
+	    sprintf(buf+strlen(buf),"%li ",state->maxoutsize);
+	else if (i == P_A_MaxSameIpSource )
+	    sprintf(buf+strlen(buf),"%li ",state->maxsameiplimit);
+	else
+	    sprintf(buf+strlen(buf),"%s ",values[i] ? values[i] : ".");
     }
     return buf;
 }
@@ -157,23 +159,26 @@ const struct policystate *state;
 {
 	int i;
 
-	printf("000- always_reject=%d\n",state->always_reject);
-	printf("000- always_freeze=%d\n",state->always_freeze);
-	printf("000- always_accept=%d\n",state->always_accept);
-	printf("000- full_trust=%d\n",   state->full_trust);
-	printf("000- trust_recipients=%d\n",state->trust_recipients);
-	printf("000- sender_reject=%d\n",state->sender_reject);
-	printf("000- sender_freeze=%d\n",state->sender_freeze);
-	printf("000- sender_norelay=%d\n",state->sender_norelay);
-	printf("000- relaycustnet=%d\n", state->relaycustnet);
-	printf("000- rcpt_nocheck=%d\n", state->rcpt_nocheck);
+	type(NULL,0,NULL," always_reject=%d",state->always_reject);
+	type(NULL,0,NULL," always_freeze=%d",state->always_freeze);
+	type(NULL,0,NULL," always_accept=%d",state->always_accept);
+	type(NULL,0,NULL," full_trust=%d",   state->full_trust);
+	type(NULL,0,NULL," trust_recipients=%d",state->trust_recipients);
+	type(NULL,0,NULL," sender_reject=%d",state->sender_reject);
+	type(NULL,0,NULL," sender_freeze=%d",state->sender_freeze);
+	type(NULL,0,NULL," sender_norelay=%d",state->sender_norelay);
+	type(NULL,0,NULL," relaycustnet=%d", state->relaycustnet);
+	type(NULL,0,NULL," rcpt_nocheck=%d", state->rcpt_nocheck);
 
 	for ( i = P_A_FirstAttr; i <= P_A_LastAttr ; ++i) {
-		printf("000- %s: %srequested, value=%s\n",
-		       KA(i),
-		       (state->origrequest & (1<<i)) ? "" : "not ",
-		       state->values[i]?state->values[i]:".");
+	    type(NULL,0,NULL," %s: %srequested, value=%s", KA(i),
+		 (state->origrequest & (1<<i)) ? "" : "not ",
+		 state->values[i]?state->values[i]:".");
 	}
+
+	type(NULL,0,NULL," maxinsize=%li", state->maxinsize);
+	type(NULL,0,NULL," maxoutsize=%li", state->maxoutsize);
+	type(NULL,0,NULL," maxsameiplimit=%li", state->maxsameiplimit);
 }
 
 static void mask_ip_bits __((unsigned char *, int, int));
@@ -225,20 +230,20 @@ const int qlen;
 int *rlenp;			/* result length ptr ! */
 {
     char *buffer;
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
     Ndatum Nkey, Nresult;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     Gdatum Gkey, Gresult;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     DBT Bkey, Bresult;
     int rc;
 #endif
 
 
     switch (rel->dbt) {
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
     case _dbt_ndbm:
 
 	Nkey.dptr = (char *) qptr;
@@ -257,7 +262,7 @@ int *rlenp;			/* result length ptr ! */
 	break; /* some compilers complain, some produce bad code
 		  without this... */
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     case _dbt_gdbm:
 
 	Gkey.dptr = (void *) qptr;
@@ -274,11 +279,11 @@ int *rlenp;			/* result length ptr ! */
 	break; /* some compilers complain, some produce bad code
 		  without this... */
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     case _dbt_btree:
 
 
-	memset(&Bkey, 0, sizeof(Bkey));
+	memset(&Bkey,    0, sizeof(Bkey));
 	memset(&Bresult, 0, sizeof(Bresult));
 
 	Bkey.data = (void *) qptr;
@@ -303,7 +308,7 @@ int *rlenp;			/* result length ptr ! */
 
     case _dbt_bhash:
 
-	memset(&Bkey, 0, sizeof(Bkey));
+	memset(&Bkey,    0, sizeof(Bkey));
 	memset(&Bresult, 0, sizeof(Bresult));
 
 	Bkey.data = (void *) qptr;
@@ -356,7 +361,7 @@ struct policystate *state;
 const char *key;
 int init;
 {
-    char *str, *str_base, pbuf[256];
+    unsigned char *str, *str_base;
     int rlen, result, interest;
     char *msgstr = NULL;
 
@@ -376,14 +381,14 @@ int init;
     --recursions;
 
     if (debug)
-       printf("000- Key: %s\n", showkey(key));
+       type(NULL,0,NULL," Key: %s", showkey(key));
 /*
     if (key[1] != P_K_IPv4 && key[1] != P_K_IPv6) {
 	if (debug)
-	  printf("000- Key: %d/%d/%s\n", key[0],key[1],key+2);
+	  type(NULL,0,NULL," Key: %d/%d/%s", key[0],key[1],key+2);
     } else
       if (debug)
-	printf("000- Key: %u.%u.%u.%u\n", key[2] & 0xff, key[3] & 0xff, 
+	type(NULL,0,NULL," Key: %u.%u.%u.%u", key[2] & 0xff, key[3] & 0xff, 
 	       key[4] & 0xff, key[5] & 0xff);
 */
 
@@ -395,7 +400,7 @@ int init;
 
     if (str == NULL) {
       if (debug)
-	printf("000-  query failed\n");
+	type(NULL,0,NULL,"  query failed");
       return -1;
     }
     /* Scan trough attribute list. Call resolveattributes recursively
@@ -403,17 +408,31 @@ int init;
 
     while (rlen > 3) {
 
+	if (str[0] < 3) {
+	  if (debug)
+	    type(NULL,0,NULL," Bad length of attrbute, under 3 bytes!  %d", str[0]);
+	  break; /* BAD ATTRIBUTE! */
+	}
+
+	/* Attribute */
+	if (debug)
+	  type(NULL,0,NULL,"   Attribute: %s", showattr(str));
+
 	/* Alias */
 	if (str[1] == P_A_ALIAS) {
 	    /* Do not continue if max recursions reached. */
 	    if (recursions < 0) {
 	      if (debug)
-		printf("000- Max recursions reached.\n");
+		type(NULL,0,NULL," Max recursions reached.");
 	    } else {
+	      char pbuf[256];
+
 	      if (debug)
-		printf("000- Alias-recursion: %d\n", recursions);
+		type(NULL,0,NULL," Alias-recursion: %d", recursions);
+
+	      strncpy(pbuf+2,str+2, sizeof(pbuf)-3);
+	      pbuf[ sizeof(pbuf)-1 ] = 0;
 
-	      strcpy(pbuf+2,str+2);
 	      strlower(pbuf+2);
 	      pbuf[0] = strlen(str+2) + 3;
 	      pbuf[1] = P_K_TAG;
@@ -424,10 +443,6 @@ int init;
 	    continue;
 	}
 
-	/* Attribute */
-	if (debug)
-	  printf("000-   Attribute: %s\n", showattr(str));
-
 	if (str[1] == P_A_MESSAGE) {
 	  if (msgstr) free(msgstr);
 	  msgstr = strdup(str+2);
@@ -438,7 +453,7 @@ int init;
 	if ((interest & state->request) == 0) {
 	    /* Not interested in this attribute, skip into next. */
 	    if (debug)
-	      printf("000-     not interested, skipped...\n");
+	      type(NULL,0,NULL,"     not interested, skipped...");
 
 	    goto nextattr;
 	} else {
@@ -462,6 +477,11 @@ int init;
 	  sscanf(str+2,"%li", &state->maxoutsize);
 	  goto nextattr;
 
+	} else if (str[1] == P_A_MaxSameIpSource) {
+
+	  sscanf(str+2,"%li", &state->maxsameiplimit);
+	  goto nextattr;
+
 	} else if ((str[2] != '+' && str[2] != '-') &&
 		   !state->values[str[1] & 0xFF]) {
 
@@ -474,7 +494,7 @@ int init;
 	} else if (str[2] != '+' && str[2] != '-') {
 
 	  if (debug)
-	    printf("000- Unknown flag: %s\n", &str[2]);
+	    type(NULL,0,NULL," Unknown flag: %s", &str[2]);
 	  goto nextattr;
 	}
 	/* Store valid attribute.
@@ -484,10 +504,10 @@ int init;
 	    if (!state->values[str[1] & 0xFF])
 		state->values[str[1] & 0xFF] = strdup(str + 2);
 	  if (debug)
-	    printf("000-     accepted!\n");
+	    type(NULL,0,NULL,"     accepted!");
 	} else {
 	  if (debug)
-	    printf("000-   Unknown attribute, number: %d\n", str[1]);
+	    type(NULL,0,NULL,"   Unknown attribute, number: %d", str[1]);
 	}
 
     nextattr:
@@ -505,9 +525,10 @@ int init;
 	/* If all requests are done, exit. */
 	if (!state->request) {
 	  if (debug)
-	    printf("000- Every request found. Finishing search.\n");
+	    type(NULL,0,NULL," Every request found. Finishing search.");
 	  break;
 	}
+
     }				/* End of while. */
 
     /* Free memory from attribute list. Allocated in dbquery. */
@@ -534,19 +555,19 @@ const char *pbuf;
 
     if (pbuf[1] == P_K_DOMAIN) {
 	if (debug)
-	  printf("000- checkaddr(): domain of '%s'\n",pbuf+2);
+	  type(NULL,0,NULL," checkaddr(): domain of '%s'",pbuf+2);
 	result = resolveattributes(rel, maxrecursions, state, pbuf, 1);
 	if (debug) {
-	  printf("000- Results: %s\n", showresults(state->values));
+	  type(NULL,0,NULL," Results: %s", showresults(state));
 	}
 	return (result);
     }
     if (pbuf[1] == P_K_USER) {
 	if (debug)
-	  printf("000- checkaddr(): user of '%s'\n",pbuf+2);
+	  type(NULL,0,NULL," checkaddr(): user of '%s'",pbuf+2);
 	result = resolveattributes(rel, maxrecursions, state, pbuf, 1);
 	if (debug) {
-	  printf("000- Results: %s\n", showresults(state->values));
+	  type(NULL,0,NULL," Results: %s", showresults(state));
 	}
 	return (result);
     } else if (pbuf[1] == P_K_IPv4)
@@ -591,11 +612,11 @@ const char *pbuf;
 
     if (result != 0) {
       if (debug)
-	printf("000- Address not found.\n");
+	type(NULL,0,NULL," Address not found.");
       return -1;
     } else
       if (debug) {
-	printf("000-   %s\n", showresults(state->values));
+	type(NULL,0,NULL," Results:  %s", showresults(state));
       }
     return 0;
 }
@@ -613,15 +634,17 @@ int whosonrc;
     if (rel == NULL)
       return -1;  /* Not defined! */
 
-#ifdef HAVE_NDBM_H
+    memset(state, 0, sizeof(*state));
+
+#ifdef HAVE_NDBM
     if (cistrcmp(rel->dbtype, "ndbm") == 0)
 	rel->dbt = _dbt_ndbm;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     if (cistrcmp(rel->dbtype, "gdbm") == 0)
 	rel->dbt = _dbt_gdbm;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     if (cistrcmp(rel->dbtype, "btree") == 0)
 	rel->dbt = _dbt_btree;
     if (cistrcmp(rel->dbtype, "bhash") == 0)
@@ -639,7 +662,7 @@ int whosonrc;
     dbname = (char*)emalloc(strlen(rel->dbpath) + 8);
 #endif
     switch (rel->dbt) {
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
     case _dbt_ndbm:
 	/*
 	   rel->ndbm = dbm_open((char*)rel->dbpath, O_RDWR|O_CREAT|O_TRUNC, 0644);
@@ -649,7 +672,7 @@ int whosonrc;
 	openok = (rel->ndbm != NULL);
 	break;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     case _dbt_gdbm:
 	/* Append '.gdbm' to the name */
 	sprintf(dbname, "%s.gdbm", rel->dbpath);
@@ -657,35 +680,77 @@ int whosonrc;
 	openok = (rel->gdbm != NULL);
 	break;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     case _dbt_btree:
 	/* Append '.db' to the name */
 	sprintf(dbname, "%s.db", rel->dbpath);
-#ifdef HAVE_DB_OPEN2
+
+#if defined(HAVE_DB3) || defined(HAVE_DB4)
+
+	rel->btree = NULL;
+	openok = db_create(&rel->btree, NULL, 0);
+	if (openok == 0)
+	  openok = rel->btree->open(rel->btree,
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+				    NULL, /* TXN id was added at SleepyDB 4.1 */
+#endif
+				    dbname, NULL,  DB_BTREE,
+				    DB_RDONLY, 0);
+	if (debug && openok)
+	  type(NULL,0,NULL," btree->open('%s',BTREE, RDONLY) ret=%d",dbname,openok);
+	openok = !openok;
+
+#else
+#if defined(HAVE_DB2)
+
 	rel->btree = NULL;
 #ifndef DB_RDONLY
 # define DB_RDONLY O_RDONLY
 #endif
-	db_open(dbname, DB_BTREE, DB_RDONLY, 0644, NULL, NULL, &rel->btree);
-#else
+	openok = db_open(dbname, DB_BTREE, DB_RDONLY, 0644,
+			 NULL, NULL, &rel->btree);
+	openok = !openok;
+#else /* HAVE_DB1 */
 	rel->btree = dbopen(dbname, O_RDONLY, 0644, DB_BTREE, NULL);
-#endif
 	openok = (rel->btree != NULL);
+#endif
+#endif
 	break;
 
     case _dbt_bhash:
 	/* Append '.db' to the name */
 	sprintf(dbname, "%s.dbh", rel->dbpath);
-#ifdef HAVE_DB_OPEN2
+
+#if defined(HAVE_DB3) || defined(HAVE_DB4)
+
+	rel->bhash = NULL;
+	openok = db_create(&rel->bhash, NULL, 0);
+	if (openok == 0)
+	  openok = rel->bhash->open(rel->bhash,
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+				    NULL, /* TXN id was added at SleepyDB 4.1 */
+#endif
+				    dbname, NULL, DB_HASH,
+				    DB_RDONLY, 0);
+	if (debug && openok)
+	  type(NULL,0,NULL," bhash->open('%s',BHASH, RDONLY) ret=%d",dbname,openok);
+	openok = !openok;
+
+#else
+#if defined(HAVE_DB2)
+
 	rel->bhash = NULL;
 #ifndef DB_RDONLY
 # define DB_RDONLY O_RDONLY
 #endif
-	db_open(dbname, DB_HASH, DB_RDONLY, 0644, NULL, NULL, &rel->bhash);
-#else
+	openok = db_open(dbname, DB_HASH, DB_RDONLY, 0644,
+			 NULL, NULL, &rel->bhash);
+	openok = !openok;
+#else /* HAVE_DB1 */
 	rel->bhash = dbopen(rel->dbpath, O_RDONLY, 0644, DB_HASH, NULL);
-#endif
 	openok = (rel->bhash != NULL);
+#endif
+#endif
 	break;
 #endif
     default:
@@ -694,7 +759,7 @@ int whosonrc;
     if (!openok) {
 	/* ERROR!  Could not open the database! */
       if (debug) {
-	printf("000- ERROR!  Could not open the database file '%s'; errno=%d!\n",
+	type(NULL,0,NULL," ERROR!  Could not open the database file '%s'; errno=%d!",
 	       dbname, errno);
 	fflush(stdout);
       }
@@ -709,10 +774,10 @@ int whosonrc;
     free(dbname);
 #endif
 
-    memset(state, 0, sizeof(*state));
 #ifdef HAVE_WHOSON_H
     state->whoson_result = whosonrc;
 #endif
+    state->maxsameiplimit = -1;
     return 0;
 }
 
@@ -748,10 +813,10 @@ int sourceaddr;
     }
 
     lcldom = (state->request & (1 << P_A_LocalDomain));
-    myaddress = matchmyaddress((struct sockaddr*)&saddr);
+    myaddress = matchmyaddress(&saddr);
 
     if (debug)
-      printf("000- policytestaddr: lcldom/myaddress=%d/%d\n",lcldom,myaddress);
+      type(NULL,0,NULL," policytestaddr: lcldom/myaddress=%d/%d",lcldom,myaddress);
 
     /* state->request initialization !! */
 
@@ -763,7 +828,9 @@ int sourceaddr;
 			 1 << P_A_OutboundSizeLimit |
 			 1 << P_A_FullTrustNet      |
 			 1 << P_A_TrustRecipients   |
-			 1 << P_A_TrustWhosOn        );
+			 1 << P_A_TrustWhosOn       |
+			 1 << P_A_Filtering         |
+			 1 << P_A_MaxSameIpSource    );
     if (!myaddress)
       state->request |= ( 1 << P_A_TestDnsRBL       |
 			  1 << P_A_RcptDnsRBL        );
@@ -785,8 +852,19 @@ int sourceaddr;
       if (state->values[P_A_RcptDnsRBL]) free(state->values[P_A_RcptDnsRBL]);
       state->values[P_A_RcptDnsRBL] = NULL;
 
+      if (state->values[P_A_Filtering]) {
+	if (debug)
+	  type(NULL,0,NULL," policytestaddr: 'filter %s' found",
+		 state->values[P_A_Filtering]);
+	if (valueeq(state->values[P_A_Filtering], "+")) {
+	  state->content_filter = 1;
+	} else {
+	  state->content_filter = 0;
+	}
+      }
+
       if (debug)
-	printf("000-   %s\n", showresults(state->values));
+	type(NULL,0,NULL," Results:  %s", showresults(state));
 
       return 0;
     }
@@ -829,7 +907,7 @@ int sourceaddr;
 
     if (valueeq(state->values[P_A_REJECTNET], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'rejectnet +' found\n");
+	type(NULL,0,NULL," policytestaddr: 'rejectnet +' found");
       PICK_PA_MSG(P_A_REJECTNET);
       if (state->message == NULL)
 	state->message = strdup("Your network address is blackholed in our static tables");
@@ -838,7 +916,7 @@ int sourceaddr;
     }
     if (valueeq(state->values[P_A_FREEZENET], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'freezenet +' found\n");
+	type(NULL,0,NULL," policytestaddr: 'freezenet +' found");
       PICK_PA_MSG(P_A_FREEZENET);
       if (state->message == NULL)
 	state->message = strdup("Your network address is blackholed in our static tables");
@@ -847,20 +925,20 @@ int sourceaddr;
     }
     if (valueeq(state->values[P_A_TrustRecipients], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'trustrecipients +' found\n");
+	type(NULL,0,NULL," policytestaddr: 'trustrecipients +' found");
       state->trust_recipients = 1;
       PICK_PA_MSG(P_A_TrustRecipients);
     }
     if (valueeq(state->values[P_A_FullTrustNet], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'fulltrustnet +' found\n");
+	type(NULL,0,NULL," policytestaddr: 'fulltrustnet +' found");
       state->full_trust = 1;
       PICK_PA_MSG(P_A_FullTrustNet);
     }
 #ifdef HAVE_WHOSON_H
     if (valueeq(state->values[P_A_TrustWhosOn], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'trust-whoson +' found, accept? = %d\n",
+	type(NULL,0,NULL," policytestaddr: 'trust-whoson +' found, accept? = %d",
 	       (state->whoson_result == 0));
       if (state->whoson_result == 0)
 	state->always_accept = 1;
@@ -869,43 +947,92 @@ int sourceaddr;
 #endif
     if (valueeq(state->values[P_A_RELAYCUSTNET], "+")) {
       if (debug)
-	printf("000- policytestaddr: 'relaycustnet +' found\n");
+	type(NULL,0,NULL," policytestaddr: 'relaycustnet +' found");
       state->always_accept = 1;
-      PICK_PA_MSG(P_A_TrustWhosOn);
+      PICK_PA_MSG(P_A_RELAYCUSTNET);
     }
+
+    if (state->values[P_A_Filtering]) {
+      if (debug)
+	type(NULL,0,NULL," policytestaddr: 'filter %s' found",
+	       state->values[P_A_Filtering]);
+      if (valueeq(state->values[P_A_Filtering], "+")) {
+	state->content_filter = 1;
+      } else {
+	state->content_filter = 0;
+      }
+    }
+
     if (state->trust_recipients || state->full_trust || state->always_accept)
       return 0;
 
-    just_rbl_checks:;
+ just_rbl_checks:;
+
+    if (state->values[P_A_Filtering]) {
+      if (debug)
+	type(NULL,0,NULL," policytestaddr: 'filter %s' found",
+	       state->values[P_A_Filtering]);
+      if (valueeq(state->values[P_A_Filtering], "+")) {
+	state->content_filter = 1;
+      } else {
+	state->content_filter = 0;
+      }
+    }
 
     if (state->values[P_A_TestDnsRBL] &&
 	!valueeq(state->values[P_A_TestDnsRBL], "-")) {
       int rc;
       if (debug)
-	printf("000- policytestaddr: 'test-dns-rbl %s' found;\n",
+	type(NULL,0,NULL," policytestaddr: 'test-dns-rbl %s' found;",
 	       state->values[P_A_TestDnsRBL]);
       rc = rbl_dns_test(ipaf, ipaddr, state->values[P_A_TestDnsRBL], &state->message);
       if (!state->message){ PICK_PA_MSG(P_A_TestDnsRBL); }
 
       if (debug)
-	printf("000-  rc=%d\n", rc);
+	type(NULL,0,NULL,"  rc=%d; msg='%s'",
+	     rc, state->message ? state->message : "<nil>");
+
       return rc;
     }
+
+    /* bag = Andrey Blochintsev <bag@iptelecom.net.ua>  */
+    /* bag + */
+    if (state->values[P_A_RcptDnsRBL] &&
+	state->values[P_A_RcptDnsRBL][0] == '_') {
+      int rc = 1;
+      if (debug)
+	type(NULL,0,NULL," policytestaddr: 'rcpt-dns-rbl %s' found;",
+	       state->values[P_A_RcptDnsRBL]);
+      if (state->values[P_A_RcptDnsRBL][1] != '+') {
+      	if (state->rblmsg != NULL)
+	  free(state->rblmsg);
+	state->rblmsg = strdup(state->values[P_A_RcptDnsRBL] + 1);
+	rc = 0;
+      }
+      if (!state->message){ PICK_PA_MSG(P_A_RcptDnsRBL); }
+      if (debug)
+	type(NULL,0,NULL,"  rc=%d", rc);
+      return 0; /* We report error LATER */
+
+    }
+    /* bag - */
+
     if (state->values[P_A_RcptDnsRBL] &&
 	!valueeq(state->values[P_A_RcptDnsRBL], "-")) {
       int rc;
       if (debug)
-	printf("000- policytestaddr: 'rcpt-dns-rbl %s' found;\n",
+	type(NULL,0,NULL," policytestaddr: 'rcpt-dns-rbl %s' found;",
 	       state->values[P_A_RcptDnsRBL]);
       rc = rbl_dns_test(ipaf, ipaddr, state->values[P_A_RcptDnsRBL], &state->rblmsg);
-#if 0
-      type(NULL, 0, NULL, "rcpt-dns-rbl test yiels: rc=%d rblmsg='%s'",rc,
-	   state->rblmsg ? state->rblmsg : "<none>");
-#endif
+
+      if (debug)
+	type(NULL, 0, NULL, "rcpt-dns-rbl test yields: rc=%d rblmsg='%s'", rc,
+	     state->rblmsg ? state->rblmsg : "<none>");
+
       if (!state->message){ PICK_PA_MSG(P_A_RcptDnsRBL); }
       if (debug)
-	printf("000-  rc=%d\n", rc);
-      return 0; /* We report error LATTER */
+	type(NULL,0,NULL,"  rc=%d", rc);
+      return 0; /* We report error LATER */
     }
     return 0;
 }
@@ -965,12 +1092,14 @@ Usockaddr *raddr;
     } else
 #endif
     {
-      printf("Unknown address format; sa_family = %d\n",
-	     raddr->v4.sin_family);
+      type(NULL,0,NULL,"Unknown address format; sa_family = %d",
+	   raddr->v4.sin_family);
       return -2;
     }
 
     state->request = 0;
+    state->content_filter = -1;
+
     rc = _addrtest_(rel, state, pbuf, 1);
     if (debug) fflush(stdout);
     return rc;
@@ -1050,7 +1179,7 @@ int inlen;
 
     while (result != 0) {
 	if (debug)
-	  printf("000- DEBUG: %s\n", showkey(pbuf));
+	  type(NULL,0,NULL," DEBUG: %s", showkey(pbuf));
 	result = checkaddr(rel, state, pbuf);
 
 	if (result == 0) /* Found! */
@@ -1174,27 +1303,50 @@ const int len;
     if (state->full_trust)
 	return 0;
 
-    /* state->request initialization !! */
-    state->request = ( 1 << P_A_REJECTNET    |
-		       1 << P_A_FREEZENET  );
+    /*
+     * This is somewhat controversial.
+     * This exists solely to allow simplification of
+     * smtpserver.conf  file by having the HELO/EHLO
+     * strings to be rejected stored in the policy
+     * database.
+     *
+     * In Sep-2001 it became apparent that very least
+     * there is no point in analyzing '[...]' numeric
+     * HELO parameter contained address data.
+     *
+     * Current code will only look for the input string
+     * in the database (by domain lookup protocol), and
+     * react on it by smalling the door shut (if any).
+     *
+     */
 
-    check_domain(rel, state, str, len);
+
+    if (*str != '[') { /* Don't test address literals! */
+
+      /* state->request initialization !! */
+      state->request = ( 1 << P_A_REJECTNET    |
+			 1 << P_A_FREEZENET  );
+
+      check_domain(rel, state, str, len);
 
 /*
    # if (name of SMTP client has 'rejectnet +' attribute) then
    #    any further conversation refused
    #      [state->always_reject = 1; return -1;]
  */
-    if (valueeq(state->values[P_A_REJECTNET], "+")) {
+      if (valueeq(state->values[P_A_REJECTNET], "+")) {
 	state->always_reject = 1;
 	PICK_PA_MSG(P_A_REJECTNET);
 	return -1;
-    }
-    if (valueeq(state->values[P_A_FREEZENET], "+")) {
+      }
+      if (valueeq(state->values[P_A_FREEZENET], "+")) {
 	state->always_freeze = 1;
 	PICK_PA_MSG(P_A_FREEZENET);
 	return  1;
+      }
+
     }
+
     return 0;
 }
 
@@ -1218,7 +1370,8 @@ const int len;
 		       1 << P_A_FREEZENET    |
 		       1 << P_A_RELAYCUSTNET |
 		       1 << P_A_InboundSizeLimit  |
-		       1 << P_A_OutboundSizeLimit   );
+		       1 << P_A_OutboundSizeLimit  );
+    state->request |= ( 1 << P_A_RcptDnsRBL );	/* bag */
 
     check_domain(rel, state, str, len);
 
@@ -1239,18 +1392,33 @@ const int len;
     }
     if (valueeq(state->values[P_A_RELAYCUSTNET], "+")) {
       if (debug)
-	printf("000- pt_sourceaddr: 'relaycustnet +' found\n");
+	type(NULL,0,NULL," pt_sourceaddr: 'relaycustnet +' found");
       state->always_accept = 1;
       PICK_PA_MSG(P_A_RELAYCUSTNET);
       return  0;
     }
     if (valueeq(state->values[P_A_FullTrustNet], "+")) {
       if (debug)
-	printf("000- pt_sourceaddr: 'fulltrustnet +' found\n");
+	type(NULL,0,NULL," pt_sourceaddr: 'fulltrustnet +' found");
       state->full_trust = 1;
       PICK_PA_MSG(P_A_FullTrustNet);
       return  0;
     }
+    /* bag + */
+    if (state->rblmsg == NULL &&	/* only if no rbl_message before (from lookup by net) */
+	state->values[P_A_RcptDnsRBL] &&
+	state->values[P_A_RcptDnsRBL][0] == '_') {
+      if (debug)
+	type(NULL,0,NULL," pt_sourceaddr: 'rcpt-dns-rbl %s' found;",
+	       state->values[P_A_RcptDnsRBL]);
+      if (state->values[P_A_RcptDnsRBL][1] != '+') {
+	state->rblmsg = strdup(state->values[P_A_RcptDnsRBL] + 1);
+      }
+      free(state->values[P_A_RcptDnsRBL]);
+      return 0; /* We report error LATER */
+
+    }
+    /* bag - */
     return 0;
 }
 
@@ -1261,6 +1429,7 @@ const char *str;
 const int len;
 {
     const char *at;
+    int requestmask = 0;
 
     state->rcpt_nocheck  = 0;
     state->sender_reject = 0;
@@ -1289,18 +1458,23 @@ const int len;
     if (check_user(rel, state, str, len) == 0) {
       if (valueeq(state->values[P_A_FREEZESOURCE], "+")) {
 	if (debug)
-	  printf("000- mailfrom: 'freezesource +'\n");
+	  type(NULL,0,NULL," mailfrom: 'freezesource +'");
 	state->sender_freeze = 1;
 	PICK_PA_MSG(P_A_FREEZESOURCE);
 	return 1;
       }
+      if (state->values[P_A_FREEZESOURCE])
+	requestmask |= 1 << P_A_FREEZESOURCE;
+
       if (valueeq(state->values[P_A_REJECTSOURCE], "+")) {
 	if (debug)
-	  printf("000- mailfrom: 'rejectsource +'\n");
+	  type(NULL,0,NULL," mailfrom: 'rejectsource +'");
 	state->sender_reject = 1;
 	PICK_PA_MSG(P_A_REJECTSOURCE);
 	return -1;
       }
+      if (state->values[P_A_REJECTSOURCE])
+	requestmask |= 1 << P_A_REJECTSOURCE;
     }
 
     state->request = ( 1 << P_A_REJECTSOURCE  |
@@ -1309,7 +1483,7 @@ const int len;
 		       1 << P_A_RELAYCUSTOMER |
 #endif
 		       1 << P_A_SENDERNoRelay |
-		       1 << P_A_SENDERokWithDNS );
+		       1 << P_A_SENDERokWithDNS ) & (~ requestmask);
 
     at = find_nonqchr(str, '@', len);
     if (at != NULL) {
@@ -1323,7 +1497,7 @@ const int len;
 
     if (valueeq(state->values[P_A_SENDERNoRelay], "+")) {
       if (debug)
-	printf("000- mailfrom: 'sendernorelay +'\n");
+	type(NULL,0,NULL," mailfrom: 'sendernorelay +'");
       state->sender_norelay = 1;
       PICK_PA_MSG(P_A_SENDERNoRelay);
     }
@@ -1333,21 +1507,21 @@ const int len;
       int rc = sender_dns_verify(state->values[P_A_SENDERokWithDNS][0],
 				 at+1, len - (1 + at - str));
       if (debug)
-	printf("000- ... returns: %d\n", rc);
+	type(NULL,0,NULL," ... returns: %d", rc);
       PICK_PA_MSG(P_A_SENDERokWithDNS);
       return rc;
     }
 
     if (valueeq(state->values[P_A_REJECTSOURCE], "+")) {
 	if (debug)
-	  printf("000- mailfrom: 'rejectsource +'\n");
+	  type(NULL,0,NULL," mailfrom: 'rejectsource +'");
 	state->sender_reject = 1;
 	PICK_PA_MSG(P_A_REJECTSOURCE);
 	return -1;
     }
     if (valueeq(state->values[P_A_FREEZESOURCE], "+")) {
 	if (debug)
-	  printf("000- mailfrom: 'freezesource +'\n");
+	  type(NULL,0,NULL," mailfrom: 'freezesource +'");
 	state->sender_freeze = 1;
 	PICK_PA_MSG(P_A_FREEZESOURCE);
 	return -1;
@@ -1357,13 +1531,13 @@ const int len;
       /* Always accept, and not an address literal! */
       int rc = sender_dns_verify('-', at+1, len - (1 + at - str));
       if (debug)
-	printf("000- ... returns: %d\n", rc);
+	type(NULL,0,NULL," ... returns: %d", rc);
       return rc;
     }
 #if 0 /* Eh..., NOT! */
     if (valueeq(state->values[P_A_RELAYCUSTOMER], "+")) {
 	if (debug)
-	  printf("000- mailfrom: 'relaycustomer +'\n");
+	  type(NULL,0,NULL," mailfrom: 'relaycustomer +'");
 	state->rcpt_nocheck = 1;
 	PICK_PA_MSG(P_A_RELAYCUSTOMER);
 	return  0;
@@ -1424,7 +1598,7 @@ const int len;
 	  if (state->message != NULL) free(state->message);
 	  state->message = strdup(state->rblmsg);
 	  if (debug)
-	    printf("000- ... TestRcptDnsRBL has a message: '%s'\n",
+	    type(NULL,0,NULL," ... TestRcptDnsRBL has a message: '%s'",
 		   state->rblmsg);
 	  return -1;
 	}
@@ -1534,6 +1708,15 @@ const int len;
       break; /* Ok, could be ok, but RBL may say differently ... */
     }
 
+
+    /* Do target specific rejects early */
+
+    if (valueeq(state->values[P_A_RELAYTARGET], "-")) {
+      PICK_PA_MSG(P_A_RELAYTARGET);
+      return -1;
+    }
+
+
     if (valueeq(state->values[P_A_ACCEPTbutFREEZE], "+")) {
 	state->sender_freeze = 1;
 	PICK_PA_MSG(P_A_ACCEPTbutFREEZE);
@@ -1552,7 +1735,7 @@ const int len;
 	if (state->message != NULL) free(state->message);
 	state->message = strdup(state->rblmsg);
 	if (debug)
-	  printf("000- ... TestRcptDnsRBL has a message: '%s'\n",
+	  type(NULL,0,NULL," ... TestRcptDnsRBL has a message: '%s'",
 		 state->rblmsg);
 	return -1;
       }
@@ -1565,19 +1748,20 @@ const int len;
 
     if (state->rcpt_nocheck) {
       if (debug)
-	printf("000- ... rcpt_nocheck is on!\n");
+	type(NULL,0,NULL," ... rcpt_nocheck is on!");
       return 0;
     }
 
     if (state->always_accept) {
       int rc, c = '-';
+
       if (state->values[P_A_ACCEPTifMX]) {
 	c = state->values[P_A_ACCEPTifMX][0];
       }
       rc = client_dns_verify(c, at+1, len - (1 + at - str));
       /* XX: state->message setup! */
       if (debug)
-	printf("000- ... returns: %d\n", rc);
+	type(NULL,0,NULL," ... returns: %d", rc);
       PICK_PA_MSG(P_A_ACCEPTifMX);
       return rc;
     }
@@ -1587,26 +1771,22 @@ const int len;
       int rc = mx_client_verify(c, at+1, len - (1 + at - str)); 
       /* XX: state->message setup! */
       if (debug)
-	printf("000- ...(mx_client_verify('%.*s')) returns: %d\n",
+	type(NULL,0,NULL," ...(mx_client_verify('%.*s')) returns: %d",
 	       (int)(len - (1 + at - str)), at+1, rc);
       PICK_PA_MSG(P_A_ACCEPTifMX);
       return rc;
     }
+
     if (state->values[P_A_ACCEPTifDNS]) {
       int rc = client_dns_verify(state->values[P_A_ACCEPTifDNS][0],
 				 at+1, len - (1 + at - str));
       /* XX: state->message setup! */
       if (debug)
-	printf("000- ... returns: %d\n", rc);
+	type(NULL,0,NULL," ... returns: %d", rc);
       PICK_PA_MSG(P_A_ACCEPTifDNS);
       return rc;
     }
 
-    if (valueeq(state->values[P_A_RELAYTARGET], "-")) {
-      PICK_PA_MSG(P_A_RELAYTARGET);
-      return -1;
-    }
-
     return 0;
 }
 
@@ -1644,7 +1824,7 @@ const int len;
       state->authuser = (char*)authuser;
 
     if (debug) {
-	printf("000- policytest what=%d\n", what);
+	type(NULL,0,NULL," policytest what=%d", what);
 	printstate(state);
     }
 
@@ -1691,3 +1871,11 @@ struct policystate *state;
 {
     return state->maxinsize;
 }
+
+long
+policysameiplimit(rel, state)
+struct policytest *rel;
+struct policystate *state;
+{
+    return state->maxsameiplimit;
+}
