@@ -3,7 +3,7 @@
  *  Added for ZMailer 2.99.44 by Matti Aarnio on 16-Dec-1996;
  *  part of ZMailer.
  *
- *  by Matti Aarnio <mea@nic.funet.fi> 1996-1999
+ *  by Matti Aarnio <mea@nic.funet.fi> 1996-1999,2003
  */
 
 /* Pre-included before including this:
@@ -25,6 +25,7 @@ struct policystate {		/* Part of SmtpState structure */
     int relaycustnet;
     int rcpt_nocheck;
     int sender_norelay;
+    int content_filter;
     char *authuser;
 
     /* This variable contains bitmapped flags of attributes to be checked. */
@@ -43,6 +44,7 @@ struct policystate {		/* Part of SmtpState structure */
     char *rblmsg;
     long maxinsize;
     long maxoutsize;
+    long maxsameiplimit;
     int  islocaldomain;
 #ifdef HAVE_WHOSON_H
     int whoson_result;
@@ -72,13 +74,13 @@ struct policytest {
     char *dbpath;
     dbtypes dbt;
     union {
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	DBM *_ndbm;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	GDBM_FILE _gdbm;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#if defined(HAVE_DB1) || defined(HAVE_DB2) || defined(HAVE_DB3)
 	DB *_db;
 #endif
     } db_;
@@ -117,6 +119,7 @@ extern int policytest __((struct policytest * rel, struct policystate * ps, Poli
 extern int policytestaddr __((struct policytest * rel, struct policystate * ps, PolicyTest how, Usockaddr * raddr));
 extern char *policymsg __((struct policytest *rel, struct policystate *ps));
 extern long  policyinsizelimit __((struct policytest *rel, struct policystate *ps));
+extern long  policysameiplimit __((struct policytest *rel, struct policystate *ps));
 
 extern struct policytest *policydb;
 
