@@ -119,6 +119,8 @@ int type;
 {
     int len = strlen(str) + 1;	/* Terminating \000 included! */
 
+    if (len > (255-2)) return -2; /* BAD BAD!! */
+
     memcpy(pbuf->data, str, len);
     pbuf->len = len + 2;
     pbuf->type = type;
@@ -142,10 +144,10 @@ char *str;
 	*s++ = 0;
 	return parse_ip_policy(pbuf, str, s);
     }
-    if (*str == '_')
-	return parse_gen_policy(pbuf, str, P_K_TAG);
     if (strchr(str,'@') != NULL)
 	return parse_gen_policy(pbuf, str, P_K_USER);
+    if (*str == '_')
+	return parse_gen_policy(pbuf, str, P_K_TAG);
 
     return parse_gen_policy(pbuf, str, P_K_DOMAIN);
 }
@@ -179,6 +181,8 @@ struct _tokenpair {
 #ifdef HAVE_WHOSON_H
   { "trust-whoson",	P_A_TrustWhosOn		},
 #endif
+  { "filtering",	P_A_Filtering		},
+  { "maxsameipsource",	P_A_MaxSameIpSource	},
   { NULL, 0 },
 };
 
@@ -188,8 +192,11 @@ struct attribute *abuf;
 char *str1, *str2;
 {
     struct _tokenpair *ap = attributes;
+    int len = 2 + strlen(str2) + 1;
 
-    abuf->len = 2 + strlen(str2) + 1;
+    if (len > (255-2)) return -2;
+
+    abuf->len = len;
     strcpy((char *)(abuf->data), str2);
 
     while (ap->name != NULL) {
