@@ -1,7 +1,7 @@
 /* Small pieces for scanning forward on a buffer of RFC-821/822 compliant
    addresses */
 
-/* (c) Matti Aarnio 1993-1997 <mea@nic.funet.fi> */
+/* (c) Matti Aarnio 1993-2000 <mea@nic.funet.fi> */
 
 /* All these routines scan over the lexical elements they are after, and
    if successfull, return pointer just AFTER such element.
@@ -100,11 +100,11 @@ static
 /*  30: `^^'   */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 0),
 /*  31: `^_'   */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 0),
 /*  32: ` '    */ CHR_ENT(0, 0, 0, 0, 0, 1, 1, 0, 0, 0),
-/*  33: `!'    */ CHR_ENT(0, 0, 0, 0, 0, 1, 1, 0, 1, 1),
+/*  33: `!'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  34: `"'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 0, 0, 1, 1),
 /*  35: `#'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  36: `$'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
-/*  37: `%'    */ CHR_ENT(0, 0, 0, 0, 0, 1, 1, 0, 1, 1),
+/*  37: `%'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  38: `&'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  39: `''    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  40: `('    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 1),
@@ -513,7 +513,7 @@ int strict;
      *    Paul Vixie, 1996.
      */
 
-#ifndef IN6ADDRSZ		/* Propably these all set at the same time.. */
+#ifndef IN6ADDRSZ		/* Probably these all set at the same time.. */
 #define IN6ADDRSZ 16
 #define INADDRSZ   4
 #define INT16SZ    2
@@ -708,9 +708,6 @@ int strict;
     }
     if (*p == '[') {
 	q = rfc821_dotnum(p + 1, strict);
-#if 0
- printf("    dotnum: p='%s', q='%s'\n",p,q); 
-#endif
 	if (q == p + 1)
 	    return s;
 	if (*q != ']') {
@@ -950,7 +947,7 @@ char *s;			/* Stretched RFC821 a bit here.. !%-hacks */
 int strict;
 {
     char *p = s, *q;
-    int _ok = 0;
+    /* int _ok = 0; */
 
     if (!s || !*s)
 	return s;		/* Pathological termination */
@@ -959,6 +956,8 @@ int strict;
 	    q = rfc821_quoted_string(p, strict);
 	    if (q == p)
 		return s;	/* Uh... */
+#if 0 /* RFC 2821 -- local part internal hack-syntax interpretation
+	 shall not be done, unless domain is known to be local */
 	    if (*q == '%' || *q == '!') {
 		p = q + 1;
 		_ok = (*q == '!' && *p == '_');
@@ -973,11 +972,14 @@ int strict;
 		}
 		continue;
 	    }
+#endif
 	    return q;
 	}
 	q = rfc821_dot_string(p, strict);
 	if (q == p)
 	    return s;		/* Uh... */
+#if 0 /* RFC 2821 -- local part internal hack-syntax interpretation
+	 shall not be done, unless domain is known to be local */
 	if (*q == '%' || *q == '!') {
 	    p = q + 1;
 	    _ok = (*q == '!' && *p == '_');
@@ -992,6 +994,7 @@ int strict;
 	    }
 	    continue;
 	}
+#endif
 	return q;
     }
     return p;			/* Run to end of string.. */
