@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-2000.
+ *      Copyright 1991-2003.
  */
 /*
  * Zmailer SMTP-server divided into bits
@@ -21,7 +21,6 @@
 
 #include "smtpserver.h"
 
-extern int netconnected_flg;
 extern int do_whoson;
 
 static const char *orcpt_string __((const char *));
@@ -55,9 +54,13 @@ const char *str;
   fputc('"', mfp);
   for ( ; *str ; ++str ) {
     int c = (*str) & 0xFF;
-    if (c == '"' || c == '\\' || c == '(' || c == ')')
-      fputc('\\', mfp);
-    fputc(c, mfp);
+    if (c < 0x20 || c > 0x7e)
+      fprintf(mfp,"\\%03o", c);
+    else {
+      if (c == '"' || c == '\\' || c == '(' || c == ')')
+	fputc('\\', mfp);
+      fputc(c, mfp);
+    }
   }
   fputc('"', mfp);
 }
@@ -134,6 +137,18 @@ SmtpState *SS;
 const char *buf, *cp;
 {
     const char *msg = NULL;
+
+    switch (SS->carp->cmd) {
+    case Hello2:
+      MIBMtaEntry->ss.IncomingSMTP_EHLO += 1;
+      break;
+    case Hello:
+      MIBMtaEntry->ss.IncomingSMTP_HELO += 1;
+      break;
+    default: /* Should not happen... */
+      break;
+    }
+
     if (SS->state != Hello && SS->state != MailOrHello) {
 	switch (SS->state) {
 	case Mail:
@@ -162,13 +177,21 @@ const char *buf, *cp;
     else
       while (*cp == ' ' || *cp == '\t') ++cp;
 
-    SS->policyresult = policytest(policydb, &SS->policystate,
-				  POLICY_HELONAME, cp, strlen(cp),
-				  SS->authuser);
+    if (debug) typeflush(SS);
+    if (SS->netconnected_flg)
+      SS->policyresult = policytest(policydb, &SS->policystate,
+				    POLICY_HELONAME, cp, strlen(cp),
+				    SS->authuser);
+    else
+      SS->policyresult = 0;
+
+    if (logfp || logfp_to_syslog) time( & now );
+
     if (logfp) {
       char *s = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || s != NULL) {
-	fprintf(logfp, "%s#\t-- policy result=%d, msg: %s\n", logtag,
+	fprintf(logfp, "%s%04d#\t-- policy result=%d, msg: %s\n", logtag,
+		(int)(now - logtagepoch),
 		SS->policyresult, (s ? s : "<NONE!>"));
 	fflush(logfp);
       }
@@ -176,7 +199,8 @@ const char *buf, *cp;
     if (logfp_to_syslog) {
       char *s = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || s != NULL) {
-	zsyslog((LOG_DEBUG, "%s # policy result=%d, msg: %s", logtag,
+	zsyslog((LOG_DEBUG, "%s%04d # policy result=%d, msg: %s", logtag,
+		(int)(now - logtagepoch),
 		SS->policyresult, (s ? s : "<NONE!>")));
       }
     }
@@ -228,10 +252,21 @@ const char *buf, *cp;
     /* Router interacting policy analysis functions */
     /* Note, these are orthogonal to those of smtp-server
        internal policy functions! */
+
+#if 0 /* Original security reviewer reported here a potential
+	 for a buffer overflow, however there are potentially
+	 more severe things lurking in the interactive router
+	 call with these parameters, than mere buffer overflow.
+	 Observe the quotation issues...
+	 This code is now buried into '#if 0', until the
+	 real fix is written -- if ever..
+	 This code has never been in active use...
+      */
+
     if (STYLE(SS->cfinfo, 'h')) {
-      char argbuf[MAXHOSTNAMELEN+30];
+      char argbuf[100+100];
       char *s;
-      sprintf(argbuf,"%s %s", SS->rhostname,
+      sprintf(argbuf,"%.99s %.99s", SS->rhostname,
 	      ((SS->ihostaddr && (SS->ihostaddr[0] != '\0'))
 	       ? SS->ihostaddr : "[0.0.0.0]"));
       if ((s = router(SS, RKEY_HELLO, 1, argbuf, strlen(argbuf))) == NULL)
@@ -249,6 +284,7 @@ const char *buf, *cp;
 	free(s);
       }
     }
+#endif
 
     /* Check `cp' corresponds to the reverse address */
     if (skeptical && SS->ihostaddr[0] != '\0'
@@ -315,13 +351,8 @@ const char *buf, *cp;
 	if (rcptlimitcnt > 100)
 	  type(SS, -250, NULL, "X-RCPTLIMIT %d", rcptlimitcnt);
 
-	if (auth_login_without_tls || SS->sslmode) {
-	  if (auth_ok)
-	    type(SS, -250, NULL, "AUTH=LOGIN"); /* RFC 2554, NetScape/
-						   Sun Solstice/ ? */
-	  if (auth_ok)
-	    type(SS, -250, NULL, "AUTH LOGIN"); /* RFC 2554, M$ Exchange ? */
-	}
+	smtpauth_ehloresponse(SS);
+
 #ifdef HAVE_OPENSSL
 	/* NOTE: This seems to require TLS and STARTTLS facilities,
 	   better known as SSL..  TLS: RFC 2246, STARTTLS: RFC 2487 */
@@ -338,7 +369,7 @@ const char *buf, *cp;
     SS->state = MailOrHello;
 }
 
-void smtp_mail(SS, buf, cp, insecure)
+int smtp_mail(SS, buf, cp, insecure)
 SmtpState *SS;
 const char *buf, *cp;
 int insecure;
@@ -355,6 +386,7 @@ int insecure;
     int strict = STYLE(SS->cfinfo, 'R');
     int sloppy = STYLE(SS->cfinfo, 'S');
 
+
     if (strict && sloppy) /* If misconfigured, SLOPPY takes precedence! */
       strict = 0;
 
@@ -393,7 +425,7 @@ int insecure;
 	}
 	smtp_tarpit(SS);
 	type(SS, 503, m551, cp);
-	return;
+	return -1;
     }
 
     if (*cp == ' ') ++cp;
@@ -401,7 +433,7 @@ int insecure;
     if (!CISTREQN(cp, "From:", 5)) {
 	smtp_tarpit(SS);
 	type(SS, 501, m552, "where is From: in that?");
-	return;
+	return -1;
     }
     cp += 5;
     if (!strict_protocol || sloppy)
@@ -411,23 +443,23 @@ int insecure;
 	  if (!sloppy) {
 	    smtp_tarpit(SS);
 	    type(SS, 501, m517, "where is <...> in that?");
-	    return;
+	    return -1;
 	  }
 	  break; /* Sigh, be sloppy.. */
 	}
     if (*cp == '\0') {
 	smtp_tarpit(SS);
 	type(SS, 501, m517, "where is <...> in this: %s", cp);
-	return;
+	return -1;
     } else if (*cp != '<' && !sloppy) {
 	smtp_tarpit(SS);
 	type(SS, 501, m517, "strangeness between ':' and '<': %s", cp);
-	return;
+	return -1;
     }
     if (*(cp + 1) == '<') {
 	smtp_tarpit(SS);
 	type(SS, 501, m517, "there are too many <'s in this: %s", cp);
-	return;
+	return -1;
     }
     /* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
     if (*cp == '<') {
@@ -437,12 +469,12 @@ int insecure;
 	  /* Failure.. ? */
 	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
-	  return;
+	  return -1;
 	}
 	if (*s == '>') {
 	  smtp_tarpit(SS);
 	  type(SS, 501, m517, "there are too many >'s in this: <%s", cp);
-	  return;
+	  return -1;
 	}
 	/* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
 	if (cp[1] == '@') {
@@ -463,7 +495,7 @@ int insecure;
 	  /* Failure.. ? */
 	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
-	  return;
+	  return -1;
 	}
 	/* Now it is a moment to see, if we have source routes: @a,@b:c@d */
 	if (*cp == '@') {
@@ -481,7 +513,7 @@ int insecure;
 	  rfc821_error_ptr = s;
 	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Missing ending '>' bracket");
-	  return;
+	  return -1;
 	}
 	++s;
       }
@@ -493,13 +525,13 @@ int insecure;
 	/* Failure.. */
 	type(SS, -501, m517, "For input: %s", cp);
 	type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
-	return;
+	return -1;
       }
 
       if (*s == '>') {
 	smtp_tarpit(SS);
 	type(SS, 501, m517, "there are too many >'s in that!");
-	return;
+	return -1;
       }
 
       /* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
@@ -531,7 +563,7 @@ int insecure;
 	    if (drpt_ret) {
 		smtp_tarpit(SS);
 		type(SS, 501, m554, "RET-param double defined!");
-		return;
+		return -1;
 	    }
 	    s += 4;
 	    drpt_ret = s;
@@ -541,9 +573,10 @@ int insecure;
 	    if (*s && *s != ' ' && *s == '\t') {
 		smtp_tarpit(SS);
 		type(SS, 501, m454, "RET-param data error");
-		return;
+		return -1;
 	    }
 	    drptret_len = (s - drpt_ret);
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_RET ++;
 	    continue;
 	}
 	if (mime8bitok && CISTREQN("BODY=", s, 5)) {
@@ -558,12 +591,15 @@ int insecure;
 	    if (CISTREQN(s, "8BITMIME", 8)) {
 		bodytype = "8BITMIME";
 		s += 8;
+		MIBMtaEntry->ss.IncomingSMTP_OPT_BODY_8BITMIME ++;
 	    } else if (CISTREQN(s, "BINARYMIME", 10)) {
 		bodytype = "BINARYMIME";
 		s += 10;
+		MIBMtaEntry->ss.IncomingSMTP_OPT_BODY_BINARYMIME ++;
 	    } else if (CISTREQN(s, "7BIT", 4)) {
 		bodytype = "7BIT";
 		s += 4;
+		MIBMtaEntry->ss.IncomingSMTP_OPT_BODY_7BIT ++;
 	    }
 	    if (*s && *s != ' ' && *s != '\t') {
 		smtp_tarpit(SS);
@@ -584,19 +620,20 @@ int insecure;
 	    /* This data we use, gather the value */
 	    SS->sizeoptval = 0;
 	    while (isascii(*s) && isdigit(*s)) {
+	        if (SS->sizeoptval > 100000000) /* Next to exceed 1G ? */
+		  break;
 		SS->sizeoptval *= 10;
 		SS->sizeoptval += (*s - '0');
 		++s;
-		if (SS->sizeoptval > 999999999) /* 1GB or more? */
-		  break;
 	    }
 	    if ((*s && *s != ' ' && *s != '\t') ||
-		(SS->sizeoptval > 999999999) /* 1GB */ ) {
+		(SS->sizeoptval > 1000000000) /* 1GB */ ) {
 		smtp_tarpit(SS);
 		type(SS, 501, m554, "SIZE-param data error");
 		rc = 1;
 		break;
 	    }
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_SIZE ++;
 	    continue;
 	}
 	/* IETF-NOTARY  SMTP-DSN extensions */
@@ -625,6 +662,7 @@ int insecure;
 		rc = 1;
 		break;
 	    }
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_ENVID ++;
 	    continue;
 	}
 	if (auth_ok && CISTREQN("AUTH=", s, 5)) {
@@ -647,6 +685,7 @@ int insecure;
 		rc = 1;
 		break;
 	    }
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_AUTH ++;
 	    continue;
 	}
 	if (deliverby_ok >= 0 && CISTREQN("BY=", s, 3)) {
@@ -711,6 +750,7 @@ int insecure;
 	    SS->deliverby_time  = time(NULL) + val;
 	    SS->deliverby_flags = neg;
 	    s = p;
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_DELIVERBY ++;
 	    continue;
 	}
 
@@ -720,15 +760,19 @@ int insecure;
 	break;
     }
     if (rc != 0)
-	return;			/* Error(s) in previous loop.. */
+	return -1;		/* Error(s) in previous loop.. */
 
     /*printf("  <path>: len=%d \"%s\"\n",cp-s,cp); */
 
     RFC821_822QUOTE(cp, newcp, addrlen);
 
-    SS->policyresult = policytest(policydb, &SS->policystate,
-				  POLICY_MAILFROM, cp, addrlen,
-				  SS->authuser);
+    if (debug) typeflush(SS);
+    if (SS->netconnected_flg)
+      SS->policyresult = policytest(policydb, &SS->policystate,
+				    POLICY_MAILFROM, cp, addrlen,
+				    SS->authuser);
+    else
+      SS->policyresult = 0;
     if (logfp || logfp_to_syslog) {
       char *ss = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || ss != NULL) {
@@ -749,80 +793,42 @@ int insecure;
 	  mcode = m571;
 	}
 	smtp_tarpit(SS);
-	type(SS, code, mcode, "For MAIL FROM address <%.*s> policy analysis reported: %s",
-	     addrlen, cp, ss);
+	type(SS, code, mcode, "Hello %s, for your MAIL FROM address <%.*s> policy analysis reported: %s",
+	     SS->ihostaddr, addrlen, cp, ss);
       } else if (SS->policyresult < -99) {
 	smtp_tarpit(SS);
 	if (SS->policyresult < -103) { /* -104 */
-	  if (!multilinereplies) {
-	    type(SS,450,m443, "For input: <%.*s> policy analysis reports temporary DNS error with your source domain.", addrlen, cp);
-	  } else {
-	    type(SS, -450, m443, "Policy analysis reports temporary DNS error");
-	    type(SS, -450, m443, "with your source domain.  Retrying may help,");
-	    type(SS, -450, m443, "or if the condition persists, you may need");
-	    type(SS,  450, m443, "to get somebody to fix your DNS servers: <%.*s>",
-		 addrlen, cp);
-	  }
+	  type(SS,450,m443, "Hello %s, for your input: <%.*s> policy analysis reports temporary DNS error with your source domain.",
+	       SS->ihostaddr, addrlen, cp);
 	} else if (SS->policyresult < -100) {
-	  if (!multilinereplies) {
-	    type(SS,450,m443, "For input: <%.*s> policy analysis reports DNS error with your source domain.", addrlen, cp);
-	  } else {
-	    type(SS, -450, m443, "Policy analysis reports DNS error with your");
-	    type(SS, -450, m443, "source domain.   Please correct your source");
-	    type(SS,  450, m443, "address and/or the info at the DNS: <%.*s>",
-		 addrlen, cp);
-	  }
+	  type(SS,450,m443, "Hello %s, for your input: <%.*s> policy analysis reports DNS error with your source domain.",
+	       SS->ihostaddr, addrlen, cp);
 	} else {
-	  if (!multilinereplies) {
-	    type(SS,450,m471, "For address <%.*s> access denied by the policy analysis functions.", addrlen, cp);
-	  } else {
-	    type(SS, -450, m471, "Access denied by the policy analysis functions.");
-	    type(SS, -450, m471, "This may be due to your source IP address,");
-	    type(SS, -450, m471, "the IP reversal domain, the data you gave for");
-	    type(SS, -450, m471, "the HELO/EHLO parameter, or address/domain you");
-	    type(SS,  450, m471, "gave at the MAIL FROM:<%.*s> address.",
-		 addrlen, cp);
-	  }
+	  type(SS,450,m471, "Hello %s, for address <%.*s> access denied by the policy analysis functions.", SS->ihostaddr, addrlen, cp);
 	}
       } else {
 	char *ss = policymsg(policydb, &SS->policystate);
 	smtp_tarpit(SS);
 	if (ss != NULL) {
-	  type(SS, 553, m571, "For input address <%.*s> Policy analysis reported: %s", addrlen, cp, ss);
+	  type(SS, 553, m571, "Hello %s, for your input address <%.*s> Policy analysis reported: %s", SS->ihostaddr, addrlen, cp, ss);
 	} else if (SS->policyresult < -1) {
-	  if (!multilinereplies) {
-	    type(SS,553,m543,"For MAIL FROM address <%.*s> the policy analysis reports DNS error with your source domain.", addrlen, cp);
-	  } else {
-	    type(SS, -553, m543, "Policy analysis reports DNS error with your");
-	    type(SS, -553, m543, "source domain.   Please correct your source");
-	    type(SS,  553, m543, "address and/or the info at the DNS: <%.*s>",
-		 addrlen, cp);
-	  }
+	  type(SS,553,m543,"Hello %s, for MAIL FROM address <%.*s> the policy analysis reports DNS error with your source domain.", SS->ihostaddr, addrlen, cp);
 	} else {
-	  if (!multilinereplies) {
-	    type(SS,553,m571,"For MAIL FROM address <%.*s> access is denied by the policy analysis functions.", addrlen, cp);
-	  } else {
-	    type(SS, -553, m571, "Access denied by the policy analysis functions.");
-	    type(SS, -553, m571, "This may be due to your source IP address,");
-	    type(SS, -553, m571, "the IP reversal domain, the data you gave for");
-	    type(SS, -553, m571, "the HELO/EHLO parameter, or address/domain you");
-	    type(SS,  553, m571, "gave at the MAIL FROM:<%.*s> address.",
-		 addrlen, cp);
-	  }
+	  type(SS,553,m571,"Hello %s, for MAIL FROM address <%.*s> access is denied by the policy analysis functions.", SS->ihostaddr, addrlen, cp);
 	}
       }
       if (newcp)
 	free((void*)newcp);
-      return;
+      return -1;
     }
     s = NULL;
-    if (/* addrlen > 0 && */ STYLE(SS->cfinfo, 'f')) {
+    if ( enable_router && /* addrlen > 0 && */ STYLE(SS->cfinfo, 'f')) {
 	s = router(SS, RKEY_FROM, 1, cp, addrlen);
 	if (s == NULL) {
 	    /* the error was printed in router() */
 	    if (newcp)
 		free((void*)newcp);
-	    return;
+	    return -1;
 	}
 	if (atoi(s) / 100 != 2) {
 	    /* verification failed */
@@ -831,10 +837,11 @@ int insecure;
 	    free((void *) s);
 	    if (newcp)
 		free((void *) newcp);
-	    return;
+	    return -1;
 	}
 	/* The 's' goes to use below */
     }
+
     if (SS->mfp == NULL &&
 	(SS->mfp = mail_open(MSG_RFC822)) == NULL) {
 	if (s)
@@ -843,17 +850,20 @@ int insecure;
 	type(SS, 452, m430, (char *) NULL);
 	if (newcp)
 	    free((void *) newcp);
-	return;
+	return -1;
     }
+    SS->messagesize = 0;
     fflush(SS->mfp);
     rewind(SS->mfp);
 #ifdef HAVE_FTRUNCATE
-    ftruncate(FILENO(SS->mfp), 0);
+    while (ftruncate(FILENO(SS->mfp), 0) < 0)
+      if (errno != EINTR && errno != EAGAIN)
+	break;
 #endif
     if (insecure)
 	fprintf(SS->mfp, "external\n");
 
-    if (netconnected_flg) {
+    if (SS->netconnected_flg) {
 
       /* Produce the 'rcvdfrom' header only when connected
 	 to network socket */
@@ -884,18 +894,19 @@ int insecure;
 	  fprintf(SS->mfp, "<none>");
 	}
       }
+#ifdef HAVE_OPENSSL
       if (SS->sslmode) {
 	if (log_rcvd_tls_mode) {
 	  fprintf(SS->mfp, " TLS-CIPHER: ");
-	  if (SS->tls_cipher_info)
-	    rfc822commentprint(SS->mfp, SS->tls_cipher_info);
+	  if (SS->TLS.cipher_info)
+	    rfc822commentprint(SS->mfp, SS->TLS.cipher_info);
 	  else
 	    fprintf(SS->mfp, "<none>");
 	}
 	if (log_rcvd_tls_peer) {
-	  fprintf(SS->mfp, " TLS-PEER: ");
-	  if (SS->tls_peer_subject)
-	    rfc822commentprint(SS->mfp, SS->tls_peer_subject);
+	  fprintf(SS->mfp, " TLS-PEER-CN1: ");
+	  if (SS->TLS.peer_subject)
+	    rfc822commentprint(SS->mfp, SS->TLS.peer_subject);
 	  else
 	    fprintf(SS->mfp, "<none>");
 	}
@@ -903,8 +914,9 @@ int insecure;
 	if (log_rcvd_tls_mode)
 	  fprintf(SS->mfp, " TLS-CIPHER: <none>");
 	if (log_rcvd_tls_peer)
-	  fprintf(SS->mfp, " TLS-PEER: <none>");
+	  fprintf(SS->mfp, " TLS-PEER-CN1: <none>");
       }
+#endif /* - HAVE_OPENSSL */
       fprintf(SS->mfp, ")\n");
 
       /* COMMENT SECTION GETTING IT ALL IN EVERY CASE! */
@@ -929,22 +941,24 @@ int insecure;
       } else {
 	fprintf(SS->mfp, "<none>");
       }
+#ifdef HAVE_OPENSSL
       if (SS->sslmode) {
 	fprintf(SS->mfp, " TLS-CIPHER: ");
-	if (SS->tls_cipher_info)
-	  rfc822commentprint(SS->mfp, SS->tls_cipher_info);
+	if (SS->TLS.cipher_info)
+	  rfc822commentprint(SS->mfp, SS->TLS.cipher_info);
 	else
 	  fprintf(SS->mfp, "<none>");
 
 	fprintf(SS->mfp, " TLS-PEER: ");
-	if (SS->tls_peer_subject)
-	  rfc822commentprint(SS->mfp, SS->tls_peer_subject);
+	if (SS->TLS.peer_subject)
+	  rfc822commentprint(SS->mfp, SS->TLS.peer_subject);
 	else
 	  fprintf(SS->mfp, "<none>");
       } else {
 	fprintf(SS->mfp, " TLS-CIPHER: <none>");
 	fprintf(SS->mfp, " TLS-PEER: <none>");
       }
+#endif /* - HAVE_OPENSSL */
       fprintf(SS->mfp, ")\n");
 
     }
@@ -984,10 +998,19 @@ int insecure;
 	fputs("\n", SS->mfp);
     }
 
-    availspace = fd_statfs(FILENO(SS->mfp));
+    availspace = used_fd_statfs(FILENO(SS->mfp));
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolUsedSpace = availspace;
+    availspace = free_fd_statfs(FILENO(SS->mfp));
     if (availspace < 0)
-	availspace = 2000000000;	/* Over 2G ? */
+	availspace = LONG_MAX / 1024;	/* Over 2G ? */
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolFreeSpace = availspace;
     availspace -= minimum_availspace;
+    if (availspace > (LONG_MAX / 1024))
+      availspace = LONG_MAX / 1024;
+    availspace *= 1024;
+
 
     if (ferror(SS->mfp)) {
 	smtp_tarpit(SS);
@@ -1024,11 +1047,14 @@ int insecure;
       SS->state = Recipient;
 
     SS->rcpt_count = 0;
+    SS->ok_rcpt_count = 0;
     SS->from_box = (addrlen == 0);
+
+    return 0; /* Is ok */
 }
 
 
-void smtp_rcpt(SS, buf, cp)
+int smtp_rcpt(SS, buf, cp)
 SmtpState *SS;
 const char *buf, *cp;
 {
@@ -1039,6 +1065,7 @@ const char *buf, *cp;
     int addrlen = 0, notifylen = 0, orcptlen = 0, notifyflgs;
     int strict = STYLE(SS->cfinfo, 'R');
     int sloppy = STYLE(SS->cfinfo, 'S');
+    int err;
 
     if (strict && sloppy) /* If misconfigured, SLOPPY takes precedence! */
       strict = 0;
@@ -1048,16 +1075,8 @@ const char *buf, *cp;
     if ( (SS->state == MailOrHello || SS->state == Mail) &&
 	 policydb != NULL && SS->policyresult < 0 ) {
       smtp_tarpit(SS);
-      if (!multilinereplies)
-	type(SS, 550, m571, "Access denied by the policy analysis functions by earlier rejection");
-      else {
-	type(SS, -550, m571, "Access denied by the policy analysis functions.");
-	type(SS, -550, m571, "This may be due to your source IP address,");
-	type(SS, -550, m571, "the IP reversal domain, the data you gave for");
-	type(SS, -550, m571, "the HELO/EHLO parameter, or address/domain");
-	type(SS,  550, m571, "you gave at the MAIL FROM:<...> address.");
-      }
-      return;
+      type(SS, 550, m571, "Hello %s, access denied by the policy analysis functions by earlier rejection", SS->ihostaddr);
+      return -1;
     }
 
     if (SS->state != Recipient && SS->state != RecipientOrData) {
@@ -1075,7 +1094,7 @@ const char *buf, *cp;
 	}
 	smtp_tarpit(SS);
 	type(SS, 503, m551, cp);
-	return;
+	return -1;
     }
 
     if (*cp == ' ') ++cp;
@@ -1084,7 +1103,7 @@ const char *buf, *cp;
     if (!CISTREQN(cp, "To:", 3)) {
 	smtp_tarpit(SS);
 	type(SS, 501, m552, "where is To: in this?  %s", cp);
-	return;
+	return -1;
     }
     cp += 3;
     if (!strict_protocol || sloppy)
@@ -1093,27 +1112,27 @@ const char *buf, *cp;
 	  if (!sloppy) {
 	    smtp_tarpit(SS);
 	    type(SS, 501, m513, "where is <...> in this?  %s", cp);
-	    return;
+	    return -1;
 	  }
 	  break; /* Sigh, be sloppy.. */
 	}
     if (*cp == '\0') {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "where is <...> in this?  %s", cp);
-	return;
+	return -1;
     } else if (*cp != '<' && !sloppy) {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "strangeness between ':' and '<': %s", cp);
-	return;
+	return -1;
     } else if (*(cp+1) == '>') {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "Null address valid only as source: %s", cp);
-	return;
+	return -1;
     }
     if (*(cp + 1) == '<') {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "there are too many <'s in this: %s", cp);
-	return;
+	return -1;
     }
     if (*cp == '<') {
       /* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
@@ -1127,13 +1146,13 @@ const char *buf, *cp;
 	  } else {
 	    /* Genuine failure.. */
 	    type821err(SS, 501, m513, buf, "Path data: %s", rfc821_error);
-	    return;
+	    return -1;
 	  }
 	}
 	if (*s == '>') {
 	  smtp_tarpit(SS);
 	  type(SS, 501, m513, "there are too many >'s in this: %s", cp);
-	  return;
+	  return -1;
 	}
 	/* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
 	if (cp[1] == '@') {
@@ -1154,7 +1173,7 @@ const char *buf, *cp;
 	if (s == cp) {
 	  /* Failure.. ? */
 	  type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
-	  return;
+	  return -1;
 	}
 	/* Now it is a moment to see, if we have source routes: @a,@b:c@d */
 	if (*cp == '@') {
@@ -1171,7 +1190,7 @@ const char *buf, *cp;
 	if (*s != '>') {
 	  rfc821_error_ptr = s;
 	  type821err(SS, 501, m517, buf, "Missing ending '>' bracket: %s", cp);
-	  return;
+	  return -1;
 	}
 	++s;
       }
@@ -1182,13 +1201,13 @@ const char *buf, *cp;
       if (s == cp) {
 	/* Failure.. */
 	type821err(SS, 501, m513, buf, "Path data: %.200s", rfc821_error);
-	return;
+	return -1;
       }
 
       if (*s == '>') {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "there are too many >'s in that!");
-	return;
+	return -1;
       }
 
       /* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
@@ -1211,7 +1230,7 @@ const char *buf, *cp;
     if (addrlen < 1) {
 	smtp_tarpit(SS);
 	type(SS, 501, m513, "What is an empty recipient?");
-	return;
+	return -1;
     }
     drpt_notify = NULL;
     notifyflgs  = 0;
@@ -1234,7 +1253,7 @@ const char *buf, *cp;
 	    if (drpt_notify) {
 		smtp_tarpit(SS);
 		type(SS, 501, m554, "NOTIFY-param double defined!");
-		return;
+		return -1;
 	    }
 	    drpt_notify = s;
 	    s += 7;
@@ -1254,16 +1273,17 @@ const char *buf, *cp;
 	    if (*s && *s != ' ' && *s != '\t') {
 		smtp_tarpit(SS);
 		type(SS, 455, m454, "NOTIFY-param data error");
-		return;
+		return -1;
 	    }
 	    notifylen = s - drpt_notify;
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_NOTIFY ++;
 	    continue;
 	}
 	if (dsn_ok && CISTREQN("ORCPT=", s, 6)) {
 	    if (drpt_orcpt) {
 		smtp_tarpit(SS);
 		type(SS, 501, m554, "ORCPT-param double defined!");
-		return;
+		return -1;
 	    }
 	    drpt_orcpt = s;
 	    s = orcpt_string(s + 6);
@@ -1271,27 +1291,32 @@ const char *buf, *cp;
 		smtp_tarpit(SS);
 		type821err(SS, -501, m454, buf, "Invalid ORCPT value '%s'", drpt_orcpt);
 		type(SS, 501, m454, "ORCPT-param data error!");
-		return;
+		return -1;
 	    }
 	    orcptlen = s - drpt_orcpt;
+	    MIBMtaEntry->ss.IncomingSMTP_OPT_ORCPT ++;
 	    continue;
 	}
 	smtp_tarpit(SS);
 	type(SS, 555, m554, "Unknown RCPT TO:<> parameter: %s", s);
-	return;
+	return -1;
     }
 
     if (SS->rcpt_count >= rcptlimitcnt) {
       smtp_tarpit(SS);
       type(SS, 452, "4.5.2", "Too many recipients in one go!");
-      return;
+      return -1;
     }
 
     RFC821_822QUOTE(cp, newcp, addrlen);
 
-    SS->policyresult = policytest(policydb, &SS->policystate,
-				  POLICY_RCPTTO, cp, addrlen,
-				  SS->authuser);
+    if (debug) typeflush(SS);
+    if (SS->netconnected_flg)
+      SS->policyresult = policytest(policydb, &SS->policystate,
+				    POLICY_RCPTTO, cp, addrlen,
+				    SS->authuser);
+    else
+      SS->policyresult = 0;
     if (logfp || logfp_to_syslog) {
       char *ss = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || ss != NULL) {
@@ -1309,9 +1334,14 @@ const char *buf, *cp;
 	  SS->policyresult = 0; /* Plain <postmaster> */
 	else
 	  if (policydb != NULL && SS->policyresult > -100) {
-	    int rc = policytest(policydb, &SS->policystate,
-				POLICY_RCPTPOSTMASTER, cp, addrlen,
-				SS->authuser);
+	    int rc;
+	    if (debug) typeflush(SS);
+	    if (SS->netconnected_flg)
+	      rc = policytest(policydb, &SS->policystate,
+			      POLICY_RCPTPOSTMASTER, cp, addrlen,
+			      SS->authuser);
+	    else
+	      rc = 0;
 	    if (rc == 0)
 	      SS->policyresult = 0;
 
@@ -1339,87 +1369,57 @@ const char *buf, *cp;
 
 	if (SS->policyresult < -99) { /* "soft error, 4XX code */
 	  if (ss != NULL) {
-	    type(SS, 450, m471, "For recipient address <%.*s> the policy analysis reported: %s", addrlen, cp, ss);
+	    type(SS, 450, m471, "Hello %s, for recipient address <%.*s> the policy analysis reported: %s", SS->ihostaddr, addrlen, cp, ss);
+
 	  } else if (SS->policyresult < -103) { /* -104 */
-	    if (!multilinereplies)
-	      type(SS, 450, m443, "Policy analysis reports temporary DNS error with the target domain: <%.*s>", addrlen, cp);
-	    else {
-	      type(SS, -450, m443, "Policy analysis reports temporary DNS error");
-	      type(SS, -450, m443, "with this target domain. Retrying may help,");
-	      type(SS, -450, m443, "or if the condition persists, some further");
-	      type(SS, -450, m443, "work may be in need with the target domain");
-	      type(SS,  450, m443, "DNS servers; RCPT TO:<%.*s>", addrlen, cp);
-	    }
+	    type(SS, 450, m443, "Hello %s, policy analysis reports temporary DNS error with the target domain: <%.*s>", SS->ihostaddr, addrlen, cp);
 
 	  } else if (SS->policyresult < -102) {
 	    /* Code: -103 */
-	    if (!multilinereplies) {
-	      type(SS,450, m471, "This target address is not our MX service client: <%.*s>", addrlen, cp);
-	    } else {
-	      type(SS,-450, m471, "This target address is not our MX service");
-	      type(SS,-450, m471, "client, nor you are connecting from address");
-	      type(SS,-450, m471, "that is allowed to openly use us to relay");
-	      type(SS,-450, m471, "to any arbitrary address thru us.");
-	      type(SS, 450, m471, "We don't accept this recipient: <%.*s>",
-		   addrlen, cp);
-	    }
+	    type(SS,450, m471, "Your IP address %s is not allowed to relay to email address <%.*s> via our server; MX rule",
+		 SS->ihostaddr, addrlen, cp);
+
 	  } else if (SS->policyresult < -100) {
 	    /* Code: -102 */
-	    if (!multilinereplies)
-	      type(SS, 450, m443, "Policy analysis found DNS error on the target address: <%.*s>", addrlen, cp);
-	    else {
-	      type(SS,-450, m443, "Policy analysis found DNS error on");
-	      type(SS,-450, m443, "the target address. This address is");
-	      type(SS, 450, m443, "not currently acceptable: <%.*s>",
-		   addrlen, cp);
-	    }
+	    type(SS, 450, m443, "Hello %s, Policy analysis found DNS error on the target address: <%.*s>", SS->ihostaddr, addrlen, cp);
+
 	  } else {
-	    type(SS,450,m443, "Policy rejection on the target address: <%.*s>",
-		 addrlen, cp);
+	    type(SS,450,m443, "Hello %s, Policy rejection on the target address: <%.*s>",
+		 SS->ihostaddr, addrlen, cp);
 	  }
 	} else {
 	  if (ss != NULL) {
-	    type(SS, 550, m571, "Policy analysis reported: %s rcpt=<%.*s>",
-		 ss, addrlen, cp);
+	    type(SS, 550, m571, "Hello %s, Policy analysis reported: %s rcpt=<%.*s>",
+		 SS->ihostaddr, ss, addrlen, cp);
+
 	  } else if (SS->policyresult < -2) {
 	    /* Code: -3 */
-	    if (!multilinereplies)
-	      type(SS,550,m571, "This target address is not our MX service client: <%.*s>", addrlen, cp);
-	    else {
-	      type(SS,-550, m571, "This target address is not our MX service");
-	      type(SS,-550, m571, "client, nor you are connecting from address");
-	      type(SS,-550, m571, "that is allowed to openly use us to relay");
-	      type(SS,-550, m571, "to any arbitrary address thru us.");
-	      type(SS, 550, m571, "We don't accept this recipient: <%.*s>",
-		   addrlen, cp);
-	    }
+	    type(SS,550, m571, "Your IP address %s is not allowed to relay to email address <%.*s> via our server; MX rule",
+		 SS->ihostaddr, addrlen, cp);
 
 	  } else if (SS->policyresult < -1) {
 	    /* Code: -2 */
-	    if (!multilinereplies) {
-	      type(SS,550,m543, "Policy analysis found DNS error on the target domain: <%.*s>", addrlen, cp);
-	    } else {
-	      type(SS,-550, m543, "Policy analysis found DNS error on");
-	      type(SS,-550, m543, "the target address. This address is");
-	      type(SS, 550, m543, "not currently acceptable: <%.*s>",
-		   addrlen, cp);
-	    }
+	    type(SS,550,m543, "Hello %s, Policy analysis found DNS error on the target domain: <%.*s>", SS->ihostaddr, addrlen, cp);
+
 	  } else {
-	    type(SS,550,m571, "Policy rejection on the target address: <%.*s>",
-		 addrlen, cp);
+	    type(SS,550,m571, "Hello %s, Policy rejection on the target address: <%.*s>",
+		 SS->ihostaddr, addrlen, cp);
 	  }
 	}
 	if (newcp)
 	    free((void *) newcp);
-	return;
+	return -1;
     }
 
     s = NULL;
-    if (STYLE(SS->cfinfo, 't')) {
+    if (enable_router && STYLE(SS->cfinfo, 't')) {
 	s = router(SS, RKEY_TO, 1, cp, addrlen);
-	if (s == NULL)
+	if (s == NULL) {
 	    /* the error was printed in router() */
-	    return;
+	    if (newcp)
+		free((void *) newcp);
+	    return -1;
+	}
 	if (atoi(s) / 100 != 2) {
 	    /* verification failed */
 	    smtp_tarpit(SS);
@@ -1427,7 +1427,7 @@ const char *buf, *cp;
 	    free((void *) s);
 	    if (newcp)
 		free((void *) newcp);
-	    return;
+	    return -1;
 	}
 	/* The 's' goes to use below */
     }
@@ -1435,7 +1435,9 @@ const char *buf, *cp;
 
     /* IETF-NOTARY DSN data: */
     fputs("todsn", SS->mfp);
-    if (drpt_notify) {
+    if (force_rcpt_notify_never) {
+      fwrite(" NOTIFY=NEVER", 1, 13, SS->mfp);
+    } else if (drpt_notify) {
 	fputc(' ', SS->mfp);
 	fwrite(drpt_notify, 1, notifylen, SS->mfp);
 	if (!(notifyflgs & NOTIFY_NEVER) /* Not 'NEVER' */ &&
@@ -1451,7 +1453,7 @@ const char *buf, *cp;
 	  fwrite(",TRACE", 1, 6, SS->mfp);
     }
     if (SS->deliverby_time) {
-      fprintf(SS->mfp, " BY=%ld;", SS->deliverby_time);
+      fprintf(SS->mfp, " BY=%ld;", (long)SS->deliverby_time);
       if (SS->deliverby_flags & DELIVERBY_R)
 	fputc('R', SS->mfp);
       if (SS->deliverby_flags & DELIVERBY_N)
@@ -1521,6 +1523,9 @@ const char *buf, *cp;
     else
       SS->sizeoptsum = SS->sizeoptval;
 
+    err = 1;
+    SS->rcpt_count += 1;
+
     if (ferror(SS->mfp)) {
 	smtp_tarpit(SS);
 	type(SS, 452, m430, (char *) NULL);
@@ -1534,9 +1539,17 @@ const char *buf, *cp;
 	if (SS->from_box && SS->rcpt_count > MaxErrorRecipients) {
 	    smtp_tarpit(SS);
 	    type(SS, 552, m571, "SPAM trap -- too many recipients for an empty source address!");
-	} else
-	    type(SS, atoi(s), s + 4, "Ok");
-	SS->rcpt_count += 1;
+	} else {
+	    err = atoi(s);
+	    if (err >= 400) {
+	      smtp_tarpit(SS);
+	      mail_abort(SS->mfp);
+	      SS->mfp = NULL;
+	    }
+	    type(SS, err, s + 4, "Ok");
+	    if (err < 400)
+	      err = 0;
+	}
     } else {
 	if (SS->from_box && SS->rcpt_count > MaxErrorRecipients) {
 	    smtp_tarpit(SS);
@@ -1547,11 +1560,16 @@ const char *buf, *cp;
 	else
 	    type(SS, 250, "2.1.5", "Recipient address syntax Ok%s; rcpt=<%.*s>",
 		 srcrtestatus, addrlen, cp);
-	SS->rcpt_count += 1;
+	err = 0;
     }
     if (s)
 	free((void *) s);
-    SS->state = RecipientOrData;
+
+    if (!err) {
+      SS->ok_rcpt_count += 1;
+      SS->state = RecipientOrData;
+    }
+    return err;
 }
 
 
@@ -1564,6 +1582,8 @@ const char *buf, *cp;
     const char *newcp = NULL;
     int addrlen;
 
+    MIBMtaEntry->ss.IncomingSMTP_VRFY ++;
+
     if (SS->state == Hello) {
 	type(SS, 503, m551, "Waiting for HELO/EHLO command");
 	return;
@@ -1595,7 +1615,7 @@ const char *buf, *cp;
 
     RFC821_822QUOTE(cp, newcp, addrlen);
 
-    if (cfi) {
+    if (cfi && enable_router) {
 	s = router(SS, RKEY_VERIFY, 0, cp, addrlen);
 	if (s != NULL) {
 	    /* printf("%s\r\n", s); */
@@ -1617,6 +1637,8 @@ const char *buf, *cp;
     int cfi, addrlen;
     char *newcp = NULL;
 
+    MIBMtaEntry->ss.IncomingSMTP_EXPN ++;
+
     if (SS->state == Hello) {
 	type(SS, 503, m551, "Waiting for HELO/EHLO command");
 	return;
@@ -1625,7 +1647,7 @@ const char *buf, *cp;
 	++cp;
 
     cfi = STYLE(SS->cfinfo, 'e');
-    if (cfi != 0) {
+    if (cfi && enable_router) {
 #if 0				/* The input string is an arbitrary string! */
 	if (*cp == '<')
 	    s = rfc821_path(cp, cfi);	/* with < > */
