@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-1999.
+ *      Copyright 1991-2003.
  */
 /*
  * Zmailer SMTP-server divided into bits
@@ -33,23 +33,66 @@
 #include <libtrans.h>
 #endif				/* USE_TRANSLATION */
 
+#define SKIPSPACE(Y) while (*Y == ' ' || *Y == '\t') ++Y
+#define SKIPDIGIT(Y) while ('0' <= *Y && *Y <= '9') ++Y
+#define SKIPTEXT(Y)  while (*Y && *Y != ' ' && *Y != '\t') ++Y
+
 static int mvdata __((SmtpState *, char *));
 static int mvbdata __((SmtpState *, char *, long));
 
+static int parsestatcode __((const char **ss, const char **statcode));
+static int parsestatcode(ssp, statcodep)
+     const char **ssp;
+     const char **statcodep;
+{
+    int code = -1;
+    const char *ss = *ssp;
+    static char statcodebuf[6];
+
+    *statcodep = NULL;
+
+    for (;'0' <= *ss && *ss <= '9'; ++ss) {
+      if (code < 0) code = 0;
+      code = code * 10 + (*ss - '0');
+    }
+    SKIPSPACE(ss);
+    if (isdigit(ss[0]) && ss[1] == '.' &&
+	isdigit(ss[2]) && ss[3] == '.' &&
+	isdigit(ss[4])) {
+      memcpy(statcodebuf, ss, 5);
+      statcodebuf[5] = 0;
+      *statcodep = statcodebuf;
+      ss += 5;
+    }
+    SKIPSPACE(ss);
+    *ssp = ss;
+    if (code < 200 || code > 599) code = 0;
+    return code;
+}
+
 int smtp_data(SS, buf, cp)
 SmtpState *SS;
 const char *buf, *cp;
 {
     int filsiz;
     long tell = 0;
-    time_t mtime;
-    int ino;
+    int i, j;
     char msg[2048];
 
+    struct stat stbuf;
+    char *fname;
+    char taspid[30];
+
+
+    MIBMtaEntry->ss.ReceivedMessagesSs  += 1;
+    MIBMtaEntry->ss.ReceivedRecipientsSs += SS->ok_rcpt_count;
+    MIBMtaEntry->ss.IncomingSMTP_DATA   += 1;
+
     while (!strict_protocol && (*cp == ' ' || *cp == '\t')) ++cp;
     if (strict_protocol && *cp != 0) {
-      type(SS, 501, m554, "Extra junk after 'DATA' verb");
-      return 0;
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
+	type(SS, 501, m554, "Extra junk after 'DATA' verb");
+	return 0;
     }
 
     if (SS->state != RecipientOrData) {
@@ -71,15 +114,18 @@ const char *buf, *cp;
 	    cp = NULL;
 	    break;
 	}
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	type(SS, 503, m552, cp);
 	typeflush(SS);
 	return 0;
     }
+
     if (ferror(SS->mfp)) {
 	type(SS, 452, m430, (char *) NULL);
 	typeflush(SS);
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "message file error");
 	return 0;
@@ -90,6 +136,7 @@ const char *buf, *cp;
 	SS->state = MailOrHello;
 	if (SS->mfp)
 	  mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	return 0;
     }
@@ -99,6 +146,7 @@ const char *buf, *cp;
 	typeflush(SS);
 	SS->state = MailOrHello;
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	return 0;
     }
@@ -108,6 +156,7 @@ const char *buf, *cp;
 	typeflush(SS);
 	SS->state = MailOrHello;
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	return 0;
     }
@@ -123,43 +172,66 @@ const char *buf, *cp;
     /* We set alarm()s inside the mvdata() */
     *msg = 0;
     filsiz = mvdata(SS, msg);
-    alarm(0);			/* cancel the alarm() */
-    tell = ftell(SS->mfp);
+    SS->messagesize = filsiz;
+
+    fflush(SS->mfp);
+    fname = mail_fname(SS->mfp);
+    fstat(FILENO(SS->mfp), &stbuf);
+
+    taspoolid(taspid, stbuf.st_mtime, stbuf.st_ino);
+    tell = stbuf.st_size,
 
     report(SS, "Got '.'; tell=%ld", tell);
 
-    availspace = fd_statfs(FILENO(SS->mfp));
+    availspace = used_fd_statfs(FILENO(SS->mfp));
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolUsedSpace = availspace;
+    availspace = free_fd_statfs(FILENO(SS->mfp));
     if (availspace < 0)
-	availspace = 2000000000;	/* Over 2G ? */
+	availspace = LONG_MAX / 1024;	/* Over 2G ? */
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolFreeSpace = availspace;
     availspace -= minimum_availspace;
+    if (availspace > (LONG_MAX / 1024))
+      availspace = LONG_MAX / 1024;
+    availspace *= 1024;
 
     if (*msg != 0) {
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 452, m430, "%s", msg);
+	if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	    type(SS, 452, m430, "%s", msg);
 	typeflush(SS);
     } else if (s_feof(SS)) {
-	/* [mea@utu.fi] says this can happen */
 	if (STYLE(SS->cfinfo,'D')) {
 	  /* Says: DON'T DISCARD -- aka DEBUG ERRORS! */
 	  mail_close_alternate(SS->mfp,"public",".DATA-EOF");
 	} else
 	  mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on DATA input");
 	typeflush(SS);
 	return -1;
     } else if (availspace < 0 || ferror(SS->mfp)) {
 	type(SS, 452, m430, NULL); /* insufficient system storage */
+	if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 452, m430, NULL); /* insufficient system storage */
 	typeflush(SS);
 	reporterr(SS, tell, ferror(SS->mfp) ? "write to spool file failed" : "system free storage under limit");
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
     } else if (maxsize > 0 && filsiz > maxsize) {
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
+	if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
 	typeflush(SS);
     } else {
 
@@ -167,65 +239,161 @@ const char *buf, *cp;
 	   the resulting file into router spool area;
 	   pending a few things we do at first.. */
 
+	const char *statcode = NULL, *ss, *ss0;
+	int code = 0;
+	const char *sslines[20];
+	int sslinecnt = 0;
+
+
 	/* Lets see what the content-policy will tell now ? */
-	char *fname = mail_fname(SS->mfp);
+
+	if (debug) typeflush(SS);
 	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
 
-	if (SS->policyresult < 0) {
-	  char *ss = policymsg(policydb, &SS->policystate);
+	ss0 = ss  = policymsg(policydb, &SS->policystate);
+
+	if (ss)
 	  type(NULL,0,NULL,
-	       "Content-policy analysis ordered message rejection. (code=%d)", SS->policyresult);
-	  type(SS, -552,m571, "Content-Policy analysis rejected this message");
-	  type(SS, 552, m571, "Content-Policy msg: %s", ss ? ss : "rejected");
+	       "Content-policy analysis ordered message %s. (code=%d); msg='%s'",
+	       (SS->policyresult < 0 ? "rejection" :
+		(SS->policyresult > 0 ? "freezing" : "acceptance")),
+	       SS->policyresult, ss);
+
+	if (ss) {
+	  char *p, *s;
+	  code = parsestatcode(&ss,&statcode);
+	  s = (char *)ss;
+	  p = strchr(s,'\r');
+	  sslinecnt = 0;
+	  if (p) {
+	    /* Multiline! CRs in message text... */
+	    while (p && sslinecnt < 17) { /* Arbitrary fixed limit.. */
+	      *p++ = '\0';
+	      sslines[sslinecnt++] = s;
+	      sslines[sslinecnt+0] = p;
+	      sslines[sslinecnt+1] = NULL;
+	      s = p;
+	      p = strchr(s,'\r');
+	    }
+	  }
+	}
+	if (!ss || *ss == 0) {
+	  if (SS->policyresult < 0)
+	    sslines[0] = ss = "rejected, no further explanations";
+	  else  if (SS->policyresult == 0)
+	    sslines[0] = ss = "accepted";
+	  else
+	    sslines[0] = ss = "accepted into freezer, no explanations";
+	  sslines[1] = NULL;
+	}
+
+	if (SS->policyresult < 0) {
+
+	  if (!statcode)  statcode = m571;
+	  if (!code)      code = 552;
+
+	  type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
+	  for (j= 1; j <= sslinecnt; ++j)
+	    type(SS, -code, statcode, "msg: %s", sslines[j]);
+	  type(SS, code, statcode, "Content-Policy analysis rejected this message");
+
+	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+	    type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "msg: %s", sslines[j]);
+	    type(SS, code, statcode, "Content-Policy analysis rejected this message");
+	  }
+
 	  mail_abort(SS->mfp);
+	  MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	  SS->mfp = NULL;
 	} else if (SS->policyresult > 0) {
 	  char polbuf[20];
-	  struct stat stbuf;
-	  char *ss = policymsg(policydb, &SS->policystate);
 
-	  fflush(SS->mfp);
-	  fstat(FILENO(SS->mfp), &stbuf);
 	  runasrootuser();
 	  sprintf(polbuf,"policy-%d",SS->policyresult);
 	  if (mail_close_alternate(SS->mfp, FREEZERDIR, polbuf) != 0) {
 	    type(NULL,0,NULL,
 		 "mail_close_alternate(..'FREEZER','%s') failed, errno=%d (%s)",
 		 polbuf, errno, strerror(errno));
-	    type(SS, 452, m430, "Message file disposition failed");
+	    type(SS, 452, m430, "Message file disposition failed; %s", taspid);
+	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	      type(SS, 452, m430, "Message file disposition failed; %s", taspid);
 	    typeflush(SS);
 	    SS->mfp = NULL;
 	    reporterr(SS, tell, "message file close failed");
 	  } else {
-	    static int freezecnt = 1;
-	    freezecnt <<= 1;
-	    sleep(freezecnt);
-	    type(SS, 250, "2.6.0", "message accepted; into freezer[%d] area; %s", SS->policyresult, ss ? ss : "");
+
+	    smtp_tarpit(SS);
+
+	    if (!statcode)  statcode = "2.7.1";
+	    if (!code)      code = 250;
+
+	    type(SS, -code, statcode, "%s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "%s", sslines[j]);
+	    type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    
+	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+	      type(SS, -code, statcode, "%s; %s", ss, taspid);
+	      for (j= 1; j <= sslinecnt; ++j)
+		type(SS, -code, statcode, "%s", sslines[j]);
+	      type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    }
+
 	    typeflush(SS);
 	    SS->mfp = NULL;
-	    zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze[%d]",
-		     (int) stbuf.st_ino, (int) stbuf.st_size,
-		     SS->rhostname, SS->rport, SS->policyresult));
+	    zsyslog((LOG_INFO, "accepted  %s (%ldc) from %s/%d into freeze[%d]",
+		     taspid, tell, SS->rhostname, SS->rport, SS->policyresult));
 	  }
+	  MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
+
 	  runastrusteduser();
 	} else {
 
 	  /*  Ok, we didn't have smtp-policy defined freezer action,
 	      lets see if we do it some other way.. */
 
-	  if (_mail_close_(SS->mfp, &ino, &mtime) == EOF) {
+	  if (mail_close(SS->mfp) == EOF) {
 	    type(SS, 452, m430, (char *) NULL);
+	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	      type(SS, 452, m430, (char *) NULL);
 	    typeflush(SS);
 	    SS->mfp = NULL;
 	    reporterr(SS, tell, "message file close failed");
+	    MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
+
 	  } else {
 	    /* Ok, build response with proper "spoolid" */
-	    char fnam[20], taspid[30];
-	    sprintf(fnam, "%d", ino);
-	    taspoolid(taspid, mtime, (long)ino);
 
 	    SS->mfp = NULL;
-	    type(SS, 250, "2.6.0", "%s message accepted", taspid);
+	    if (!ss || *ss == 0) {
+	      type(SS, 250, "2.0.0", "Message accepted; %s", taspid);
+	      if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+		type(SS, 250, "2.0.0", "Message accepted; %s", taspid);
+	    } else {
+	      if (!statcode)  statcode = "2.0.0";
+	      if (!code)      code = 250;
+
+	      if (sslinecnt < 1)
+		type(SS,  code, statcode, "%s; %s", ss, taspid);
+	      else
+		type(SS, -code, statcode, "%s; %s", ss, taspid);
+	      for (j= 1; j <= sslinecnt; ++j)
+		type(SS, -code, statcode, "%s", sslines[j]);
+	      if (sslinecnt >= 1)
+		type(SS, code, statcode, "Content-Policy accepted this message; %s", taspid);
+	      if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+		if (sslinecnt < 1)
+		  type(SS,  code, statcode, "%s; %s", ss, taspid);
+		else
+		  type(SS, -code, statcode, "%s; %s", ss, taspid);
+		for (j= 1; j <= sslinecnt; ++j)
+		  type(SS, -code, statcode, "%s", sslines[j]);
+		if (sslinecnt >= 1)
+		  type(SS, code, statcode, "Content-Policy accepted this message; %s", taspid);
+	      }
+	    }
 	    typeflush(SS);
 
 	    if (smtp_syslog)
@@ -233,6 +401,14 @@ const char *buf, *cp;
 		       "%s: (%ldc) accepted from %s/%d", taspid, tell,
 		       SS->rhostname, SS->rport));
 		
+	    MIBMtaEntry->ss.IncomingSMTP_DATA_ok    += 1;
+
+	    MIBMtaEntry->ss.TransmittedMessagesSs   += 1;
+	    MIBMtaEntry->ss.TransmittedRecipientsSs += SS->ok_rcpt_count;
+
+	    MIBMtaEntry->ss.IncomingSMTP_DATA_KBYTES  += (SS->messagesize+1023)/1024;
+	    MIBMtaEntry->ss.IncomingSMTP_spool_KBYTES += (tell + 1023)/1024;
+
 	    type(NULL,0,NULL,"%s: %ld bytes", taspid, tell);
 	    if (logfp)
 	      fflush(logfp);
@@ -253,7 +429,16 @@ const char *buf, *cp;
     long tell;
     char msg[2048];
     long bdata_chunksize;
-    int bdata_last;
+    int bdata_last, i, j;
+
+    struct stat stbuf;
+    char *fname;
+    char taspid[30];
+
+    
+    MIBMtaEntry->ss.ReceivedMessagesSs  += 1;
+    MIBMtaEntry->ss.ReceivedRecipientsSs += SS->ok_rcpt_count;
+    MIBMtaEntry->ss.IncomingSMTP_BDAT   += 1;
 
     if (SS->state == RecipientOrData) {
 	SS->state = BData;
@@ -264,10 +449,11 @@ const char *buf, *cp;
     rc = sscanf(cp, "%ld %7s %7s", &bdata_chunksize, msg, msg + 20);
     SS->bdata_blocknum += 1;
     bdata_last = CISTREQ(msg, "LAST");
-    if (!(bdata_chunksize > 0L
+    if (!(bdata_chunksize >= 0L
 	  && (rc == 1 || (rc == 2 && bdata_last)))) {
 	type(SS, 501, m552, NULL);
 	typeflush(SS);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	return 0;
     }
     if (SS->bdata_blocknum == 1 && SS->mfp) {
@@ -279,13 +465,29 @@ const char *buf, *cp;
     /* We set alarm()s inside the mvbdata() */
     *msg = 0;
     filsiz = mvbdata(SS, msg, bdata_chunksize);
-    alarm(0);			/* cancel the alarm() */
-    if (SS->mfp)
-	tell = ftell(SS->mfp);
-    else
-	tell = 0;
+    SS->messagesize += filsiz;
 
-    report(SS, "BDAT; last=%d tell=%ld", bdata_last, tell);
+    tell = 0;
+
+    if (SS->mfp) {
+
+      fflush(SS->mfp);
+      fname = mail_fname(SS->mfp);
+      fstat(FILENO(SS->mfp), &stbuf);
+
+      taspoolid(taspid, stbuf.st_mtime, stbuf.st_ino);
+      tell = stbuf.st_size;
+
+    } else {
+
+      tell = 0;
+      fname = "<NIL>";
+      strcpy(taspid, "<NIL>");
+
+    }
+
+    report(SS, "BDAT %ld%s; tell=%ld", bdata_chunksize,
+	   bdata_last ? " LAST":"", tell);
 
     if (SS->state != BData) {
 	switch (SS->state) {
@@ -303,10 +505,15 @@ const char *buf, *cp;
 	    cp = NULL;
 	    break;
 	}
+
 	type(SS, 503, m552, cp);
+	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 503, m552, cp);
+
 	typeflush(SS);
 	if (SS->mfp)
 	    mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	return 0;
     }
@@ -315,20 +522,30 @@ const char *buf, *cp;
 	    cp = "No valid sender, rejecting all recipients";
 	    if (SS->sender_ok != 0)
 		cp = "No valid recipient at RCPT addresses, or no RCPT addresses at all";
+
 	    type(SS, 550, "5.1.3", cp);
+	    if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	      type(SS, 550, "5.1.3", cp);
+
 	    typeflush(SS);
 	    SS->state = MailOrHello;
 	    if (SS->mfp)
 	      mail_abort(SS->mfp);
+	    MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	    SS->mfp = NULL;
 	    return 0;
 	}
 	if ((SS->from_box != 0) && (SS->rcpt_count > MaxErrorRecipients)) {
+
 	  /* Too many recipients for a  "MAIL FROM:<>" */
 	  type(SS, 550, "5.7.1", "SPAM trap -- too many recipients for an empty source address!");
+	  if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	    type(SS, 550, "5.7.1", "SPAM trap -- too many recipients for an empty source address!");
+
 	  typeflush(SS);
 	  SS->state = MailOrHello;
 	  mail_abort(SS->mfp);
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	  SS->mfp = NULL;
 	  return 0;
 	}
@@ -336,18 +553,32 @@ const char *buf, *cp;
 	SS->sender_ok = 0;
     }
 
-    availspace = fd_statfs(FILENO(SS->mfp));
+    availspace = used_fd_statfs(FILENO(SS->mfp));
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolUsedSpace = availspace;
+    availspace = free_fd_statfs(FILENO(SS->mfp));
     if (availspace < 0)
-	availspace = 2000000000;	/* Over 2G ? */
+	availspace = LONG_MAX / 1024;	/* Over 2G ? */
+    if (availspace >= 0)
+      MIBMtaEntry->sys.SpoolFreeSpace = availspace;
     availspace -= minimum_availspace;
+    if (availspace > (LONG_MAX / 1024))
+      availspace = LONG_MAX / 1024;
+    availspace *= 1024;
 
     /* The common typeflush() is at the end... */
     if (SS->mfp == NULL) {
-	type(SS, 503, m552, "BDAT block discarded due to earlier error");
+      type(SS, 452, m430, "BDAT block discarded due to earlier error");
+	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 452, m430, "BDAT block discarded due to earlier error");
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
     } else if (*msg != 0) {
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
 	type(SS, 452, "%s", msg);
+	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 452, "%s", msg);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
     } else if (s_feof(SS)) {
 	/* [mea@utu.fi] says this can happen */
 	if (STYLE(SS->cfinfo,'D')) {
@@ -355,45 +586,105 @@ const char *buf, *cp;
 	  mail_close_alternate(SS->mfp,"public",".BDAT-EOF");
 	} else
 	  mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on BDAT input");
-	typeflush(SS);
+	typeflush(SS); /* Pointless ?? */
 	return -1;
     } else if (availspace < 0 || ferror(SS->mfp)) {
 	type(SS, 452, m400, (char *) NULL);
-	reporterr(SS, tell, ferror(SS->mfp) ? "write to spool file failed" : "system free storage under limit");
+	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 452, m400, (char *) NULL);
+	reporterr(SS, tell,
+		  ferror(SS->mfp) ? "write to spool file failed" :
+				    "system free storage under limit");
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
     } else if (maxsize > 0 && tell > maxsize) {
 	mail_abort(SS->mfp);
+	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
+	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	  type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
+	
     } else if (bdata_last) {
-	time_t mtime;
-	int inum;
 
 	/* Things have been good thus far, now we store
 	   the resulting file into router spool area;
 	   pending a few things we do at first.. */
 
+	const char *statcode = NULL, *ss, *ss0;
+	int code = 0;
+	const char *sslines[20];
+	int sslinecnt = 0;
+
 	/* Lets see what the content-policy will tell now ? */
-	char *fname = mail_fname(SS->mfp);
+
+	if (debug) typeflush(SS);
 	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
+	ss0 = ss  = policymsg(policydb, &SS->policystate);
+
+	if (ss)
+	  type(NULL,0,NULL,
+	       "Content-policy analysis ordered message %s. (code=%d); msg='%s'",
+	       (SS->policyresult < 0 ? "rejection" :
+		(SS->policyresult > 0 ? "freezing" : "acceptance")),
+	       SS->policyresult, ss);
+
+	if (ss) {
+	  char *p, *s;
+	  code = parsestatcode(&ss,&statcode);
+	  s = (char *) ss;
+	  p = strchr(s,'\r');
+	  sslinecnt = 0;
+	  if (p) {
+	    /* Multiline! CRs in message text... */
+	    while (p && sslinecnt < 17) { /* Arbitrary fixed limit.. */
+	      *p++ = '\0';
+	      sslines[sslinecnt++] = s;
+	      sslines[sslinecnt+0] = p;
+	      sslines[sslinecnt+1] = NULL;
+	      s = p;
+	      p = strchr(s,'\r');
+	    }
+	  }
+	}
+	if (!ss || *ss == 0) {
+	  if (SS->policyresult < 0)
+	    sslines[0] = ss = "rejected, no further explanations";
+	  else  if (SS->policyresult == 0)
+	    sslines[0] = ss = "accepted";
+	  else
+	    sslines[0] = ss = "accepted into freezer, no explanations";
+	  sslines[1] = NULL;
+	}
 
 	if (SS->policyresult < 0) {
-	  char *ss = policymsg(policydb, &SS->policystate);
-type(NULL,0,NULL,
-  "Content-policy analysis ordered message rejection. (code=%d)", SS->policyresult);
-	  type(SS, -552,m571, "Content-Policy analysis rejected this message");
-	  type(SS, 552, m571, "Content-Policy msg: %s", ss ? ss : "rejected");
+	  
+	  if (!statcode)  statcode = m571;
+	  if (!code)      code = 552;
+
+	  type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
+	  for (j= 1; j <= sslinecnt; ++j)
+	    type(SS, -code, statcode, "msg: %s", sslines[j]);
+	  type(SS, code, statcode, "Content-Policy analysis rejected this message");
+
+	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+	    type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "msg: %s", sslines[j]);
+	    type(SS, code, statcode, "Content-Policy analysis rejected this message");
+	  }
+
 	  mail_abort(SS->mfp);
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	  SS->mfp = NULL;
+
 	} else if (SS->policyresult > 0) {
-	  struct stat stbuf;
-	  char *ss = policymsg(policydb, &SS->policystate);
-	  fflush(SS->mfp);
-	  fstat(FILENO(SS->mfp), &stbuf);
+
 	  runasrootuser();
 	  if (mail_close_alternate(SS->mfp, FREEZERDIR, "policy") != 0) {
 	    type(NULL,0,NULL,
@@ -401,42 +692,108 @@ type(NULL,0,NULL,
 		 "policy", errno, strerror(errno));
 	    if (logfp)
 	      fflush(logfp);
-	    type(SS, 452, m430, "Message file disposition failed");
+
+	    type(SS, 452, m430, "Message file disposition failed; %s", taspid);
+	    if (lmtp_mode) for(i = 0; i < SS->ok_rcpt_count; ++i)
+	      type(SS, 452, m430, "Message file disposition failed; %s",taspid);
+
 	    SS->mfp = NULL;
 	    reporterr(SS, tell, "message file close failed");
 	  } else {
-	    type(SS, 250, "2.6.0", "message accepted; into freezer[%d] area; %s", SS->policyresult, ss ? ss : "");
+
+	    smtp_tarpit(SS);
+
+	    if (!statcode)  statcode = "2.7.1";
+	    if (!code)      code = 250;
+
+	    type(SS, -code, statcode, "%s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "%s", sslines[j]);
+	    type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    
+	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+	      type(SS, -code, statcode, "%s; %s", ss, taspid);
+	      for (j= 1; j <= sslinecnt; ++j)
+		type(SS, -code, statcode, "%s", sslines[j]);
+	      type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    }
+
+	    typeflush(SS);
 	    SS->mfp = NULL;
-	    zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze",
-		     (int) stbuf.st_ino, (int) stbuf.st_size,
-		     SS->rhostname, SS->rport));
+	    zsyslog((LOG_INFO, "accepted  %s (%ldc) from %s/%d into freeze[%d]",
+		     taspid, tell, SS->rhostname, SS->rport, SS->policyresult));
 	  }
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	  runastrusteduser();
-	} else if (_mail_close_(SS->mfp, &inum, &mtime) == EOF) {
+	} else if (mail_close(SS->mfp) == EOF) {
+
 	  type(SS, 452, m400, (char *) NULL);
-	  SS->mfp = NULL;
-	  reporterr(SS, tell, "message file close failed");
-	} else {
-	  /* Ok, build response with proper "spoolid" */
-	  char taspid[30];
-	  taspoolid(taspid, mtime, inum);
+	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	    type(SS, 452, m400, (char *) NULL);
 
 	  SS->mfp = NULL;
-	  type(SS, 250, "2.6.0", "%s Roger, got %ld bytes in the last chunk, stored %ld bytes into spool",
+	  reporterr(SS, tell, "message file close failed");
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
+	} else {
+	  /* Ok, build response with proper "spoolid" */
+
+	  SS->mfp = NULL;
+
+#if 1
+	  type(SS, 250, "2.0.0", "%s Roger, got %ld bytes in the last chunk, stored %ld bytes into spool",
 	       taspid, bdata_chunksize, (long) tell);
+	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	    type(SS, 250, "2.0.0", "%s Roger, got %ld bytes in the last chunk, stored %ld bytes into spool",
+	       taspid, bdata_chunksize, (long) tell);
+
 	  type(NULL,0,NULL,"-- pipeline input: %d bytes",s_hasinput(SS));
 
+#else
+	  if (!statcode)  statcode = "2.0.0";
+	  if (!code)      code = 250;
+
+	  if (sslinecnt < 1)
+	    type(SS,  code, statcode, "%s; %s", ss, taspid);
+	  else
+	    type(SS, -code, statcode, "%s; %s", ss, taspid);
+	  for (j= 1; j <= sslinecnt; ++j)
+	    type(SS, -code, statcode, "%s", sslines[j]);
+	  if (sslinecnt >= 1)
+	    type(SS, code, statcode, "Content-Policy accepted this message; %s", taspid);
+	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
+	    if (sslinecnt < 1)
+	      type(SS,  code, statcode, "%s; %s", ss, taspid);
+	    else
+	      type(SS, -code, statcode, "%s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "%s", sslines[j]);
+	    if (sslinecnt >= 1)
+	      type(SS, code, statcode, "Content-Policy accepted this message; %s", taspid);
+	  }
+#endif
+
+
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_ok    += 1;
+
+	  MIBMtaEntry->ss.TransmittedMessagesSs   += 1;
+	  MIBMtaEntry->ss.TransmittedRecipientsSs += SS->ok_rcpt_count;
+
+	  MIBMtaEntry->ss.IncomingSMTP_BDAT_KBYTES  += (SS->messagesize+1023)/1024;
+	  MIBMtaEntry->ss.IncomingSMTP_spool_KBYTES += (tell + 1023)/1024;
+
 	  if (smtp_syslog)
 	    zsyslog((LOG_INFO,
 		     "%s: (%ldc) accepted from %s/%d", taspid, tell,
 		     SS->rhostname, SS->rport));
-
 	  type(NULL,0,NULL,"%s: %ld bytes", taspid, tell);
+
 	  if (logfp)
 	    fflush(logfp);
 	}
     } else {			/* Not last chunk! */
-	type(SS, 250, "2.6.0", "Received %ld bytes", bdata_chunksize);
+      type(SS, 250, "2.0.0", "Received %ld bytes", bdata_chunksize);
+      if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
+	type(SS, 250, "2.0.0", "Received %ld bytes", bdata_chunksize);
     }
     if (bdata_last) {
 	SS->state = MailOrHello;
@@ -623,6 +980,7 @@ int encodebase64string(instr,inlen,outstr,outspc)
  * Rayan back in 1988:
  *  "If you can improve on this heavily optimized routine, I'd like to see it.
  *   This version goes at better than 100kB/cpu-sec on a Sun 3/180."
+ *   (with 68030 CPU running at 33 MHz -- about 10-15 MIPS)
  */
 
 static int /* count of bytes */ mvdata(SS, msg)
@@ -631,7 +989,6 @@ char *msg;
 {
     register int c, state, *sts, endstate, cnt;
     register char *idxnum;
-    int timercnt = 0;
 
 #ifdef NO_INCOMING_HEADER_PROCESSING
     idxnum = indexnum + 1;
@@ -641,7 +998,7 @@ char *msg;
     sts = states;
     cnt = 0;
 
-    alarm(SMTP_DATA_TIME_PER_LINE);
+    SS->read_alarm_ival = SMTP_DATA_TIME_PER_LINE;
 #else
     char linebuf[4000], *s, *eol;
     int col;
@@ -675,19 +1032,14 @@ char *msg;
     cnt = 0;
     col = 0;
 
-    alarm(SMTP_DATA_TIME_PER_LINE);
+    SS->read_alarm_ival = SMTP_DATA_TIME_PER_LINE;
 
     /* ================ Input the email headers ================ */
     /*           and analyze them a bit (Precedence:)            */
     /*        ... that only "Subject:" has 8-bit chars ...       */
     mail_priority = _MAILPRIO_NORMAL;
     for (;;) {
-	if (--timercnt <= 0) {
-	    /* Re-arm for every kilobyte */
-	    alarm(SMTP_DATA_TIME_PER_LINE);
-	    timercnt = 1024;
-	}
-	c = s_getc(SS);
+	c = s_getc(SS, 1);
 	/* An EOF in here is an error! */
 #if EOF != -1
 	if (c == EOF)
@@ -1048,18 +1400,12 @@ char *msg;
 
     /* ================ Normal email BODY input.. ================ */
     for (;;) {
-	c = s_getc(SS);
+	c = s_getc(SS, 1);
 #if EOF != -1
 	if (c == EOF)		/* a little slower... */
 	    break;
 #endif
 
-	--timercnt;
-	if (timercnt <= 0) {
-	    /* Re-arm for every kilobyte */
-	    alarm(SMTP_DATA_TIME_PER_LINE);
-	    timercnt = 1024;
-	}
 	++cnt;
 	state = sts[state + idxnum[c]];
 	if (state & ~0xff) {
@@ -1152,20 +1498,16 @@ char *msg;
 register long incount;
 {
     register int c, cnt;
-    int timercnt = 0;
 
     cnt = 0;
 
     /* XX: header processing REMOVED from BDAT processing */
 
+    SS->read_alarm_ival = SMTP_DATA_TIME_PER_LINE;
+
     /* ================ Normal email BODY input.. ================ */
     for (; incount > 0; --incount) {
-	if (--timercnt <= 0) {
-	    /* Re-arm for every kilobyte */
-	    alarm(SMTP_DATA_TIME_PER_LINE);
-	    timercnt = 1024;
-	}
-	c = s_getc(SS);
+	c = s_getc(SS, 1);
 	if (c == EOF)
 	    break;
 	++cnt;
