@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-2000.
+ *      Copyright 1991-2001.
  */
 /*
  * Zmailer SMTP-server divided into bits
@@ -40,7 +40,7 @@ int silence;
 	return -1;
     }
 
-    fprintf(mfp, "%c%c%s\n", _CF_TURNME, _CFTAG_NORMAL, cp);
+    fprintf(mfp, "%c%c%s %s\n", _CF_TURNME,  _CFTAG_NORMAL, cp, SS->ihostaddr);
     /* printf("050-My uid=%d/%d\r\n",getuid(),geteuid()); */
     runasrootuser();
     rc = mail_close_alternate(mfp, TRANSPORTDIR, "");
@@ -96,7 +96,7 @@ static int _getline(buf, bufsize, bufspace, fp)
   (*buf)[*bufsize] = 0;
 
   if (c == EOF && *bufsize != 0) {
-    fprintf(stderr, "%s: no input from scheduler\n", progname);
+    fprintf(stderr," no input from scheduler");
     (*buf)[0] = '\0';
     return -1;
   }
@@ -143,7 +143,7 @@ const char *name, *cp;
       if (port && isdigit(port[1])) {
 	portnum = atol(port+1);
       } else if (port == NULL) {
-	serv = getservbyname(port ? port : "mailq", "tcp");
+	serv = getservbyname("mailq", "tcp");
 	if (serv == NULL) {
 
 	  type(SS,-250,m200,"Cannot find 'mailq' tcp service");
@@ -163,7 +163,7 @@ const char *name, *cp;
       req.ai_family   = PF_INET;
       ai = NULL;
 
-      if (debug) fprintf(stderr,"INET6 lookup for '%s'\n",node->nodename);
+      if (debug) fprintf(stderr,"INET lookup for '%s'\n",node->nodename);
 
 #ifdef HAVE_GETADDRINFO
       rc = getaddrinfo(node->nodename, "0", &req, &ai);
@@ -262,14 +262,14 @@ const char *name, *cp;
 
 #ifdef HAVE_OPENSSL
     MD5_Init(&CTX);
-    MD5_Update(&CTX, challenge, strlen(challenge));
-    MD5_Update(&CTX, node->password, strlen(node->password));
+    MD5_Update(&CTX, (const void *)challenge, strlen(challenge));
+    MD5_Update(&CTX, (const void *)(node->password), strlen(node->password));
     MD5_Final(digbuf, &CTX);
 #endif /* - HAVE_OPENSSL */
 #ifndef HAVE_OPENSSL
     MD5Init(&CTX);
-    MD5Update(&CTX, challenge, strlen(challenge));
-    MD5Update(&CTX, node->password, strlen(node->password));
+    MD5Update(&CTX, (const void *)challenge, strlen(challenge));
+    MD5Update(&CTX, (const void *)(node->password), strlen(node->password));
     MD5Final(digbuf, &CTX);
 #endif /* --HAVE_OPENSSL */
 
@@ -291,7 +291,7 @@ const char *name, *cp;
       goto failure_exit;
     }
 
-    fprintf(fpo,"ETRN %s\n", cp);
+    fprintf(fpo,"ETRN %s %s@%s\n", cp, SS->ihostaddr, SS->myhostname);
     if (fflush(fpo) || ferror(fpo)) {
       type(SS,-250,m200,"MQ2-ETRN write failure occurred");
       goto failure_exit;
@@ -349,6 +349,9 @@ void smtp_turnme(SS, name, cp)
 SmtpState *SS;
 const char *name, *cp;
 {
+
+    MIBMtaEntry->ss.IncomingSMTP_ETRN += 1;
+
     while (*cp == ' ' || *cp == '\t') ++cp;
     if (*cp == 0) {
 	type(SS, 552, "5.0.0", "ETRN needs target domain name parameter.");
