@@ -1,12 +1,13 @@
 /*
  *  Globals of the ZMailer  smtp-server
  *
- *  Matti Aarnio <mea@nic.funet.fi> 1997-2000
+ *  Matti Aarnio <mea@nic.funet.fi> 1997-2001
  */
 
 
 #define SMTP_COMMAND_ALARM_IVAL 1200	/* 20 minutes.. */
 #define SMTP_DATA_TIME_PER_LINE  600	/* 10 minutes of life.. */
+#define SMTP_REPLY_ALARM_IVAL    300	/*  5 minutes to write to socket.. */
 
 /*
  * The smtpserver connects to the router to ask it various questions, like,
@@ -57,6 +58,13 @@
 #include <varargs.h>		/* If no  <stdarg.h>,  then presume <varargs.h> ... */
 #endif
 
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
+#ifndef LONG_MAX
+# define LONG_MAX 2147483647L /* For 32 bit machine! */
+#endif
+
 #include <netinet/in.h> /* In some systems needed before <arpa/inet.h> */
 #include <arpa/inet.h>
 
@@ -67,6 +75,9 @@
 #ifdef  HAVE_WAITPID
 #include <sys/wait.h>
 #else
+#ifdef HAVE_WAIT4
+#include <sys/wait.h>		/* Has BSD wait4() */
+#else
 #ifdef HAVE_WAIT3
 #include <sys/wait.h>		/* Has BSD wait3() */
 #else
@@ -77,6 +88,7 @@ extern int wait();
 #endif
 #endif
 #endif
+#endif
 
 #ifdef HAVE_NETDB_H
 #include <netdb.h>
@@ -86,6 +98,9 @@ extern int wait();
 #endif
 
 #include <sys/socket.h>
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 #include <netinet/in.h>
 #ifdef HAVE_NETINET_IN6_H
 #include <netinet/in6.h>
@@ -103,6 +118,8 @@ extern int wait();
 #include "libc.h"
 #include "libz.h"
 
+#include "shmmib.h"
+
 #ifndef	SIGCHLD
 #define	SIGCHLD	SIGCLD
 #endif				/* SIGCHLD */
@@ -114,14 +131,54 @@ extern int wait();
 #include "zsyslog.h"
 
 
-#ifndef __Usockaddr__
-typedef union {
-    struct sockaddr_in v4;
-#ifdef INET6
-    struct sockaddr_in6 v6;
+
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
 #endif
-} Usockaddr;
-#define __Usockaddr__
+
+#include <sys/time.h>
+
+#ifndef	NFDBITS
+/*
+ * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
+ * assumption we are dealing with pre-4.3bsd select().
+ */
+
+/* #error "FDSET macro susceptible" */
+
+typedef long	fd_mask;
+
+#ifndef	NBBY
+#define	NBBY	8
+#endif	/* NBBY */
+#define	NFDBITS		((sizeof fd_mask) * NBBY)
+
+/* SunOS 3.x and 4.x>2 BSD already defines this in /usr/include/sys/types.h */
+#ifdef	notdef
+typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
+#endif	/* notdef */
+
+#ifndef	_Z_FD_SET
+/* #warning "_Z_FD_SET[1]" */
+#define	_Z_FD_SET(n, p)   ((p)->fds_bits[0] |= (1 << (n)))
+#define	_Z_FD_CLR(n, p)   ((p)->fds_bits[0] &= ~(1 << (n)))
+#define	_Z_FD_ISSET(n, p) ((p)->fds_bits[0] & (1 << (n)))
+#define _Z_FD_ZERO(p)	  memset((char *)(p), 0, sizeof(*(p)))
+#endif	/* !FD_SET */
+#endif	/* !NFDBITS */
+
+#ifdef FD_SET
+/* #warning "_Z_FD_SET[2]" */
+#define _Z_FD_SET(sock,var) FD_SET(sock,&var)
+#define _Z_FD_CLR(sock,var) FD_CLR(sock,&var)
+#define _Z_FD_ZERO(var) FD_ZERO(&var)
+#define _Z_FD_ISSET(i,var) FD_ISSET(i,&var)
+#else
+/* #warning "_Z_FD_SET[3]" */
+#define _Z_FD_SET(sock,var) var |= (1 << sock)
+#define _Z_FD_CLR(sock,var) var &= ~(1 << sock)
+#define _Z_FD_ZERO(var) var = 0
+#define _Z_FD_ISSET(i,var) ((var & (1 << i)) != 0)
 #endif
 
 
@@ -130,8 +187,19 @@ typedef union {
 #ifdef HAVE_OPENSSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/rand.h>
 #endif /* - HAVE_OPENSSL */
 
+#ifdef HAVE_SASL2
+#ifdef HAVE_SASL_SASL_H
+#include <sasl/sasl.h>
+#include <sasl/saslutil.h>
+#else
+# error "No <sasl/sasl.h> available!"
+#endif
+#endif
+
+
 struct smtpconf {
     const char *pattern;
     int maxloadavg;
@@ -149,6 +217,7 @@ typedef enum {
     StartTLS,
 #endif /* - HAVE_OPENSSL */
     Hello2, Mail2, Send2, Verify2,	/* 8-bit extensions */
+    HelloL,		/* RFC 2033 LHLO -- sort of */
     Silent		/* One particular client error trap.. */
 } Command;
 
@@ -160,10 +229,59 @@ struct command {
 
 extern struct command command_list[];
 
-typedef struct {
-    FILE *outfp;		/* stdout */
+#ifdef HAVE_OPENSSL
+
+#define CCERT_BUFSIZ 256
+
+struct smtpserver_ssl_subset {
+    SSL * ssl;
+
+    const char *protocol;
+    const char *cipher_name;
+    int         cipher_usebits;
+    int         cipher_algbits;
+
+    int  peer_verified;
+
+    const char *cipher_info;
+    const char *issuer_CN;
+    const char *peer_issuer;
+    const char *peer_CN;
+    const char *peer_subject;
+    const char *peer_fingerprint;
+
+    unsigned char peer_md[EVP_MAX_MD_SIZE];
+};
+
+
+#endif /* - HAVE_OPENSSL */
+
+#ifdef HAVE_SASL2
+struct SmtpSASLState {
+    char  *volatile auth_type;
+    const char *mechlist;
+    sasl_conn_t *conn;
+    volatile int sasl_ok;
+    volatile unsigned int n_auth;	/* count of AUTH commands */
+    volatile unsigned int n_mechs;
+    unsigned int len;
+    sasl_security_properties_t ssp;
+#if 0 /* Is in SASL-1, different/not in SASL-2 */
+#ifdef SASL_SSF_EXTERNAL
+    sasl_external_properties_t ext_ssf;
+#endif
+#endif
+    sasl_ssf_t *ssf;
+};
+#endif /* HAVE_SASL2 */
+
+
+
+typedef struct SmtpState {
+    int  outputfd;		/* stdout */
     int  inputfd;		/* stdin  */
     FILE *mfp;			/* Storage-bound mail-file fp */
+    long messagesize;		/* Reset at MAIL, add at BDATs / DATA */
     long sizeoptval;		/* "MAIL FROM:<xxx> SIZE=nnn" -value    */
     long sizeoptsum;
     char myhostname[MAXHOSTNAMELEN + 1];
@@ -176,7 +294,9 @@ typedef struct {
     struct policystate policystate;
     int  policyresult, reject_net;
     int  postmasteronly;
+    int  netconnected_flg;
     int  tarpit;
+    int  tarpit_cval;		/* current tarpit value */
     int  rport;
     char ihostaddr[sizeof("[ipv6.ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255]") + 8];
     Usockaddr raddr;
@@ -189,25 +309,28 @@ typedef struct {
 #define DELIVERBY_N  2
 #define DELIVERBY_T  4
 
-    int   sslmode;		/* Set, when SSL/TLS in running */
-#ifdef HAVE_OPENSSL
-    SSL * ssl;
-    char *sslwrbuf;
+    char *sslwrbuf;		/* Despite of the name, all modes
+				   use this write-out buffer.. */
     int   sslwrspace, sslwrin, sslwrout;
     /* space, how much stuff in, where the output cursor is */
-#endif /* - HAVE_OPENSSL */
-    char *tls_cipher_info;
-    char *tls_peer_subject;
-    char *tls_peer_issuer;
-    char *tls_peer_fingerprint;
 
+    int   sslmode;		/* Set, when SSL/TLS in running */
+#ifdef HAVE_OPENSSL
+    struct smtpserver_ssl_subset TLS; /* TLS specific things */
+#endif /* - HAVE_OPENSSL */
+
+    int  read_alarm_ival;
     int  s_bufread;
     int  s_readout;
     int  s_status;
+    int  s_readerrno;
     char s_buffer[SMTPLINESIZE];
+    int  s_ungetcbuf;
+    int  s_seen_pipeline;
 
     int  from_box;		/* Set when:  MAIL FROM:<>  */
     int  rcpt_count;
+    int  ok_rcpt_count;
     int  sender_ok;
     /* For BDAT -command */
     int  bdata_blocknum;
@@ -224,9 +347,15 @@ typedef struct {
 #endif
 
     int unknown_cmd_count;
+    long sameipcount;
+
+#ifdef HAVE_SASL2
+    struct SmtpSASLState sasl;
+#endif
 
 } SmtpState;
 
+
 #define STYLE(i,c)	(strchr(((i)==NULL ? style : (i)->flags), (c)) != NULL)
 
 #define	WITH_SMTP	"SMTP"
@@ -238,12 +367,15 @@ extern char *helplines[];
 #define HDR220MAX 4
 extern char *hdr220lines[];
 extern char logtag[];
+extern time_t logtagepoch, now;
 
 extern long availspace;
 extern long minimum_availspace;
 extern long maxsize;
+extern int use_ipv6;
 extern int tarpit_initial;
 extern int tarpit_exponent;
+extern int tarpit_toplimit;
 extern int MaxErrorRecipients;
 extern int TcpRcvBufferSize;
 extern int TcpXmitBufferSize;
@@ -266,6 +398,7 @@ extern int auth_ok;
 extern int ehlo_ok;
 extern int etrn_ok;
 extern int starttls_ok;
+extern int ssmtp_listen;
 extern int msa_mode;
 extern int deliverby_ok;
 #define MAX_ETRN_CLUSTER_IDX 40
@@ -273,7 +406,12 @@ typedef struct {
   char *nodename; char *username; char *password;
 } etrn_cluster_ent;
 extern etrn_cluster_ent etrn_cluster[];
-extern char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
+extern const char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
+extern const char *tls_dcert_file, *tls_dkey_file, *tls_dh1024_param;
+extern const char *tls_dh512_param;
+extern const char *tls_random_source;
+extern const char *tls_cipherlist;
+extern const char *SASL_Auth_Mechanisms;
 extern int tls_loglevel, tls_enforce_tls, tls_ccert_vd, tls_use_scache;
 extern int tls_ask_cert, tls_req_cert, tls_scache_timeout;
 extern int log_rcvd_whoson, log_rcvd_ident, log_rcvd_authuser;
@@ -281,6 +419,7 @@ extern int log_rcvd_tls_mode, log_rcvd_tls_peer;
 extern int auth_login_without_tls;
 extern char *smtpauth_via_pipe;
 extern char *contentfilter;
+extern int debug_content_filter;
 extern int strict_protocol;
 extern int rcptlimitcnt;
 extern int enable_router;
@@ -288,10 +427,17 @@ extern int use_tcpwrapper;
 extern int configuration_ok;
 extern int unknown_cmd_limit;
 extern int sum_sizeoption_value;
+extern int lmtp_mode;
+extern int do_sasl;
+extern int MaxSLBits;
+extern char *AuthMechanisms;
+extern int detect_incorrect_tls_use;
+extern int force_rcpt_notify_never;
 
 extern int bindaddr_set, bindport_set, testaddr_set;
 extern u_short   bindport;
-extern Usockaddr bindaddr;
+extern Usockaddr *bindaddrs;
+extern int        bindaddrs_count;
 extern Usockaddr testaddr;
 
 extern const char *progname;
@@ -303,7 +449,6 @@ extern struct smtpconf *readcffile __((const char *fname));
 extern struct smtpconf *findcf __((const char *host));
 
 extern int loadavg_current __((void));
-extern long fd_statfs __((int));
 
 extern const char *rfc821_domain __((const char *s, int strict));
 extern const char *rfc821_path __((const char *s, int strict));
@@ -372,11 +517,14 @@ extern void header_from_mime __((char *, int *, int));
 #endif				/* USE_TRANSLATION */
 
 extern void killr __((SmtpState * SS, int rpid));
+extern void killcfilter __((SmtpState * SS, int rpid));
 extern void typeflush __((SmtpState *));
 #if defined(HAVE_STDARG_H) && defined(HAVE_VPRINTF)
-extern void type __((SmtpState *, const int code, const char *status, const char *fmt,...));
+extern void type __((SmtpState *, int code, const char *status, const char *fmt,...));
+extern void Z_printf __(( SmtpState *, const char *fmt, ... ));
 #else
 extern void type __(( /* SmtpState *SS, int code, const char *status, const char *fmt, ... */ ));
+extern void Z_printf __(( /* SmtpState *, const char *fmt, ... */ ));
 #endif
 extern void debug_report __((SmtpState *, int, const char *, const char *));
 extern void header_to_mime __((char *, int *, int));
@@ -397,9 +545,10 @@ extern int kill __((pid_t, int));
 extern const char *rfc822atom __((const char *str));
 extern const char *xtext_string __((const char *str));
 
-extern void s_setup __((SmtpState * SS, int infd, int outfd));
+extern void s_setup  __((SmtpState * SS, int infd, int outfd));
+extern void s_ungetc __((SmtpState *SS, int ch));
 extern int s_feof __((SmtpState * SS));
-extern int s_getc __((SmtpState * SS));
+extern int s_getc __((SmtpState * SS, int timeout_is_fatal));
 extern int s_hasinput __((SmtpState * SS));
 extern int s_gets __((SmtpState *SS, char *buf, int buflen, int *rcp, char *cop, char *cp));
 
@@ -420,12 +569,12 @@ extern char *optarg;
 #endif				/* lint */
 
 extern int  childsameip __((Usockaddr *addr, int *childcntp));
-extern void childregister __((int cpid, Usockaddr *addr));
+extern void childregister __((int cpid, Usockaddr *addr, int tag));
 extern void childreap   __((int cpid));
 
 extern void smtp_helo   __((SmtpState * SS, const char *buf, const char *cp));
-extern void smtp_mail   __((SmtpState * SS, const char *buf, const char *cp, int insecure));
-extern void smtp_rcpt   __((SmtpState * SS, const char *buf, const char *cp));
+extern int  smtp_mail   __((SmtpState * SS, const char *buf, const char *cp, int insecure));
+extern int  smtp_rcpt   __((SmtpState * SS, const char *buf, const char *cp));
 extern void smtp_turnme __((SmtpState * SS, const char *name, const char *cp));
 extern void smtp_verify __((SmtpState * SS, const char *buf, const char *cp));
 extern void smtp_expand __((SmtpState * SS, const char *buf, const char *cp));
@@ -435,17 +584,21 @@ extern void add_to_toplevels __((char *str));
 extern void smtp_tarpit __((SmtpState * SS));
 
 extern void smtp_auth __((SmtpState * SS, const char *buf, const char *cp));
+extern void smtpauth_init __((SmtpState * SS));
+extern void smtpauth_ehloresponse __((SmtpState * SS));
 
 #ifdef HAVE_OPENSSL
+extern int tls_start_servertls __((SmtpState *SS));
 extern void smtp_starttls __((SmtpState * SS, const char *buf, const char *cp));
 extern void Z_init    __(( void ));
 extern void Z_cleanup __(( SmtpState * ));
+extern int  Z_SSL_flush __(( SmtpState * ));
 #endif /* - HAVE_OPENSSL */
 extern int  Z_pending __(( SmtpState * ));
 extern int  Z_write   __(( SmtpState *, const void *, int ));
 extern int  Z_read    __(( SmtpState *, void *, int ));
 
-#if 0 /* NOT ANYMORE! */
+#ifdef USE_TCPWRAPPER
 #ifdef HAVE_TCPD_H		/* The hall-mark of having tcp-wrapper things around */
 extern int wantconn __((int sock, char *prgname));
 #endif
@@ -461,3 +614,8 @@ extern char *rfc822date __((time_t *));
 
 extern int encodebase64string __((const char *instr, int inlen, char *outstr, int outspc));
 extern int decodebase64string __((const char *instr, int inlen, char *outstr, int outspc, const char **inleftover));
+
+/* transports/libta/nonblocking.c */
+extern int  fd_nonblockingmode __((int fd));
+extern int  fd_blockingmode __((int fd));
+extern void fd_restoremode __((int fd, int mode));
