@@ -1,6 +1,8 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Copyright 1992-2003 Matti Aarnio -- this & that smaller, and
+ *	larger changes...
  */
 
 #include "hostenv.h"
@@ -18,6 +20,9 @@
 #include "zsyslog.h"
 
 #include "libz.h"
+#include "libc.h"
+
+#include "shmmib.h"
 
 #define	PROGNAME	"errormail"
 #define	CHANNEL		"error"	/* the default channel name we deliver for */
@@ -34,12 +39,14 @@ const char *dfltform[] = {
 		NULL
 };
 
-char *progname;
+const char *progname;
 
 extern char *optarg;
 extern int  optind;
 extern void process __((struct ctldesc *));
 
+FILE *verboselog = NULL;
+
 #ifndef strchr
 extern char *strrchr(), *strchr();
 #endif
@@ -51,6 +58,43 @@ extern char *strrchr(), *strchr();
 int D_alloc = 0; /* For tmalloc() from libz.a ... */
 
 
+static void MIBcountCleanup __((void))
+{
+	MIBMtaEntry->taerrm.TaProcCountG -= 1;
+}
+
+static void SHM_MIB_diag(rc)
+     const int rc;
+{
+  switch (rc) {
+  case EX_OK:
+    /* OK */
+    MIBMtaEntry->taerrm.TaRcptsOk ++;
+    break;
+  case EX_TEMPFAIL:
+  case EX_IOERR:
+  case EX_OSERR:
+  case EX_CANTCREAT:
+  case EX_SOFTWARE:
+  case EX_DEFERALL:
+    /* DEFER */
+    MIBMtaEntry->taerrm.TaRcptsRetry ++;
+    break;
+  case EX_NOPERM:
+  case EX_PROTOCOL:
+  case EX_USAGE:
+  case EX_NOUSER:
+  case EX_NOHOST:
+  case EX_UNAVAILABLE:
+  default:
+    /* FAIL */
+    MIBMtaEntry->taerrm.TaRcptsFail ++;
+    break;
+  }
+}
+
+
+
 int
 main(argc, argv)
 	int argc;
@@ -76,6 +120,17 @@ main(argc, argv)
 	  SIGNAL_HANDLE(SIGQUIT, wantout);
 	SIGNAL_IGNORE(SIGPIPE);
 
+	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
+
+
+	Z_SHM_MIB_Attach(1); /* we don't care if it succeeds or fails.. */
+
+	MIBMtaEntry->taerrm.TaProcessStarts += 1;
+	MIBMtaEntry->taerrm.TaProcCountG    += 1;
+
+	atexit(MIBcountCleanup);
+
+
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
 	else
@@ -105,7 +160,7 @@ main(argc, argv)
 	  exit(EX_USAGE);
 	}
 
-	/* We need this latter on .. */
+	/* We need this later on .. */
 	zopenlog("errormail", LOG_PID, LOG_MAIL);
 
 	while (!getout) {
@@ -118,8 +173,10 @@ main(argc, argv)
 	  if (strchr(msgfilename, '\n') == NULL) break; /* No ending '\n' !
 							   Must have been
 							   partial input! */
-	  if (strcmp(msgfilename, "#idle\n") == 0)
+	  if (strcmp(msgfilename, "#idle\n") == 0) {
+	    MIBMtaEntry->taerrm.TaIdleStates += 1;
 	    continue; /* Ah well, we can stay idle.. */
+	  }
 
 	  /* Input:
 	       spool/file/name [ \t host.info ] \n
@@ -128,11 +185,13 @@ main(argc, argv)
 	  if (emptyline(msgfilename, sizeof msgfilename))
 	    break;
 
+	  MIBMtaEntry->taerrm.TaMessages += 1;
+
 	  host = strchr(msgfilename,'\t');
 	  if (host != NULL)
 	    *host++ = 0;
 
-	  dp = ctlopen(msgfilename, channel, host, &getout, ctlsticky, NULL, NULL, NULL);
+	  dp = ctlopen(msgfilename, channel, host, &getout, ctlsticky, NULL);
 	  if (dp != NULL) {
 	    process(dp);
 	    ctlclose(dp);
@@ -236,10 +295,14 @@ process(dp)
 	char lastchar;
 	int reportcnt = 0;
 	struct stat stbuf;
+	long inum;
+	time_t mtime;
 
 	if (fstat(dp->msgfd, &stbuf) != 0)
 	  abort(); /* This is a "CAN'T FAIL" case.. */
 
+	MIBMtaEntry->taerrm.TaDeliveryStarts += 1;
+
 	/* recipient host field is the error message file name in FORMSDIR */
 	/* recipient user field is the address causing the error */
 
@@ -251,14 +314,18 @@ process(dp)
 	   * We do NOT want to bounce this, but
 	   * instead just drop it on the floor.
 	   */
-	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    diagnostic(rp, EX_OK, 0, "error bounce dropped");
+	  for (rp = dp->recipients; rp != NULL; rp = rp->next) {
+	    diagnostic(verboselog, rp, EX_OK, 0, "error bounce dropped");
+	    SHM_MIB_diag(EX_OK);
+	  }
 	  return;
 	}
 
 	if ((mfp = sfmail_open(MSG_RFC822)) == NULL) {
-	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    diagnostic(rp, EX_TEMPFAIL, 0, "sfmail_open failure");
+	  for (rp = dp->recipients; rp != NULL; rp = rp->next) {
+	    diagnostic(verboselog, rp, EX_TEMPFAIL, 0, "sfmail_open failure");
+	    SHM_MIB_diag(EX_TEMPFAIL);
+	  }
 	  warning("Cannot open mail file!");
 	  return;
 	}
@@ -274,11 +341,16 @@ process(dp)
 	}
 
 	sfprintf(mfp, "channel error\n");
+	sfprintf(mfp, "errormsg\n");
 
 	rp = dp->recipients;
 
 	sfprintf(mfp, "todsn NOTIFY=NEVER ORCPT=RFC822;");
 	encodeXtext(mfp, rp->addr->link->user, strlen(rp->addr->link->user));
+
+	if (STREQ(rp->addr->link->channel,"error"))
+	  rp->addr->link->user = "postmaster";
+
 	sfprintf(mfp, "\nto <%s>\n",rp->addr->link->user);
 
 	/* copy error message file itself */
@@ -356,8 +428,10 @@ process(dp)
 	  /* No, throw it away and ack success. */
 	  sfmail_abort(mfp);
 	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    if (!(rp->notifyflgs & _DSN_NOTIFY_FAILURE))
-	      diagnostic(rp, EX_OK, 0, NULL);
+	    if (!(rp->notifyflgs & _DSN_NOTIFY_FAILURE)) {
+	      diagnostic(verboselog, rp, EX_OK, 0, "discarded, reportcnt=0");
+	      SHM_MIB_diag(EX_OK);
+	    }
 	  return;
 	}
 
@@ -420,6 +494,9 @@ process(dp)
 	lseek(dp->msgfd, (off_t)(dp->msgbodyoffset), SEEK_SET);
 
 	/* write the (new) headers with local "Received:"-line.. */
+
+	header_received_for_clause(rp, 0, verboselog);
+
 	swriteheaders(rp, mfp, "\n", 0, 0, NULL);
 	sfprintf(mfp,"\n");
 
@@ -440,11 +517,18 @@ process(dp)
 	if (sferror(mfp)) {
 	  sfmail_abort(mfp);
 	  n = EX_IOERR;
-	} else if (sfmail_close(mfp) == EOF)
+	} else if (_sfmail_close_(mfp, &inum, &mtime) == EOF)
 	  n = EX_IOERR;
 	else
 	  n = EX_OK;
-	for (rp = dp->recipients; rp != NULL; rp = rp->next) {
-	  diagnostic(rp, n, 0, (char *)NULL);
+	{
+	  /* Ok, build response with proper "spoolid" */
+	  char taspid[30];
+	  taspoolid(taspid, mtime, inum);
+
+	  for (rp = dp->recipients; rp != NULL; rp = rp->next) {
+	    diagnostic(verboselog, rp, n, 0, taspid);
+	    SHM_MIB_diag(n);
+	  }
 	}
 }
