@@ -3,7 +3,7 @@
  *
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1992-1997 Matti Aarnio.
+ *	Copyright 1992-2003 Matti Aarnio.
  */
 
 /*
@@ -18,6 +18,8 @@
  *  - optional host selector (empty if third parameter is given)
  *  - optional explanatory string (up to newline)
  *
+ * The scheduler will supply '-c' option to drive the channel selection.
+ *
  */
 
 #include "hostenv.h"
@@ -42,8 +44,9 @@
 
 #include "zmalloc.h"
 #include "libz.h"
-#include "splay.h"
+#include "libc.h"
 
+#include "shmmib.h"
 
 
 const char *defcharset;
@@ -51,13 +54,12 @@ const char *progname;
 const char *channel;
 const char *logfile;
 FILE *logfp = NULL;
-FILE *verboselog = NULL;
 
 extern RETSIGTYPE wantout __((int));
 extern int optind;
 extern char *optarg;
 extern void prversion __((const char *));
-extern void process __((struct ctldesc *dp, const char *optmsg, const int));
+extern void process __((struct ctldesc *dp, const char *optmsg, const int, FILE *verboselog));
 extern void deliver __((struct ctldesc *dp, struct rcpt *rp));
 
 extern void hp_init();
@@ -83,6 +85,42 @@ int sig;
 	/* Sigh, actually dummy routine.. */
 }
 
+static void MIBcountCleanup __((void))
+{
+	MIBMtaEntry->taexpi.TaProcCountG -= 1;
+}
+
+static void SHM_MIB_diag(rc)
+     const int rc;
+{
+  switch (rc) {
+  case EX_OK:
+    /* OK */
+    MIBMtaEntry->taexpi.TaRcptsOk ++;
+    break;
+  case EX_TEMPFAIL:
+  case EX_IOERR:
+  case EX_OSERR:
+  case EX_CANTCREAT:
+  case EX_SOFTWARE:
+  case EX_DEFERALL:
+    /* DEFER */
+    MIBMtaEntry->taexpi.TaRcptsRetry ++;
+    break;
+  case EX_NOPERM:
+  case EX_PROTOCOL:
+  case EX_USAGE:
+  case EX_NOUSER:
+  case EX_NOHOST:
+  case EX_UNAVAILABLE:
+  default:
+    /* FAIL */
+    MIBMtaEntry->taexpi.TaRcptsFail ++;
+    break;
+  }
+}
+
+
 
 #ifndef	MAXPATHLEN
 #define	MAXPATHLEN 1024
@@ -101,6 +139,7 @@ main(argc, argv)
 	char *host = NULL;	/* .. and what is my host ? */
 	int matchhost = 0;
 	struct ctldesc *dp;
+	FILE *verboselog = NULL;
 
 	RETSIGTYPE (*oldsig) __((int));
 
@@ -121,6 +160,17 @@ main(argc, argv)
 
 	SIGNAL_IGNORE(SIGPIPE);
 
+	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
+
+
+	Z_SHM_MIB_Attach(1); /* we don't care if it succeeds or fails.. */
+
+	MIBMtaEntry->taexpi.TaProcessStarts += 1;
+	MIBMtaEntry->taexpi.TaProcCountG    += 1;
+
+	atexit(MIBcountCleanup);
+
+
 	progname = strrchr(argv[0], '/');
 	if (progname == NULL)
 	  progname = argv[0];
@@ -171,7 +221,8 @@ main(argc, argv)
 	  exit(EX_NOPERM);
 	}
 
-	setreuid(0, 0);		/* make us root all over */
+	SETUID(0);		/* make us root all over */
+	SETEUID(0);		/* make us root all over */
 
 
 	logfp = NULL;
@@ -187,7 +238,7 @@ main(argc, argv)
 	  }
 	}
 
-	/* We need this latter on .. */
+	/* We need this later on .. */
 	zopenlog("expirer", LOG_PID, LOG_MAIL);
 
 	getnobody();
@@ -195,7 +246,7 @@ main(argc, argv)
 	while (!getout) {
 
 	  /* Input:
-	       spool/file/name [ \t host.info ] \n
+	       spool/file/name [ \t host.info [ \t explanation ]] \n
 	   */
 
 	  printf("#hungry\n");
@@ -205,11 +256,16 @@ main(argc, argv)
 	    break;
 	  if (strchr(file, '\n') == NULL) break; /* No ending '\n' !  Must
 						    have been partial input! */
-	  if (strcmp(file, "#idle\n") == 0)
+	  if (strcmp(file, "#idle\n") == 0) {
+	    MIBMtaEntry->taexpi.TaIdleStates += 1;
 	    continue; /* Ah well, we can stay idle.. */
+	  }
 	  if (emptyline(file, sizeof file))
 	    break;
 
+	  MIBMtaEntry->taexpi.TaMessages += 1;
+
+
 	  s = strchr(file,'\t');
 	  if (s != NULL) {
 	    if (host) free(host);
@@ -217,12 +273,13 @@ main(argc, argv)
 	    *s = 0;
 	  }
 
-	  setreuid(0,0); /* We begin as roots..  process() may change us */
+	  SETUID(0); /* We begin as roots..  process() may change us */
+	  SETEUID(0); /* We begin as roots..  process() may change us */
 
 	  notary_setxdelay(0); /* Our initial speed estimate is
 				  overtly optimistic.. */
 
-	  dp = ctlopen(file, channel, host, &getout, NULL, NULL, NULL, NULL);
+	  dp = ctlopen(file, channel, host, &getout, NULL, NULL);
 	  if (dp == NULL) {
 	    printf("#resync %s\n",file);
 	    fflush(stdout);
@@ -239,7 +296,7 @@ main(argc, argv)
 	      fcntl(FILENO(verboselog), F_SETFD, 1);
 	    }
 	  }
-	  process(dp, optmsg, silent);
+	  process(dp, optmsg, silent, verboselog);
 	  ctlclose(dp);
 
 	}
@@ -250,25 +307,31 @@ main(argc, argv)
 
 
 void
-process(dp, optmsg, silent)
+process(dp, optmsg, silent, verboselog)
 	struct ctldesc *dp;
 	const char *optmsg;
 	const int silent;
+	FILE *verboselog;
 {
 	struct rcpt *rp;
 
 	if (optmsg == NULL || *optmsg == 0)
 	  optmsg = "x-local; 500 (Administrative message deletion from delivery queue)";
 
+	MIBMtaEntry->taexpi.TaDeliveryStarts += 1;
+
+
 	for (rp = dp->recipients; rp != NULL; rp = rp->next) {
 
 	  notary_setxdelay(0);
 	  if (silent) {
 	    notaryreport(rp->addr->user, "delivered", "2.0.0 (Silent ok)", "");
-	    diagnostic(rp, EX_OK, 0, "");
+	    diagnostic(verboselog, rp, EX_OK, 0, "");
+	    SHM_MIB_diag(EX_OK);
 	  } else {
 	    notaryreport(rp->addr->user, "failed", "5.7.0 (Administrative deletion command)", optmsg);
-	    diagnostic(rp, EX_UNAVAILABLE, 0, "%s", optmsg);
+	    diagnostic(verboselog, rp, EX_UNAVAILABLE, 0, "%s", optmsg);
+	    SHM_MIB_diag(EX_UNAVAILABLE);
 	  }
 	}
 }
