@@ -1,6 +1,7 @@
 /*
  *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Copyright 1992-2003 Matti Aarnio -- MIME processing et.al.
  */
 
 #include "mailer.h"
@@ -20,13 +21,17 @@
 #include "zsyslog.h"
 
 #include "ta.h"
+#include "libz.h"
+#include "libc.h"
+
+#include "shmmib.h"
 
 #ifdef	HAVE_RESOLVER
 #include "netdb.h"
 #include <sys/socket.h>
 #include <netinet/in.h>
 #endif
-#include "libz.h"
+#include "dnsgetrr.h"
 
 #if	defined(TRY_AGAIN) && defined(HAVE_RESOLVER)
 #define	BIND		/* Want BIND (named) nameserver support enabled */
@@ -112,7 +117,7 @@ typedef char msgdata;
 #define	CHANNEL		"hold"	/* the default channel name we look at */
 
 char errormsg[BUFSIZ];
-char *progname;
+const char *progname;
 char *cmdline, *eocmdline;
 int pid;
 
@@ -138,6 +143,46 @@ extern char *strchr(), *strrchr();
 #define	putc	fputc
 #endif	/* lint */
 
+
+
+static void MIBcountCleanup __((void))
+{
+	MIBMtaEntry->tahold.TaProcCountG -= 1;
+}
+
+static void SHM_MIB_diag(rc)
+     const int rc;
+{
+  switch (rc) {
+  case EX_OK:
+    /* OK */
+    MIBMtaEntry->tahold.TaRcptsOk ++;
+    break;
+  case EX_TEMPFAIL:
+  case EX_IOERR:
+  case EX_OSERR:
+  case EX_CANTCREAT:
+  case EX_SOFTWARE:
+  case EX_DEFERALL:
+    /* DEFER */
+    MIBMtaEntry->tahold.TaRcptsRetry ++;
+    break;
+  case EX_NOPERM:
+  case EX_PROTOCOL:
+  case EX_USAGE:
+  case EX_NOUSER:
+  case EX_NOHOST:
+  case EX_UNAVAILABLE:
+  default:
+    /* FAIL */
+    MIBMtaEntry->tahold.TaRcptsFail ++;
+    break;
+  }
+}
+
+
+FILE *verboselog = NULL;
+
 static char filename[MAXPATHLEN+8000];
 
 int
@@ -169,6 +214,17 @@ main(argc, argv)
 	if (oldsig != SIG_IGN)
 	  SIGNAL_HANDLE(SIGHUP, wantout);
 
+	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
+
+
+	Z_SHM_MIB_Attach(1); /* we don't care if it succeeds or fails.. */
+
+	MIBMtaEntry->tahold.TaProcessStarts += 1;
+	MIBMtaEntry->tahold.TaProcCountG    += 1;
+
+	atexit(MIBcountCleanup);
+
+
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
 	else
@@ -198,7 +254,7 @@ main(argc, argv)
 	  exit(EX_USAGE);
 	}
 
-	/* We need this latter on .. */
+	/* We need this later on .. */
 	zopenlog("hold", LOG_PID, LOG_MAIL);
 
 	while (!getout) {
@@ -213,11 +269,16 @@ main(argc, argv)
 	    break;
 	  if (strchr(filename, '\n') == NULL) break; /* No ending '\n' !  Must
 						    have been partial input! */
-	  if (strcmp(filename, "#idle\n") == 0)
+	  if (strcmp(filename, "#idle\n") == 0) {
+	    MIBMtaEntry->tahold.TaIdleStates += 1;
 	    continue; /* Ah well, we can stay idle.. */
+	  }
 	  if (emptyline(filename, sizeof(filename)))
 	    break;
 
+	  MIBMtaEntry->tahold.TaMessages += 1;
+
+
 	  s = strchr(filename,'\t');
 	  host = NULL;
 	  if (s != NULL) {
@@ -226,7 +287,7 @@ main(argc, argv)
 	    host = s;
 	  }
 
-	  dp = ctlopen(filename, channel, host, &getout, NULL, NULL, NULL, NULL);
+	  dp = ctlopen(filename, channel, host, &getout, NULL, NULL);
 	  if (dp != NULL) {
 	    process(dp);
 	    ctlclose(dp);
@@ -254,6 +315,8 @@ process(dp)
 	const char *cp;
 	char buf[BUFSIZ];
 
+	MIBMtaEntry->tahold.TaDeliveryStarts += 1;
+
 	sawok = 0;
 	for (rp = dp->recipients; rp != NULL; rp = rp->next) {
 	  cp = rp->addr->user;
@@ -284,7 +347,8 @@ process(dp)
 	      break;
 	    }
 	    notaryreport(rp->addr->user,action,status,diagnostics);
-	    diagnostic(rp, rp->status, 0, "%s", rp->addr->user);
+	    diagnostic(verboselog, rp, rp->status, 0, "%s", rp->addr->user);
+	    SHM_MIB_diag(rp->status);
 	  }
 	}
 
@@ -294,7 +358,7 @@ process(dp)
 	if (lseek(dp->msgfd, (off_t)(dp->msgbodyoffset), SEEK_SET) < 0L)
 	  warning("Cannot seek to message body! (%m)", (char *)NULL);
 
-	setreuid(-1, atoi(dp->senders->misc));
+	SETEUID(atoi(dp->senders->misc));
 
 	mfp = mail_open(MSG_RFC822);
 	if (mfp == NULL) {
@@ -303,16 +367,17 @@ process(dp)
 	      notaryreport(rp->addr->user,"delayed",
 			   "4.3.1 (System spool full?)",
 			   "x-local; 400 (Cannot resubmit anything, out of spool space?)");
-	      diagnostic(rp, EX_TEMPFAIL, 0,
+	      diagnostic(verboselog, rp, EX_TEMPFAIL, 0,
 			 "cannot resubmit anything!");
+	      SHM_MIB_diag(EX_TEMPFAIL);
 	    }
-	  setreuid(-1, getuid());
+	  SETEUID(getuid());
 	  return;
 	}
-	setreuid(-1, getuid());
+	SETEUID(getuid());
 
 	fprintf(mfp, "via suspension\n");
-	if (dp->senders->user[0] == 0)
+	if (STREQ(dp->senders->channel,"error"))
 	  fprintf(mfp, "channel error\n");
 	else
 	  fprintf(mfp, "from <%s>\n", dp->senders->user);
@@ -323,18 +388,28 @@ process(dp)
 
 	for (rp = dp->recipients; rp != NULL; rp = rp->next)
 	  if (rp->status == EX_OK) {
-	    fprintf(mfp, "to <%s>\n", rp->addr->user);
-	    if (rp->notify || rp->orcpt) {
-
+	    if ( rp->notify || rp->orcpt  ||
+		 rp->inrcpt || rp->infrom ||
+		 rp->deliverby                 ) {
+	      
 	      fputs("todsn",mfp);
 	      if (rp->orcpt != NULL)
 		fprintf(mfp," ORCPT=%s",rp->orcpt);
 
+	      if (rp->inrcpt != NULL)
+		fprintf(mfp," INRCPT=%s",rp->inrcpt);
+
+	      if (rp->infrom != NULL)
+		fprintf(mfp," INFROM=%s",rp->infrom);
+
+	      if (rp->ezmlm != NULL)
+		fprintf(mfp," EZMLM=%s",rp->ezmlm);
+
 	      if (rp->notify)
 		fprintf(mfp," NOTIFY=%s", rp->notify);
 
 	      if (rp->deliverby) {
-		fprintf(mfp," BY=%ld;", rp->deliverby);
+		fprintf(mfp," BY=%ld;", (long)rp->deliverby);
 		if (rp->deliverbyflgs & _DELIVERBY_R) fputc('R',mfp);
 		if (rp->deliverbyflgs & _DELIVERBY_N) fputc('N',mfp);
 		if (rp->deliverbyflgs & _DELIVERBY_T) fputc('T',mfp);
@@ -342,7 +417,9 @@ process(dp)
 
 	      putc('\n',mfp);
 	    }
+	    fprintf(mfp, "to <%s>\n", rp->addr->user);
 	  }
+	fprintf(mfp,"env-end\n");
 
 	fwriteheaders(dp->recipients,mfp,"\n",0,0,NULL);
 	fprintf(mfp,"\n");
@@ -368,7 +445,8 @@ process(dp)
 	    notaryreport(rp->addr->user,"relayed",
 			 "2.2.0 (Relayed via deferral channel)",
 			 "x-local; 250 (Relayed via deferral channel)");
-	    diagnostic(rp, code, 0, cp);
+	    diagnostic(verboselog, rp, code, 0, cp);
+	    SHM_MIB_diag(code);
 	  }
 }
 
