@@ -1,4 +1,4 @@
-/* Merged into Zmailer by Edwin Allum -- and a bit latter with some
+/* Merged into Zmailer by Edwin Allum -- and a bit later with some
    modifications -- to make the compile optional -- by Matti Aarnio */
 
 /*
@@ -25,6 +25,7 @@
 #define MAXFAIL		5	/* max # checks before we give up */
 
 
+#ifndef HAVE_GETHOSTID
 /*
  * Solaris 2.x equivalent of gethostid().
  * This should be configured in differently, but for
@@ -33,6 +34,7 @@
 #if defined(sun) && defined(__svr4__)
 #include <sys/systeminfo.h>
 
+static int gethostid __((void));
 static int gethostid()
 {
 	char buf[1024];
@@ -41,9 +43,11 @@ static int gethostid()
 	return atoi(buf);
 }
 #endif
+#endif
 
 #if defined(HAVE_DOTLOCK) /* Well, we DO use DOTLOCK scheme! */
 
+#ifndef HAVE_GETHOSTID
 #ifdef __hpux
 /*
  * From: dd@mv.us.adobe.com (David DiGiacomo)
@@ -52,6 +56,7 @@ static int gethostid()
 #define _INCLUDE_HPUX_SOURCE
 #include <sys/utsname.h>
 
+static long gethostid __((void));
 static long gethostid()
 {
         struct utsname uts;
@@ -62,22 +67,22 @@ static long gethostid()
         return atoi(uts.idnumber);
 }
 #endif /* __hpux */
+#endif
 
-
+int dotlock __((const char *file));
 int
 dotlock(file)
-	char *file;
+	const char *file;
 {
 	char lockname[MAXPATHLEN];
 	char temp[MAXPATHLEN];
 	struct stat st;
 	time_t now;
 	int fd, i, fail = 0;
-	extern int errno;
 
 	sprintf(lockname, "%s.lock", file);
 	sprintf(temp, "%s.L%x.%lx.%x",
-		file, getpid(), time(NULL), gethostid());
+		file, (int)getpid(), (long)time(NULL), (int)gethostid());
 	unlink(temp);
 	for (;;) {
 		if ((fd = open(temp, O_WRONLY|O_CREAT|O_EXCL, 0666)) < 0)
@@ -104,9 +109,10 @@ dotlock(file)
 	}
 }
 
+int dotunlock __((const char *));
 int
 dotunlock(file)
-	char *file;
+	const char *file;
 {
 	char lockname[MAXPATHLEN];
 
