@@ -1,7 +1,8 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1994-1997 by Matti Aarnio -- MIME processings
+ *	Copyright 1994-2000,2002-2003 by Matti Aarnio -- MIME processings,
+ *	etc.
  */
 
 #define DefCharset "ISO-8859-1"
@@ -22,21 +23,20 @@
 #include <unistd.h>
 #endif
 #include <errno.h>
+
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
 #include "libz.h"
+#include "libc.h"
 
+#include "shmmib.h"
 
-#ifdef  HAVE_WAITPID
+#ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
 # include <sys/wait.h>
-#else
-# ifdef HAVE_WAIT3
-#  include <sys/wait.h> /* Has BSD wait3() */
-# else
-#  ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
-#   include <sys/wait.h>
-#  else /* Not POSIX.1 compatible, lets fake it.. */
+#else /* Not POSIX.1 compatible, lets fake it.. */
 extern int wait();
-#  endif
-# endif
 #endif
 
 #ifndef WEXITSTATUS
@@ -51,7 +51,8 @@ extern int wait();
 #endif	/* !SEEK_SET */
 
 /* as in: SKIPWHILE(isascii,cp) */
-#define	SKIPWHILE(X,Y)	while (*Y != '\0' && isascii(*Y) && X(*Y)) { ++Y; }
+#define SKIPSPACE(Y) while (*Y == ' ' || *Y == '\t' || *Y == '\n') ++Y
+#define SKIPTEXT(Y)  while (*Y && *Y != ' ' && *Y != '\t' && *Y != '\n') ++Y
 
 #define	FROM_	"From "
 
@@ -61,19 +62,17 @@ extern int wait();
 
 char uucpname[MAXHOSTNAMELEN+1];
 
-char	*progname;
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
+const char	*progname;
 int	readalready = 0;	/* does buffer contain valid message data? */
-#endif
 int	mimeqpnarrow = 0;	/* Can't send TAB thru without MIME-QP */
-FILE	*verboselog = NULL;
 FILE	*logfp   = NULL;
 int	maxwidth = 0;
 int	can_8bit = 0;		/* Can do 8-bit stuff! */
+int	decode_qp = 0;
 int	keep_header8 = 0;	/* Don't do "MIME-2" to the headers */
 
 int	D_alloc = 0;		/* Memory debugging */
-char    *defcharset;
+const char *defcharset;
 
 
 extern RETSIGTYPE sigpipe();
@@ -87,6 +86,15 @@ int sig;
 	SIGNAL_HANDLE(SIGPIPE, sigpipe);
 }
 
+struct maildesc {
+	char	*name;
+	long	flags;
+	char	*command;
+#define MD_ARGVMAX 20
+	char	*argv[MD_ARGVMAX];
+};
+
+
 extern RETSIGTYPE wantout();
 #ifndef MALLOC_TRACE
 extern univptr_t emalloc();
@@ -115,15 +123,62 @@ static int zsfsetfd(fp, fd)
   return fd;
 }
 
-extern int check_7bit_cleanness __((struct ctldesc *dp));
+static void
+decodeXtext(mfp,xtext)
+	FILE *mfp;
+	const char *xtext;
+{
+	for (;*xtext;++xtext) {
+	  if (*xtext == '+') {
+	    int c = '?';
+	    sscanf(xtext+1,"%02X",&c);
+	    fputc(c, mfp);
+	    if (*xtext) ++xtext;
+	    if (*xtext) ++xtext;
+	  } else
+	    fputc(*xtext, mfp);
+	}
+}
 
-struct maildesc {
-	char	*name;
-	short	flags;
-	char	*command;
-#define MD_ARGVMAX 20
-	char	*argv[MD_ARGVMAX];
-};
+
+static void MIBcountCleanup __((void))
+{
+	MIBMtaEntry->tasmcm.TaProcCountG -= 1;
+}
+
+static void SM_MIB_diag(rc)
+     const int rc;
+{
+  switch (rc) {
+  case EX_OK:
+    /* OK */
+    MIBMtaEntry->tasmcm.TaRcptsOk ++;
+    break;
+  case EX_TEMPFAIL:
+  case EX_IOERR:
+  case EX_OSERR:
+  case EX_CANTCREAT:
+  case EX_SOFTWARE:
+  case EX_DEFERALL:
+    /* DEFER */
+    MIBMtaEntry->tasmcm.TaRcptsRetry ++;
+    break;
+  case EX_NOPERM:
+  case EX_PROTOCOL:
+  case EX_USAGE:
+  case EX_NOUSER:
+  case EX_NOHOST:
+  case EX_UNAVAILABLE:
+  default:
+    /* FAIL */
+    MIBMtaEntry->tasmcm.TaRcptsFail ++;
+    break;
+  }
+}
+
+
+
+extern int check_7bit_cleanness __((struct ctldesc *dp));
 
 extern int writemimeline __(( struct maildesc *mp, FILE *fp, const char *buf, int len, int convertmode));
 
@@ -147,6 +202,9 @@ extern int writemimeline __(( struct maildesc *mp, FILE *fp, const char *buf, in
 #define MO_WANTSFROM		0x10000 /* Wants "From:" -header */
 #define MO_BSMTPHELO		0x20000 /* Add HELO/EHLO to the BSMTP */
 
+#define MO_XENVELOPES		0x40000 /* Write various X-Envelope-*: headers to mesage */
+#define MO_XORCPT		0x80000 /* Write X-Orcpt : header to message */
+
 struct exmapinfo {
 	int	origstatus;
 	const char *statusmsg;
@@ -194,6 +252,7 @@ main(argc, argv)
 	int errflg, c;
 	struct maildesc *mp;
 	RETSIGTYPE (*oldsig)();
+	FILE *verboselog = NULL;
 
 	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
 	if (oldsig != SIG_IGN)
@@ -210,6 +269,20 @@ main(argc, argv)
 
 	SIGNAL_HANDLE(SIGPIPE, sigpipe);
 
+#if defined(HAVE_LOCALE_H) && defined(HAVE_SETLOCALE) && defined(LC_ALL)
+	setlocale(LC_ALL, "C");
+#endif
+
+	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
+
+	Z_SHM_MIB_Attach(1); /* we don't care if it succeeds or fails.. */
+
+	MIBMtaEntry->tasmcm.TaProcessStarts += 1;
+	MIBMtaEntry->tasmcm.TaProcCountG    += 1;
+
+	atexit(MIBcountCleanup);
+
+
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
 	else
@@ -244,7 +317,7 @@ main(argc, argv)
 	    verboselog = stdout;
 	    break;
 	  case '8':
-	    can_8bit = 1;
+	    can_8bit = decode_qp = 1;
 	    break;
 	  case 'H':
 	    keep_header8 = 1;
@@ -265,7 +338,7 @@ main(argc, argv)
 	}
 	mailer = argv[optind];
 
-	/* We need this latter on .. */
+	/* We need this later on .. */
 	zopenlog("sm", LOG_PID, LOG_MAIL);
 
 	defcharset = getzenv("DEFCHARSET");
@@ -290,11 +363,15 @@ main(argc, argv)
 	    break;
 	  if (strchr(file, '\n') == NULL) break; /* No ending '\n' !  Must
 						    have been partial input! */
-	  if (strcmp(file, "#idle\n") == 0)
+	  if (strcmp(file, "#idle\n") == 0) {
+	    MIBMtaEntry->tasmcm.TaIdleStates += 1;
 	    continue; /* Ah well, we can stay idle.. */
+	  }
 	  if (emptyline(file, sizeof file))
 	    break;
 
+	  MIBMtaEntry->tasmcm.TaMessages += 1;
+
 	  s = strchr(file,'\t');
 	  if (s != NULL) {
 	    if (host) free(host);
@@ -303,7 +380,7 @@ main(argc, argv)
 	  }
 
 	  ctlsticky(NULL, NULL, NULL); /* reset */
-	  dp = ctlopen(file, channel, host, &getout, ctlsticky, NULL, NULL, NULL);
+	  dp = ctlopen(file, channel, host, &getout, ctlsticky, NULL);
 	  if (dp == NULL) {
 	    printf("#resync %s\n",file);
 	    fflush(stdout);
@@ -337,9 +414,7 @@ process(dp, mp, verboselog)
 {
 	struct rcpt *rp, *rphead;
 
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
 	readalready = 0; /* ignore any previous message data cache */
-#endif
 
 	if (mp->flags & MO_MANYUSERS) {
 	  for (rp = rphead = dp->recipients; rp != NULL; rp = rp->next) {
@@ -380,7 +455,16 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	const char *ds, **av, *s;
 	int status;
 	int content_kind, conversion_prohibited, ascii_clean = 0;
+	time_t now;
+	char *timestring;
 	CONVERTMODE convertmode = _CONVERT_NONE;
+	char *lineendseq = "\n";
+
+	MIBMtaEntry->tasmcm.TaDeliveryStarts += 1;
+
+	now = time((time_t *)0);
+	timestring = ctime(&now);
+	*(timestring+strlen(timestring)-1) = '\0';
 
 	if (lseek(dp->msgfd, (off_t)(dp->msgbodyoffset), SEEK_SET) < 0L)
 		warning("Cannot seek to message body! (%m)", (char *)NULL);
@@ -501,9 +585,10 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    notaryreport(rp->addr->user,"failed",
 			 "5.3.0 (Out of system resources, pipe creation failed)",
 			 "x-local; 500 (pipe creation error, out of system resources ?)");
-	    diagnostic(rp, EX_OSERR, 0,
+	    diagnostic(verboselog, rp, EX_OSERR, 0,
 		       "cannot create pipe from \"%s\"",
 		       mp->command);
+	    SM_MIB_diag(EX_OSERR);
 	  }
 	  return;
 	}
@@ -512,9 +597,10 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    notaryreport(rp->addr->user,"failed",
 			 "5.3.0 (Out of system resources, pipe creation failed)",
 			 "x-local; 500 (pipe creation error, out of system resources ?)");
-	    diagnostic(rp, EX_OSERR, 0,
+	    diagnostic(verboselog, rp, EX_OSERR, 0,
 		       "cannot create pipe to \"%s\"",
 		       mp->command);
+	    SM_MIB_diag(EX_OSERR);
 	  }
 	  return;
 	}
@@ -562,7 +648,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    notaryreport(rp->addr->user,"failed",
 			 "5.3.0 (Out of system resources, fork failed)",
 			 "x-local; 500 (fork failure, out of system resources ?)");
-	    diagnostic(rp, EX_OSERR, 0, "cannot fork");
+	    diagnostic(verboselog, rp, EX_OSERR, 0, "cannot fork");
+	    SM_MIB_diag(EX_OSERR);
 	  }
 	  return;
 	}
@@ -610,8 +697,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    if (startrp->desc->dsnretmode != NULL)
 	      fprintf(tafp, " RET=%s", startrp->desc->dsnretmode);
 	  }
-	  if (mp->flags & MO_CRLF) putc('\r',tafp);
-	  putc('\n',tafp);
+	  if (mp->flags & MO_CRLF) lineendseq = "\r\n";
+	  fputs(lineendseq, tafp);
 	  for (rp = startrp; rp != endrp; rp = rp->next) {
 	    fprintf(tafp,"RCPT TO:<%s>",rp->addr->user);
 	    /* if (mp->flags & MO_BESMTP) { } */
@@ -639,31 +726,24 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    }
 	    if (mp->flags & MO_BEBSMTP) {
 	      if (rp->deliverby) {
-		fprintf(tafp," BY=%ld;", rp->deliverby);
+		fprintf(tafp," BY=%ld;", (long)(rp->deliverby - now));
 		if (rp->deliverbyflgs & _DELIVERBY_R) fputc('R',tafp);
 		if (rp->deliverbyflgs & _DELIVERBY_N) fputc('N',tafp);
 		if (rp->deliverbyflgs & _DELIVERBY_T) fputc('T',tafp);
 	      }
 	    }
-	    if (mp->flags & MO_CRLF) putc('\r',tafp);
-	    putc('\n',tafp);
+	    fputs(lineendseq, tafp);
 	  }
 
 	  fprintf(tafp,"DATA");
-	  if (mp->flags & MO_CRLF) putc('\r',tafp);
-	  putc('\n',tafp);
+	  fputs(lineendseq, tafp);
 	}
 
 	/* Now continue with inside stuff -- well, normal UUCP stuff */
 
 	if (mp->flags & (MO_UNIXFROM|MO_REMOTEFROM)) {
-	  char *timestring;
-	  time_t now;
 	  const char *uu = startrp->addr->link->user;
 
-	  now = time((time_t *)0);
-	  timestring = ctime(&now);
-	  *(timestring+strlen(timestring)-1) = '\0';
 	  if (strcmp(startrp->addr->link->channel,"error")==0)
 	    uu = "<>";
 	  fprintf(tafp, "%s%s %s", FROM_, uu, timestring);
@@ -713,7 +793,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	    }
 	    break;
 	  case 9:		/* QUOTED-PRINTABLE */
-	    if (can_8bit) {
+	    if (decode_qp) {
 	      /* Force(d) to decode Q-P while transfer.. */
 	      convertmode = _CONVERT_8BIT;
 	      /*  UPGRADE TO 8BIT !  */
@@ -732,33 +812,126 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  }
 	}
 
- 	/* Add the "Return-Path:" is it is desired, but does not yet
-	   exist.. */
-	if (mp->flags & MO_RETURNPATH) {
-	  const char *uu = startrp->addr->link->user;
+	/* Snub the "Return-Path:" header, if it exists.. */
+	if (1) {
 	  char **hdrs;
 	  do {
 	    hdrs = has_header(startrp,"Return-Path:");
 	    if (hdrs) delete_header(startrp, hdrs);
 	  } while (hdrs);
+	}
+
+	/* Snub stuff that we add later below.. */
+
+	if (mp->flags & MO_XENVELOPES) {
+	  char **hdrs;
+
+	  do {
+	    hdrs = has_header(startrp,"X-Envelope-To:");
+	    if (hdrs) delete_header(startrp, hdrs);
+	  } while (hdrs);
+
+	  do {
+	    hdrs = has_header(startrp,"X-Envid:");
+	    if (hdrs) delete_header(rp,hdrs);
+	  } while (hdrs);
+
+	  do {
+	    hdrs = has_header(startrp,"Envelope-Id:");
+	    if (hdrs) delete_header(rp,hdrs);
+	  } while (hdrs);
+
+	}
+
+/* PERT-NB X-Orcpt if O option */
+	if (mp->flags & (MO_XORCPT|MO_XENVELOPES)) {
+	  char **hdrs;
+	  do {
+	    hdrs = has_header(startrp,"X-Orcpt:");
+	    if (hdrs) delete_header(startrp, hdrs);
+	  } while (hdrs);
+	  do { /* RFC 2298 section  2.3 */
+	    hdrs = has_header(startrp,"Original-Recipient:");
+	    if (hdrs) delete_header(startrp, hdrs);
+	  } while (hdrs);
+	}
+/* PERT-NB END */
+
+
+	/* Write headers: */
+
+	header_received_for_clause(startrp, 0, verboselog);
+
+	fwriteheaders(startrp, tafp, lineendseq, convertmode, maxwidth, NULL);
+	if (verboselog)
+	  fwriteheaders(startrp, verboselog, "\n", convertmode, maxwidth, NULL);
+
+	/*
+	 * NOTE: Following header writers make sense only for SINGLE
+	 *       RECIPIENT DELIVERIES!
+	 *
+	 */
+
+	if (mp->flags & MO_XENVELOPES) {
+	  if (dp->envid) {
+	    fprintf(tafp, "Envelope-Id: ");
+	    decodeXtext(tafp, dp->envid);
+	    fputs(lineendseq, tafp);
+
+	    if (verboselog) {
+	      fprintf(verboselog, "Envelope-Id: ");
+	      decodeXtext(verboselog, dp->envid);
+	      fputs("\n", verboselog);
+	    }
+	  }
+	}
+	if (mp->flags & (MO_XORCPT|MO_XENVELOPES)) {
+	  /* Put out a sequence of  X-Envelope-To: and
+	     Original-Recipient:  headers -- in order.. */
+	  for (rp = startrp; rp != endrp; rp = rp->next) {
+
+	    const char *uu = rp->addr->user;
+	    if (strcmp(rp->addr->link->channel,"error")==0)
+	      uu = "";
+	    fprintf(tafp, "X-Envelope-To: <%s> (uid %s)%s",
+		    uu, rp->addr->misc, lineendseq);
+	    if (rp->orcpt) {
+	      /* RFC 2298: section 2.3 */
+	      fprintf(tafp, "Original-Recipient: ");
+	      decodeXtext(tafp, rp->orcpt);
+	      fputs(lineendseq, tafp);
+	    }
+	    if (verboselog) {
+	      fprintf(verboselog, "X-Envelope-To: <%s> (uid %s)\n",
+		    uu, rp->addr->misc);
+	      if (rp->orcpt) {
+		/* RFC 2298: section 2.3 */
+		fprintf(verboselog, "Original-Recipient: ");
+		decodeXtext(verboselog, rp->orcpt);
+		fputs("\n", verboselog);
+	      }
+	    }
+	  }
+	}
+
+	if (mp->flags & MO_RETURNPATH) {
+	  const char *uu = startrp->addr->link->user;
 	  if (strcmp(startrp->addr->link->channel,"error")==0)
 	    uu = "";
-	  append_header(startrp,"Return-Path: <%.999s>", uu);
-	}
-	if (mp->flags & MO_CRLF) {
-	  fwriteheaders(startrp, tafp, "\r\n", convertmode, maxwidth, NULL);
-	  fprintf(tafp, "\r\n");
-	} else {
-	  fwriteheaders(startrp, tafp, "\n",   convertmode, maxwidth, NULL);
-	  fprintf(tafp, "\n");
+
+	  fprintf(tafp, "Return-Path: <%s>%s", uu, lineendseq);
 	}
 
-	if (verboselog) {
-	  fwriteheaders(startrp, verboselog, "\n", convertmode, maxwidth, NULL);
-	  fprintf(verboselog, "\n");
-	}
 
-	/* append message body itself */
+	/* Header-Body separator: */
+
+	fputs(lineendseq, tafp);
+
+	if (verboselog) fprintf(verboselog, "\n");
+
+
+	/* Append message body itself */
+
 	i = appendlet(dp, mp, tafp, verboselog, convertmode);
 	if (i != EX_OK && !gotsigpipe) {
 	  for (rp = startrp; rp != endrp; rp = rp->next) {
@@ -766,7 +939,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 			 /* Could indicate: 4.3.1 - mail system full ?? */
 			 "5.3.0 (Write to target failed for some reason)",
 			 "x-local; 500 (Write to target failed for some reason)");
-	    diagnostic(rp, i, 0, "write error");
+	    diagnostic(verboselog, rp, i, 0, "write error");
+	    SM_MIB_diag(i);
 	  }
 	  /* just to make sure nothing will get delivered */
 	  kill(pid, SIGTERM);
@@ -779,10 +953,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	}
 
 	if (mp->flags & MO_BSMTP) {
-	  if (mp->flags & MO_CRLF)
-	    fprintf(tafp,".\r\n");
-	  else
-	    fprintf(tafp,".\n");
+	  fputs(lineendseq, tafp);
 	}
 	
 	fclose(tafp);
@@ -830,15 +1001,14 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  /* sprintf(cp+strlen(cp), " of command: %s", mp->command); */
 	  strcat(cp, "]");
 	}
-	if (verboselog)
-	  fprintf(verboselog,"Diagnostic: %s\n",cp);
 	for (rp = startrp; rp != endrp; rp = rp->next) {
 	  if (i == EX_OK)
 	    notaryreport(rp->addr->user, "relayed",
 			 "2.5.0", "smtp;250 (Delivered)");
 	  else
 	    notaryreport(rp->addr->user, "failed", exs, exd);
-	  diagnostic(rp, i, 0, "%s", buf);
+	  diagnostic(verboselog, rp, i, 0, "%s", buf);
+	  SM_MIB_diag(i);
 	}
 	/* XX: still need to deal with MO_STRIPQUOTES */
 }
@@ -847,10 +1017,6 @@ deliver(dp, mp, startrp, endrp, verboselog)
  * appendlet - append letter to file pointed at by fd
  */
 
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-static char let_buffer[BUFSIZ*8];
-#endif
-
 int
 appendlet(dp, mp, fp, verboselog, convertmode)
 	struct ctldesc *dp;
@@ -859,124 +1025,155 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	FILE *verboselog;
 	int convertmode;
 {
-	/* `convertmode' controls the behaviour of the message conversion:
-	     _CONVERT_NONE (0): send as is
-	     _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
-	     _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
-	     _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
-	 */
+      /* `convertmode' controls the behaviour of the message conversion:
+	 _CONVERT_NONE (0): send as is
+	 _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
+	 _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
+	 _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
+      */
 
-	register int i;
-	int lastch;
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	register int bufferfull;
-	int mfd = dp->msgfd;
-#endif
+      register int i;
+      int lastch, rc;
+      register int bufferfull;
+      int mfd = dp->msgfd;
 
-	writebuf(mp, fp, (char *)NULL, 0);  /* magic initialization */
+      writebuf(mp, fp, (char *)NULL, 0);  /* magic initialization */
 
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	/* can we use cache of message body data */
-	if (convertmode == _CONVERT_NONE && readalready != 0) {
-	  lastch = let_buffer[readalready-1];
-	  if (writebuf(mp, fp, let_buffer, readalready) != readalready)
+      lastch = -1;
+      if (convertmode == _CONVERT_NONE) {
+
+	if (ta_use_mmap <= 0) {
+
+	  /* can we use cache of message body data */
+	  if (readalready > 0) {
+	    lastch = dp->let_buffer[readalready-1];
+	    if (writebuf(mp, fp, dp->let_buffer, readalready) != readalready)
+	      return EX_IOERR;
+	    if (lastch != '\n')
+	      if (writebuf(mp, fp, "\n", 1) != 1)
+		return EX_IOERR;
+	    return EX_OK;
+	  }
+
+	  bufferfull = 0;
+	  readalready = 0;
+	  lastch = -256;
+	  lseek(mfd, dp->msgbodyoffset, SEEK_SET);
+	  for (;;) {
+	    i = read(mfd, (void*)(dp->let_buffer), dp->let_buffer_size);
+	    if (i == 0) break; /* EOF */
+	    if (i < 0 && (errno == EINTR || errno == EAGAIN))
+	      continue;
+	    if (i < 0)
+	      return EX_IOERR;
+	    lastch = dp->let_buffer[i-1];
+	    if (writebuf(mp, fp, dp->let_buffer, i) != i)
+	      return EX_IOERR;
+	    readalready = i;
+	    bufferfull++;
+	  }
+	  if (lastch != '\n')
+	    if (writebuf(mp, fp, "\n", 1) != 1)
+	      return EX_IOERR;
+	  if (bufferfull > 1)	/* not all in memory, need to reread */
+	    readalready = 0;
+
+	} else { /* mmap()ed message buffer */
+
+	  const char *p = dp->let_buffer + dp->msgbodyoffset;
+	  i = dp->let_end - p;
+	  if (i < 0)
+	    return EX_IOERR;
+	  lastch = dp->let_end[-1];
+	  if (writebuf(mp, fp, p, i) != i)
 	    return EX_IOERR;
 	  if (lastch != '\n')
 	    if (writebuf(mp, fp, "\n", 1) != 1)
 	      return EX_IOERR;
-	  return EX_OK;
 	}
-#endif
 
+	rc = EX_OK;
+
+      } else {
+
+	/* convertmode something else, than _CONVERT_NONE */
+	/* Various esoteric conversion modes..
+	   We are better to feed writemimeline() with LINES
+	   instead of blocks of data.. */
+
+	writemimeline(mp, fp, (char *)NULL, 0, 0); /* reset */
+
+	/*
+	  if(verboselog) fprintf(verboselog,
+	  "sm: Convert mode: %d, fd=%d, fdoffset=%d, bodyoffset=%d\n",
+	  convertmode, mfd, (int)lseek(mfd, (off_t)0, SEEK_CUR),
+	  dp->msgbodyoffset);
+	*/
+	
+	/* we are assuming to be positioned properly
+	   at the start of the message body */
 	lastch = -1;
-	if (convertmode == _CONVERT_NONE) {
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	  bufferfull = 0;
-	  readalready = 0;
-	  lseek(mfd, dp->msgbodyoffset, SEEK_SET);
-	  while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
-#else
-	    const char *let_buffer = dp->let_buffer + dp->msgbodyoffset;
-	    i = dp->let_end - (dp->let_buffer + dp->msgbodyoffset);
-#endif
-	    if (i < 0)
-	      return EX_IOERR;
-	    lastch = let_buffer[i-1];
-	    if (writebuf(mp, fp, let_buffer, i) != i)
-	      return EX_IOERR;
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	    readalready = i;
-	    bufferfull++;
-	  }
-	  if (bufferfull > 1)	/* not all in memory, need to reread */
-	    readalready = 0;
-#endif
+	i = 0;
 
-	} else {
-	  /* convertmode something else, than _CONVERT_NONE */
-	  /* Various esoteric conversion modes..
-	     We are better to feed writemimeline() with LINES
-	     instead of blocks of data.. */
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	  Sfio_t *mfp = sfnew(NULL, NULL, 64*1024, mfd, SF_READ|SF_WHOLE);
+	if (ta_use_mmap <= 0) {
+	  /* Locally read, line by line */
+
+	  char sfio_buf[16*1024];
+	  Sfio_t *mfp = NULL;
+
+	  mfp = sfnew(NULL, sfio_buf, sizeof(sfio_buf), mfd, SF_READ);
 	  sfseek(mfp, dp->msgbodyoffset, SEEK_SET);
-
-#define MFPCLOSE zsfsetfd(mfp, -1); sfclose(mfp);
-
 	  readalready = 0;
-#else
-#define MFPCLOSE
-	  const char *s = dp->let_buffer + dp->msgbodyoffset;
-#endif
-	  writemimeline(mp, fp, (char *)NULL, 0, 0);
-
-	  /*
-	     if(verboselog) fprintf(verboselog,
-	     "sm: Convert mode: %d, fd=%d, fdoffset=%d, bodyoffset=%d\n",
-	     convertmode, mfd, (int)lseek(mfd, (off_t)0, SEEK_CUR),
-	     dp->msgbodyoffset);
-	   */
-
-	  /* we are assuming to be positioned properly
-	     at the start of the message body */
-	  lastch = -1;
-	  i = 0;
 
 	  for (;;) {
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	    if ((i = cfgets(let_buffer, sizeof(let_buffer), mfp)) == EOF)
+	    i = csfgets((void*)(dp->let_buffer), dp->let_buffer_size, mfp);
+	    if (i == EOF)
 	      break;
-#else
-	    const char *let_buffer = s, *s2 = s;
+	    lastch = dp->let_buffer[i-1];
+	    /* It MAY be malformed -- if it has a BUFSIZ length
+	       line in it, IT CAN'T BE MIME  :-/		*/
+	    /* Ok, write the line */
+	    if (writemimeline(mp, fp, dp->let_buffer, i, convertmode) != i)
+	      return EX_IOERR;
+	  }
+	  if (mfp) {
+	    if (i == EOF && !sfeof(mfp) && !sferror(mfp)) {
+	      rc = EX_IOERR;
+	    }
+	    zsfsetfd(mfp, -1);
+	    sfclose(mfp);
+	  }
+	  rc = EX_OK;
+
+	} else {
+	  /* MMAP()ED buffer */
+	  const char *s = dp->let_buffer + dp->msgbodyoffset;
+
+	  for (;;) {
+	    const char *s2 = s;
 	    i = 0;
 	    if (s >= dp->let_end) break;	/* "EOF" */
 	    while (s2 < dp->let_end && *s2 != '\n')
 	      ++s2, ++i;
 	    if ((lastch = *s2) == '\n')
 	      ++s2, ++i;
-	    s = s2;
-#endif
-	    /* It MAY be malformed -- if it has a BUFSIZ length
-	       line in it, IT CAN'T BE MIME  :-/		*/
 	    /* Ok, write the line */
-	    if (writemimeline(mp, fp, let_buffer, i, convertmode) != i) {
+	    if (writemimeline(mp, fp, s, i, convertmode) != i)
 	      return EX_IOERR;
-	    }
+	    s = s2;
 	  }
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	  if (i == EOF && !sfeof(mfp) && !sferror(mfp)) {
-	    MFPCLOSE
-	    return EX_IOERR;
-	  }
-	  MFPCLOSE
-#endif
 	}
-
-	/* we must make sure the last thing we transmit is a CRLF sequence */
-	if (lastch != '\n')
-	  writebuf(mp, fp, "\n", 1);
-
-	return EX_OK;
+	rc = EX_OK;
+      }
+     
+      /* we must make sure the last thing we transmit is a CRLF sequence */
+      if (lastch != '\n')
+	writebuf(mp, fp, "\n", 1);
+      
+      /* Failure in that last CRLF writing does not
+	 affect our return code -- should it ?? */
+      
+      return rc;
 }
 
 /*
@@ -1205,7 +1402,7 @@ readsmcf(file, mailer)
 	static struct maildesc m;
 
 	if (file == NULL) {
-	  char *mailshare = getzenv("MAILSHARE");
+	  const char *mailshare = getzenv("MAILSHARE");
 
 	  if (mailshare == NULL)
 	    mailshare = MAILSHARE;
@@ -1217,7 +1414,8 @@ readsmcf(file, mailer)
 	  perror(file);
 	  exit(EX_OSFILE);
 	}
-	entry = cp = NULL;
+	entry = NULL;
+	cp    = NULL;
 	while (fgets(buf, sizeof buf, fp) != NULL) {
 	  if (buf[0] == '#' || buf[0] == '\n')
 	    continue;
@@ -1228,7 +1426,7 @@ readsmcf(file, mailer)
 	  }
 	  entry = (char*)cp;
 	  strcpy(entry, buf);
-	  SKIPWHILE(!isspace, cp);
+	  SKIPTEXT(cp);
 	  if (isascii(*cp) && isspace(*cp)) {
 	    if (*cp == '\n') {
 	      fprintf(stderr, "%s: %s: bad entry: %s",
@@ -1250,67 +1448,74 @@ readsmcf(file, mailer)
 	m.name = entry;
 	m.flags = MO_UNIXFROM;
 	++cp;
-	SKIPWHILE(isspace, cp);
+	SKIPSPACE(cp);
 	/* process mailer option flags */
-	for (;isascii(*cp) && !isspace(*cp); ++cp) {
+	for (;*cp && *cp != ' ' && *cp != '\t' && *cp != '\n'; ++cp) {
+	  int no = 0;
 	  switch (*cp) {
-	  case 'D':		/* this mailer wants a Date: line */
-	    	m.flags |= MO_WANTSDATE;	break;
-	  case 'F':		/* this mailer wants a From: line */
-	    	m.flags |= MO_WANTSFROM;	break;
 	  case '7':	m.flags |= MO_STRIPHIBIT;	break;
-	  case 'E':	m.flags |= MO_ESCAPEFROM;	break;
-	  case 'P':	m.flags |= MO_RETURNPATH;	break;
-	  case 'R':	m.flags |= MO_CRLF;		break;
-	  case 'S':	m.flags |= MO_NORESETUID;	break;
-	  case 'U':	m.flags |= MO_REMOTEFROM;	break;
-	  case 'X':	m.flags |= MO_HIDDENDOT;	break;
-	  case 'f':	m.flags |= MO_FFROMFLAG;	break;
-	  case 'm':	m.flags |= MO_MANYUSERS;	break;
-	  case 'n':	m.flags &= ~MO_UNIXFROM;	break;
-	  case 'r':	m.flags |= MO_RFROMFLAG;	break;
-	  case 's':	m.flags |= MO_STRIPQUOTES;	break;
-	  case 'H':     m.flags |= MO_BSMTPHELO;	break;
+	  case '8':	can_8bit = 1;			break;
+	  case '9':	decode_qp = 1;			break;
+	  case 'A': no=*cp; break;	/* arpanet-compatibility */
 	  case 'b':	m.flags |= (MO_BSMTP|MO_HIDDENDOT); break;
 	  case 'B':	if (m.flags & MO_BESMTP) /* -BB */
 			    m.flags |= MO_BEDSMTP;
 			else
 			    m.flags |= MO_BESMTP|MO_BSMTP|MO_HIDDENDOT;
 			break;
-	  case 'A':		/* arpanet-compatibility */
-	  case 'C':		/* canonicalize remote hostnames */
-	  case 'I':		/* talking to a clone of I */
-	  case 'L':		/* limit line length */
-	  case 'M':		/* this mailer wants a Message-Id: line */
-	  case 'e':		/* expensive mailer */
-	  case 'h':		/* preserve upper case in host names */
-	  case 'l':		/* this is a local mailer */
-	  case 'p':		/* use SMTP return path */
-	  case 'u':		/* preserve upper case in user names */
-	  case 'x':		/* this mailer wants a Full-Name: line */
-	    fprintf(stderr,
-		    "%s: the '%c' sendmail mailer option does not make sense in this environment\n",
-		    progname, *cp);
-	    break;
-	  case '-':		/* ignore */
-	    break;
+	  case 'C': no=*cp; break; /* canonicalize remote hostnames */
+	  case 'D':		/* this mailer wants a Date: line */
+	    	m.flags |= MO_WANTSDATE;	break;
+	  case 'e':	m.flags |= MO_XENVELOPES;	break;
+/* PERT-NB O option for X-Orcpt */
+	  case 'O':	m.flags |= MO_XORCPT;		break;
+	  case 'E':	m.flags |= MO_ESCAPEFROM;	break;
+	  case 'f':	m.flags |= MO_FFROMFLAG;	break;
+	  case 'F':		/* this mailer wants a From: line */
+	    	m.flags |= MO_WANTSFROM;	break;
+	  case 'h': no=*cp; break; /* preserve upper case in host names */
+	  case 'H':     m.flags |= MO_BSMTPHELO;	break;
+	  case 'I': no=*cp; break; /* talking to a clone of I */
+	  case 'l': no=*cp; break; /* this is a local mailer */
+	  case 'L': no=*cp; break; /* limit line length */
+	  case 'm':	m.flags |= MO_MANYUSERS;	break;
+	  case 'M': no=*cp; break; /* this mailer wants a Message-Id: line */
+	  case 'n':	m.flags &= ~MO_UNIXFROM;	break;
+	  case 'p': no=*cp; break; /* use SMTP return path */
+	  case 'P':	m.flags |= MO_RETURNPATH;	break;
+	  case 'r':	m.flags |= MO_RFROMFLAG;	break;
+	  case 'R':	m.flags |= MO_CRLF;		break;
+	  case 's':	m.flags |= MO_STRIPQUOTES;	break;
+	  case 'S':	m.flags |= MO_NORESETUID;	break;
+	  case 'u': no=*cp; break; /* preserve upper case in user names */
+	  case 'U':	m.flags |= MO_REMOTEFROM;	break;
+	  case 'x': no=*cp; break; /* this mailer wants a Full-Name: line */
+	  case 'X':	m.flags |= MO_HIDDENDOT;	break;
+
+	  case '-':	break;	/* ignore */
 	  default:
 	    fprintf(stderr,
 		    "%s: unknown sendmail mailer option '%c'\n",
 		    progname, *cp);
 	    break;
 	  }
+	  if (no) {
+	    fprintf(stderr,
+		    "%s: the '%c' sendmail mailer option does not make sense in this environment\n",
+		    progname,no);
+	  }
 	}
-	SKIPWHILE(isspace, cp);
+	SKIPSPACE(cp);
 	m.command = (char*) cp;
-	SKIPWHILE(!isspace, cp);
+	SKIPTEXT(cp);
 	if ((char*)cp == m.command) {
 		fprintf(stderr,"%s: bad entry for %s\n",progname, m.name);
 		return NULL;
 	}
 	*cp++ = '\0';
 	if (*m.command != '/') {
-	  char *nmc, *mailbin = getzenv("MAILBIN");
+	  char *nmc;
+	  const char *mailbin = getzenv("MAILBIN");
 
 	  if (mailbin == NULL)
 	    mailbin = MAILBIN;
@@ -1319,16 +1524,16 @@ readsmcf(file, mailer)
 	  sprintf(nmc, "%s/%s", mailbin, m.command);
 	  m.command = nmc;
 	}
-	SKIPWHILE(isspace, cp);
+	SKIPSPACE(cp);
 	i = 0;
 	while (isascii(*cp) && !isspace(*cp) && i < MD_ARGVMAX) {
 	  if (*cp == '\0')
 	    break;
 	  m.argv[i++] = (char*) cp;
-	  SKIPWHILE(!isspace, cp);
-	  if (isascii(*cp)) {
+	  SKIPTEXT(cp);
+	  if (*cp) {
 	    *cp++ = '\0';
-	    SKIPWHILE(isspace, cp);
+	    SKIPSPACE(cp);
 	  }
 	}
 	if (i == 0) {
@@ -1345,8 +1550,10 @@ int
 check_7bit_cleanness(dp)
 struct ctldesc *dp;
 {
-#if (defined(HAVE_MMAP) && defined(TA_USE_MMAP))
+      if (ta_use_mmap > 0) {
+
 	/* With MMAP()ed spool file it is sweet and simple.. */
+
 	register const char *s = dp->let_buffer + dp->msgbodyoffset;
 	while (s < dp->let_end)
 	  if (128 & *s) {
@@ -1359,8 +1566,11 @@ struct ctldesc *dp;
 	    return 0; /* Not clean ! */
 	  }
 	  else ++s;
-	return 1;
-#else
+
+      } else {
+
+	/* Without MMAP() we work... */
+
 	register int i;
 	register int bufferfull;
 	int lastwasnl;
@@ -1369,29 +1579,31 @@ struct ctldesc *dp;
 /* can we use cache of message body data */
 	if (readalready != 0) {
 	  for (i=0; i<readalready; ++i)
-	    if (128 & (let_buffer[i]))
+	    if (128 & (dp->let_buffer[i]))
 	      return 0;		/* Not clean ! */
 	}
 
 	/* we are assumed to be positioned properly at start of message body */
 	bufferfull = 0;
 
-	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
+	for (;;) {
+	  i = read(mfd, (void*)(dp->let_buffer), dp->let_buffer_size);
+	  if (i == 0) break; /* EOF */
 	  if (i < 0) {
 	    /* ERROR ?!?!? */
-	    if (errno == EINTR)
+	    if (errno == EINTR || errno == EAGAIN)
 	      continue;
 	    readalready = 0;
 	    lseek(mfd, dp->msgbodyoffset, SEEK_SET);
 	    return 0;
 	  }
-	  lastwasnl = (let_buffer[i-1] == '\n');
+	  lastwasnl = (dp->let_buffer[i-1] == '\n');
 	  readalready = i;
 	  bufferfull++;
 	  for (i=0; i < readalready; ++i)
-	    if (128 & (let_buffer[i])) {
+	    if (128 & (dp->let_buffer[i])) {
 	      lseek(mfd, dp->msgbodyoffset, SEEK_SET);
-	      /* We propably have not read everything of the file! */
+	      /* We probably have not read everything of the file! */
 	      readalready = 0;
 	      return 0;		/* Not clean ! */
 	    }
@@ -1402,6 +1614,6 @@ struct ctldesc *dp;
 	if (bufferfull > 1)	/* not all in memory, need to reread */
 	  readalready = 0;
 
-	return 1;
-#endif
+      }
+      return 1;
 }
