@@ -1,40 +1,42 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1991-2000 by Matti Aarnio -- modifications, including MIME
+ *	Copyright 1991-2003 by Matti Aarnio -- modifications,
+ *	including MIME things.
  */
 
 #include "smtp.h"
 
-typedef union {
-	HEADER qb1;
-	char qb2[8000];
-} querybuf;
-
 int
-getmxrr(SS, host, mx, maxmx, depth)
+getmxrr(SS, host, mx, maxmx, depth, realname, realnamesize, realnamettlp)
 	SmtpState *SS;
 	const char *host;
 	struct mxdata mx[];
 	int maxmx, depth;
+	char *realname;
+	const int realnamesize;
+	time_t *realnamettlp;
 {
 	HEADER *hp;
 	msgdata *eom, *cp;
 	struct mxdata mxtemp;
 	int qlen, n, i, j, nmx, qdcount, ancount, nscount, arcount, maxpref;
 	int class;
+	long ttl;
 	u_short type;
 	int saw_cname = 0;
 	int had_eai_again = 0;
 	querybuf qbuf, answer;
-	msgdata buf[8192], realname[8192];
+	msgdata buf[8192];
 	char mxtype[MAXFORWARDERS];
 
 	h_errno = 0;
 
 #ifndef TEST
-	notary_setwtt  (NULL);
-	notary_setwttip(NULL);
+	if (maxmx > 2) {
+	  notary_setwtt  (NULL);
+	  notary_setwttip(NULL);
+	}
 #endif
 
 	if (depth == 0) {
@@ -83,7 +85,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  notary_setxdelay((int)(endtime-starttime));
 	  notaryreport(NULL,FAILED,"5.4.3 (DNS-failure)",SS->remotemsg);
 #endif
-	  return EX_TEMPFAIL;
+	  return EX_DEFERALL;
 	}
 
 	time(&now);
@@ -108,9 +110,10 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  fprintf(SS->verboselog, "\n");
 	}
 
-	rmsgappend(SS, 1,
-		   "\r-> DNSreply: len=%d rcode=%d qd=%d an=%d ns=%d ar=%d",
-		   n, hp->rcode, qdcount, ancount, nscount, arcount);
+	if (maxmx > 2)
+	  rmsgappend(SS, 1,
+		     "\r-> DNSreply: len=%d rcode=%d qd=%d an=%d ns=%d ar=%d",
+		     n, hp->rcode, qdcount, ancount, nscount, arcount);
 
 	if (hp->rcode != NOERROR || ancount == 0) {
 	  switch (hp->rcode) {
@@ -138,7 +141,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    notary_setxdelay((int)(endtime-starttime));
 	    notaryreport(NULL,FAILED,"5.4.4 (DNS lookup report)",SS->remotemsg);
 #endif
-	    return EX_TEMPFAIL;
+	    return EX_DEFERALL;
 	  case NOERROR:
 	    mx[0].host = NULL;
 	    return EX_OK;
@@ -166,7 +169,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 #endif
 
 	  if (had_eai_again)
-	    return EX_TEMPFAIL;
+	    return EX_DEFERALL;
 	  return EX_UNAVAILABLE;
 	}
 	nmx = 0;
@@ -186,16 +189,14 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    break;
 	  cp += n;
 	  if (cp+10 > eom) { cp = eom; break; }
-	  type = _getshort(cp);
-	  cp += 2;
-	  class = _getshort(cp);
-	  cp += 2;
-	  mx[nmx].expiry = now + _getlong(cp); /* TTL */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones, I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+
+	  NS_GET16(type,  cp);
+	  NS_GET16(class, cp);
+	  NS_GET32(ttl,   cp);	/* TTL */
+	  NS_GET16(n, cp);	/* dlen */
+
+	  mx[nmx].expiry = now + ttl;
+
 	  if (cp + n > eom) { cp = eom; break; }
 
 	  if (class != C_IN) {
@@ -206,12 +207,13 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  }
 	  if (type == T_CNAME) {
 	    cp += dn_expand((msgdata *)&answer, eom, cp,
-			    (void*)realname, sizeof realname);
+			    (void*)realname, realnamesize);
 	    if (cp > eom) break;
 	    saw_cname = 1;
 	    --ancount;
 	    if (SS->verboselog)
 	      fprintf(SS->verboselog, " -> CNAME: '%s'\n", realname);
+	    if (ttl < *realnamettlp) *realnamettlp = ttl;
 	    continue;
 	  } else if (type != T_MX)  {
 	    cp += n;
@@ -220,24 +222,26 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    continue;
 	  }
 	  if (cp + n /* dlen */ > eom) { cp = eom; break; }
-	  mx[nmx].pref = _getshort(cp);
-	  cp += 2; /* MX preference value */
+
+	  NS_GET16(mx[nmx].pref, cp);  /* MX preference value */
+
 	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	  if (n < 0) break;
 	  cp += n;
 	  if (cp > eom) break;
 
 	  mx[nmx].ai   = NULL;
-	  mx[nmx].host = (msgdata *)strdup((void*)buf);
+	  mx[nmx].host = (char *)strdup((void*)buf);
 	  if (mx[nmx].host == NULL) {
 	    fprintf(stderr, "Out of virtual memory!\n");
 	    exit(EX_OSERR);
 	  }
 	  if (SS->verboselog)
 	    fprintf(SS->verboselog, " -> (%lds) MX[%d] pref=%d host=%s\n",
-		    mx[nmx].expiry - now, nmx, mx[nmx].pref, buf);
-	  rmsgappend(SS, 1, "\r-> %lds MX[%d] p=%d '%s'",
-		     mx[nmx].expiry - now, nmx, mx[nmx].pref, mx[nmx].host);
+		    (long)(mx[nmx].expiry - now), nmx, mx[nmx].pref, buf);
+	  if (maxmx > 2)
+	    rmsgappend(SS, 1, "\r-> %lds MX[%d] p=%d '%s'",
+		       mx[nmx].expiry - now, nmx, mx[nmx].pref, mx[nmx].host);
 	  mxtype[nmx] = 0;
 	  ++nmx;
 	  --ancount;
@@ -268,13 +272,12 @@ getmxrr(SS, host, mx, maxmx, depth)
 	      break;
 	    cp += n;
 	    if (cp+10 > eom) { cp = eom; break; }
-	    cp += 2;
-	    cp += 2;
-	    cp += 4; /* "long" -- but keep in mind that some machines
-			have "funny" ideas about "long" -- those 64-bit
-			ones, I mean ... */
-	    n = _getshort(cp); /* dlen */
-	    cp += 2;
+
+	    cp += NS_INT16SZ; /* type */
+	    cp += NS_INT16SZ; /* class */
+	    cp += NS_INT32SZ; /* ttl */
+	    NS_GET16(n, cp); /* dlen */
+
 	    cp += n;
 	    --ancount;
 	  } /* Skipped thru all remaining answers */
@@ -290,24 +293,27 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    /* Yes, it is TRUNCATED reply!   Must retry with e.g.
 	       by using TCP! */
 	    /* FIXME: FIXME! FIXME! Truncated reply handling! */
-	    rmsgappend(SS, 1, "\r   TRUNCATED REPLY!");
+	    if (maxmx > 2)
+	      rmsgappend(SS, 1, "\r   TRUNCATED REPLY!");
 	  }
 
 	  if (SS->verboselog)
 	    fprintf(SS->verboselog,"  left-over ANCOUNT=%d != 0! TC=%d\n",
 		    ancount, hp->tc);
 
-	  rmsgappend(SS, 1, "\r   AnswerCount  %d > 0!!", ancount);
+	  if (maxmx > 2)
+	    rmsgappend(SS, 1, "\r   AnswerCount  %d > 0!!", ancount);
 
-
-	  return EX_TEMPFAIL; /* FIXME?? FIXME?? */
+	  return EX_DEFERALL; /* FIXME?? FIXME?? */
 	}
 
 	if (nmx == 0 && realname[0] != 0) {
 	  /* do it recursively for the real name */
-	  n = getmxrr(SS, (char *)realname, mx, maxmx, depth+1);
+	  n = getmxrr(SS, (char *)realname, mx, maxmx, depth+1,
+		      realname, realnamesize, realnamettlp);
+
 	  if (had_eai_again)
-	    return EX_TEMPFAIL;
+	    return EX_DEFERALL;
 	  return n;
 	} else if (nmx == 0) {
 	  /* "give it the benefit of doubt" */
@@ -315,7 +321,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  mx[0].ai   = NULL;
 	  SS->mxcount = 0;
 	  if (had_eai_again)
-	    return EX_TEMPFAIL;
+	    return EX_DEFERALL;
 	  return EX_OK;
 	}
 
@@ -329,138 +335,24 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    break;
 	  cp += n;
 	  if (cp+10 > eom) { cp = eom; break; }
-	  /* type = _getshort(cp); */
-	  cp += 2;
-	  /* class = _getshort(cp); */
-	  cp += 2;
-	  /* mx[nmx].expiry = now + _getlong(cp); */ /* TTL */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones, I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+
+	  cp += NS_INT16SZ; /* type  */
+	  cp += NS_INT16SZ; /* class */
+	  cp += NS_INT32SZ; /* ttl   */
+	  NS_GET16(n, cp);  /* dlen  */
+
 	  cp += n; /* We simply skip this data.. */
 	  if (cp <= eom)
 	    --nscount;
 	}
 
-	/* If nscount isn't zero here, then (cp >= eom) is true ... */
 
-	/* Ok, can continue to pick the ADDITIONAL SECTION data */
-	/* BUT ONLY IF THE REPLY HAD 'AA' BIT SET! If it didn't,
-	   we must always ask A/AAAA separately.. */
-	while (hp->aa && arcount > 0 && cp < eom) {
-	  int ttl;
-	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
-	  if (n < 0) { cp = eom; break; }
-	  cp += n;
-	  if (cp+10 > eom) { cp = eom; break; }
-	  type  = _getshort(cp);
-	  cp += 2;
-	  class = _getshort(cp);
-	  cp += 2;
-	  ttl   = _getlong(cp);  /* TTL */
-	  cp += 4; /* "long" -- but keep in mind that some machines
-		      have "funny" ideas about "long" -- those 64-bit
-		      ones, I mean ... */
-	  n = _getshort(cp); /* dlen */
-	  cp += 2;
+	/* =========================================================== */
+#if 0 /* Bloody Linux vs. FreeBSD implementation differences... (addrinfo internal handling) */
+#include "getmxrr-removed.txt"
 
-	  if (cp + n > eom) { cp = eom; break; }
-
-	  if (class != C_IN) {
-	    cp += n; --nscount;
-	    continue;
-	  }
-
-	  /* Ok, we have Type IN data in the ADDITIONAL SECTION */
-
-
-	  /* A and AAAA are known here! */
-
-	  if (type == T_A
-#if defined(AF_INET6) && defined(INET6)
-	      || (type == T_AAAA && use_ipv6)
-#endif
-	      ) {
-
-	    struct addrinfo *ai;
-	    Usockaddr *usa;
-	    char *canon;
-	    int   nlen = strlen((const char*)buf);
-
-	    /* Pick the address data */
-	    for (i = 0; i < nmx; ++i) {
-	      /* Is this known (wanted) name ?? */
-	      if (CISTREQ(buf, mx[i].host)) {
-		/* YES! */
-
-		/* We do have a wanted name! */
-
-		/* build addrinfo block, pick addresses */
-
-		/* WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
-		   This assumes intimate knowledge of the system
-		   implementation of the  ``struct addrinfo'' !
-
-		   Linux and FreeBSD have DIFFERENT implementatoin!
-		   Other operating systems are likely different as well! */
-
-		if (SS->verboselog)
-		  fprintf(SS->verboselog,"  host='%s' AF=%s TTL=%d\n",
-			  buf, (type == T_A) ? "INET" : "INET6", ttl);
-
-		ai = (void*)malloc(sizeof(*ai) + sizeof(*usa) + nlen + 1);
-		if (ai == NULL) exit(EX_OSERR);
-		memset(ai, 0, sizeof(*ai) + sizeof(*usa) + nlen + 1);
-
-		usa   = (void*)((char *) ai + sizeof(*ai));
-
-		canon = ((char *)usa) + sizeof(*usa);
-		memcpy(canon, buf, nlen);
-
-		ai->ai_flags    = 0;
-		ai->ai_socktype = SOCK_STREAM;
-		ai->ai_protocol = IPPROTO_TCP;
-		ai->ai_addr     = (struct sockaddr *)usa;
-		ai->ai_addrlen  = sizeof(*usa);
-		ai->ai_canonname = /* canon */ NULL;
-
-		ai->ai_next     = mx[i].ai;
-		mx[i].ai        = ai;
-
-		mxtype[i] |= (type == T_A) ? 1 : 2;
-
-		switch (type) {
-#if defined(AF_INET6) && defined(INET6)
-		case T_AAAA:
-		  ai->ai_family = PF_INET6;
-		  usa->v6.sin6_family = PF_INET6;
-		  memcpy(&usa->v6.sin6_addr, cp, 16);
-		  break;
-#endif
-		case T_A:
-		  ai->ai_family = PF_INET;
-		  usa->v4.sin_family = PF_INET;
-		  memcpy(&usa->v4.sin_addr, cp, 4);
-		  break;
-		default:
-		  break;
-		}
-	      } /* Matched name! */
-	    } /* Name matching loop */
-	  } /* type = T_A or T_AAAA */
-
-	  cp += n;
-	  --arcount;
-	} /* Additional data collected! */
-
-	if (SS->verboselog)
-	  for (i = 0; i < nmx; ++i) {
-	    if (mx[i].ai == NULL)
-	      fprintf(SS->verboselog, " MX lookup lacked ADDITIONAL SECTION Address for entry: MX %d %s\n",
-		      mx[i].pref, mx[i].host);
-	  }
+#endif /* Linux vs. FreeBSD implementation difference... */
+	/* =========================================================== */
 
 	/* Collect addresses for all those who don't have them from
 	   the ADDITIONAL SECTION data */
@@ -481,7 +373,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 
 	  if (mxtype[i] == 3)
 	    continue; /* Have both A and AAAA */
-#endif
+#endif /* INET6 */
 	  
 	  memset(&req, 0, sizeof(req));
 	  req.ai_socktype = SOCK_STREAM;
@@ -499,19 +391,20 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    n = getaddrinfo((const char*)mx[i].host, "0", &req, &ai);
 #else
 	    n = _getaddrinfo_((const char*)mx[i].host, "0", &req, &ai, SS->verboselog);
-#endif
+#endif /* HAVE_GETADDRINFO */
 	    if (SS->verboselog)
 	      fprintf(SS->verboselog,"  getaddrinfo('%s','0') (PF_INET) -> r=%d (%s), ai=%p\n",
 		      mx[i].host, n, gai_strerror(n), ai);
 	    if (n != 0)
-	      rmsgappend(SS, 1, "\r-> getaddrinfo(INET, '%s','0') -> r=%d (%s); ai=%p",
-			 mx[i].host, n, gai_strerror(n), ai);
+	      if (maxmx > 2)
+		rmsgappend(SS, 1, "\r-> getaddrinfo(INET, '%s','0') -> r=%d (%s); ai=%p",
+			   mx[i].host, n, gai_strerror(n), ai);
 #if 0
 	    if (n) {
 	      zsyslog((LOG_INFO,"getmxrr('%s') mx[%d]='%s' getaddrinfo(INET) rc=%d",
 		       host, i, mx[i].host, n));
 	    }
-#endif
+#endif /* .. 0 */
 	    switch (n) {
 	    case 0:
 	      break;
@@ -551,20 +444,21 @@ getmxrr(SS, host, mx, maxmx, depth)
 #else
 	    n2 = _getaddrinfo_((const char *)mx[i].host, "0", &req, &ai2,
 			       SS->verboselog);
-#endif
+#endif /* HAVE_GETADDRINFO */
 	    if (SS->verboselog)
 	      fprintf(SS->verboselog,"  getaddrinfo('%s','0') (PF_INET6) -> r=%d (%s), ai=%p\n",
 		      mx[i].host, n2, gai_strerror(n2), ai2);
 
 	    if (n2 != 0)
-	      rmsgappend(SS, 1, "\r-> getaddrinfo(INET6,'%s') -> r=%d (%s); ai=%p",
-			 mx[i].host, n2, gai_strerror(n2), ai2);
+	      if (maxmx > 2)
+		rmsgappend(SS, 1, "\r-> getaddrinfo(INET6,'%s') -> r=%d (%s); ai=%p",
+			   mx[i].host, n2, gai_strerror(n2), ai2);
 #if 0
 	    if (n) {
 	      zsyslog((LOG_INFO,"getmxrr('%s') mx[%d]='%s' getaddrinfo(INET6) rc=%d",
 		       host, i, mx[i].host, n));
 	    }
-#endif
+#endif /* .. 0 */
 	    switch (n2) {
 	    case 0:
 	      break;
@@ -594,7 +488,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	      *aip = ai2;
 	    }
 	  }
-#endif
+#endif /* INET6 */
 
 	  /* Catenate new stuff into the tail of the old ... */
 	  aip = &(mx[i].ai);
@@ -608,7 +502,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 #ifndef TEST
 	      notary_setxdelay((int)(endtime-starttime));
 	      notaryreport(NULL,FAILED,"4.4.4 (DNS lookup report)",SS->remotemsg);
-#endif
+#endif /* .. TEST */
 
 	      had_eai_again = 1;
 	    }
@@ -626,7 +520,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	    mx[nmx].ai = ai;
 	    ai         = ai->ai_next;
 	    mx[nmx].ai->ai_next = NULL;
-	    mx[nmx].host = (msgdata *)strdup((const char *)mx[i].host);
+	    mx[nmx].host = (char *)strdup((const char *)mx[i].host);
 	    if (mx[nmx].host == NULL) {
 	      fprintf(stderr, "Out of virtual memory!\n");
 	      exit(EX_OSERR);
@@ -643,7 +537,8 @@ getmxrr(SS, host, mx, maxmx, depth)
 
 	for (i = 0; i < nmx; ++i) {
 	  if (mx[i].ai == NULL && mx[i].host != NULL) {
-	    rmsgappend(SS, 1, "\r-> No addresses for '%s'[%d]", mx[i].host, i);
+	    if (maxmx > 2)
+	      rmsgappend(SS, 1, "\r-> No addresses for '%s'[%d]", mx[i].host, i);
 	    free(mx[i].host);
 	    mx[i].host = NULL;
 	    continue;
@@ -653,9 +548,10 @@ getmxrr(SS, host, mx, maxmx, depth)
 	       CISTREQ(mx[i].ai->ai_canonname, myhostname)) ||
 	      matchmyaddresses(mx[i].ai) == 1) {
 
-	    rmsgappend(SS, 1, "\r-> Selfmatch '%s'(%s)[%d]",
-		       mx[i].host, ((mx[i].ai->ai_canonname) ?
-				    mx[i].ai->ai_canonname : "-"), i);
+	    if (maxmx > 2)
+	      rmsgappend(SS, 1, "\r-> Selfmatch '%s'(%s)[%d]",
+			 mx[i].host, ((mx[i].ai->ai_canonname) ?
+				      mx[i].ai->ai_canonname : "-"), i);
 
 	    if (SS->verboselog)
 	      fprintf(SS->verboselog,"  matchmyaddresses(): matched!  canon='%s', myname='%s'\n", mx[i].ai->ai_canonname ? mx[i].ai->ai_canonname : "<NIL>", myhostname);
@@ -668,8 +564,9 @@ getmxrr(SS, host, mx, maxmx, depth)
 
 	if (SS->verboselog)
 	  fprintf(SS->verboselog,"  getmxrr('%s') -> nmx=%d, maxpref=%d, realname='%s'\n", host, nmx, maxpref, realname);
-	rmsgappend(SS, 1, "\r-> nmx=%d maxpref=%d realname='%s'",
-		   nmx, maxpref, realname);
+	if (maxmx > 2)
+	  rmsgappend(SS, 1, "\r-> nmx=%d maxpref=%d realname='%s'",
+		     nmx, maxpref, realname);
 
 	/* discard MX RRs with a value >= that of  myhost */
 	for (n = i = 0; n < nmx; ++n) {
@@ -686,7 +583,7 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  mx[0].host = NULL;
 	  SS->mxcount = 0;
 	  if (had_eai_again)
-	    return EX_TEMPFAIL;
+	    return EX_DEFERALL;
 	  return EX_OK;
 	}
 #ifndef TEST
@@ -720,21 +617,25 @@ getmxrr(SS, host, mx, maxmx, depth)
 	  }
 	  ++n;			/* found one! */
 	}
-	if (n == 0 && had_eai_again)
-	  return EX_TEMPFAIL;
 
-	if (n == 0) {/* MX's exist, but their WKS's show no TCP smtp service */
-	  rmsgappend(SS, 1, "\r=> NONE of MXes support SMTP!");
-	  time(&endtime);
-#ifndef TEST
-	  notary_setxdelay((int)(endtime-starttime));
-	  notaryreport(NULL,FAILED,"5.4.4 (DNS lookup report)",SS->remotemsg);
-#endif
-	  return EX_UNAVAILABLE;
-	}
 	nmx = n;
 	SS->mxcount = nmx;
 
+	if (n == 0 && had_eai_again)
+	  return EX_DEFERALL;
+
+	if (n == 0) {/* MX's exist, but their WKS's show no TCP smtp service */
+	  if (maxmx > 2) {
+	    rmsgappend(SS, 1, "\r=> NONE of MXes support SMTP!");
+	    time(&endtime);
+#ifndef TEST
+	    notary_setxdelay((int)(endtime-starttime));
+	    notaryreport(NULL,FAILED,"5.4.4 (DNS lookup report)",SS->remotemsg);
+#endif
+	  }
+	  return EX_UNAVAILABLE;
+	}
+
 	/* sort the records per preferrence value */
 	for (i = 0; i < nmx; i++) {
 	  for (j = i + 1; j < nmx; j++) {
@@ -800,11 +701,13 @@ getmxrr(SS, host, mx, maxmx, depth)
 	/* Even with errors in data retrieval, if we get
 	   ANY   MX entries, we are a happy camper! */
 	if (had_eai_again && nmx == 0)
-	  return EX_TEMPFAIL;
+	  return EX_DEFERALL;
 	return EX_OK;
 }
 
 
+/* rmsgappend() is here because of getmxrr() test facility */
+
 #ifdef HAVE_STDARG_H
 #ifdef __STDC__
 void
@@ -913,6 +816,8 @@ int main(argc, argv)
 	int rc;
 	SmtpState SS;
 	char *host, *s;
+	char realname[1024];
+	time_t realnamettl;
 
 	progname = argv[0];
 
@@ -923,11 +828,28 @@ int main(argc, argv)
 
 	if (argc != 2) {
 	  printf("Usage: getmxrr-test domain.name\n");
+	  printf("\n");
+	  printf("Std tests:\n");
+	  printf("     getmxrr-test  timeout-mx.zmailer.org\n");
+	  printf("  -> GETMXRR() rc=100 EX_DEFERALL; mxcount=0\n");
+	  printf("  -> NO SUCCESSFULLY COLLECTED MX DATA, LOOKING FOR A/AAAA DATA:\n");
+	  printf("  -> .. getaddrinfo() (INET*) r=-2 (no address)\n");
+	  printf("  ->  GOT NO ADDRESSES!\n");
+	  printf("\n");
+	  printf("     getmxrr-test  timeout-zone.zmailer.org\n");
+	  printf("  -> GETMXRR() rc=100 EX_DEFERALL; mxcount=0\n");
+	  printf("  -> NO SUCCESSFULLY COLLECTED MX DATA, LOOKING FOR A/AAAA DATA:\n");
+	  printf("  -> .. getaddrinfo() (INET*) r=-3 (temporary failure in name resolution)\n");
+	  printf("  ->  GOT NO ADDRESSES!\n");
 	  exit(EX_USAGE);
 	}
 
+	printf("ZMAILER GETMXRR() TEST HARNESS\n");
 
-	rc = getmxrr(&SS, host, SS.mxh, MAXFORWARDERS, 0);
+	realname[0] = 0;
+	realnamettl = 84600;
+
+	rc = getmxrr(&SS, host, SS.mxh, MAXFORWARDERS, 0, realname, sizeof(realname), &realnamettl);
 
 	switch (rc) {
 	case EX_OK:
@@ -975,12 +897,15 @@ int main(argc, argv)
 	case EX_NOPERM:
 	  s = "EX_NOPERM";
 	  break;
+	case EX_DEFERALL:
+	  s = "EX_DEFERALL";
+	  break;
 	default:
 	  s = "UNKNOWN!";
 	}
 
 
-	printf("getmxrr() rc=%d %s; mxcount=%d\n", rc, s, SS.mxcount);
+	printf("GETMXRR() rc=%d %s; mxcount=%d\n", rc, s, SS.mxcount);
 
 
 	if (SS.mxcount == 0 || SS.mxh[0].host == NULL) {
@@ -988,6 +913,8 @@ int main(argc, argv)
 	  struct addrinfo req, *ai;
 	  int r;
 
+	  printf("NO SUCCESSFULLY COLLECTED MX DATA, LOOKING FOR A/AAAA DATA:\n");
+
 	  memset(&req, 0, sizeof(req));
 	  req.ai_socktype = SOCK_STREAM;
 	  req.ai_protocol = IPPROTO_TCP;
@@ -999,15 +926,15 @@ int main(argc, argv)
 	  errno = 0;
 	  /* Either forbidden MX usage, or does not have MX entries! */
 
-#ifdef HAVE_GETADDRINFO
-	  r = getaddrinfo(host, "smtp", &req, &ai);
+#ifdef HAVE__GETADDRINFO_
+	  r = _getaddrinfo_(host, "0", &req, &ai, SS.verboselog);
 #else
-	  r = _getaddrinfo_(host, "smtp", &req, &ai, SS.verboselog);
+	  r = getaddrinfo(host, "0", &req, &ai);
 #endif
 
 	  if (SS.verboselog)
 	    fprintf(SS.verboselog,
-		    "getaddrinfo('%s','smtp') -> r=%d (%s), ai=%p\n",
+		    "getaddrinfo('%s','0' (INET)) -> r=%d (%s), ai=%p\n",
 		    host, r, gai_strerror(r), ai);
 #if defined(AF_INET6) && defined(INET6)
 	  if (use_ipv6) {
@@ -1022,14 +949,14 @@ int main(argc, argv)
 
 	    /* This resolves CNAME, it should not happen in case
 	       of MX server, though..    */
-#ifdef HAVE_GETADDRINFO
-	    i2 = getaddrinfo(host, "0", &req, &ai2);
-#else
+#ifdef HAVE__GETADDRINFO_
 	    i2 = _getaddrinfo_(host, "0", &req, &ai2, SS.verboselog);
+#else
+	    i2 = getaddrinfo(host, "0", &req, &ai2);
 #endif
 	    if (SS.verboselog)
 	      fprintf(SS.verboselog,
-		      "  getaddrinfo('%s','smtp') -> r=%d (%s), ai=%p\n",
+		      "getaddrinfo('%s','0' (INET6)) -> r=%d (%s), ai=%p\n",
 		      host,i2, gai_strerror(i2), ai2);
 
 	    if (r != 0 && i2 == 0) {
@@ -1045,6 +972,10 @@ int main(argc, argv)
 	    }
 	  }
 #endif
+	  if (ai)
+	    printf("  GOT SOME ADDRESS, SUCCESS!\n");
+	  else
+	    printf("  GOT NO ADDRESSES!\n");
 	}
 
 	return 0;
