@@ -1,7 +1,8 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1991-2000 by Matti Aarnio -- modifications, including MIME
+ *	Copyright 1991-2003 by Matti Aarnio -- modifications, including
+ *	MIME things...
  */
 
 #define	RFC974		/* If BIND, check that TCP SMTP service is enabled */
@@ -33,34 +34,22 @@
 #include <setjmp.h>
 #include <string.h>
 
-#include "ta.h"
+#include "zresolv.h"
 
+#include "ta.h"
 #include "mail.h"
 #include "zsyslog.h"
-
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#ifdef HAVE_NETINET_IN6_H
-# include <netinet/in6.h>
-#endif
-#ifdef HAVE_NETINET6_IN6_H
-# include <netinet6/in6.h>
-#endif
-#ifdef HAVE_LINUX_IN6_H
-# include <linux/in6.h>
-#endif
-#include <netdb.h>
-#ifndef EAI_AGAIN
-# include "netdb6.h"
-#endif
-
+#include "dnsgetrr.h"
 #include "zmalloc.h"
 #include "libz.h"
 #include "libc.h"
 
-#include "dnsgetrr.h"
+#include "shmmib.h"
 
+#include <sys/socket.h>
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 
 #ifdef _AIX /* Defines NFDBITS, et.al. */
 #include <sys/types.h>
@@ -72,6 +61,14 @@
 
 #include <sys/time.h>
 
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
+#ifdef HAVE_NETINET_TCP_H
+#include <netinet/tcp.h>
+#endif
+
 #ifndef	NFDBITS
 /*
  * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
@@ -121,72 +118,9 @@ typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
 #ifdef HAVE_OPENSSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/rand.h>
 #endif /* - HAVE_OPENSSL */
 
-#if	defined(TRY_AGAIN) && defined(HAVE_RESOLVER)
-#define	BIND		/* Want BIND (named) nameserver support enabled */
-#endif	/* TRY_AGAIN */
-#ifdef	BIND
-#ifdef NOERROR
-#undef NOERROR		/* Several SysV-streams using systems have NOERROR,
-			   which is not the same as  <arpa/nameser.h> has! */
-#endif
-#include <arpa/nameser.h>
-#include <resolv.h>
-
-#ifndef	BIND_VER
-#ifdef	GETLONG
-/* 4.7.3 introduced the {GET,PUT}{LONG,SHORT} macros in nameser.h */
-#define	BIND_VER	473
-#else	/* !GETLONG */
-#define	BIND_VER	472
-#endif	/* GETLONG */
-#endif	/* !BIND_VER */
-#endif	/* BIND */
-
-/* Define all those things which exist on newer BINDs, and which may
-   get returned to us, when we make a query with  T_ANY ... */
-
-#ifndef	T_TXT
-# define T_TXT 16	/* Text strings */
-#endif
-#ifndef T_RP
-# define T_RP 17	/* Responsible person */
-#endif
-#ifndef T_AFSDB
-# define T_AFSDB 18	/* AFS cell database */
-#endif
-#ifndef T_X25
-# define T_X25 19	/* X.25 calling address */
-#endif
-#ifndef T_ISDN
-# define T_ISDN 20	/* ISDN calling address */
-#endif
-#ifndef T_RT
-# define T_RT 21	/* router */
-#endif
-#ifndef T_NSAP
-# define T_NSAP 22	/* NSAP address */
-#endif
-#ifndef T_NSAP_PTR
-# define T_NSAP_PTR 23	/* reverse NSAP lookup (depreciated) */
-#endif
-#ifndef	T_UINFO
-# define T_UINFO 100
-#endif
-#ifndef T_UID
-# define T_UID 101
-#endif
-#ifndef T_GID
-# define T_GID 102
-#endif
-#ifndef T_UNSPEC
-# define T_UNSPEC 103
-#endif
-#ifndef T_SA
-# define T_SA 200		/* Shuffle addresses */
-#endif
-
 
 
 #ifndef	SEEK_SET
@@ -235,9 +169,7 @@ extern int noalarmjmp;		/* Don't long-jmp on alarm! */
 extern jmp_buf alarmjmp;
 extern jmp_buf procabortjmp;
 extern int procabortset;
-#if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
 extern int readalready;		/* does buffer contain valid message data? */
-#endif
 extern int wantreserved;	/* open connection on secure (reserved) port */
 extern int statusreport;	/* put status reports on the command line */
 extern int force_8bit;		/* Claim to the remote to be 8-bit system, even
@@ -283,26 +215,13 @@ extern int use_ipv6;
 #define ESMTP_STARTTLS    0x0040 /* RFC 2487	 */
 #endif /* - HAVE_OPENSSL */
 #define ESMTP_DELIVERBY   0x0080 /* RFC 2852      */
-
-
-typedef union {
-  struct sockaddr_in  v4;
-#if defined(AF_INET6) && defined(INET6)
-  struct sockaddr_in6 v6;
-#endif
-} Usockaddr;
+#define ESMTP_AUTH        0x0100 /* RFC 2554+++   */
 
 
 # ifdef RFC974
 
-#if	defined(BIND_VER) && (BIND_VER >= 473)
-typedef u_char msgdata;
-#else	/* !defined(BIND_VER) || (BIND_VER < 473) */
-typedef char msgdata;
-#endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
-
 struct mxdata {
-	msgdata		*host;
+	char		*host;
 	int		 pref;
 	time_t		 expiry;
 	struct addrinfo *ai;
@@ -316,6 +235,8 @@ struct smtpdisc {
 
 
 typedef enum {
+  SMTPSTATE_CONNECT  = 0,
+  SMTPSTATE_HELO     = 0,
   SMTPSTATE_MAILFROM = 0,
   SMTPSTATE_RCPTTO   = 1,
   SMTPSTATE_DATA     = 2,
@@ -324,7 +245,51 @@ typedef enum {
   SMTPSTATE99        = 99
 } SMTPSTATES;
 
-typedef struct {
+
+#ifdef HAVE_OPENSSL
+struct _SmtpState_SSL_aux {
+  int   sslmode;		/* Set, when SSL/TLS in running */
+  SSL * ssl;
+  SSL_CTX * ctx;
+
+  int   wantreadwrite;		/* <0: read, =0: nothing, >0: write */
+#if 0
+  char *sslwrbuf;
+  int   sslwrspace, sslwrin, sslwrout;
+  /* space, how much stuff in, where the output cursor is */
+#endif
+
+  const char *peername_save;	 /* strdup()ed string */
+  const char *peer_subject;      /* strdup()ed string */
+  const char *peer_issuer;       /* strdup()ed string */
+  const char *peer_fingerprint;  /* strdup()ed string */
+  const char *peer_CN;           /* strdup()ed string */
+  const char *issuer_CN;         /* strdup()ed string */
+  const char *peer_CN1;          /* strdup()ed string */
+  const char *issuer_CN1;        /* strdup()ed string */
+  const char *notBefore;         /* strdup()ed string */
+  const char *notAfter;          /* strdup()ed string */
+
+  unsigned char peername_md5[MD5_DIGEST_LENGTH];
+
+  int	peer_verified;
+
+  const char *protocol;		/* from SSL_get_version() */
+  const char *cipher_name;	/* from SSL_CIPHER_get_name() */
+  int	cipher_usebits;
+  int	cipher_algbits;
+
+  int verify_depth;
+  int verify_error; /*  = X509_V_OK; */
+
+  int enforce_verify_errors;
+  int enforce_CN;
+};
+
+#endif /* - HAVE_OPENSSL */
+
+
+typedef struct { /* SmtpState */
   int  ehlo_capabilities;	/* Capabilities of the remote system */
   int  esmtp_on_banner;
   int  within_ehlo;
@@ -349,9 +314,12 @@ typedef struct {
 
   int   smtpfd;			/* FD from the remote host		*/
   Sfio_t *smtpfp;		/* Sfio_t* to the remote host           */
+  int   writeclosed;		/* The SMTP socket is closed for write  */
   struct smtpdisc smtpdisc;	/* SMTP outstream discipline data	*/
+  int   lasterrno;		/* Last errno value			*/
+  int	do_rset;		/* Will possibly need to do RSET	*/
+  time_t lastactiontime;	/* When last (write) action was ?	*/
 
-  char *smtphost;		/* strdup()ed name of the remote host */
   char *myhostname;		/* strdup()ed name of my outbound interface */
 
   FILE *verboselog;		/* verboselogfile */
@@ -374,15 +342,18 @@ typedef struct {
   int rcptcnt;			/* PIPELINING variables */
   int rcptstates;
 
-#define RCPTSTATE_OK   0x001  /* At least one OK   state   */
-#define FROMSTATE_OK   0x002  /* MAIL FROM --> 2XX code */
-#define DATASTATE_OK   0x004  /* DATA/BDAT --> 2/3XX code */
-#define RCPTSTATE_400  0x010  /* At least one TEMP failure */
-#define FROMSTATE_400  0x020  /* MAIL FROM --> 4XX code */
-#define DATASTATE_400  0x040  /* DATA/BDAT --> 4XX code */
-#define RCPTSTATE_500  0x100  /* At least one PERM failure */
-#define FROMSTATE_500  0x200  /* MAIL FROM --> 5XX code */
-#define DATASTATE_500  0x400  /* DATA/BDAT --> 5XX code */
+#define HELOSTATE_OK   0x0001
+#define HELOSTATE_400  0x0002
+#define HELOSTATE_500  0x0004
+#define FROMSTATE_OK   0x0010  /* MAIL FROM --> 2XX code */
+#define FROMSTATE_400  0x0020  /* MAIL FROM --> 4XX code */
+#define FROMSTATE_500  0x0040  /* MAIL FROM --> 5XX code */
+#define RCPTSTATE_OK   0x0100  /* At least one OK   state   */
+#define RCPTSTATE_400  0x0200  /* At least one TEMP failure */
+#define RCPTSTATE_500  0x0400  /* At least one PERM failure */
+#define DATASTATE_OK   0x1000  /* DATA/BDAT --> 2/3XX code */
+#define DATASTATE_400  0x2000  /* DATA/BDAT --> 4XX code */
+#define DATASTATE_500  0x4000  /* DATA/BDAT --> 5XX code */
 
   int state;
   int alarmcnt;
@@ -394,7 +365,7 @@ typedef struct {
   int   chunksize, chunkspace;
 
   char remotemsg[2*ZBUFSIZ];
-  char *remotemsgs[5];
+  char *remotemsgs[7];
 
   SMTPSTATES cmdstate, prevcmdstate;
 
@@ -410,19 +381,10 @@ typedef struct {
   int  stdincurs; /* Consumed  */
 
 #ifdef HAVE_OPENSSL
-  int   sslmode;		/* Set, when SSL/TLS in running */
-  SSL * ssl;
-  SSL_CTX * ctx;
-
-  char *sslwrbuf;
-  int   sslwrspace, sslwrin, sslwrout;
-  /* space, how much stuff in, where the output cursor is */
-  int   wantreadwrite;		/* <0: read, =0: nothing, >0: write */
+  struct _SmtpState_SSL_aux TLS;
 #endif /* - HAVE_OPENSSL */
-  char *tls_cipher_info;
-  char *tls_ccert_subject;
-  char *tls_ccert_issuer;
-  char *tls_ccert_fingerprint;
+
+  const char *taspoolid;
 
 } SmtpState;
 
@@ -449,7 +411,7 @@ extern int  optind;
 
 extern char **environ;
 
-extern int  deliver    __((SmtpState *SS, struct ctldesc *dp, struct rcpt *startrp, struct rcpt *endrp));
+extern int  deliver    __((SmtpState *SS, struct ctldesc *dp, struct rcpt *startrp, struct rcpt *endrp, const char *host, const int noMX));
 extern int  writebuf   __((SmtpState *SS, const char *buf, int len));
 extern int  writemimeline __((SmtpState *SS, const char *buf, int len, CONVERTMODE cvtmode));
 extern int  appendlet  __((SmtpState *SS, struct ctldesc *dp, CONVERTMODE convertmode, struct ct_data *));
@@ -466,7 +428,7 @@ extern int  process    __((SmtpState *SS, struct ctldesc*, int, const char*, int
 extern int  check_7bit_cleanness __((struct ctldesc *dp));
 extern void notarystatsave __((SmtpState *SS, char *smtpstatline, char *status));
 
-extern int  makeconn  __((SmtpState *SS, struct addrinfo *, int));
+extern int  makeconn  __((SmtpState *SS, const char *, struct addrinfo *, int));
 extern int  makereconn __((SmtpState *SS));
 extern int  vcsetup  __((SmtpState *SS, struct sockaddr *, int*, char*));
 #ifdef	BIND
@@ -474,7 +436,7 @@ extern int  rightmx  __((const char*, const char*, void*));
 extern int  h_errno;
 extern int  res_mkquery(), res_send(), dn_skipname(), dn_expand();
 # ifdef RFC974
-extern int  getmxrr __((SmtpState *, const char*, struct mxdata*, int, int));
+extern int  getmxrr __((SmtpState *SS, const char *host, struct mxdata mx[], int maxmx, int depth, char *realname, const int realnamesize, time_t *realnamettl));
 # endif /* RFC974 */
 extern void mxsetsave __((SmtpState *SS, const char *));
 #endif	/* BIND */
@@ -491,7 +453,6 @@ extern ssize_t smtp_sfwrite __((Sfio_t *, const void *, size_t, Sfdisc_t *));
 extern int  zsfsetfd     __((Sfio_t *, int));
 extern int  smtp_nbread  __((SmtpState *, void *, int));
 
-extern int  getmxrr __((SmtpState *SS, const char *host, struct mxdata mx[], int maxmx, int depth));
 
 extern void rmsgappend __((SmtpState *, int, const char *, ...));
 
