@@ -1,4 +1,4 @@
-/* Copyright 1993-1997 - Matti Aarnio
+/* Copyright 1993-2002 - Matti Aarnio
 
    The way the Zmailer uses DBM entries is by using strings with
    their terminating NULL as keys, and as data..  Thus the length
@@ -15,33 +15,18 @@
 #include <ctype.h>
 #include <sys/types.h>
 #include <fcntl.h>
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 #define datum Ndatum
 #include <ndbm.h>
 #undef datum
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 #define datum Gdatum
 #include <gdbm.h>
 #undef datum
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
-# include <db_185.h>
-#else
-#ifdef HAVE_DB2_DB_H
-# include <db2/db.h>
-#else
-#ifdef HAVE_DB_H
-# include <db.h>
-#else
-#ifdef HAVE_DB1_DB_H
-# include <db1/db.h>
-#endif
-#endif
-#endif
-#endif
-#endif
+
+#include "sleepycatdb.h"
 
 #define _POLICYTEST_INTERNAL_
 #include "policy.h"
@@ -76,13 +61,13 @@ int err;
 {
 	fprintf(stderr, "Usage: %s [-l|-u]] [-A][-a|-p][-s] dbtype database.name [infilename|-]\n", prog);
 	fprintf(stderr, "  where supported dbtypes are:");
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	fprintf(stderr, " ndbm");
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	fprintf(stderr, " gdbm");
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
 	fprintf(stderr, " btree bhash");
 #endif
 	fprintf(stderr, "\n");
@@ -93,15 +78,20 @@ int err;
 	  exit(1);
 	}
 	fprintf(stderr, "\n\n  If no infilename is defined, database.name is assumed.\n");
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	fprintf(stderr, "  (NDBM appends  .pag, and .dir  into actual db file names..)\n");
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	fprintf(stderr, "  (GDBM appends .gdbm, to the actual db file name..)\n");
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
 	fprintf(stderr, "  (BTREE appends .db, to the actual db file name..)\n");
 	fprintf(stderr, "  (BHASH appends .pag, and .dir into actual db file names..)\n");
+#if defined(HAVE_DB2) || defined(HAVE_DB3) || defined(HAVE_DB4)
+	fprintf(stderr, "    (Version: %s)\n", db_version(NULL,NULL,NULL));
+#else
+	fprintf(stderr, "  (Version 1.x ?)\n");
+#endif
 #endif
 	fprintf(stderr, "\n");
 
@@ -158,13 +148,13 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
      const void *s, *t;
      const int slen, tlen;
 {
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	DBM *ndbmfile = dbf;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	GDBM_FILE gdbmfile = dbf;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
 	DB *dbfile = dbf;
 #endif
 	int rc = -2;
@@ -178,58 +168,68 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 		    (const char*)t, (const char *)s);
 	}
 
-#ifdef HAVE_NDBM_H
-	if (typ == 1) {
-		Ndatum Ndat, Nkey;
-		Nkey.dptr = (void*)t;
-		Nkey.dsize = tlen;
-		Ndat.dptr = (void*)s;
-		Ndat.dsize = slen;
-		rc = dbm_store(ndbmfile, Nkey, Ndat,
-			       overwritemode ? DBM_REPLACE : DBM_INSERT);
 
-		if (rc < 0 && errno == ENOSPC) {
-		  fprintf(stderr,"NDBM does not allow KEY.LEN + DATA.LEN to exceed 1024 bytes!  linenro=%d\n",linenum);
-		}
-	}
+	switch (typ) {
+#ifdef HAVE_NDBM
+	case 1:
+	  {
+	    Ndatum Ndat, Nkey;
+	    Nkey.dptr = (void*)t;
+	    Nkey.dsize = tlen;
+	    Ndat.dptr = (void*)s;
+	    Ndat.dsize = slen;
+	    rc = dbm_store(ndbmfile, Nkey, Ndat,
+			   overwritemode ? DBM_REPLACE : DBM_INSERT);
+
+	    if (rc < 0 && errno == ENOSPC) {
+	      fprintf(stderr,"NDBM does not allow KEY.LEN + DATA.LEN to exceed 1024 bytes!  linenro=%d\n",linenum);
+	    }
+	  }
+	  break;
 #endif
-#ifdef HAVE_GDBM_H
-	if (typ == 2) {
-		Gdatum Gdat, Gkey;
-		Gkey.dptr = (void*)t;
-		Gkey.dsize = tlen;
-		Gdat.dptr = (void*)s;
-		Gdat.dsize = slen;
-		rc = gdbm_store(gdbmfile, Gkey, Gdat,
-				overwritemode ? GDBM_REPLACE : GDBM_INSERT);
-	}
+#ifdef HAVE_GDBM
+	case 2:
+	  {
+	    Gdatum Gdat, Gkey;
+	    Gkey.dptr = (void*)t;
+	    Gkey.dsize = tlen;
+	    Gdat.dptr = (void*)s;
+	    Gdat.dsize = slen;
+	    rc = gdbm_store(gdbmfile, Gkey, Gdat,
+			    overwritemode ? GDBM_REPLACE : GDBM_INSERT);
+	  }
+	  break;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-	if (typ == 3 || typ == 4) {
-		DBT Bkey, Bdat;
+#ifdef HAVE_DB
+	case 3: case 4:
+	  {
+	    DBT Bkey, Bdat;
 
-		memset(&Bkey,0,sizeof(Bkey));
-		memset(&Bdat,0,sizeof(Bdat));
+	    memset(&Bkey,0,sizeof(Bkey));
+	    memset(&Bdat,0,sizeof(Bdat));
 
-		Bkey.data = (void*)t;
-		Bkey.size = tlen;
+	    Bkey.data = (void*)t;
+	    Bkey.size = tlen;
 
-		Bdat.data = (void*)s;
-		Bdat.size = slen;
+	    Bdat.data = (void*)s;
+	    Bdat.size = slen;
 
 #ifdef DB_INIT_TXN
-		rc = (dbfile->put) (dbfile, NULL, &Bkey, &Bdat,
-				    overwritemode ? 0: DB_NOOVERWRITE);
+	    rc = (dbfile->put) (dbfile, NULL, &Bkey, &Bdat,
+				overwritemode ? 0: DB_NOOVERWRITE);
 
-		/* emulate BSD DB 1.85 return values */
-		rc = -rc;
+	    /* emulate BSD DB 1.85 return values */
+	    rc = -rc;
 
 #else
-		rc = (dbfile->put) (dbfile, &Bkey, &Bdat,
-				    overwritemode ? 0: R_NOOVERWRITE);
+	    rc = (dbfile->put) (dbfile, &Bkey, &Bdat,
+				overwritemode ? 0: R_NOOVERWRITE);
 #endif
-	}
+	  }
+	  break;
 #endif
+	} /* end of switch(typ) ... */
+
 
 	if (rc > 0 && append_mode == 0)
 	  return rc; /* Duh! Duplicate, and no append_mode :-( */
@@ -243,46 +243,55 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 	  char *newptr = NULL;
 	  int   newlen = 0;
 
-#ifdef HAVE_NDBM_H
-	  if (typ == 1) {
-	    Ndatum Ndat, Nkey;
-	    Nkey.dptr = (void*)t;
-	    Nkey.dsize = tlen;
-	    Ndat = dbm_fetch(ndbmfile, Nkey);
-	    dataptr = Ndat.dptr;
-	    datalen = Ndat.dsize;
-	  }
-#endif
-#ifdef HAVE_GDBM_H
-	  if (typ == 2) {
-	    Gdatum Gdat, Gkey;
-	    Gkey.dptr = (void*)t;
-	    Gkey.dsize = tlen;
-	    Gdat = gdbm_fetch(gdbmfile, Gkey);
-	    dataptr = Gdat.dptr; /* Must free() this */
-	    datalen = Gdat.dsize;
-	  }
-#endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-	  if (typ == 3 || typ == 4) {
-	    DBT Bkey, Bdat;
+	  switch (typ) {
 
-	    memset(&Bkey,0,sizeof(Bkey));
-	    memset(&Bdat,0,sizeof(Bdat));
+#ifdef HAVE_NDBM
+	  case 1:
+	    {
+	      Ndatum Ndat, Nkey;
+	      Nkey.dptr = (void*)t;
+	      Nkey.dsize = tlen;
+	      Ndat = dbm_fetch(ndbmfile, Nkey);
+	      dataptr = Ndat.dptr;
+	      datalen = Ndat.dsize;
+	    }
+	    break;
+#endif
+#ifdef HAVE_GDBM
+	  case 2:
+	    {
+	      Gdatum Gdat, Gkey;
+	      Gkey.dptr = (void*)t;
+	      Gkey.dsize = tlen;
+	      Gdat = gdbm_fetch(gdbmfile, Gkey);
+	      dataptr = Gdat.dptr; /* Must free() this */
+	      datalen = Gdat.dsize;
+	    }
+	    break;
+#endif
+#ifdef HAVE_DB
+	  case 3: case 4:
+	    {
+	      DBT Bkey, Bdat;
 
-	    Bkey.data = (void*)t;
-	    Bkey.size = tlen;
+	      memset(&Bkey,0,sizeof(Bkey));
+	      memset(&Bdat,0,sizeof(Bdat));
+
+	      Bkey.data = (void*)t;
+	      Bkey.size = tlen;
 #ifdef DB_INIT_TXN
-	    rc = (dbfile->get) (dbfile, NULL, &Bkey, &Bdat, 0);
+	      rc = (dbfile->get) (dbfile, NULL, &Bkey, &Bdat, 0);
 #else
-	    rc = (dbfile->get) (dbfile, &Bkey, &Bdat, 0);
+	      rc = (dbfile->get) (dbfile, &Bkey, &Bdat, 0);
 #endif
-	    if (rc != 0)
-	      memset(&Bdat, 0, sizeof(Bdat));
-	    dataptr = Bdat.data;
-	    datalen = Bdat.size;
-	  }
+	      if (rc != 0)
+		memset(&Bdat, 0, sizeof(Bdat));
+	      dataptr = Bdat.data;
+	      datalen = Bdat.size;
+	    }
+	    break;
 #endif
+	  } /* end of .. switch(typ) ... */
 
 	  /* Ok, now we have  dataptr  and  datalen -- we should have
 	     non-NULL dataptr.  We need to build a new block with
@@ -443,7 +452,9 @@ const int typ;
 		/* Collect attribute pairs */
 		str1 = tokskip(s, &s);
 
-		if (*str1 == '#') continue; /* Begins a within line comment */
+		if (!str1 || *str1 == '#')
+		  continue; /* Data begins a within line comment, or
+			     is completely void.. */
 
 		if (str1 == NULL) {
 			fprintf(stderr,
@@ -782,18 +793,19 @@ char *argv[];
     FILE *infile = NULL;
     int c, rc;
     int typ = 0;
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
     DBM *ndbmfile = NULL;
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     GDBM_FILE gdbmfile = NULL;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     DB *dbfile = NULL;
 #endif
     char *dbtype = NULL;
     void *dbf = NULL;
     char *argv0 = argv[0];
+    int err;
 
     progname = argv[0];
 
@@ -851,71 +863,127 @@ char *argv[];
 	usage(argv0, "bad infile", errno);
 
     typ = 0;
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
     if (cistrcmp(dbtype, "ndbm") == 0)
 	typ = 1;
-    else
 #endif
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
     if (cistrcmp(dbtype, "gdbm") == 0)
 	typ = 2;
-    else
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
+#ifdef HAVE_DB
     if (cistrcmp(dbtype, "btree") == 0)
 	typ = 3;
     if (cistrcmp(dbtype, "bhash") == 0)
 	typ = 4;
 #endif
-    if (typ == 0)
-	usage(argv0, "unknown dbtype", 0);
 
-#ifdef HAVE_NDBM_H
-    if (typ == 1) {
+    switch (typ) {
+    case 0:
+	usage(argv0, "unknown dbtype", 0);
+	break;
+
+#ifdef HAVE_NDBM
+    case 1:
 	ndbmfile = dbm_open(dbasename, O_RDWR | O_CREAT | O_TRUNC, 0644);
 	dbf = ndbmfile;
-    }
+	break;
 #endif
-#ifdef HAVE_GDBM_H
-    if (typ == 2) {
+
+#ifdef HAVE_GDBM
+    case 2:
 	/* Play loose .. don't do syncs while writing */
 	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
 	strcat(dbasename, ".gdbm");	/* ALWAYS append this */
 	gdbmfile = gdbm_open(dbasename, 0, GDBM_NEWDB | GDBM_FAST, 0644, NULL);
 	dbf = gdbmfile;
-    }
+	break;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-#ifdef HAVE_DB_OPEN2
-    if (typ == 3) {
+#ifdef HAVE_DB
+#if defined(HAVE_DB3) || defined(HAVE_DB4)
+
+    case 3:
 	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
 	strcat(dbasename, ".db");	/* ALWAYS append this */
+
 	dbfile = NULL;
-	db_open(dbasename, DB_BTREE,  DB_CREATE|DB_TRUNCATE,
-		0644, NULL, NULL, &dbfile);
-	dbf = dbfile;
-    }
-    if (typ == 4) {
+	err = db_create(&dbfile, NULL, 0);
+	if (! err)
+	  err = dbfile->open(dbfile,
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+			     NULL, /* TXN id was added at SleepyDB 4.1 */
+#endif
+			     dbasename, NULL, DB_BTREE,
+			     DB_CREATE|DB_TRUNCATE, 0644);
+	if (!err)
+	  dbf = dbfile;
+	break;
+
+    case 4:
+
+	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
+	strcat(dbasename, ".db");	/* ALWAYS append this */
+
 	dbfile = NULL;
-	db_open(dbasename, DB_HASH,  DB_CREATE|DB_TRUNCATE,
-		0644, NULL, NULL, &dbfile);
-	dbf = dbfile;
-    }
+	err = db_create(&dbfile, NULL, 0);
+	if (! err)
+	  err = dbfile->open(dbfile,
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+			     NULL, /* TXN id was added at SleepyDB 4.1 */
+#endif
+			     dbasename, NULL, DB_HASH,
+			     DB_CREATE|DB_TRUNCATE, 0644);
+	if (!err)
+	  dbf = dbfile;
+	break;
+
 #else
-    if (typ == 3) {
+#if defined(HAVE_DB2)
+    case 3:
+
 	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
 	strcat(dbasename, ".db");	/* ALWAYS append this */
+
+	dbfile = NULL;
+	err = db_open(dbasename, DB_BTREE,  DB_CREATE|DB_TRUNCATE,
+		      0644, NULL, NULL, &dbfile);
+	if (! err)
+	  dbf = dbfile;
+	break;
+
+    case 4:
+
+	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
+	strcat(dbasename, ".db");	/* ALWAYS append this */
+
+	dbfile = NULL;
+	err = db_open(dbasename, DB_HASH,  DB_CREATE|DB_TRUNCATE,
+		      0644, NULL, NULL, &dbfile);
+	if (! err)
+	  dbf = dbfile;
+	break;
+#else
+    case 3:
+	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
+	strcat(dbasename, ".db");	/* ALWAYS append this */
+
 	dbfile = dbopen(dbasename, O_RDWR | O_CREAT | O_TRUNC, 0644,
 			DB_BTREE, NULL);
 	dbf = dbfile;
-    }
-    if (typ == 4) {
+	break;
+
+    case 4:
+	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
+	strcat(dbasename, ".db");	/* ALWAYS append this */
+
 	dbfile = dbopen(dbasename, O_RDWR | O_CREAT | O_TRUNC, 0644,
 			DB_HASH, NULL);
 	dbf = dbfile;
+	break;
+#endif
+#endif
+#endif
     }
-#endif
-#endif
     if (dbf == NULL)
 	usage(argv0, "Can't open dbase file", errno);
 
@@ -928,24 +996,28 @@ char *argv[];
     else
 	    rc = create_keyed_dbase(infile, dbf, typ);
 
-#ifdef HAVE_NDBM_H
-    if (typ == 1)
-	dbm_close(ndbmfile);
+    switch (typ) {
+#ifdef HAVE_NDBM
+    case 1:
+      dbm_close(ndbmfile);
+      break;
 #endif
-#ifdef HAVE_GDBM_H
-    if (typ == 2)
-	gdbm_close(gdbmfile);
+#ifdef HAVE_GDBM
+    case 2:
+      gdbm_close(gdbmfile);
+      break;
 #endif
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-    if (typ == 3 || typ == 4) {
-	(dbfile->sync) (dbfile, 0);
-#ifdef HAVE_DB_OPEN2
-	(dbfile->close) (dbfile, 0);
+#ifdef HAVE_DB
+    case 3: case 4:
+      (dbfile->sync) (dbfile, 0);
+#if defined(HAVE_DB_CLOSE2)
+      (dbfile->close) (dbfile, 0);
 #else
-	(dbfile->close) (dbfile);
+      (dbfile->close) (dbfile);
 #endif
-    }
+      break;
 #endif
+    }  /* end of .. switch(typ) .. */
 
     if (store_errors) {
       fprintf(stderr,"STORE ERRORS DURING DATABASE WRITE!\n");
